/* Generated by Cython 0.29 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "C:\\bld\\scikit-garden_1541891273783\\_h_env\\lib\\site-packages\\numpy\\core\\include\\numpy\\arrayobject.h",
            "C:\\bld\\scikit-garden_1541891273783\\_h_env\\lib\\site-packages\\numpy\\core\\include\\numpy\\ufuncobject.h"
        ],
        "extra_compile_args": [
            "-O3"
        ],
        "include_dirs": [
            "C:\\bld\\scikit-garden_1541891273783\\_h_env\\lib\\site-packages\\numpy\\core\\include"
        ],
        "name": "skgarden.mondrian.tree._tree",
        "sources": [
            "skgarden/mondrian/tree/_tree.pyx"
        ]
    },
    "module_name": "skgarden.mondrian.tree._tree"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29"
#define CYTHON_HEX_VERSION 0x001D00F0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_USE_DICT_VERSIONS
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
        static PY_UINT64_T __pyx_dict_version = 0;\
        static PyObject *__pyx_dict_cached_value = NULL;\
        if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
            (VAR) = __pyx_dict_cached_value;\
        } else {\
            (VAR) = __pyx_dict_cached_value = (LOOKUP);\
            __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
        }\
    }
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0; // PyThread_create_key reports success always
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif // TSS (Thread Specific Storage) API
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
  #define PyObject_Unicode             PyObject_Str
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : (Py_INCREF(func), func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__skgarden__mondrian__tree___tree
#define __PYX_HAVE_API__skgarden__mondrian__tree___tree
/* Early includes */
#include <string.h>
#include <stdio.h>
#include "numpy/arrayobject.h"
#include "numpy/ufuncobject.h"
#include "pythread.h"
#include <math.h>
#include <stdlib.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "skgarden\\mondrian\\tree\\_tree.pyx",
  "stringsource",
  "skgarden\\mondrian\\tree\\_tree.pxd",
  "__init__.pxd",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
  "skgarden\\mondrian\\tree\\_criterion.pxd",
  "skgarden\\mondrian\\tree\\_splitter.pxd",
  "skgarden\\mondrian\\tree\\_utils.pxd",
};
/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()

/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* BufferFormatStructs.proto */
#define IS_UNSIGNED(type) (((type) -1) > 0)
struct __Pyx_StructField_;
#define __PYX_BUF_FLAGS_PACKED_STRUCT (1 << 0)
typedef struct {
  const char* name;
  struct __Pyx_StructField_* fields;
  size_t size;
  size_t arraysize[8];
  int ndim;
  char typegroup;
  char is_unsigned;
  int flags;
} __Pyx_TypeInfo;
typedef struct __Pyx_StructField_ {
  __Pyx_TypeInfo* type;
  const char* name;
  size_t offset;
} __Pyx_StructField;
typedef struct {
  __Pyx_StructField* field;
  size_t parent_offset;
} __Pyx_BufFmt_StackElem;
typedef struct {
  __Pyx_StructField root;
  __Pyx_BufFmt_StackElem* head;
  size_t fmt_offset;
  size_t new_count, enc_count;
  size_t struct_alignment;
  int is_complex;
  char enc_type;
  char new_packmode;
  char enc_packmode;
  char is_valid_array;
} __Pyx_BufFmt_Context;


/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":776
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":777
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":778
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":779
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":783
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":784
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":785
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":786
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":790
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":791
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":800
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":801
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":802
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":804
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":805
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":806
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":808
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":809
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":811
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":812
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":813
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;

/* "_criterion.pxd":15
 * cimport numpy as np
 * 
 * ctypedef np.npy_float32 DTYPE_t          # Type of X             # <<<<<<<<<<<<<<
 * ctypedef np.npy_float64 DOUBLE_t         # Type of y, sample_weight
 * ctypedef np.npy_intp SIZE_t              # Type for indices and counters
 */
typedef npy_float32 __pyx_t_8skgarden_8mondrian_4tree_10_criterion_DTYPE_t;

/* "_criterion.pxd":16
 * 
 * ctypedef np.npy_float32 DTYPE_t          # Type of X
 * ctypedef np.npy_float64 DOUBLE_t         # Type of y, sample_weight             # <<<<<<<<<<<<<<
 * ctypedef np.npy_intp SIZE_t              # Type for indices and counters
 * ctypedef np.npy_int32 INT32_t            # Signed 32 bit integer
 */
typedef npy_float64 __pyx_t_8skgarden_8mondrian_4tree_10_criterion_DOUBLE_t;

/* "_criterion.pxd":17
 * ctypedef np.npy_float32 DTYPE_t          # Type of X
 * ctypedef np.npy_float64 DOUBLE_t         # Type of y, sample_weight
 * ctypedef np.npy_intp SIZE_t              # Type for indices and counters             # <<<<<<<<<<<<<<
 * ctypedef np.npy_int32 INT32_t            # Signed 32 bit integer
 * ctypedef np.npy_uint32 UINT32_t          # Unsigned 32 bit integer
 */
typedef npy_intp __pyx_t_8skgarden_8mondrian_4tree_10_criterion_SIZE_t;

/* "_criterion.pxd":18
 * ctypedef np.npy_float64 DOUBLE_t         # Type of y, sample_weight
 * ctypedef np.npy_intp SIZE_t              # Type for indices and counters
 * ctypedef np.npy_int32 INT32_t            # Signed 32 bit integer             # <<<<<<<<<<<<<<
 * ctypedef np.npy_uint32 UINT32_t          # Unsigned 32 bit integer
 * 
 */
typedef npy_int32 __pyx_t_8skgarden_8mondrian_4tree_10_criterion_INT32_t;

/* "_criterion.pxd":19
 * ctypedef np.npy_intp SIZE_t              # Type for indices and counters
 * ctypedef np.npy_int32 INT32_t            # Signed 32 bit integer
 * ctypedef np.npy_uint32 UINT32_t          # Unsigned 32 bit integer             # <<<<<<<<<<<<<<
 * 
 * cdef class Criterion:
 */
typedef npy_uint32 __pyx_t_8skgarden_8mondrian_4tree_10_criterion_UINT32_t;

/* "_splitter.pxd":17
 * from ._criterion cimport Criterion
 * 
 * ctypedef np.npy_float32 DTYPE_t          # Type of X             # <<<<<<<<<<<<<<
 * ctypedef np.npy_float64 DOUBLE_t         # Type of y, sample_weight
 * ctypedef np.npy_intp SIZE_t              # Type for indices and counters
 */
typedef npy_float32 __pyx_t_8skgarden_8mondrian_4tree_9_splitter_DTYPE_t;

/* "_splitter.pxd":18
 * 
 * ctypedef np.npy_float32 DTYPE_t          # Type of X
 * ctypedef np.npy_float64 DOUBLE_t         # Type of y, sample_weight             # <<<<<<<<<<<<<<
 * ctypedef np.npy_intp SIZE_t              # Type for indices and counters
 * ctypedef np.npy_int32 INT32_t            # Signed 32 bit integer
 */
typedef npy_float64 __pyx_t_8skgarden_8mondrian_4tree_9_splitter_DOUBLE_t;

/* "_splitter.pxd":19
 * ctypedef np.npy_float32 DTYPE_t          # Type of X
 * ctypedef np.npy_float64 DOUBLE_t         # Type of y, sample_weight
 * ctypedef np.npy_intp SIZE_t              # Type for indices and counters             # <<<<<<<<<<<<<<
 * ctypedef np.npy_int32 INT32_t            # Signed 32 bit integer
 * ctypedef np.npy_uint32 UINT32_t          # Unsigned 32 bit integer
 */
typedef npy_intp __pyx_t_8skgarden_8mondrian_4tree_9_splitter_SIZE_t;

/* "_splitter.pxd":20
 * ctypedef np.npy_float64 DOUBLE_t         # Type of y, sample_weight
 * ctypedef np.npy_intp SIZE_t              # Type for indices and counters
 * ctypedef np.npy_int32 INT32_t            # Signed 32 bit integer             # <<<<<<<<<<<<<<
 * ctypedef np.npy_uint32 UINT32_t          # Unsigned 32 bit integer
 * 
 */
typedef npy_int32 __pyx_t_8skgarden_8mondrian_4tree_9_splitter_INT32_t;

/* "_splitter.pxd":21
 * ctypedef np.npy_intp SIZE_t              # Type for indices and counters
 * ctypedef np.npy_int32 INT32_t            # Signed 32 bit integer
 * ctypedef np.npy_uint32 UINT32_t          # Unsigned 32 bit integer             # <<<<<<<<<<<<<<
 * 
 * cdef struct SplitRecord:
 */
typedef npy_uint32 __pyx_t_8skgarden_8mondrian_4tree_9_splitter_UINT32_t;

/* "_utils.pxd":15
 * from _tree cimport Node
 * 
 * ctypedef np.npy_float32 DTYPE_t          # Type of X             # <<<<<<<<<<<<<<
 * ctypedef np.npy_float64 DOUBLE_t         # Type of y, sample_weight
 * ctypedef np.npy_intp SIZE_t              # Type for indices and counters
 */
typedef npy_float32 __pyx_t_8skgarden_8mondrian_4tree_6_utils_DTYPE_t;

/* "_utils.pxd":16
 * 
 * ctypedef np.npy_float32 DTYPE_t          # Type of X
 * ctypedef np.npy_float64 DOUBLE_t         # Type of y, sample_weight             # <<<<<<<<<<<<<<
 * ctypedef np.npy_intp SIZE_t              # Type for indices and counters
 * ctypedef np.npy_int32 INT32_t            # Signed 32 bit integer
 */
typedef npy_float64 __pyx_t_8skgarden_8mondrian_4tree_6_utils_DOUBLE_t;

/* "_utils.pxd":17
 * ctypedef np.npy_float32 DTYPE_t          # Type of X
 * ctypedef np.npy_float64 DOUBLE_t         # Type of y, sample_weight
 * ctypedef np.npy_intp SIZE_t              # Type for indices and counters             # <<<<<<<<<<<<<<
 * ctypedef np.npy_int32 INT32_t            # Signed 32 bit integer
 * ctypedef np.npy_uint32 UINT32_t          # Unsigned 32 bit integer
 */
typedef npy_intp __pyx_t_8skgarden_8mondrian_4tree_6_utils_SIZE_t;

/* "_utils.pxd":18
 * ctypedef np.npy_float64 DOUBLE_t         # Type of y, sample_weight
 * ctypedef np.npy_intp SIZE_t              # Type for indices and counters
 * ctypedef np.npy_int32 INT32_t            # Signed 32 bit integer             # <<<<<<<<<<<<<<
 * ctypedef np.npy_uint32 UINT32_t          # Unsigned 32 bit integer
 * 
 */
typedef npy_int32 __pyx_t_8skgarden_8mondrian_4tree_6_utils_INT32_t;

/* "_utils.pxd":19
 * ctypedef np.npy_intp SIZE_t              # Type for indices and counters
 * ctypedef np.npy_int32 INT32_t            # Signed 32 bit integer
 * ctypedef np.npy_uint32 UINT32_t          # Unsigned 32 bit integer             # <<<<<<<<<<<<<<
 * 
 * cdef enum:
 */
typedef npy_uint32 __pyx_t_8skgarden_8mondrian_4tree_6_utils_UINT32_t;

/* "skgarden/mondrian/tree/_tree.pxd":16
 * cimport numpy as np
 * 
 * ctypedef np.npy_float32 DTYPE_t          # Type of X             # <<<<<<<<<<<<<<
 * ctypedef np.npy_float64 DOUBLE_t         # Type of y, sample_weight
 * ctypedef np.npy_intp SIZE_t              # Type for indices and counters
 */
typedef npy_float32 __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t;

/* "skgarden/mondrian/tree/_tree.pxd":17
 * 
 * ctypedef np.npy_float32 DTYPE_t          # Type of X
 * ctypedef np.npy_float64 DOUBLE_t         # Type of y, sample_weight             # <<<<<<<<<<<<<<
 * ctypedef np.npy_intp SIZE_t              # Type for indices and counters
 * ctypedef np.npy_int32 INT32_t            # Signed 32 bit integer
 */
typedef npy_float64 __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t;

/* "skgarden/mondrian/tree/_tree.pxd":18
 * ctypedef np.npy_float32 DTYPE_t          # Type of X
 * ctypedef np.npy_float64 DOUBLE_t         # Type of y, sample_weight
 * ctypedef np.npy_intp SIZE_t              # Type for indices and counters             # <<<<<<<<<<<<<<
 * ctypedef np.npy_int32 INT32_t            # Signed 32 bit integer
 * ctypedef np.npy_uint32 UINT32_t          # Unsigned 32 bit integer
 */
typedef npy_intp __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t;

/* "skgarden/mondrian/tree/_tree.pxd":19
 * ctypedef np.npy_float64 DOUBLE_t         # Type of y, sample_weight
 * ctypedef np.npy_intp SIZE_t              # Type for indices and counters
 * ctypedef np.npy_int32 INT32_t            # Signed 32 bit integer             # <<<<<<<<<<<<<<
 * ctypedef np.npy_uint32 UINT32_t          # Unsigned 32 bit integer
 * 
 */
typedef npy_int32 __pyx_t_8skgarden_8mondrian_4tree_5_tree_INT32_t;

/* "skgarden/mondrian/tree/_tree.pxd":20
 * ctypedef np.npy_intp SIZE_t              # Type for indices and counters
 * ctypedef np.npy_int32 INT32_t            # Signed 32 bit integer
 * ctypedef np.npy_uint32 UINT32_t          # Unsigned 32 bit integer             # <<<<<<<<<<<<<<
 * 
 * from ._splitter cimport Splitter
 */
typedef npy_uint32 __pyx_t_8skgarden_8mondrian_4tree_5_tree_UINT32_t;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);


/*--- Type declarations ---*/
struct __pyx_obj_8skgarden_8mondrian_4tree_10_criterion_Criterion;
struct __pyx_obj_8skgarden_8mondrian_4tree_9_splitter_Splitter;
struct __pyx_obj_8skgarden_8mondrian_4tree_6_utils_Stack;
struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree;
struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder;
struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder;
struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder;

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":815
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":816
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":817
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":819
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;
struct __pyx_t_8skgarden_8mondrian_4tree_9_splitter_SplitRecord;
struct __pyx_opt_args_8skgarden_8mondrian_4tree_9_splitter_8Splitter_init;

/* "_splitter.pxd":23
 * ctypedef np.npy_uint32 UINT32_t          # Unsigned 32 bit integer
 * 
 * cdef struct SplitRecord:             # <<<<<<<<<<<<<<
 *     # Data to track sample split
 *     SIZE_t feature         # Which feature to split on.
 */
struct __pyx_t_8skgarden_8mondrian_4tree_9_splitter_SplitRecord {
  __pyx_t_8skgarden_8mondrian_4tree_9_splitter_SIZE_t feature;
  __pyx_t_8skgarden_8mondrian_4tree_9_splitter_SIZE_t pos;
  double threshold;
  double improvement;
  double impurity_left;
  double impurity_right;
  double E;
};

/* "_splitter.pxd":86
 * 
 *     # Methods
 *     cdef int init(self, object X, np.ndarray y,             # <<<<<<<<<<<<<<
 *                   DOUBLE_t* sample_weight,
 *                   np.ndarray X_idx_sorted=*) except -1
 */
struct __pyx_opt_args_8skgarden_8mondrian_4tree_9_splitter_8Splitter_init {
  int __pyx_n;
  PyArrayObject *X_idx_sorted;
};
struct __pyx_t_8skgarden_8mondrian_4tree_6_utils_StackRecord;

/* "_utils.pxd":21
 * ctypedef np.npy_uint32 UINT32_t          # Unsigned 32 bit integer
 * 
 * cdef enum:             # <<<<<<<<<<<<<<
 *     # Max value for our rand_r replacement (near the bottom).
 *     # We don't use RAND_MAX because it's different across platforms and
 */
enum  {
  __pyx_e_8skgarden_8mondrian_4tree_6_utils_RAND_R_MAX = 0x7FFFFFFF
};

/* "_utils.pxd":69
 * 
 * # A record on the stack for depth-first tree growing
 * cdef struct StackRecord:             # <<<<<<<<<<<<<<
 *     SIZE_t start
 *     SIZE_t end
 */
struct __pyx_t_8skgarden_8mondrian_4tree_6_utils_StackRecord {
  __pyx_t_8skgarden_8mondrian_4tree_6_utils_SIZE_t start;
  __pyx_t_8skgarden_8mondrian_4tree_6_utils_SIZE_t end;
  __pyx_t_8skgarden_8mondrian_4tree_6_utils_SIZE_t depth;
  __pyx_t_8skgarden_8mondrian_4tree_6_utils_SIZE_t parent;
  int is_left;
  double impurity;
  __pyx_t_8skgarden_8mondrian_4tree_6_utils_SIZE_t n_constant_features;
};
struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node;
struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_4Tree__resize_c;
struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_4Tree_predict;
struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_4Tree_set_node_attributes;
struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_build;
struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder_build;
struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder_build;

/* "skgarden/mondrian/tree/_tree.pxd":25
 * from ._splitter cimport SplitRecord
 * 
 * cdef struct Node:             # <<<<<<<<<<<<<<
 *     # Base storage structure for the nodes in a Tree object
 * 
 */
struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node {
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t left_child;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t right_child;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t feature;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t threshold;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t impurity;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t n_node_samples;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t weighted_n_node_samples;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *lower_bounds;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *upper_bounds;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t tau;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t variance;
};

/* "skgarden/mondrian/tree/_tree.pxd":71
 *                           double E) nogil except -1
 *     cdef int _resize(self, SIZE_t capacity) nogil except -1
 *     cdef int _resize_c(self, SIZE_t capacity=*) nogil except -1             # <<<<<<<<<<<<<<
 * 
 *     cdef np.ndarray _get_value_ndarray(self)
 */
struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_4Tree__resize_c {
  int __pyx_n;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t capacity;
};

/* "skgarden/mondrian/tree/_tree.pxd":76
 *     cdef np.ndarray _get_node_ndarray(self)
 * 
 *     cpdef tuple predict(self, object X, bint return_std=*, bint is_regression=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef np.ndarray apply(self, object X)
 */
struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_4Tree_predict {
  int __pyx_n;
  int return_std;
  int is_regression;
};

/* "skgarden/mondrian/tree/_tree.pxd":88
 *                      SIZE_t X_f_stride, SIZE_t y_stride, UINT32_t random_state,
 *                      SIZE_t min_samples_split)
 *     cdef void set_node_attributes(self, SIZE_t node_ind, SIZE_t left_child,             # <<<<<<<<<<<<<<
 *                                   SIZE_t right_child, SIZE_t feature, DOUBLE_t threshold,
 *                                   DTYPE_t tau, SIZE_t n_node_samples,
 */
struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_4Tree_set_node_attributes {
  int __pyx_n;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t child_ind;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t y_start;
};

/* "skgarden/mondrian/tree/_tree.pxd":120
 *     cdef object random_state
 * 
 *     cpdef build(self, Tree tree, object X, np.ndarray y,             # <<<<<<<<<<<<<<
 *                 np.ndarray sample_weight=*,
 *                 np.ndarray X_idx_sorted=*)
 */
struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_build {
  int __pyx_n;
  PyArrayObject *sample_weight;
  PyArrayObject *X_idx_sorted;
};

/* "skgarden/mondrian/tree/_tree.pyx":137
 *         self.random_state = random_state
 * 
 *     cpdef build(self, Tree tree, object X, np.ndarray y,             # <<<<<<<<<<<<<<
 *                 np.ndarray sample_weight=None,
 *                 np.ndarray X_idx_sorted=None):
 */
struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder_build {
  int __pyx_n;
  PyArrayObject *sample_weight;
  PyArrayObject *X_idx_sorted;
};

/* "skgarden/mondrian/tree/_tree.pyx":184
 *         self.max_depth = max_depth
 * 
 *     cpdef build(self, Tree tree, object X, np.ndarray y,             # <<<<<<<<<<<<<<
 *                 np.ndarray sample_weight=None,
 *                 np.ndarray X_idx_sorted=None):
 */
struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder_build {
  int __pyx_n;
  PyArrayObject *sample_weight;
  PyArrayObject *X_idx_sorted;
};

/* "_criterion.pxd":21
 * ctypedef np.npy_uint32 UINT32_t          # Unsigned 32 bit integer
 * 
 * cdef class Criterion:             # <<<<<<<<<<<<<<
 *     # The criterion computes the impurity of a node and the reduction of
 *     # impurity of a split on that node. It also computes the output statistics
 */
struct __pyx_obj_8skgarden_8mondrian_4tree_10_criterion_Criterion {
  PyObject_HEAD
  struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_10_criterion_Criterion *__pyx_vtab;
  __pyx_t_8skgarden_8mondrian_4tree_10_criterion_DOUBLE_t *y;
  __pyx_t_8skgarden_8mondrian_4tree_10_criterion_SIZE_t y_stride;
  __pyx_t_8skgarden_8mondrian_4tree_10_criterion_DOUBLE_t *sample_weight;
  __pyx_t_8skgarden_8mondrian_4tree_10_criterion_SIZE_t *samples;
  __pyx_t_8skgarden_8mondrian_4tree_10_criterion_SIZE_t start;
  __pyx_t_8skgarden_8mondrian_4tree_10_criterion_SIZE_t pos;
  __pyx_t_8skgarden_8mondrian_4tree_10_criterion_SIZE_t end;
  __pyx_t_8skgarden_8mondrian_4tree_10_criterion_SIZE_t n_outputs;
  __pyx_t_8skgarden_8mondrian_4tree_10_criterion_SIZE_t n_samples;
  __pyx_t_8skgarden_8mondrian_4tree_10_criterion_SIZE_t n_node_samples;
  double weighted_n_samples;
  double weighted_n_node_samples;
  double weighted_n_left;
  double weighted_n_right;
  double *sum_total;
  double *sum_left;
  double *sum_right;
};


/* "_splitter.pxd":35
 *     double E               # Time of split.
 * 
 * cdef class Splitter:             # <<<<<<<<<<<<<<
 *     # The splitter searches in the input space for a feature and a threshold
 *     # to split the samples samples[start:end].
 */
struct __pyx_obj_8skgarden_8mondrian_4tree_9_splitter_Splitter {
  PyObject_HEAD
  struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_9_splitter_Splitter *__pyx_vtab;
  struct __pyx_obj_8skgarden_8mondrian_4tree_10_criterion_Criterion *criterion;
  __pyx_t_8skgarden_8mondrian_4tree_9_splitter_SIZE_t max_features;
  PyObject *random_state;
  __pyx_t_8skgarden_8mondrian_4tree_9_splitter_UINT32_t rand_r_state;
  __pyx_t_8skgarden_8mondrian_4tree_9_splitter_SIZE_t *samples;
  __pyx_t_8skgarden_8mondrian_4tree_9_splitter_SIZE_t n_samples;
  double weighted_n_samples;
  __pyx_t_8skgarden_8mondrian_4tree_9_splitter_SIZE_t *features;
  __pyx_t_8skgarden_8mondrian_4tree_9_splitter_SIZE_t *constant_features;
  __pyx_t_8skgarden_8mondrian_4tree_9_splitter_SIZE_t n_features;
  __pyx_t_8skgarden_8mondrian_4tree_9_splitter_DTYPE_t *feature_values;
  __pyx_t_8skgarden_8mondrian_4tree_9_splitter_SIZE_t start;
  __pyx_t_8skgarden_8mondrian_4tree_9_splitter_SIZE_t end;
  int presort;
  __pyx_t_8skgarden_8mondrian_4tree_9_splitter_DOUBLE_t *y;
  __pyx_t_8skgarden_8mondrian_4tree_9_splitter_SIZE_t y_stride;
  __pyx_t_8skgarden_8mondrian_4tree_9_splitter_DOUBLE_t *sample_weight;
  __pyx_t_8skgarden_8mondrian_4tree_9_splitter_DTYPE_t *upper_bounds;
  __pyx_t_8skgarden_8mondrian_4tree_9_splitter_DTYPE_t *lower_bounds;
};


/* "_utils.pxd":78
 *     SIZE_t n_constant_features
 * 
 * cdef class Stack:             # <<<<<<<<<<<<<<
 *     cdef SIZE_t capacity
 *     cdef SIZE_t top
 */
struct __pyx_obj_8skgarden_8mondrian_4tree_6_utils_Stack {
  PyObject_HEAD
  struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_6_utils_Stack *__pyx_vtab;
  __pyx_t_8skgarden_8mondrian_4tree_6_utils_SIZE_t capacity;
  __pyx_t_8skgarden_8mondrian_4tree_6_utils_SIZE_t top;
  struct __pyx_t_8skgarden_8mondrian_4tree_6_utils_StackRecord *stack_;
};


/* "skgarden/mondrian/tree/_tree.pxd":41
 * 
 * 
 * cdef class Tree:             # <<<<<<<<<<<<<<
 *     # The Tree object is a binary tree structure constructed by the
 *     # TreeBuilder. The tree structure is used for predictions and
 */
struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree {
  PyObject_HEAD
  struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_vtab;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t n_features;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t *n_classes;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t n_outputs;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t max_n_classes;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t max_depth;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t node_count;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t capacity;
  struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node *nodes;
  double *value;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t value_stride;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t root;
};


/* "skgarden/mondrian/tree/_tree.pxd":104
 * # =============================================================================
 * 
 * cdef class TreeBuilder:             # <<<<<<<<<<<<<<
 *     # The TreeBuilder recursively builds a Tree object from training samples,
 *     # using a Splitter object for splitting internal nodes and assigning
 */
struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder {
  PyObject_HEAD
  struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *__pyx_vtab;
  struct __pyx_obj_8skgarden_8mondrian_4tree_9_splitter_Splitter *splitter;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t min_samples_split;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t min_samples_leaf;
  double min_weight_leaf;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t max_depth;
  PyObject *random_state;
};


/* "skgarden/mondrian/tree/_tree.pyx":128
 *         return X, y, sample_weight
 * 
 * cdef class PartialFitTreeBuilder(TreeBuilder):             # <<<<<<<<<<<<<<
 *     """Build a decision tree incrementally."""
 * 
 */
struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder {
  struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder __pyx_base;
};


/* "skgarden/mondrian/tree/_tree.pyx":175
 * # Depth first builder ---------------------------------------------------------
 * 
 * cdef class DepthFirstTreeBuilder(TreeBuilder):             # <<<<<<<<<<<<<<
 *     """Build a decision tree in depth-first fashion."""
 * 
 */
struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder {
  struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder __pyx_base;
};



/* "_criterion.pxd":21
 * ctypedef np.npy_uint32 UINT32_t          # Unsigned 32 bit integer
 * 
 * cdef class Criterion:             # <<<<<<<<<<<<<<
 *     # The criterion computes the impurity of a node and the reduction of
 *     # impurity of a split on that node. It also computes the output statistics
 */

struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_10_criterion_Criterion {
  int (*init)(struct __pyx_obj_8skgarden_8mondrian_4tree_10_criterion_Criterion *, __pyx_t_8skgarden_8mondrian_4tree_10_criterion_DOUBLE_t *, __pyx_t_8skgarden_8mondrian_4tree_10_criterion_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_10_criterion_DOUBLE_t *, double, __pyx_t_8skgarden_8mondrian_4tree_10_criterion_SIZE_t *, __pyx_t_8skgarden_8mondrian_4tree_10_criterion_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_10_criterion_SIZE_t);
  int (*reset)(struct __pyx_obj_8skgarden_8mondrian_4tree_10_criterion_Criterion *);
  int (*reverse_reset)(struct __pyx_obj_8skgarden_8mondrian_4tree_10_criterion_Criterion *);
  int (*update)(struct __pyx_obj_8skgarden_8mondrian_4tree_10_criterion_Criterion *, __pyx_t_8skgarden_8mondrian_4tree_10_criterion_SIZE_t);
  double (*node_impurity)(struct __pyx_obj_8skgarden_8mondrian_4tree_10_criterion_Criterion *);
  void (*children_impurity)(struct __pyx_obj_8skgarden_8mondrian_4tree_10_criterion_Criterion *, double *, double *);
  void (*node_value)(struct __pyx_obj_8skgarden_8mondrian_4tree_10_criterion_Criterion *, double *);
  double (*impurity_improvement)(struct __pyx_obj_8skgarden_8mondrian_4tree_10_criterion_Criterion *, double);
  double (*proxy_impurity_improvement)(struct __pyx_obj_8skgarden_8mondrian_4tree_10_criterion_Criterion *);
  int (*is_pure)(struct __pyx_obj_8skgarden_8mondrian_4tree_10_criterion_Criterion *);
};
static struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_10_criterion_Criterion *__pyx_vtabptr_8skgarden_8mondrian_4tree_10_criterion_Criterion;


/* "_splitter.pxd":35
 *     double E               # Time of split.
 * 
 * cdef class Splitter:             # <<<<<<<<<<<<<<
 *     # The splitter searches in the input space for a feature and a threshold
 *     # to split the samples samples[start:end].
 */

struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_9_splitter_Splitter {
  int (*init)(struct __pyx_obj_8skgarden_8mondrian_4tree_9_splitter_Splitter *, PyObject *, PyArrayObject *, __pyx_t_8skgarden_8mondrian_4tree_9_splitter_DOUBLE_t *, struct __pyx_opt_args_8skgarden_8mondrian_4tree_9_splitter_8Splitter_init *__pyx_optional_args);
  int (*node_reset)(struct __pyx_obj_8skgarden_8mondrian_4tree_9_splitter_Splitter *, __pyx_t_8skgarden_8mondrian_4tree_9_splitter_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_9_splitter_SIZE_t, double *);
  int (*node_split)(struct __pyx_obj_8skgarden_8mondrian_4tree_9_splitter_Splitter *, double, struct __pyx_t_8skgarden_8mondrian_4tree_9_splitter_SplitRecord *, __pyx_t_8skgarden_8mondrian_4tree_9_splitter_SIZE_t *);
  void (*node_value)(struct __pyx_obj_8skgarden_8mondrian_4tree_9_splitter_Splitter *, double *);
  double (*node_impurity)(struct __pyx_obj_8skgarden_8mondrian_4tree_9_splitter_Splitter *);
  void (*set_bounds)(struct __pyx_obj_8skgarden_8mondrian_4tree_9_splitter_Splitter *);
};
static struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_9_splitter_Splitter *__pyx_vtabptr_8skgarden_8mondrian_4tree_9_splitter_Splitter;


/* "_utils.pxd":78
 *     SIZE_t n_constant_features
 * 
 * cdef class Stack:             # <<<<<<<<<<<<<<
 *     cdef SIZE_t capacity
 *     cdef SIZE_t top
 */

struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_6_utils_Stack {
  int (*is_empty)(struct __pyx_obj_8skgarden_8mondrian_4tree_6_utils_Stack *);
  int (*push)(struct __pyx_obj_8skgarden_8mondrian_4tree_6_utils_Stack *, __pyx_t_8skgarden_8mondrian_4tree_6_utils_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_6_utils_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_6_utils_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_6_utils_SIZE_t, int, double, __pyx_t_8skgarden_8mondrian_4tree_6_utils_SIZE_t);
  int (*pop)(struct __pyx_obj_8skgarden_8mondrian_4tree_6_utils_Stack *, struct __pyx_t_8skgarden_8mondrian_4tree_6_utils_StackRecord *);
};
static struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_6_utils_Stack *__pyx_vtabptr_8skgarden_8mondrian_4tree_6_utils_Stack;


/* "skgarden/mondrian/tree/_tree.pyx":315
 * # =============================================================================
 * 
 * cdef class Tree:             # <<<<<<<<<<<<<<
 *     """Array-based representation of a binary decision tree.
 * 
 */

struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree {
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t (*_add_node)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, int, int, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, double, double, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, double, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *, double);
  int (*_resize)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t);
  int (*_resize_c)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_4Tree__resize_c *__pyx_optional_args);
  PyArrayObject *(*_get_value_ndarray)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *);
  PyArrayObject *(*_get_node_ndarray)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *);
  PyObject *(*predict)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_4Tree_predict *__pyx_optional_args);
  PyArrayObject *(*apply)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, PyObject *, int __pyx_skip_dispatch);
  PyArrayObject *(*_apply_dense)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, PyObject *);
  PyObject *(*decision_path)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_decision_path_dense)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, PyObject *);
  PyObject *(*weighted_decision_path)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, PyObject *, int __pyx_skip_dispatch);
  void (*_init)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t *, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t);
  void (*extend)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t *, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_UINT32_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t);
  void (*set_node_attributes)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t *, struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_4Tree_set_node_attributes *__pyx_optional_args);
  void (*update_node_extent)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t);
  void (*_update_node_info)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t *, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t);
};
static struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_vtabptr_8skgarden_8mondrian_4tree_5_tree_Tree;
static PyArrayObject *__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree__apply_dense(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, PyObject *);
static PyObject *__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree__decision_path_dense(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, PyObject *);


/* "skgarden/mondrian/tree/_tree.pyx":101
 * # =============================================================================
 * 
 * cdef class TreeBuilder:             # <<<<<<<<<<<<<<
 *     """Interface for different tree building strategies."""
 * 
 */

struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_TreeBuilder {
  PyObject *(*build)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *, struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, PyObject *, PyArrayObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_build *__pyx_optional_args);
  PyObject *(*_check_input)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *, PyObject *, PyArrayObject *, PyArrayObject *);
};
static struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *__pyx_vtabptr_8skgarden_8mondrian_4tree_5_tree_TreeBuilder;
static PyObject *__pyx_f_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder__check_input(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *, PyObject *, PyArrayObject *, PyArrayObject *);


/* "skgarden/mondrian/tree/_tree.pyx":128
 *         return X, y, sample_weight
 * 
 * cdef class PartialFitTreeBuilder(TreeBuilder):             # <<<<<<<<<<<<<<
 *     """Build a decision tree incrementally."""
 * 
 */

struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder {
  struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_TreeBuilder __pyx_base;
};
static struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder *__pyx_vtabptr_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder;


/* "skgarden/mondrian/tree/_tree.pyx":175
 * # Depth first builder ---------------------------------------------------------
 * 
 * cdef class DepthFirstTreeBuilder(TreeBuilder):             # <<<<<<<<<<<<<<
 *     """Build a decision tree in depth-first fashion."""
 * 
 */

struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder {
  struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_TreeBuilder __pyx_base;
};
static struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder *__pyx_vtabptr_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = PyFrame_Type.tp_basicsize - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* IsLittleEndian.proto */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void);

/* BufferFormatCheck.proto */
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts);
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type);

/* BufferGetAndValidate.proto */
#define __Pyx_GetBufferAndValidate(buf, obj, dtype, flags, nd, cast, stack)\
    ((obj == Py_None || obj == NULL) ?\
    (__Pyx_ZeroBuffer(buf), 0) :\
    __Pyx__GetBufferAndValidate(buf, obj, dtype, flags, nd, cast, stack))
static int  __Pyx__GetBufferAndValidate(Py_buffer* buf, PyObject* obj,
    __Pyx_TypeInfo* dtype, int flags, int nd, int cast, __Pyx_BufFmt_StackElem* stack);
static void __Pyx_ZeroBuffer(Py_buffer* buf);
static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info);
static Py_ssize_t __Pyx_minusones[] = { -1, -1, -1, -1, -1, -1, -1, -1 };
static Py_ssize_t __Pyx_zeros[] = { 0, 0, 0, 0, 0, 0, 0, 0 };

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

#define __Pyx_BufPtrStrided1d(type, buf, i0, s0) (type)((char*)buf + i0 * s0)
/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_NeObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

#define __Pyx_BufPtrStrided2d(type, buf, i0, s0, i1, s1) (type)((char*)buf + i0 * s0 + i1 * s1)
/* BufferFallbackError.proto */
static void __Pyx_RaiseBufferFallbackError(void);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_Py_intptr_t(Py_intptr_t value);

/* BufferStructDeclare.proto */
typedef struct {
  Py_ssize_t shape, strides, suboffsets;
} __Pyx_Buf_DimInfo;
typedef struct {
  size_t refcount;
  Py_buffer pybuffer;
} __Pyx_Buffer;
typedef struct {
  __Pyx_Buffer *rcbuffer;
  char *data;
  __Pyx_Buf_DimInfo diminfo[8];
} __Pyx_LocalBuf_ND;

#if PY_MAJOR_VERSION < 3
    static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags);
    static void __Pyx_ReleaseBuffer(Py_buffer *view);
#else
    #define __Pyx_GetBuffer PyObject_GetBuffer
    #define __Pyx_ReleaseBuffer PyBuffer_Release
#endif


/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* None.proto */
static CYTHON_INLINE long __Pyx_pow_long(long, long);

/* None.proto */
static CYTHON_INLINE int __Pyx_ErrOccurredWithGIL(void); /* proto */

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__NPY_TYPES(enum NPY_TYPES value);

/* CIntFromPy.proto */
static CYTHON_INLINE Py_intptr_t __Pyx_PyInt_As_Py_intptr_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE npy_uint32 __Pyx_PyInt_As_npy_uint32(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static PyObject *__pyx_f_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_build(CYTHON_UNUSED struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_tree, CYTHON_UNUSED PyObject *__pyx_v_X, CYTHON_UNUSED PyArrayObject *__pyx_v_y, int __pyx_skip_dispatch, struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_build *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder__check_input(CYTHON_UNUSED struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *__pyx_v_self, PyObject *__pyx_v_X, PyArrayObject *__pyx_v_y, PyArrayObject *__pyx_v_sample_weight); /* proto*/
static PyObject *__pyx_f_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder_build(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder *__pyx_v_self, struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_tree, PyObject *__pyx_v_X, PyArrayObject *__pyx_v_y, int __pyx_skip_dispatch, struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder_build *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder_build(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder *__pyx_v_self, struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_tree, PyObject *__pyx_v_X, PyArrayObject *__pyx_v_y, int __pyx_skip_dispatch, struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder_build *__pyx_optional_args); /* proto*/
static int __pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree__resize(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_capacity); /* proto*/
static int __pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree__resize_c(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_4Tree__resize_c *__pyx_optional_args); /* proto*/
static void __pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree_update_node_extent(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_node_ind, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_child_ind, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *__pyx_v_X_ptr, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_X_start, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_X_f_stride); /* proto*/
static void __pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree__update_node_info(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_parent_id, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_child_id, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t *__pyx_v_y_ptr, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_y_start); /* proto*/
static void __pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree_set_node_attributes(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_node_ind, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_left_child, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_right_child, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_feature, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t __pyx_v_threshold, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t __pyx_v_tau, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_n_node_samples, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t __pyx_v_weighted_n_node_samples, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t __pyx_v_impurity, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t __pyx_v_variance, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_X_start, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_X_f_stride, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *__pyx_v_X_ptr, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t *__pyx_v_y_ptr, struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_4Tree_set_node_attributes *__pyx_optional_args); /* proto*/
static void __pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree__init(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *__pyx_v_X_ptr, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t *__pyx_v_y_ptr, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_X_f_stride); /* proto*/
static void __pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree_extend(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *__pyx_v_X_ptr, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t *__pyx_v_y_ptr, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_X_start, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_X_f_stride, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_y_start, __pyx_t_8skgarden_8mondrian_4tree_5_tree_UINT32_t __pyx_v_random_state, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_min_samples_split); /* proto*/
static __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree__add_node(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_parent, int __pyx_v_is_left, int __pyx_v_is_leaf, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_feature, double __pyx_v_threshold, double __pyx_v_impurity, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_n_node_samples, double __pyx_v_weighted_n_node_samples, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *__pyx_v_lower_bounds, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *__pyx_v_upper_bounds, double __pyx_v_E); /* proto*/
static PyArrayObject *__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree_apply(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_X, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree_predict(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_X, int __pyx_skip_dispatch, struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_4Tree_predict *__pyx_optional_args); /* proto*/
static PyArrayObject *__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree__apply_dense(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_X); /* proto*/
static PyObject *__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree_decision_path(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_X, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree_weighted_decision_path(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_X, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree__decision_path_dense(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_X); /* proto*/
static PyArrayObject *__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree__get_value_ndarray(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self); /* proto*/
static PyArrayObject *__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree__get_node_ndarray(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self); /* proto*/

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython.version' */

/* Module declarations from 'cpython.exc' */

/* Module declarations from 'cpython.module' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'cpython.tuple' */

/* Module declarations from 'cpython.list' */

/* Module declarations from 'cpython.sequence' */

/* Module declarations from 'cpython.mapping' */

/* Module declarations from 'cpython.iterator' */

/* Module declarations from 'cpython.number' */

/* Module declarations from 'cpython.int' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.bool' */
static PyTypeObject *__pyx_ptype_7cpython_4bool_bool = 0;

/* Module declarations from 'cpython.long' */

/* Module declarations from 'cpython.float' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.complex' */
static PyTypeObject *__pyx_ptype_7cpython_7complex_complex = 0;

/* Module declarations from 'cpython.string' */

/* Module declarations from 'cpython.unicode' */

/* Module declarations from 'cpython.dict' */

/* Module declarations from 'cpython.instance' */

/* Module declarations from 'cpython.function' */

/* Module declarations from 'cpython.method' */

/* Module declarations from 'cpython.weakref' */

/* Module declarations from 'cpython.getargs' */

/* Module declarations from 'cpython.pythread' */

/* Module declarations from 'cpython.pystate' */

/* Module declarations from 'cpython.cobject' */

/* Module declarations from 'cpython.oldbuffer' */

/* Module declarations from 'cpython.set' */

/* Module declarations from 'cpython.bytes' */

/* Module declarations from 'cpython.pycapsule' */

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;
static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *, char *, char *, int *); /*proto*/
static CYTHON_INLINE int __pyx_f_5numpy_import_array(void); /*proto*/

/* Module declarations from 'skgarden.mondrian.tree._criterion' */
static PyTypeObject *__pyx_ptype_8skgarden_8mondrian_4tree_10_criterion_Criterion = 0;

/* Module declarations from 'skgarden.mondrian.tree._splitter' */
static PyTypeObject *__pyx_ptype_8skgarden_8mondrian_4tree_9_splitter_Splitter = 0;

/* Module declarations from 'libc.math' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'skgarden.mondrian.tree._utils' */
static PyTypeObject *__pyx_ptype_8skgarden_8mondrian_4tree_6_utils_Stack = 0;
static PyArrayObject *(*__pyx_f_8skgarden_8mondrian_4tree_6_utils_sizet_ptr_to_ndarray)(__pyx_t_8skgarden_8mondrian_4tree_6_utils_SIZE_t *, __pyx_t_8skgarden_8mondrian_4tree_6_utils_SIZE_t); /*proto*/
static double (*__pyx_f_8skgarden_8mondrian_4tree_6_utils_rand_uniform)(double, double, __pyx_t_8skgarden_8mondrian_4tree_6_utils_UINT32_t *); /*proto*/
static __pyx_t_8skgarden_8mondrian_4tree_6_utils_SIZE_t (*__pyx_f_8skgarden_8mondrian_4tree_6_utils_rand_multinomial)(__pyx_t_8skgarden_8mondrian_4tree_6_utils_DTYPE_t *, __pyx_t_8skgarden_8mondrian_4tree_6_utils_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_6_utils_UINT32_t *); /*proto*/
static double (*__pyx_f_8skgarden_8mondrian_4tree_6_utils_rand_exponential)(__pyx_t_8skgarden_8mondrian_4tree_6_utils_DTYPE_t, __pyx_t_8skgarden_8mondrian_4tree_6_utils_UINT32_t *); /*proto*/
static __pyx_t_8skgarden_8mondrian_4tree_6_utils_DTYPE_t *(*__pyx_fuse_0__pyx_f_8skgarden_8mondrian_4tree_6_utils_safe_realloc)(__pyx_t_8skgarden_8mondrian_4tree_6_utils_DTYPE_t **, size_t); /*proto*/
static __pyx_t_8skgarden_8mondrian_4tree_6_utils_SIZE_t *(*__pyx_fuse_1__pyx_f_8skgarden_8mondrian_4tree_6_utils_safe_realloc)(__pyx_t_8skgarden_8mondrian_4tree_6_utils_SIZE_t **, size_t); /*proto*/
static unsigned char *(*__pyx_fuse_2__pyx_f_8skgarden_8mondrian_4tree_6_utils_safe_realloc)(unsigned char **, size_t); /*proto*/
static __pyx_t_8skgarden_8mondrian_4tree_6_utils_DOUBLE_t *(*__pyx_fuse_3__pyx_f_8skgarden_8mondrian_4tree_6_utils_safe_realloc)(__pyx_t_8skgarden_8mondrian_4tree_6_utils_DOUBLE_t **, size_t); /*proto*/
static __pyx_t_8skgarden_8mondrian_4tree_6_utils_DOUBLE_t **(*__pyx_fuse_4__pyx_f_8skgarden_8mondrian_4tree_6_utils_safe_realloc)(__pyx_t_8skgarden_8mondrian_4tree_6_utils_DOUBLE_t ***, size_t); /*proto*/
static struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node *(*__pyx_fuse_5__pyx_f_8skgarden_8mondrian_4tree_6_utils_safe_realloc)(struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node **, size_t); /*proto*/
static struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node **(*__pyx_fuse_6__pyx_f_8skgarden_8mondrian_4tree_6_utils_safe_realloc)(struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node ***, size_t); /*proto*/
static struct __pyx_t_8skgarden_8mondrian_4tree_6_utils_StackRecord *(*__pyx_fuse_7__pyx_f_8skgarden_8mondrian_4tree_6_utils_safe_realloc)(struct __pyx_t_8skgarden_8mondrian_4tree_6_utils_StackRecord **, size_t); /*proto*/

/* Module declarations from 'skgarden.mondrian.tree._tree' */
static PyTypeObject *__pyx_ptype_8skgarden_8mondrian_4tree_5_tree_Tree = 0;
static PyTypeObject *__pyx_ptype_8skgarden_8mondrian_4tree_5_tree_TreeBuilder = 0;
static PyTypeObject *__pyx_ptype_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder = 0;
static PyTypeObject *__pyx_ptype_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder = 0;
static double __pyx_v_8skgarden_8mondrian_4tree_5_tree_INFINITY;
static int __pyx_v_8skgarden_8mondrian_4tree_5_tree_IS_FIRST;
static int __pyx_v_8skgarden_8mondrian_4tree_5_tree_IS_NOT_FIRST;
static int __pyx_v_8skgarden_8mondrian_4tree_5_tree_IS_LEFT;
static int __pyx_v_8skgarden_8mondrian_4tree_5_tree_IS_NOT_LEFT;
static __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_8skgarden_8mondrian_4tree_5_tree__TREE_LEAF;
static __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_8skgarden_8mondrian_4tree_5_tree__TREE_UNDEFINED;
static __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_8skgarden_8mondrian_4tree_5_tree_INITIAL_STACK_SIZE;
static CYTHON_INLINE double __pyx_f_8skgarden_8mondrian_4tree_5_tree_fmax(double, double); /*proto*/
static CYTHON_INLINE double __pyx_f_8skgarden_8mondrian_4tree_5_tree_fmin(double, double); /*proto*/
static PyObject *__pyx_f_8skgarden_8mondrian_4tree_5_tree___pyx_unpickle_TreeBuilder__set_state(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *, PyObject *); /*proto*/
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t = { "SIZE_t", NULL, sizeof(__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t), { 0 }, 0, IS_UNSIGNED(__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t) ? 'U' : 'I', IS_UNSIGNED(__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t), 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t = { "DOUBLE_t", NULL, sizeof(__pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t), { 0 }, 0, 'R', 0, 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t = { "DTYPE_t", NULL, sizeof(__pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t), { 0 }, 0, 'R', 0, 0 };
#define __Pyx_MODULE_NAME "skgarden.mondrian.tree._tree"
extern int __pyx_module_is_main_skgarden__mondrian__tree___tree;
int __pyx_module_is_main_skgarden__mondrian__tree___tree = 0;

/* Implementation of 'skgarden.mondrian.tree._tree' */
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_ImportError;
static const char __pyx_k_C[] = "C";
static const char __pyx_k_X[] = "X";
static const char __pyx_k_y[] = "y";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_inf[] = "inf";
static const char __pyx_k_max[] = "max";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_tau[] = "tau";
static const char __pyx_k_Tree[] = "Tree";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_intp[] = "intp";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_ndim[] = "ndim";
static const char __pyx_k_ones[] = "ones";
static const char __pyx_k_root[] = "root";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_tree[] = "tree";
static const char __pyx_k_DTYPE[] = "DTYPE";
static const char __pyx_k_apply[] = "apply";
static const char __pyx_k_build[] = "build";
static const char __pyx_k_dtype[] = "dtype";
static const char __pyx_k_flags[] = "flags";
static const char __pyx_k_names[] = "names";
static const char __pyx_k_nodes[] = "nodes";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_order[] = "order";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_ravel[] = "ravel";
static const char __pyx_k_shape[] = "shape";
static const char __pyx_k_zeros[] = "zeros";
static const char __pyx_k_DOUBLE[] = "DOUBLE";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_asarray[] = "asarray";
static const char __pyx_k_feature[] = "feature";
static const char __pyx_k_float32[] = "float32";
static const char __pyx_k_float64[] = "float64";
static const char __pyx_k_formats[] = "formats";
static const char __pyx_k_offsets[] = "offsets";
static const char __pyx_k_predict[] = "predict";
static const char __pyx_k_randint[] = "randint";
static const char __pyx_k_strides[] = "strides";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_impurity[] = "impurity";
static const char __pyx_k_itemsize[] = "itemsize";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_splitter[] = "splitter";
static const char __pyx_k_variance[] = "variance";
static const char __pyx_k_TREE_LEAF[] = "TREE_LEAF";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_max_depth[] = "max_depth";
static const char __pyx_k_n_classes[] = "n_classes";
static const char __pyx_k_n_outputs[] = "n_outputs";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_threshold[] = "threshold";
static const char __pyx_k_NODE_DTYPE[] = "NODE_DTYPE";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_contiguous[] = "contiguous";
static const char __pyx_k_csr_matrix[] = "csr_matrix";
static const char __pyx_k_left_child[] = "left_child";
static const char __pyx_k_n_features[] = "n_features";
static const char __pyx_k_node_count[] = "node_count";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_return_std[] = "return_std";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_TreeBuilder[] = "TreeBuilder";
static const char __pyx_k_right_child[] = "right_child";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_X_idx_sorted[] = "X_idx_sorted";
static const char __pyx_k_c_contiguous[] = "c_contiguous";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_random_state[] = "random_state";
static const char __pyx_k_scipy_sparse[] = "scipy.sparse";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_decision_path[] = "decision_path";
static const char __pyx_k_is_regression[] = "is_regression";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_sample_weight[] = "sample_weight";
static const char __pyx_k_TREE_UNDEFINED[] = "TREE_UNDEFINED";
static const char __pyx_k_asfortranarray[] = "asfortranarray";
static const char __pyx_k_n_node_samples[] = "n_node_samples";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_ascontiguousarray[] = "ascontiguousarray";
static const char __pyx_k_min_samples_split[] = "min_samples_split";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_resizing_tree_to_d[] = "resizing tree to %d";
static const char __pyx_k_DepthFirstTreeBuilder[] = "DepthFirstTreeBuilder";
static const char __pyx_k_PartialFitTreeBuilder[] = "PartialFitTreeBuilder";
static const char __pyx_k_weighted_decision_path[] = "weighted_decision_path";
static const char __pyx_k_weighted_n_node_samples[] = "weighted_n_node_samples";
static const char __pyx_k_pyx_unpickle_TreeBuilder[] = "__pyx_unpickle_TreeBuilder";
static const char __pyx_k_ndarray_is_not_C_contiguous[] = "ndarray is not C contiguous";
static const char __pyx_k_skgarden_mondrian_tree__tree[] = "skgarden.mondrian.tree._tree";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_unknown_dtype_code_in_numpy_pxd[] = "unknown dtype code in numpy.pxd (%d)";
static const char __pyx_k_Did_not_recognise_loaded_array_l[] = "Did not recognise loaded array layout";
static const char __pyx_k_Format_string_allocated_too_shor[] = "Format string allocated too short, see comment in numpy.pxd";
static const char __pyx_k_Incompatible_checksums_s_vs_0xed[] = "Incompatible checksums (%s vs 0xedae8f4 = (max_depth, min_samples_leaf, min_samples_split, min_weight_leaf, random_state, splitter))";
static const char __pyx_k_Non_native_byte_order_not_suppor[] = "Non-native byte order not supported";
static const char __pyx_k_X_dtype_should_be_np_float32_got[] = "X.dtype should be np.float32, got %s";
static const char __pyx_k_X_should_be_in_np_ndarray_format[] = "X should be in np.ndarray format, got %s";
static const char __pyx_k_You_have_loaded_Tree_version_whi[] = "You have loaded Tree version which cannot be imported";
static const char __pyx_k_ndarray_is_not_Fortran_contiguou[] = "ndarray is not Fortran contiguous";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static const char __pyx_k_Format_string_allocated_too_shor_2[] = "Format string allocated too short.";
static PyObject *__pyx_n_s_C;
static PyObject *__pyx_n_s_DOUBLE;
static PyObject *__pyx_n_s_DTYPE;
static PyObject *__pyx_n_s_DepthFirstTreeBuilder;
static PyObject *__pyx_kp_s_Did_not_recognise_loaded_array_l;
static PyObject *__pyx_kp_u_Format_string_allocated_too_shor;
static PyObject *__pyx_kp_u_Format_string_allocated_too_shor_2;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0xed;
static PyObject *__pyx_n_s_MemoryError;
static PyObject *__pyx_n_s_NODE_DTYPE;
static PyObject *__pyx_kp_u_Non_native_byte_order_not_suppor;
static PyObject *__pyx_n_s_PartialFitTreeBuilder;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_n_s_TREE_LEAF;
static PyObject *__pyx_n_s_TREE_UNDEFINED;
static PyObject *__pyx_n_s_Tree;
static PyObject *__pyx_n_s_TreeBuilder;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_X;
static PyObject *__pyx_kp_s_X_dtype_should_be_np_float32_got;
static PyObject *__pyx_n_s_X_idx_sorted;
static PyObject *__pyx_kp_s_X_should_be_in_np_ndarray_format;
static PyObject *__pyx_kp_s_You_have_loaded_Tree_version_whi;
static PyObject *__pyx_n_s_apply;
static PyObject *__pyx_n_s_asarray;
static PyObject *__pyx_n_s_ascontiguousarray;
static PyObject *__pyx_n_s_asfortranarray;
static PyObject *__pyx_n_s_build;
static PyObject *__pyx_n_s_c_contiguous;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_contiguous;
static PyObject *__pyx_n_s_csr_matrix;
static PyObject *__pyx_n_s_decision_path;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_dtype;
static PyObject *__pyx_n_s_feature;
static PyObject *__pyx_n_s_flags;
static PyObject *__pyx_n_s_float32;
static PyObject *__pyx_n_s_float64;
static PyObject *__pyx_n_s_formats;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_impurity;
static PyObject *__pyx_n_s_inf;
static PyObject *__pyx_n_s_intp;
static PyObject *__pyx_n_s_is_regression;
static PyObject *__pyx_n_s_itemsize;
static PyObject *__pyx_n_s_left_child;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_max;
static PyObject *__pyx_n_s_max_depth;
static PyObject *__pyx_n_s_min_samples_split;
static PyObject *__pyx_n_s_n_classes;
static PyObject *__pyx_n_s_n_features;
static PyObject *__pyx_n_s_n_node_samples;
static PyObject *__pyx_n_s_n_outputs;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_names;
static PyObject *__pyx_kp_u_ndarray_is_not_C_contiguous;
static PyObject *__pyx_kp_u_ndarray_is_not_Fortran_contiguou;
static PyObject *__pyx_n_s_ndim;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_n_s_node_count;
static PyObject *__pyx_n_s_nodes;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_kp_s_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_s_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_offsets;
static PyObject *__pyx_n_s_ones;
static PyObject *__pyx_n_s_order;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_predict;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_TreeBuilder;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_randint;
static PyObject *__pyx_n_s_random_state;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_ravel;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_kp_s_resizing_tree_to_d;
static PyObject *__pyx_n_s_return_std;
static PyObject *__pyx_n_s_right_child;
static PyObject *__pyx_n_s_root;
static PyObject *__pyx_n_s_sample_weight;
static PyObject *__pyx_n_s_scipy_sparse;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_shape;
static PyObject *__pyx_n_s_skgarden_mondrian_tree__tree;
static PyObject *__pyx_n_s_splitter;
static PyObject *__pyx_n_s_strides;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_tau;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_threshold;
static PyObject *__pyx_n_s_tree;
static PyObject *__pyx_kp_u_unknown_dtype_code_in_numpy_pxd;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_values;
static PyObject *__pyx_n_s_variance;
static PyObject *__pyx_n_s_weighted_decision_path;
static PyObject *__pyx_n_s_weighted_n_node_samples;
static PyObject *__pyx_n_s_y;
static PyObject *__pyx_n_s_zeros;
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_build(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *__pyx_v_self, struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_tree, PyObject *__pyx_v_X, PyArrayObject *__pyx_v_y, PyArrayObject *__pyx_v_sample_weight, PyArrayObject *__pyx_v_X_idx_sorted); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_2__reduce_cython__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_4__setstate_cython__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder___cinit__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder *__pyx_v_self, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_min_samples_split, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_max_depth, PyObject *__pyx_v_random_state); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder_2build(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder *__pyx_v_self, struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_tree, PyObject *__pyx_v_X, PyArrayObject *__pyx_v_y, PyArrayObject *__pyx_v_sample_weight, PyArrayObject *__pyx_v_X_idx_sorted); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder___cinit__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder *__pyx_v_self, struct __pyx_obj_8skgarden_8mondrian_4tree_9_splitter_Splitter *__pyx_v_splitter, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_min_samples_split, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_max_depth); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder_2build(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder *__pyx_v_self, struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_tree, PyObject *__pyx_v_X, PyArrayObject *__pyx_v_y, PyArrayObject *__pyx_v_sample_weight, PyArrayObject *__pyx_v_X_idx_sorted); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_9n_classes___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_13children_left___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_14children_right___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_7feature___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_9threshold___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_8impurity___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_14n_node_samples___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_23weighted_n_node_samples___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_5value___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_3tau___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_4mean___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_8variance___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_4root___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static int __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree___cinit__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, int __pyx_v_n_features, PyArrayObject *__pyx_v_n_classes, int __pyx_v_n_outputs); /* proto */
static void __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_2__dealloc__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_4__reduce__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_6__getstate__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_8__setstate__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_d); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_10apply(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_X); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_12predict(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_X, int __pyx_v_return_std, int __pyx_v_is_regression); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_14decision_path(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_X); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_16weighted_decision_path(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_X); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_10n_features___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static int __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_10n_features_2__set__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_9n_outputs___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static int __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_9n_outputs_2__set__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_13max_n_classes___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static int __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_13max_n_classes_2__set__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_9max_depth___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static int __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_9max_depth_2__set__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_10node_count___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static int __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_10node_count_2__set__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_8capacity___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self); /* proto */
static int __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_8capacity_2__set__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree___pyx_unpickle_TreeBuilder(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info); /* proto */
static PyObject *__pyx_tp_new_8skgarden_8mondrian_4tree_5_tree_Tree(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8skgarden_8mondrian_4tree_5_tree_TreeBuilder(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_249227508;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_int_neg_2;
static __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_k__7;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_slice__8;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_codeobj__18;
/* Late includes */

/* "skgarden/mondrian/tree/_tree.pyx":89
 * })
 * 
 * cdef inline double fmax(double left, double right) nogil:             # <<<<<<<<<<<<<<
 *     return left if left > right else right
 * 
 */

static CYTHON_INLINE double __pyx_f_8skgarden_8mondrian_4tree_5_tree_fmax(double __pyx_v_left, double __pyx_v_right) {
  double __pyx_r;
  double __pyx_t_1;

  /* "skgarden/mondrian/tree/_tree.pyx":90
 * 
 * cdef inline double fmax(double left, double right) nogil:
 *     return left if left > right else right             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (((__pyx_v_left > __pyx_v_right) != 0)) {
    __pyx_t_1 = __pyx_v_left;
  } else {
    __pyx_t_1 = __pyx_v_right;
  }
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "skgarden/mondrian/tree/_tree.pyx":89
 * })
 * 
 * cdef inline double fmax(double left, double right) nogil:             # <<<<<<<<<<<<<<
 *     return left if left > right else right
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":93
 * 
 * 
 * cdef inline double fmin(double left, double right) nogil:             # <<<<<<<<<<<<<<
 *     return left if left < right else right
 * 
 */

static CYTHON_INLINE double __pyx_f_8skgarden_8mondrian_4tree_5_tree_fmin(double __pyx_v_left, double __pyx_v_right) {
  double __pyx_r;
  double __pyx_t_1;

  /* "skgarden/mondrian/tree/_tree.pyx":94
 * 
 * cdef inline double fmin(double left, double right) nogil:
 *     return left if left < right else right             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (((__pyx_v_left < __pyx_v_right) != 0)) {
    __pyx_t_1 = __pyx_v_left;
  } else {
    __pyx_t_1 = __pyx_v_right;
  }
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "skgarden/mondrian/tree/_tree.pyx":93
 * 
 * 
 * cdef inline double fmin(double left, double right) nogil:             # <<<<<<<<<<<<<<
 *     return left if left < right else right
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":104
 *     """Interface for different tree building strategies."""
 * 
 *     cpdef build(self, Tree tree, object X, np.ndarray y,             # <<<<<<<<<<<<<<
 *                 np.ndarray sample_weight=None,
 *                 np.ndarray X_idx_sorted=None):
 */

static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_1build(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_build(CYTHON_UNUSED struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_tree, CYTHON_UNUSED PyObject *__pyx_v_X, CYTHON_UNUSED PyArrayObject *__pyx_v_y, int __pyx_skip_dispatch, struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_build *__pyx_optional_args) {

  /* "skgarden/mondrian/tree/_tree.pyx":105
 * 
 *     cpdef build(self, Tree tree, object X, np.ndarray y,
 *                 np.ndarray sample_weight=None,             # <<<<<<<<<<<<<<
 *                 np.ndarray X_idx_sorted=None):
 *         """Build a decision tree from the training set (X, y)."""
 */
  PyArrayObject *__pyx_v_sample_weight = ((PyArrayObject *)Py_None);

  /* "skgarden/mondrian/tree/_tree.pyx":106
 *     cpdef build(self, Tree tree, object X, np.ndarray y,
 *                 np.ndarray sample_weight=None,
 *                 np.ndarray X_idx_sorted=None):             # <<<<<<<<<<<<<<
 *         """Build a decision tree from the training set (X, y)."""
 *         pass
 */
  PyArrayObject *__pyx_v_X_idx_sorted = ((PyArrayObject *)Py_None);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("build", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_sample_weight = __pyx_optional_args->sample_weight;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_X_idx_sorted = __pyx_optional_args->X_idx_sorted;
      }
    }
  }

  /* "skgarden/mondrian/tree/_tree.pyx":104
 *     """Interface for different tree building strategies."""
 * 
 *     cpdef build(self, Tree tree, object X, np.ndarray y,             # <<<<<<<<<<<<<<
 *                 np.ndarray sample_weight=None,
 *                 np.ndarray X_idx_sorted=None):
 */
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    static PY_UINT64_T tp_dict_version = 0, obj_dict_version = 0;
    if (likely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict && tp_dict_version == __PYX_GET_DICT_VERSION(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict) && (!Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset || obj_dict_version == __PYX_GET_DICT_VERSION(_PyObject_GetDictPtr(((PyObject *)__pyx_v_self))))));
    else {
      PY_UINT64_T type_dict_guard = (likely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict)) ? __PYX_GET_DICT_VERSION(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict) : 0;
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_build); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 104, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_1build)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[6] = {__pyx_t_4, ((PyObject *)__pyx_v_tree), __pyx_v_X, ((PyObject *)__pyx_v_y), ((PyObject *)__pyx_v_sample_weight), ((PyObject *)__pyx_v_X_idx_sorted)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 5+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 104, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[6] = {__pyx_t_4, ((PyObject *)__pyx_v_tree), __pyx_v_X, ((PyObject *)__pyx_v_y), ((PyObject *)__pyx_v_sample_weight), ((PyObject *)__pyx_v_X_idx_sorted)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 5+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 104, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(5+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 104, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_tree));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_tree));
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_tree));
          __Pyx_INCREF(__pyx_v_X);
          __Pyx_GIVEREF(__pyx_v_X);
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_X);
          __Pyx_INCREF(((PyObject *)__pyx_v_y));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_y));
          PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_5, ((PyObject *)__pyx_v_y));
          __Pyx_INCREF(((PyObject *)__pyx_v_sample_weight));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_sample_weight));
          PyTuple_SET_ITEM(__pyx_t_6, 3+__pyx_t_5, ((PyObject *)__pyx_v_sample_weight));
          __Pyx_INCREF(((PyObject *)__pyx_v_X_idx_sorted));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_X_idx_sorted));
          PyTuple_SET_ITEM(__pyx_t_6, 4+__pyx_t_5, ((PyObject *)__pyx_v_X_idx_sorted));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 104, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
      tp_dict_version = likely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict) ? __PYX_GET_DICT_VERSION(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict) : 0;
      obj_dict_version = likely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset) ? __PYX_GET_DICT_VERSION(_PyObject_GetDictPtr(((PyObject *)__pyx_v_self))) : 0;
      if (unlikely(type_dict_guard != tp_dict_version)) {
        tp_dict_version = obj_dict_version = 0;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    }
    #endif
  }

  /* "skgarden/mondrian/tree/_tree.pyx":107
 *                 np.ndarray sample_weight=None,
 *                 np.ndarray X_idx_sorted=None):
 *         """Build a decision tree from the training set (X, y)."""             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.TreeBuilder.build", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":104
 *     """Interface for different tree building strategies."""
 * 
 *     cpdef build(self, Tree tree, object X, np.ndarray y,             # <<<<<<<<<<<<<<
 *                 np.ndarray sample_weight=None,
 *                 np.ndarray X_idx_sorted=None):
 */

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_1build(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_build[] = "Build a decision tree from the training set (X, y).";
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_1build(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_tree = 0;
  PyObject *__pyx_v_X = 0;
  PyArrayObject *__pyx_v_y = 0;
  PyArrayObject *__pyx_v_sample_weight = 0;
  PyArrayObject *__pyx_v_X_idx_sorted = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("build (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_tree,&__pyx_n_s_X,&__pyx_n_s_y,&__pyx_n_s_sample_weight,&__pyx_n_s_X_idx_sorted,0};
    PyObject* values[5] = {0,0,0,0,0};

    /* "skgarden/mondrian/tree/_tree.pyx":105
 * 
 *     cpdef build(self, Tree tree, object X, np.ndarray y,
 *                 np.ndarray sample_weight=None,             # <<<<<<<<<<<<<<
 *                 np.ndarray X_idx_sorted=None):
 *         """Build a decision tree from the training set (X, y)."""
 */
    values[3] = (PyObject *)((PyArrayObject *)Py_None);

    /* "skgarden/mondrian/tree/_tree.pyx":106
 *     cpdef build(self, Tree tree, object X, np.ndarray y,
 *                 np.ndarray sample_weight=None,
 *                 np.ndarray X_idx_sorted=None):             # <<<<<<<<<<<<<<
 *         """Build a decision tree from the training set (X, y)."""
 *         pass
 */
    values[4] = (PyObject *)((PyArrayObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tree)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("build", 0, 3, 5, 1); __PYX_ERR(0, 104, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("build", 0, 3, 5, 2); __PYX_ERR(0, 104, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sample_weight);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X_idx_sorted);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "build") < 0)) __PYX_ERR(0, 104, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_tree = ((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)values[0]);
    __pyx_v_X = values[1];
    __pyx_v_y = ((PyArrayObject *)values[2]);
    __pyx_v_sample_weight = ((PyArrayObject *)values[3]);
    __pyx_v_X_idx_sorted = ((PyArrayObject *)values[4]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("build", 0, 3, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 104, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.TreeBuilder.build", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_tree), __pyx_ptype_8skgarden_8mondrian_4tree_5_tree_Tree, 1, "tree", 0))) __PYX_ERR(0, 104, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_y), __pyx_ptype_5numpy_ndarray, 1, "y", 0))) __PYX_ERR(0, 104, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_sample_weight), __pyx_ptype_5numpy_ndarray, 1, "sample_weight", 0))) __PYX_ERR(0, 105, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_X_idx_sorted), __pyx_ptype_5numpy_ndarray, 1, "X_idx_sorted", 0))) __PYX_ERR(0, 106, __pyx_L1_error)
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_build(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *)__pyx_v_self), __pyx_v_tree, __pyx_v_X, __pyx_v_y, __pyx_v_sample_weight, __pyx_v_X_idx_sorted);

  /* "skgarden/mondrian/tree/_tree.pyx":104
 *     """Interface for different tree building strategies."""
 * 
 *     cpdef build(self, Tree tree, object X, np.ndarray y,             # <<<<<<<<<<<<<<
 *                 np.ndarray sample_weight=None,
 *                 np.ndarray X_idx_sorted=None):
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_build(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *__pyx_v_self, struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_tree, PyObject *__pyx_v_X, PyArrayObject *__pyx_v_y, PyArrayObject *__pyx_v_sample_weight, PyArrayObject *__pyx_v_X_idx_sorted) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_build __pyx_t_2;
  __Pyx_RefNannySetupContext("build", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 2;
  __pyx_t_2.sample_weight = __pyx_v_sample_weight;
  __pyx_t_2.X_idx_sorted = __pyx_v_X_idx_sorted;
  __pyx_t_1 = __pyx_vtabptr_8skgarden_8mondrian_4tree_5_tree_TreeBuilder->build(__pyx_v_self, __pyx_v_tree, __pyx_v_X, __pyx_v_y, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.TreeBuilder.build", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":110
 *         pass
 * 
 *     cdef inline _check_input(self, object X, np.ndarray y,             # <<<<<<<<<<<<<<
 *                              np.ndarray sample_weight):
 *         """Check input dtype, layout and format"""
 */

static PyObject *__pyx_f_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder__check_input(CYTHON_UNUSED struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *__pyx_v_self, PyObject *__pyx_v_X, PyArrayObject *__pyx_v_y, PyArrayObject *__pyx_v_sample_weight) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("_check_input", 0);
  __Pyx_INCREF(__pyx_v_X);
  __Pyx_INCREF((PyObject *)__pyx_v_y);
  __Pyx_INCREF((PyObject *)__pyx_v_sample_weight);

  /* "skgarden/mondrian/tree/_tree.pyx":113
 *                              np.ndarray sample_weight):
 *         """Check input dtype, layout and format"""
 *         if X.dtype != DTYPE:             # <<<<<<<<<<<<<<
 *             # since we have to copy we will make it fortran for efficiency
 *             X = np.asfortranarray(X, dtype=DTYPE)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_dtype); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_DTYPE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 113, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 113, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "skgarden/mondrian/tree/_tree.pyx":115
 *         if X.dtype != DTYPE:
 *             # since we have to copy we will make it fortran for efficiency
 *             X = np.asfortranarray(X, dtype=DTYPE)             # <<<<<<<<<<<<<<
 * 
 *         if y.dtype != DOUBLE or not y.flags.contiguous:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_asfortranarray); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_X);
    __Pyx_GIVEREF(__pyx_v_X);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_X);
    __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_DTYPE); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(0, 115, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_X, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "skgarden/mondrian/tree/_tree.pyx":113
 *                              np.ndarray sample_weight):
 *         """Check input dtype, layout and format"""
 *         if X.dtype != DTYPE:             # <<<<<<<<<<<<<<
 *             # since we have to copy we will make it fortran for efficiency
 *             X = np.asfortranarray(X, dtype=DTYPE)
 */
  }

  /* "skgarden/mondrian/tree/_tree.pyx":117
 *             X = np.asfortranarray(X, dtype=DTYPE)
 * 
 *         if y.dtype != DOUBLE or not y.flags.contiguous:             # <<<<<<<<<<<<<<
 *             y = np.ascontiguousarray(y, dtype=DOUBLE)
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_y), __pyx_n_s_dtype); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_DOUBLE); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_5, __pyx_t_1, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_y), __pyx_n_s_flags); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_contiguous); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = ((!__pyx_t_6) != 0);
  __pyx_t_4 = __pyx_t_7;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_4) {

    /* "skgarden/mondrian/tree/_tree.pyx":118
 * 
 *         if y.dtype != DOUBLE or not y.flags.contiguous:
 *             y = np.ascontiguousarray(y, dtype=DOUBLE)             # <<<<<<<<<<<<<<
 * 
 *         if (sample_weight is not None and
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ascontiguousarray); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)__pyx_v_y));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_y));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_y));
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_DOUBLE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 118, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 118, __pyx_L1_error)
    __Pyx_DECREF_SET(__pyx_v_y, ((PyArrayObject *)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "skgarden/mondrian/tree/_tree.pyx":117
 *             X = np.asfortranarray(X, dtype=DTYPE)
 * 
 *         if y.dtype != DOUBLE or not y.flags.contiguous:             # <<<<<<<<<<<<<<
 *             y = np.ascontiguousarray(y, dtype=DOUBLE)
 * 
 */
  }

  /* "skgarden/mondrian/tree/_tree.pyx":120
 *             y = np.ascontiguousarray(y, dtype=DOUBLE)
 * 
 *         if (sample_weight is not None and             # <<<<<<<<<<<<<<
 *             (sample_weight.dtype != DOUBLE or
 *             not sample_weight.flags.contiguous)):
 */
  __pyx_t_7 = (((PyObject *)__pyx_v_sample_weight) != Py_None);
  __pyx_t_6 = (__pyx_t_7 != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L8_bool_binop_done;
  }

  /* "skgarden/mondrian/tree/_tree.pyx":121
 * 
 *         if (sample_weight is not None and
 *             (sample_weight.dtype != DOUBLE or             # <<<<<<<<<<<<<<
 *             not sample_weight.flags.contiguous)):
 *                 sample_weight = np.asarray(sample_weight, dtype=DOUBLE,
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_sample_weight), __pyx_n_s_dtype); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_DOUBLE); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_2, __pyx_t_5, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L8_bool_binop_done;
  }

  /* "skgarden/mondrian/tree/_tree.pyx":122
 *         if (sample_weight is not None and
 *             (sample_weight.dtype != DOUBLE or
 *             not sample_weight.flags.contiguous)):             # <<<<<<<<<<<<<<
 *                 sample_weight = np.asarray(sample_weight, dtype=DOUBLE,
 *                                            order="C")
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_sample_weight), __pyx_n_s_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_contiguous); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_7 = ((!__pyx_t_6) != 0);
  __pyx_t_4 = __pyx_t_7;
  __pyx_L8_bool_binop_done:;

  /* "skgarden/mondrian/tree/_tree.pyx":120
 *             y = np.ascontiguousarray(y, dtype=DOUBLE)
 * 
 *         if (sample_weight is not None and             # <<<<<<<<<<<<<<
 *             (sample_weight.dtype != DOUBLE or
 *             not sample_weight.flags.contiguous)):
 */
  if (__pyx_t_4) {

    /* "skgarden/mondrian/tree/_tree.pyx":123
 *             (sample_weight.dtype != DOUBLE or
 *             not sample_weight.flags.contiguous)):
 *                 sample_weight = np.asarray(sample_weight, dtype=DOUBLE,             # <<<<<<<<<<<<<<
 *                                            order="C")
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_asarray); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)__pyx_v_sample_weight));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_sample_weight));
    PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)__pyx_v_sample_weight));
    __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_DOUBLE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_3) < 0) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_order, __pyx_n_s_C) < 0) __PYX_ERR(0, 123, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_5, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_DECREF_SET(__pyx_v_sample_weight, ((PyArrayObject *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "skgarden/mondrian/tree/_tree.pyx":120
 *             y = np.ascontiguousarray(y, dtype=DOUBLE)
 * 
 *         if (sample_weight is not None and             # <<<<<<<<<<<<<<
 *             (sample_weight.dtype != DOUBLE or
 *             not sample_weight.flags.contiguous)):
 */
  }

  /* "skgarden/mondrian/tree/_tree.pyx":126
 *                                            order="C")
 * 
 *         return X, y, sample_weight             # <<<<<<<<<<<<<<
 * 
 * cdef class PartialFitTreeBuilder(TreeBuilder):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_X);
  __Pyx_GIVEREF(__pyx_v_X);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_X);
  __Pyx_INCREF(((PyObject *)__pyx_v_y));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_y));
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_y));
  __Pyx_INCREF(((PyObject *)__pyx_v_sample_weight));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_sample_weight));
  PyTuple_SET_ITEM(__pyx_t_3, 2, ((PyObject *)__pyx_v_sample_weight));
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "skgarden/mondrian/tree/_tree.pyx":110
 *         pass
 * 
 *     cdef inline _check_input(self, object X, np.ndarray y,             # <<<<<<<<<<<<<<
 *                              np.ndarray sample_weight):
 *         """Check input dtype, layout and format"""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.TreeBuilder._check_input", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_X);
  __Pyx_XDECREF((PyObject *)__pyx_v_y);
  __Pyx_XDECREF((PyObject *)__pyx_v_sample_weight);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef bint use_setstate
 *     state = (self.max_depth, self.min_samples_leaf, self.min_samples_split, self.min_weight_leaf, self.random_state, self.splitter)
 */

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_2__reduce_cython__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_2__reduce_cython__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *__pyx_v_self) {
  int __pyx_v_use_setstate;
  PyObject *__pyx_v_state = NULL;
  PyObject *__pyx_v__dict = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     cdef bint use_setstate
 *     state = (self.max_depth, self.min_samples_leaf, self.min_samples_split, self.min_weight_leaf, self.random_state, self.splitter)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_Py_intptr_t(__pyx_v_self->max_depth); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_Py_intptr_t(__pyx_v_self->min_samples_leaf); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_Py_intptr_t(__pyx_v_self->min_samples_split); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->min_weight_leaf); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);
  __Pyx_INCREF(__pyx_v_self->random_state);
  __Pyx_GIVEREF(__pyx_v_self->random_state);
  PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_v_self->random_state);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->splitter));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->splitter));
  PyTuple_SET_ITEM(__pyx_t_5, 5, ((PyObject *)__pyx_v_self->splitter));
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "(tree fragment)":4
 *     cdef bint use_setstate
 *     state = (self.max_depth, self.min_samples_leaf, self.min_samples_split, self.min_weight_leaf, self.random_state, self.splitter)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_5 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v__dict = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "(tree fragment)":5
 *     state = (self.max_depth, self.min_samples_leaf, self.min_samples_split, self.min_weight_leaf, self.random_state, self.splitter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_6 = (__pyx_v__dict != Py_None);
  __pyx_t_7 = (__pyx_t_6 != 0);
  if (__pyx_t_7) {

    /* "(tree fragment)":6
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":7
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.random_state is not None or self.splitter is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":5
 *     state = (self.max_depth, self.min_samples_leaf, self.min_samples_split, self.min_weight_leaf, self.random_state, self.splitter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":9
 *         use_setstate = True
 *     else:
 *         use_setstate = self.random_state is not None or self.splitter is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_TreeBuilder, (type(self), 0xedae8f4, None), state
 */
  /*else*/ {
    __pyx_t_6 = (__pyx_v_self->random_state != Py_None);
    __pyx_t_8 = (__pyx_t_6 != 0);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_8 = (((PyObject *)__pyx_v_self->splitter) != Py_None);
    __pyx_t_6 = (__pyx_t_8 != 0);
    __pyx_t_7 = __pyx_t_6;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_7;
  }
  __pyx_L3:;

  /* "(tree fragment)":10
 *     else:
 *         use_setstate = self.random_state is not None or self.splitter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_TreeBuilder, (type(self), 0xedae8f4, None), state
 *     else:
 */
  __pyx_t_7 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_7) {

    /* "(tree fragment)":11
 *         use_setstate = self.random_state is not None or self.splitter is not None
 *     if use_setstate:
 *         return __pyx_unpickle_TreeBuilder, (type(self), 0xedae8f4, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_TreeBuilder, (type(self), 0xedae8f4, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_TreeBuilder); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 11, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 11, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_249227508);
    __Pyx_GIVEREF(__pyx_int_249227508);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_249227508);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_5, 2, Py_None);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 11, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":10
 *     else:
 *         use_setstate = self.random_state is not None or self.splitter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_TreeBuilder, (type(self), 0xedae8f4, None), state
 *     else:
 */
  }

  /* "(tree fragment)":13
 *         return __pyx_unpickle_TreeBuilder, (type(self), 0xedae8f4, None), state
 *     else:
 *         return __pyx_unpickle_TreeBuilder, (type(self), 0xedae8f4, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_TreeBuilder__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_TreeBuilder); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_249227508);
    __Pyx_GIVEREF(__pyx_int_249227508);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_249227508);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);
    __pyx_t_3 = 0;
    __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef bint use_setstate
 *     state = (self.max_depth, self.min_samples_leaf, self.min_samples_split, self.min_weight_leaf, self.random_state, self.splitter)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.TreeBuilder.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":14
 *     else:
 *         return __pyx_unpickle_TreeBuilder, (type(self), 0xedae8f4, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_TreeBuilder__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_4__setstate_cython__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_4__setstate_cython__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":15
 *         return __pyx_unpickle_TreeBuilder, (type(self), 0xedae8f4, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_TreeBuilder__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 15, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_8skgarden_8mondrian_4tree_5_tree___pyx_unpickle_TreeBuilder__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":14
 *     else:
 *         return __pyx_unpickle_TreeBuilder, (type(self), 0xedae8f4, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_TreeBuilder__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.TreeBuilder.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":131
 *     """Build a decision tree incrementally."""
 * 
 *     def __cinit__(self, SIZE_t min_samples_split, SIZE_t max_depth,             # <<<<<<<<<<<<<<
 *                   object random_state):
 *         self.min_samples_split = min_samples_split
 */

/* Python wrapper */
static int __pyx_pw_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_min_samples_split;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_max_depth;
  PyObject *__pyx_v_random_state = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_min_samples_split,&__pyx_n_s_max_depth,&__pyx_n_s_random_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_samples_split)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_depth)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 1); __PYX_ERR(0, 131, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_random_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 2); __PYX_ERR(0, 131, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 131, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_min_samples_split = __Pyx_PyInt_As_Py_intptr_t(values[0]); if (unlikely((__pyx_v_min_samples_split == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 131, __pyx_L3_error)
    __pyx_v_max_depth = __Pyx_PyInt_As_Py_intptr_t(values[1]); if (unlikely((__pyx_v_max_depth == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 131, __pyx_L3_error)
    __pyx_v_random_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 131, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.PartialFitTreeBuilder.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder___cinit__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder *)__pyx_v_self), __pyx_v_min_samples_split, __pyx_v_max_depth, __pyx_v_random_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder___cinit__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder *__pyx_v_self, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_min_samples_split, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_max_depth, PyObject *__pyx_v_random_state) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "skgarden/mondrian/tree/_tree.pyx":133
 *     def __cinit__(self, SIZE_t min_samples_split, SIZE_t max_depth,
 *                   object random_state):
 *         self.min_samples_split = min_samples_split             # <<<<<<<<<<<<<<
 *         self.max_depth = max_depth
 *         self.random_state = random_state
 */
  __pyx_v_self->__pyx_base.min_samples_split = __pyx_v_min_samples_split;

  /* "skgarden/mondrian/tree/_tree.pyx":134
 *                   object random_state):
 *         self.min_samples_split = min_samples_split
 *         self.max_depth = max_depth             # <<<<<<<<<<<<<<
 *         self.random_state = random_state
 * 
 */
  __pyx_v_self->__pyx_base.max_depth = __pyx_v_max_depth;

  /* "skgarden/mondrian/tree/_tree.pyx":135
 *         self.min_samples_split = min_samples_split
 *         self.max_depth = max_depth
 *         self.random_state = random_state             # <<<<<<<<<<<<<<
 * 
 *     cpdef build(self, Tree tree, object X, np.ndarray y,
 */
  __Pyx_INCREF(__pyx_v_random_state);
  __Pyx_GIVEREF(__pyx_v_random_state);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.random_state);
  __Pyx_DECREF(__pyx_v_self->__pyx_base.random_state);
  __pyx_v_self->__pyx_base.random_state = __pyx_v_random_state;

  /* "skgarden/mondrian/tree/_tree.pyx":131
 *     """Build a decision tree incrementally."""
 * 
 *     def __cinit__(self, SIZE_t min_samples_split, SIZE_t max_depth,             # <<<<<<<<<<<<<<
 *                   object random_state):
 *         self.min_samples_split = min_samples_split
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":137
 *         self.random_state = random_state
 * 
 *     cpdef build(self, Tree tree, object X, np.ndarray y,             # <<<<<<<<<<<<<<
 *                 np.ndarray sample_weight=None,
 *                 np.ndarray X_idx_sorted=None):
 */

static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder_3build(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder_build(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder *__pyx_v_self, struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_tree, PyObject *__pyx_v_X, PyArrayObject *__pyx_v_y, int __pyx_skip_dispatch, struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder_build *__pyx_optional_args) {

  /* "skgarden/mondrian/tree/_tree.pyx":138
 * 
 *     cpdef build(self, Tree tree, object X, np.ndarray y,
 *                 np.ndarray sample_weight=None,             # <<<<<<<<<<<<<<
 *                 np.ndarray X_idx_sorted=None):
 *         X, y, sample_weight = self._check_input(X, y, None)
 */
  PyArrayObject *__pyx_v_sample_weight = ((PyArrayObject *)Py_None);

  /* "skgarden/mondrian/tree/_tree.pyx":139
 *     cpdef build(self, Tree tree, object X, np.ndarray y,
 *                 np.ndarray sample_weight=None,
 *                 np.ndarray X_idx_sorted=None):             # <<<<<<<<<<<<<<
 *         X, y, sample_weight = self._check_input(X, y, None)
 * 
 */
  PyArrayObject *__pyx_v_X_idx_sorted = ((PyArrayObject *)Py_None);
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_UINT32_t __pyx_v_rand_r_state;
  int __pyx_v_n_samples;
  int __pyx_v_init_capacity;
  PyArrayObject *__pyx_v_X_ndarray = 0;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *__pyx_v_X_ptr;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t *__pyx_v_y_ptr;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_X_f_stride;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_X_s_stride;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_y_stride;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_sample_ind;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_start;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *(*__pyx_t_7)(PyObject *);
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_UINT32_t __pyx_t_8;
  int __pyx_t_9;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_10;
  int __pyx_t_11;
  __Pyx_RefNannySetupContext("build", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_sample_weight = __pyx_optional_args->sample_weight;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_X_idx_sorted = __pyx_optional_args->X_idx_sorted;
      }
    }
  }
  __Pyx_INCREF(__pyx_v_X);
  __Pyx_INCREF((PyObject *)__pyx_v_y);
  __Pyx_INCREF((PyObject *)__pyx_v_sample_weight);

  /* "skgarden/mondrian/tree/_tree.pyx":137
 *         self.random_state = random_state
 * 
 *     cpdef build(self, Tree tree, object X, np.ndarray y,             # <<<<<<<<<<<<<<
 *                 np.ndarray sample_weight=None,
 *                 np.ndarray X_idx_sorted=None):
 */
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    static PY_UINT64_T tp_dict_version = 0, obj_dict_version = 0;
    if (likely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict && tp_dict_version == __PYX_GET_DICT_VERSION(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict) && (!Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset || obj_dict_version == __PYX_GET_DICT_VERSION(_PyObject_GetDictPtr(((PyObject *)__pyx_v_self))))));
    else {
      PY_UINT64_T type_dict_guard = (likely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict)) ? __PYX_GET_DICT_VERSION(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict) : 0;
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_build); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 137, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder_3build)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[6] = {__pyx_t_4, ((PyObject *)__pyx_v_tree), __pyx_v_X, ((PyObject *)__pyx_v_y), ((PyObject *)__pyx_v_sample_weight), ((PyObject *)__pyx_v_X_idx_sorted)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 5+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 137, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[6] = {__pyx_t_4, ((PyObject *)__pyx_v_tree), __pyx_v_X, ((PyObject *)__pyx_v_y), ((PyObject *)__pyx_v_sample_weight), ((PyObject *)__pyx_v_X_idx_sorted)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 5+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 137, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(5+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 137, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_tree));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_tree));
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_tree));
          __Pyx_INCREF(__pyx_v_X);
          __Pyx_GIVEREF(__pyx_v_X);
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_X);
          __Pyx_INCREF(((PyObject *)__pyx_v_y));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_y));
          PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_5, ((PyObject *)__pyx_v_y));
          __Pyx_INCREF(((PyObject *)__pyx_v_sample_weight));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_sample_weight));
          PyTuple_SET_ITEM(__pyx_t_6, 3+__pyx_t_5, ((PyObject *)__pyx_v_sample_weight));
          __Pyx_INCREF(((PyObject *)__pyx_v_X_idx_sorted));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_X_idx_sorted));
          PyTuple_SET_ITEM(__pyx_t_6, 4+__pyx_t_5, ((PyObject *)__pyx_v_X_idx_sorted));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 137, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
      tp_dict_version = likely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict) ? __PYX_GET_DICT_VERSION(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict) : 0;
      obj_dict_version = likely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset) ? __PYX_GET_DICT_VERSION(_PyObject_GetDictPtr(((PyObject *)__pyx_v_self))) : 0;
      if (unlikely(type_dict_guard != tp_dict_version)) {
        tp_dict_version = obj_dict_version = 0;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    }
    #endif
  }

  /* "skgarden/mondrian/tree/_tree.pyx":140
 *                 np.ndarray sample_weight=None,
 *                 np.ndarray X_idx_sorted=None):
 *         X, y, sample_weight = self._check_input(X, y, None)             # <<<<<<<<<<<<<<
 * 
 *         cdef UINT32_t rand_r_state = self.random_state.randint(0, RAND_R_MAX)
 */
  __pyx_t_1 = __pyx_f_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder__check_input(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *)__pyx_v_self), __pyx_v_X, __pyx_v_y, ((PyArrayObject *)Py_None)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 3)) {
      if (size > 3) __Pyx_RaiseTooManyValuesError(3);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 140, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_6 = PyTuple_GET_ITEM(sequence, 2); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_6 = PyList_GET_ITEM(sequence, 2); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_6);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_7(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_7(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    index = 2; __pyx_t_6 = __pyx_t_7(__pyx_t_4); if (unlikely(!__pyx_t_6)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_6);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_4), 3) < 0) __PYX_ERR(0, 140, __pyx_L1_error)
    __pyx_t_7 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_7 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 140, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 140, __pyx_L1_error)
  if (!(likely(((__pyx_t_6) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_6, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 140, __pyx_L1_error)
  __Pyx_DECREF_SET(__pyx_v_X, __pyx_t_2);
  __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_y, ((PyArrayObject *)__pyx_t_3));
  __pyx_t_3 = 0;
  __Pyx_DECREF_SET(__pyx_v_sample_weight, ((PyArrayObject *)__pyx_t_6));
  __pyx_t_6 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":142
 *         X, y, sample_weight = self._check_input(X, y, None)
 * 
 *         cdef UINT32_t rand_r_state = self.random_state.randint(0, RAND_R_MAX)             # <<<<<<<<<<<<<<
 *         cdef int n_samples = X.shape[0]
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.random_state, __pyx_n_s_randint); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_e_8skgarden_8mondrian_4tree_6_utils_RAND_R_MAX); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_int_0, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_int_0, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_4 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_5, __pyx_int_0);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_5, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_8 = __Pyx_PyInt_As_npy_uint32(__pyx_t_1); if (unlikely((__pyx_t_8 == ((npy_uint32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_rand_r_state = __pyx_t_8;

  /* "skgarden/mondrian/tree/_tree.pyx":143
 * 
 *         cdef UINT32_t rand_r_state = self.random_state.randint(0, RAND_R_MAX)
 *         cdef int n_samples = X.shape[0]             # <<<<<<<<<<<<<<
 * 
 *         # Allocate memory for tree.
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_n_samples = __pyx_t_5;

  /* "skgarden/mondrian/tree/_tree.pyx":147
 *         # Allocate memory for tree.
 *         cdef int init_capacity
 *         if tree.max_depth <= 10:             # <<<<<<<<<<<<<<
 *             init_capacity = (2 ** (tree.max_depth + 1)) - 1
 *             tree._resize(init_capacity)
 */
  __pyx_t_9 = ((__pyx_v_tree->max_depth <= 10) != 0);
  if (__pyx_t_9) {

    /* "skgarden/mondrian/tree/_tree.pyx":148
 *         cdef int init_capacity
 *         if tree.max_depth <= 10:
 *             init_capacity = (2 ** (tree.max_depth + 1)) - 1             # <<<<<<<<<<<<<<
 *             tree._resize(init_capacity)
 * 
 */
    __pyx_v_init_capacity = (__Pyx_pow_long(2, (__pyx_v_tree->max_depth + 1)) - 1);

    /* "skgarden/mondrian/tree/_tree.pyx":149
 *         if tree.max_depth <= 10:
 *             init_capacity = (2 ** (tree.max_depth + 1)) - 1
 *             tree._resize(init_capacity)             # <<<<<<<<<<<<<<
 * 
 *         cdef np.ndarray X_ndarray = X
 */
    __pyx_t_5 = ((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_tree->__pyx_vtab)->_resize(__pyx_v_tree, __pyx_v_init_capacity); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 149, __pyx_L1_error)

    /* "skgarden/mondrian/tree/_tree.pyx":147
 *         # Allocate memory for tree.
 *         cdef int init_capacity
 *         if tree.max_depth <= 10:             # <<<<<<<<<<<<<<
 *             init_capacity = (2 ** (tree.max_depth + 1)) - 1
 *             tree._resize(init_capacity)
 */
  }

  /* "skgarden/mondrian/tree/_tree.pyx":151
 *             tree._resize(init_capacity)
 * 
 *         cdef np.ndarray X_ndarray = X             # <<<<<<<<<<<<<<
 *         cdef DTYPE_t* X_ptr = <DTYPE_t*> X_ndarray.data
 *         cdef DOUBLE_t* y_ptr = <DOUBLE_t*> y.data
 */
  if (!(likely(((__pyx_v_X) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_X, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 151, __pyx_L1_error)
  __pyx_t_6 = __pyx_v_X;
  __Pyx_INCREF(__pyx_t_6);
  __pyx_v_X_ndarray = ((PyArrayObject *)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":152
 * 
 *         cdef np.ndarray X_ndarray = X
 *         cdef DTYPE_t* X_ptr = <DTYPE_t*> X_ndarray.data             # <<<<<<<<<<<<<<
 *         cdef DOUBLE_t* y_ptr = <DOUBLE_t*> y.data
 *         cdef SIZE_t X_f_stride = X.strides[1] / X.itemsize
 */
  __pyx_v_X_ptr = ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *)__pyx_v_X_ndarray->data);

  /* "skgarden/mondrian/tree/_tree.pyx":153
 *         cdef np.ndarray X_ndarray = X
 *         cdef DTYPE_t* X_ptr = <DTYPE_t*> X_ndarray.data
 *         cdef DOUBLE_t* y_ptr = <DOUBLE_t*> y.data             # <<<<<<<<<<<<<<
 *         cdef SIZE_t X_f_stride = X.strides[1] / X.itemsize
 *         cdef SIZE_t X_s_stride = X.strides[0] / X.itemsize
 */
  __pyx_v_y_ptr = ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t *)__pyx_v_y->data);

  /* "skgarden/mondrian/tree/_tree.pyx":154
 *         cdef DTYPE_t* X_ptr = <DTYPE_t*> X_ndarray.data
 *         cdef DOUBLE_t* y_ptr = <DOUBLE_t*> y.data
 *         cdef SIZE_t X_f_stride = X.strides[1] / X.itemsize             # <<<<<<<<<<<<<<
 *         cdef SIZE_t X_s_stride = X.strides[0] / X.itemsize
 *         cdef SIZE_t y_stride = y.strides[0] / y.itemsize
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_strides); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_6, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_itemsize); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = __Pyx_PyNumber_Divide(__pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_10 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_4); if (unlikely((__pyx_t_10 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 154, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_X_f_stride = __pyx_t_10;

  /* "skgarden/mondrian/tree/_tree.pyx":155
 *         cdef DOUBLE_t* y_ptr = <DOUBLE_t*> y.data
 *         cdef SIZE_t X_f_stride = X.strides[1] / X.itemsize
 *         cdef SIZE_t X_s_stride = X.strides[0] / X.itemsize             # <<<<<<<<<<<<<<
 *         cdef SIZE_t y_stride = y.strides[0] / y.itemsize
 *         cdef SIZE_t sample_ind
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_strides); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_4, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_itemsize); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_6, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_10 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_1); if (unlikely((__pyx_t_10 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 155, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_X_s_stride = __pyx_t_10;

  /* "skgarden/mondrian/tree/_tree.pyx":156
 *         cdef SIZE_t X_f_stride = X.strides[1] / X.itemsize
 *         cdef SIZE_t X_s_stride = X.strides[0] / X.itemsize
 *         cdef SIZE_t y_stride = y.strides[0] / y.itemsize             # <<<<<<<<<<<<<<
 *         cdef SIZE_t sample_ind
 *         cdef SIZE_t start
 */
  __pyx_t_1 = __Pyx_PyInt_From_Py_intptr_t((__pyx_v_y->strides[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_y), __pyx_n_s_itemsize); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyNumber_Divide(__pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_10 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_6); if (unlikely((__pyx_t_10 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_y_stride = __pyx_t_10;

  /* "skgarden/mondrian/tree/_tree.pyx":161
 * 
 *         # Initialize the tree when the first sample is inserted.
 *         if tree.node_count == 0:             # <<<<<<<<<<<<<<
 *             tree._init(X_ptr, y_ptr, X_f_stride)
 *             start = 1
 */
  __pyx_t_9 = ((__pyx_v_tree->node_count == 0) != 0);
  if (__pyx_t_9) {

    /* "skgarden/mondrian/tree/_tree.pyx":162
 *         # Initialize the tree when the first sample is inserted.
 *         if tree.node_count == 0:
 *             tree._init(X_ptr, y_ptr, X_f_stride)             # <<<<<<<<<<<<<<
 *             start = 1
 *         else:
 */
    ((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_tree->__pyx_vtab)->_init(__pyx_v_tree, __pyx_v_X_ptr, __pyx_v_y_ptr, __pyx_v_X_f_stride);

    /* "skgarden/mondrian/tree/_tree.pyx":163
 *         if tree.node_count == 0:
 *             tree._init(X_ptr, y_ptr, X_f_stride)
 *             start = 1             # <<<<<<<<<<<<<<
 *         else:
 *             start = 0
 */
    __pyx_v_start = 1;

    /* "skgarden/mondrian/tree/_tree.pyx":161
 * 
 *         # Initialize the tree when the first sample is inserted.
 *         if tree.node_count == 0:             # <<<<<<<<<<<<<<
 *             tree._init(X_ptr, y_ptr, X_f_stride)
 *             start = 1
 */
    goto __pyx_L6;
  }

  /* "skgarden/mondrian/tree/_tree.pyx":165
 *             start = 1
 *         else:
 *             start = 0             # <<<<<<<<<<<<<<
 * 
 *         for sample_ind in range(start, n_samples):
 */
  /*else*/ {
    __pyx_v_start = 0;
  }
  __pyx_L6:;

  /* "skgarden/mondrian/tree/_tree.pyx":167
 *             start = 0
 * 
 *         for sample_ind in range(start, n_samples):             # <<<<<<<<<<<<<<
 *             tree.extend(X_ptr, y_ptr, sample_ind*X_s_stride,
 *                         X_f_stride,
 */
  __pyx_t_5 = __pyx_v_n_samples;
  __pyx_t_11 = __pyx_t_5;
  for (__pyx_t_10 = __pyx_v_start; __pyx_t_10 < __pyx_t_11; __pyx_t_10+=1) {
    __pyx_v_sample_ind = __pyx_t_10;

    /* "skgarden/mondrian/tree/_tree.pyx":168
 * 
 *         for sample_ind in range(start, n_samples):
 *             tree.extend(X_ptr, y_ptr, sample_ind*X_s_stride,             # <<<<<<<<<<<<<<
 *                         X_f_stride,
 *                         sample_ind*y_stride, rand_r_state,
 */
    ((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_tree->__pyx_vtab)->extend(__pyx_v_tree, __pyx_v_X_ptr, __pyx_v_y_ptr, (__pyx_v_sample_ind * __pyx_v_X_s_stride), __pyx_v_X_f_stride, (__pyx_v_sample_ind * __pyx_v_y_stride), __pyx_v_rand_r_state, __pyx_v_self->__pyx_base.min_samples_split);
  }

  /* "skgarden/mondrian/tree/_tree.pyx":137
 *         self.random_state = random_state
 * 
 *     cpdef build(self, Tree tree, object X, np.ndarray y,             # <<<<<<<<<<<<<<
 *                 np.ndarray sample_weight=None,
 *                 np.ndarray X_idx_sorted=None):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.PartialFitTreeBuilder.build", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_X_ndarray);
  __Pyx_XDECREF(__pyx_v_X);
  __Pyx_XDECREF((PyObject *)__pyx_v_y);
  __Pyx_XDECREF((PyObject *)__pyx_v_sample_weight);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder_3build(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder_3build(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_tree = 0;
  PyObject *__pyx_v_X = 0;
  PyArrayObject *__pyx_v_y = 0;
  PyArrayObject *__pyx_v_sample_weight = 0;
  PyArrayObject *__pyx_v_X_idx_sorted = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("build (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_tree,&__pyx_n_s_X,&__pyx_n_s_y,&__pyx_n_s_sample_weight,&__pyx_n_s_X_idx_sorted,0};
    PyObject* values[5] = {0,0,0,0,0};

    /* "skgarden/mondrian/tree/_tree.pyx":138
 * 
 *     cpdef build(self, Tree tree, object X, np.ndarray y,
 *                 np.ndarray sample_weight=None,             # <<<<<<<<<<<<<<
 *                 np.ndarray X_idx_sorted=None):
 *         X, y, sample_weight = self._check_input(X, y, None)
 */
    values[3] = (PyObject *)((PyArrayObject *)Py_None);

    /* "skgarden/mondrian/tree/_tree.pyx":139
 *     cpdef build(self, Tree tree, object X, np.ndarray y,
 *                 np.ndarray sample_weight=None,
 *                 np.ndarray X_idx_sorted=None):             # <<<<<<<<<<<<<<
 *         X, y, sample_weight = self._check_input(X, y, None)
 * 
 */
    values[4] = (PyObject *)((PyArrayObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tree)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("build", 0, 3, 5, 1); __PYX_ERR(0, 137, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("build", 0, 3, 5, 2); __PYX_ERR(0, 137, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sample_weight);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X_idx_sorted);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "build") < 0)) __PYX_ERR(0, 137, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_tree = ((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)values[0]);
    __pyx_v_X = values[1];
    __pyx_v_y = ((PyArrayObject *)values[2]);
    __pyx_v_sample_weight = ((PyArrayObject *)values[3]);
    __pyx_v_X_idx_sorted = ((PyArrayObject *)values[4]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("build", 0, 3, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 137, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.PartialFitTreeBuilder.build", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_tree), __pyx_ptype_8skgarden_8mondrian_4tree_5_tree_Tree, 1, "tree", 0))) __PYX_ERR(0, 137, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_y), __pyx_ptype_5numpy_ndarray, 1, "y", 0))) __PYX_ERR(0, 137, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_sample_weight), __pyx_ptype_5numpy_ndarray, 1, "sample_weight", 0))) __PYX_ERR(0, 138, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_X_idx_sorted), __pyx_ptype_5numpy_ndarray, 1, "X_idx_sorted", 0))) __PYX_ERR(0, 139, __pyx_L1_error)
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder_2build(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder *)__pyx_v_self), __pyx_v_tree, __pyx_v_X, __pyx_v_y, __pyx_v_sample_weight, __pyx_v_X_idx_sorted);

  /* "skgarden/mondrian/tree/_tree.pyx":137
 *         self.random_state = random_state
 * 
 *     cpdef build(self, Tree tree, object X, np.ndarray y,             # <<<<<<<<<<<<<<
 *                 np.ndarray sample_weight=None,
 *                 np.ndarray X_idx_sorted=None):
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder_2build(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder *__pyx_v_self, struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_tree, PyObject *__pyx_v_X, PyArrayObject *__pyx_v_y, PyArrayObject *__pyx_v_sample_weight, PyArrayObject *__pyx_v_X_idx_sorted) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_build __pyx_t_2;
  __Pyx_RefNannySetupContext("build", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 2;
  __pyx_t_2.sample_weight = __pyx_v_sample_weight;
  __pyx_t_2.X_idx_sorted = __pyx_v_X_idx_sorted;
  __pyx_t_1 = __pyx_vtabptr_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder->__pyx_base.build(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *)__pyx_v_self), __pyx_v_tree, __pyx_v_X, __pyx_v_y, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.PartialFitTreeBuilder.build", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder_4__reduce_cython__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.PartialFitTreeBuilder.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder_6__setstate_cython__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.PartialFitTreeBuilder.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":178
 *     """Build a decision tree in depth-first fashion."""
 * 
 *     def __cinit__(self, Splitter splitter, SIZE_t min_samples_split,             # <<<<<<<<<<<<<<
 *                   SIZE_t max_depth):
 *         self.splitter = splitter
 */

/* Python wrapper */
static int __pyx_pw_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_8skgarden_8mondrian_4tree_9_splitter_Splitter *__pyx_v_splitter = 0;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_min_samples_split;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_max_depth;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_splitter,&__pyx_n_s_min_samples_split,&__pyx_n_s_max_depth,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_splitter)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_samples_split)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 1); __PYX_ERR(0, 178, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_depth)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 2); __PYX_ERR(0, 178, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 178, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_splitter = ((struct __pyx_obj_8skgarden_8mondrian_4tree_9_splitter_Splitter *)values[0]);
    __pyx_v_min_samples_split = __Pyx_PyInt_As_Py_intptr_t(values[1]); if (unlikely((__pyx_v_min_samples_split == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 178, __pyx_L3_error)
    __pyx_v_max_depth = __Pyx_PyInt_As_Py_intptr_t(values[2]); if (unlikely((__pyx_v_max_depth == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 179, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 178, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.DepthFirstTreeBuilder.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_splitter), __pyx_ptype_8skgarden_8mondrian_4tree_9_splitter_Splitter, 1, "splitter", 0))) __PYX_ERR(0, 178, __pyx_L1_error)
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder___cinit__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder *)__pyx_v_self), __pyx_v_splitter, __pyx_v_min_samples_split, __pyx_v_max_depth);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder___cinit__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder *__pyx_v_self, struct __pyx_obj_8skgarden_8mondrian_4tree_9_splitter_Splitter *__pyx_v_splitter, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_min_samples_split, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_max_depth) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "skgarden/mondrian/tree/_tree.pyx":180
 *     def __cinit__(self, Splitter splitter, SIZE_t min_samples_split,
 *                   SIZE_t max_depth):
 *         self.splitter = splitter             # <<<<<<<<<<<<<<
 *         self.min_samples_split = min_samples_split
 *         self.max_depth = max_depth
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_splitter));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_splitter));
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.splitter);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->__pyx_base.splitter));
  __pyx_v_self->__pyx_base.splitter = __pyx_v_splitter;

  /* "skgarden/mondrian/tree/_tree.pyx":181
 *                   SIZE_t max_depth):
 *         self.splitter = splitter
 *         self.min_samples_split = min_samples_split             # <<<<<<<<<<<<<<
 *         self.max_depth = max_depth
 * 
 */
  __pyx_v_self->__pyx_base.min_samples_split = __pyx_v_min_samples_split;

  /* "skgarden/mondrian/tree/_tree.pyx":182
 *         self.splitter = splitter
 *         self.min_samples_split = min_samples_split
 *         self.max_depth = max_depth             # <<<<<<<<<<<<<<
 * 
 *     cpdef build(self, Tree tree, object X, np.ndarray y,
 */
  __pyx_v_self->__pyx_base.max_depth = __pyx_v_max_depth;

  /* "skgarden/mondrian/tree/_tree.pyx":178
 *     """Build a decision tree in depth-first fashion."""
 * 
 *     def __cinit__(self, Splitter splitter, SIZE_t min_samples_split,             # <<<<<<<<<<<<<<
 *                   SIZE_t max_depth):
 *         self.splitter = splitter
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":184
 *         self.max_depth = max_depth
 * 
 *     cpdef build(self, Tree tree, object X, np.ndarray y,             # <<<<<<<<<<<<<<
 *                 np.ndarray sample_weight=None,
 *                 np.ndarray X_idx_sorted=None):
 */

static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder_3build(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder_build(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder *__pyx_v_self, struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_tree, PyObject *__pyx_v_X, PyArrayObject *__pyx_v_y, int __pyx_skip_dispatch, struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder_build *__pyx_optional_args) {

  /* "skgarden/mondrian/tree/_tree.pyx":185
 * 
 *     cpdef build(self, Tree tree, object X, np.ndarray y,
 *                 np.ndarray sample_weight=None,             # <<<<<<<<<<<<<<
 *                 np.ndarray X_idx_sorted=None):
 *         """Build a decision tree from the training set (X, y)."""
 */
  PyArrayObject *__pyx_v_sample_weight = ((PyArrayObject *)Py_None);

  /* "skgarden/mondrian/tree/_tree.pyx":186
 *     cpdef build(self, Tree tree, object X, np.ndarray y,
 *                 np.ndarray sample_weight=None,
 *                 np.ndarray X_idx_sorted=None):             # <<<<<<<<<<<<<<
 *         """Build a decision tree from the training set (X, y)."""
 * 
 */
  PyArrayObject *__pyx_v_X_idx_sorted = ((PyArrayObject *)Py_None);
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t *__pyx_v_sample_weight_ptr;
  int __pyx_v_init_capacity;
  struct __pyx_obj_8skgarden_8mondrian_4tree_9_splitter_Splitter *__pyx_v_splitter = 0;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_max_depth;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_min_samples_split;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_start;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_end;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_depth;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_parent;
  int __pyx_v_is_left;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_n_node_samples;
  CYTHON_UNUSED double __pyx_v_weighted_n_samples;
  double __pyx_v_weighted_n_node_samples;
  struct __pyx_t_8skgarden_8mondrian_4tree_9_splitter_SplitRecord __pyx_v_split;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_node_id;
  double __pyx_v_impurity;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_n_constant_features;
  int __pyx_v_is_leaf;
  int __pyx_v_first;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_max_depth_seen;
  int __pyx_v_rc;
  struct __pyx_obj_8skgarden_8mondrian_4tree_6_utils_Stack *__pyx_v_stack = 0;
  struct __pyx_t_8skgarden_8mondrian_4tree_6_utils_StackRecord __pyx_v_stack_record;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_t_8;
  int __pyx_t_9;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_10;
  struct __pyx_opt_args_8skgarden_8mondrian_4tree_9_splitter_8Splitter_init __pyx_t_11;
  __pyx_t_8skgarden_8mondrian_4tree_9_splitter_SIZE_t __pyx_t_12;
  double __pyx_t_13;
  __pyx_t_8skgarden_8mondrian_4tree_6_utils_SIZE_t __pyx_t_14;
  struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_4Tree__resize_c __pyx_t_15;
  __Pyx_RefNannySetupContext("build", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_sample_weight = __pyx_optional_args->sample_weight;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_X_idx_sorted = __pyx_optional_args->X_idx_sorted;
      }
    }
  }
  __Pyx_INCREF(__pyx_v_X);
  __Pyx_INCREF((PyObject *)__pyx_v_y);
  __Pyx_INCREF((PyObject *)__pyx_v_sample_weight);

  /* "skgarden/mondrian/tree/_tree.pyx":184
 *         self.max_depth = max_depth
 * 
 *     cpdef build(self, Tree tree, object X, np.ndarray y,             # <<<<<<<<<<<<<<
 *                 np.ndarray sample_weight=None,
 *                 np.ndarray X_idx_sorted=None):
 */
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    static PY_UINT64_T tp_dict_version = 0, obj_dict_version = 0;
    if (likely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict && tp_dict_version == __PYX_GET_DICT_VERSION(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict) && (!Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset || obj_dict_version == __PYX_GET_DICT_VERSION(_PyObject_GetDictPtr(((PyObject *)__pyx_v_self))))));
    else {
      PY_UINT64_T type_dict_guard = (likely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict)) ? __PYX_GET_DICT_VERSION(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict) : 0;
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_build); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 184, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder_3build)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[6] = {__pyx_t_4, ((PyObject *)__pyx_v_tree), __pyx_v_X, ((PyObject *)__pyx_v_y), ((PyObject *)__pyx_v_sample_weight), ((PyObject *)__pyx_v_X_idx_sorted)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 5+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 184, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[6] = {__pyx_t_4, ((PyObject *)__pyx_v_tree), __pyx_v_X, ((PyObject *)__pyx_v_y), ((PyObject *)__pyx_v_sample_weight), ((PyObject *)__pyx_v_X_idx_sorted)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 5+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 184, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(5+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 184, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_tree));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_tree));
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_tree));
          __Pyx_INCREF(__pyx_v_X);
          __Pyx_GIVEREF(__pyx_v_X);
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_X);
          __Pyx_INCREF(((PyObject *)__pyx_v_y));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_y));
          PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_5, ((PyObject *)__pyx_v_y));
          __Pyx_INCREF(((PyObject *)__pyx_v_sample_weight));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_sample_weight));
          PyTuple_SET_ITEM(__pyx_t_6, 3+__pyx_t_5, ((PyObject *)__pyx_v_sample_weight));
          __Pyx_INCREF(((PyObject *)__pyx_v_X_idx_sorted));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_X_idx_sorted));
          PyTuple_SET_ITEM(__pyx_t_6, 4+__pyx_t_5, ((PyObject *)__pyx_v_X_idx_sorted));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 184, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
      tp_dict_version = likely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict) ? __PYX_GET_DICT_VERSION(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict) : 0;
      obj_dict_version = likely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset) ? __PYX_GET_DICT_VERSION(_PyObject_GetDictPtr(((PyObject *)__pyx_v_self))) : 0;
      if (unlikely(type_dict_guard != tp_dict_version)) {
        tp_dict_version = obj_dict_version = 0;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    }
    #endif
  }

  /* "skgarden/mondrian/tree/_tree.pyx":190
 * 
 *         # check input
 *         X, y, sample_weight = self._check_input(X, y, sample_weight)             # <<<<<<<<<<<<<<
 * 
 *         cdef DOUBLE_t* sample_weight_ptr = NULL
 */
  __pyx_t_1 = __pyx_f_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder__check_input(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *)__pyx_v_self), __pyx_v_X, __pyx_v_y, __pyx_v_sample_weight); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 3)) {
      if (size > 3) __Pyx_RaiseTooManyValuesError(3);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 190, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_6 = PyTuple_GET_ITEM(sequence, 2); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_6 = PyList_GET_ITEM(sequence, 2); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_6);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_7(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_7(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    index = 2; __pyx_t_6 = __pyx_t_7(__pyx_t_4); if (unlikely(!__pyx_t_6)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_6);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_4), 3) < 0) __PYX_ERR(0, 190, __pyx_L1_error)
    __pyx_t_7 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_7 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 190, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 190, __pyx_L1_error)
  if (!(likely(((__pyx_t_6) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_6, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_DECREF_SET(__pyx_v_X, __pyx_t_2);
  __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_y, ((PyArrayObject *)__pyx_t_3));
  __pyx_t_3 = 0;
  __Pyx_DECREF_SET(__pyx_v_sample_weight, ((PyArrayObject *)__pyx_t_6));
  __pyx_t_6 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":192
 *         X, y, sample_weight = self._check_input(X, y, sample_weight)
 * 
 *         cdef DOUBLE_t* sample_weight_ptr = NULL             # <<<<<<<<<<<<<<
 *         if sample_weight is not None:
 *             sample_weight_ptr = <DOUBLE_t*> sample_weight.data
 */
  __pyx_v_sample_weight_ptr = NULL;

  /* "skgarden/mondrian/tree/_tree.pyx":193
 * 
 *         cdef DOUBLE_t* sample_weight_ptr = NULL
 *         if sample_weight is not None:             # <<<<<<<<<<<<<<
 *             sample_weight_ptr = <DOUBLE_t*> sample_weight.data
 * 
 */
  __pyx_t_8 = (((PyObject *)__pyx_v_sample_weight) != Py_None);
  __pyx_t_9 = (__pyx_t_8 != 0);
  if (__pyx_t_9) {

    /* "skgarden/mondrian/tree/_tree.pyx":194
 *         cdef DOUBLE_t* sample_weight_ptr = NULL
 *         if sample_weight is not None:
 *             sample_weight_ptr = <DOUBLE_t*> sample_weight.data             # <<<<<<<<<<<<<<
 * 
 *         # Initial capacity
 */
    __pyx_v_sample_weight_ptr = ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t *)__pyx_v_sample_weight->data);

    /* "skgarden/mondrian/tree/_tree.pyx":193
 * 
 *         cdef DOUBLE_t* sample_weight_ptr = NULL
 *         if sample_weight is not None:             # <<<<<<<<<<<<<<
 *             sample_weight_ptr = <DOUBLE_t*> sample_weight.data
 * 
 */
  }

  /* "skgarden/mondrian/tree/_tree.pyx":198
 *         # Initial capacity
 *         cdef int init_capacity
 *         if tree.max_depth <= 10:             # <<<<<<<<<<<<<<
 *             init_capacity = (2 ** (tree.max_depth + 1)) - 1
 *         else:
 */
  __pyx_t_9 = ((__pyx_v_tree->max_depth <= 10) != 0);
  if (__pyx_t_9) {

    /* "skgarden/mondrian/tree/_tree.pyx":199
 *         cdef int init_capacity
 *         if tree.max_depth <= 10:
 *             init_capacity = (2 ** (tree.max_depth + 1)) - 1             # <<<<<<<<<<<<<<
 *         else:
 *             init_capacity = 2047
 */
    __pyx_v_init_capacity = (__Pyx_pow_long(2, (__pyx_v_tree->max_depth + 1)) - 1);

    /* "skgarden/mondrian/tree/_tree.pyx":198
 *         # Initial capacity
 *         cdef int init_capacity
 *         if tree.max_depth <= 10:             # <<<<<<<<<<<<<<
 *             init_capacity = (2 ** (tree.max_depth + 1)) - 1
 *         else:
 */
    goto __pyx_L6;
  }

  /* "skgarden/mondrian/tree/_tree.pyx":201
 *             init_capacity = (2 ** (tree.max_depth + 1)) - 1
 *         else:
 *             init_capacity = 2047             # <<<<<<<<<<<<<<
 *         tree._resize(init_capacity)
 * 
 */
  /*else*/ {
    __pyx_v_init_capacity = 0x7FF;
  }
  __pyx_L6:;

  /* "skgarden/mondrian/tree/_tree.pyx":202
 *         else:
 *             init_capacity = 2047
 *         tree._resize(init_capacity)             # <<<<<<<<<<<<<<
 * 
 *         # Parameters
 */
  __pyx_t_5 = ((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_tree->__pyx_vtab)->_resize(__pyx_v_tree, __pyx_v_init_capacity); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 202, __pyx_L1_error)

  /* "skgarden/mondrian/tree/_tree.pyx":205
 * 
 *         # Parameters
 *         cdef Splitter splitter = self.splitter             # <<<<<<<<<<<<<<
 *         cdef SIZE_t max_depth = self.max_depth
 *         cdef SIZE_t min_samples_split = self.min_samples_split
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->__pyx_base.splitter);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_splitter = ((struct __pyx_obj_8skgarden_8mondrian_4tree_9_splitter_Splitter *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":206
 *         # Parameters
 *         cdef Splitter splitter = self.splitter
 *         cdef SIZE_t max_depth = self.max_depth             # <<<<<<<<<<<<<<
 *         cdef SIZE_t min_samples_split = self.min_samples_split
 * 
 */
  __pyx_t_10 = __pyx_v_self->__pyx_base.max_depth;
  __pyx_v_max_depth = __pyx_t_10;

  /* "skgarden/mondrian/tree/_tree.pyx":207
 *         cdef Splitter splitter = self.splitter
 *         cdef SIZE_t max_depth = self.max_depth
 *         cdef SIZE_t min_samples_split = self.min_samples_split             # <<<<<<<<<<<<<<
 * 
 *         # Recursive partition (without actual recursion)
 */
  __pyx_t_10 = __pyx_v_self->__pyx_base.min_samples_split;
  __pyx_v_min_samples_split = __pyx_t_10;

  /* "skgarden/mondrian/tree/_tree.pyx":210
 * 
 *         # Recursive partition (without actual recursion)
 *         splitter.init(X, y, sample_weight_ptr, X_idx_sorted)             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t start
 */
  __pyx_t_11.__pyx_n = 1;
  __pyx_t_11.X_idx_sorted = __pyx_v_X_idx_sorted;
  __pyx_t_5 = ((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_9_splitter_Splitter *)__pyx_v_splitter->__pyx_vtab)->init(__pyx_v_splitter, __pyx_v_X, __pyx_v_y, __pyx_v_sample_weight_ptr, &__pyx_t_11); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 210, __pyx_L1_error)

  /* "skgarden/mondrian/tree/_tree.pyx":217
 *         cdef SIZE_t parent
 *         cdef bint is_left
 *         cdef SIZE_t n_node_samples = splitter.n_samples             # <<<<<<<<<<<<<<
 *         cdef double weighted_n_samples = splitter.weighted_n_samples
 *         cdef double weighted_n_node_samples
 */
  __pyx_t_12 = __pyx_v_splitter->n_samples;
  __pyx_v_n_node_samples = __pyx_t_12;

  /* "skgarden/mondrian/tree/_tree.pyx":218
 *         cdef bint is_left
 *         cdef SIZE_t n_node_samples = splitter.n_samples
 *         cdef double weighted_n_samples = splitter.weighted_n_samples             # <<<<<<<<<<<<<<
 *         cdef double weighted_n_node_samples
 *         cdef SplitRecord split
 */
  __pyx_t_13 = __pyx_v_splitter->weighted_n_samples;
  __pyx_v_weighted_n_samples = __pyx_t_13;

  /* "skgarden/mondrian/tree/_tree.pyx":224
 * 
 *         cdef double threshold
 *         cdef double impurity = INFINITY             # <<<<<<<<<<<<<<
 *         cdef SIZE_t n_constant_features
 *         cdef bint is_leaf
 */
  __pyx_v_impurity = __pyx_v_8skgarden_8mondrian_4tree_5_tree_INFINITY;

  /* "skgarden/mondrian/tree/_tree.pyx":227
 *         cdef SIZE_t n_constant_features
 *         cdef bint is_leaf
 *         cdef bint first = 1             # <<<<<<<<<<<<<<
 *         cdef SIZE_t max_depth_seen = -1
 *         cdef int rc = 0
 */
  __pyx_v_first = 1;

  /* "skgarden/mondrian/tree/_tree.pyx":228
 *         cdef bint is_leaf
 *         cdef bint first = 1
 *         cdef SIZE_t max_depth_seen = -1             # <<<<<<<<<<<<<<
 *         cdef int rc = 0
 * 
 */
  __pyx_v_max_depth_seen = -1;

  /* "skgarden/mondrian/tree/_tree.pyx":229
 *         cdef bint first = 1
 *         cdef SIZE_t max_depth_seen = -1
 *         cdef int rc = 0             # <<<<<<<<<<<<<<
 * 
 *         cdef Stack stack = Stack(INITIAL_STACK_SIZE)
 */
  __pyx_v_rc = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":231
 *         cdef int rc = 0
 * 
 *         cdef Stack stack = Stack(INITIAL_STACK_SIZE)             # <<<<<<<<<<<<<<
 *         cdef StackRecord stack_record
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_From_Py_intptr_t(__pyx_v_8skgarden_8mondrian_4tree_5_tree_INITIAL_STACK_SIZE); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_8skgarden_8mondrian_4tree_6_utils_Stack), __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_stack = ((struct __pyx_obj_8skgarden_8mondrian_4tree_6_utils_Stack *)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":234
 *         cdef StackRecord stack_record
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             # push root node onto stack
 *             rc = stack.push(0, n_node_samples, 0, _TREE_UNDEFINED, 0, INFINITY, 0)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "skgarden/mondrian/tree/_tree.pyx":236
 *         with nogil:
 *             # push root node onto stack
 *             rc = stack.push(0, n_node_samples, 0, _TREE_UNDEFINED, 0, INFINITY, 0)             # <<<<<<<<<<<<<<
 *             if rc == -1:
 *                 # got return code -1 - out-of-memory
 */
        __pyx_t_5 = ((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_6_utils_Stack *)__pyx_v_stack->__pyx_vtab)->push(__pyx_v_stack, 0, __pyx_v_n_node_samples, 0, __pyx_v_8skgarden_8mondrian_4tree_5_tree__TREE_UNDEFINED, 0, __pyx_v_8skgarden_8mondrian_4tree_5_tree_INFINITY, 0); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 236, __pyx_L8_error)
        __pyx_v_rc = __pyx_t_5;

        /* "skgarden/mondrian/tree/_tree.pyx":237
 *             # push root node onto stack
 *             rc = stack.push(0, n_node_samples, 0, _TREE_UNDEFINED, 0, INFINITY, 0)
 *             if rc == -1:             # <<<<<<<<<<<<<<
 *                 # got return code -1 - out-of-memory
 *                 with gil:
 */
        __pyx_t_9 = ((__pyx_v_rc == -1L) != 0);
        if (__pyx_t_9) {

          /* "skgarden/mondrian/tree/_tree.pyx":239
 *             if rc == -1:
 *                 # got return code -1 - out-of-memory
 *                 with gil:             # <<<<<<<<<<<<<<
 *                     raise MemoryError()
 * 
 */
          {
              #ifdef WITH_THREAD
              PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
              #endif
              /*try:*/ {

                /* "skgarden/mondrian/tree/_tree.pyx":240
 *                 # got return code -1 - out-of-memory
 *                 with gil:
 *                     raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *             while not stack.is_empty():
 */
                PyErr_NoMemory(); __PYX_ERR(0, 240, __pyx_L12_error)
              }

              /* "skgarden/mondrian/tree/_tree.pyx":239
 *             if rc == -1:
 *                 # got return code -1 - out-of-memory
 *                 with gil:             # <<<<<<<<<<<<<<
 *                     raise MemoryError()
 * 
 */
              /*finally:*/ {
                __pyx_L12_error: {
                  #ifdef WITH_THREAD
                  __Pyx_PyGILState_Release(__pyx_gilstate_save);
                  #endif
                  goto __pyx_L8_error;
                }
              }
          }

          /* "skgarden/mondrian/tree/_tree.pyx":237
 *             # push root node onto stack
 *             rc = stack.push(0, n_node_samples, 0, _TREE_UNDEFINED, 0, INFINITY, 0)
 *             if rc == -1:             # <<<<<<<<<<<<<<
 *                 # got return code -1 - out-of-memory
 *                 with gil:
 */
        }

        /* "skgarden/mondrian/tree/_tree.pyx":242
 *                     raise MemoryError()
 * 
 *             while not stack.is_empty():             # <<<<<<<<<<<<<<
 *                 stack.pop(&stack_record)
 * 
 */
        while (1) {
          __pyx_t_9 = ((!(((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_6_utils_Stack *)__pyx_v_stack->__pyx_vtab)->is_empty(__pyx_v_stack) != 0)) != 0);
          if (!__pyx_t_9) break;

          /* "skgarden/mondrian/tree/_tree.pyx":243
 * 
 *             while not stack.is_empty():
 *                 stack.pop(&stack_record)             # <<<<<<<<<<<<<<
 * 
 *                 start = stack_record.start
 */
          (void)(((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_6_utils_Stack *)__pyx_v_stack->__pyx_vtab)->pop(__pyx_v_stack, (&__pyx_v_stack_record)));

          /* "skgarden/mondrian/tree/_tree.pyx":245
 *                 stack.pop(&stack_record)
 * 
 *                 start = stack_record.start             # <<<<<<<<<<<<<<
 *                 end = stack_record.end
 *                 depth = stack_record.depth
 */
          __pyx_t_14 = __pyx_v_stack_record.start;
          __pyx_v_start = __pyx_t_14;

          /* "skgarden/mondrian/tree/_tree.pyx":246
 * 
 *                 start = stack_record.start
 *                 end = stack_record.end             # <<<<<<<<<<<<<<
 *                 depth = stack_record.depth
 *                 parent = stack_record.parent
 */
          __pyx_t_14 = __pyx_v_stack_record.end;
          __pyx_v_end = __pyx_t_14;

          /* "skgarden/mondrian/tree/_tree.pyx":247
 *                 start = stack_record.start
 *                 end = stack_record.end
 *                 depth = stack_record.depth             # <<<<<<<<<<<<<<
 *                 parent = stack_record.parent
 *                 is_left = stack_record.is_left
 */
          __pyx_t_14 = __pyx_v_stack_record.depth;
          __pyx_v_depth = __pyx_t_14;

          /* "skgarden/mondrian/tree/_tree.pyx":248
 *                 end = stack_record.end
 *                 depth = stack_record.depth
 *                 parent = stack_record.parent             # <<<<<<<<<<<<<<
 *                 is_left = stack_record.is_left
 *                 impurity = stack_record.impurity
 */
          __pyx_t_14 = __pyx_v_stack_record.parent;
          __pyx_v_parent = __pyx_t_14;

          /* "skgarden/mondrian/tree/_tree.pyx":249
 *                 depth = stack_record.depth
 *                 parent = stack_record.parent
 *                 is_left = stack_record.is_left             # <<<<<<<<<<<<<<
 *                 impurity = stack_record.impurity
 *                 n_constant_features = stack_record.n_constant_features
 */
          __pyx_t_9 = __pyx_v_stack_record.is_left;
          __pyx_v_is_left = __pyx_t_9;

          /* "skgarden/mondrian/tree/_tree.pyx":250
 *                 parent = stack_record.parent
 *                 is_left = stack_record.is_left
 *                 impurity = stack_record.impurity             # <<<<<<<<<<<<<<
 *                 n_constant_features = stack_record.n_constant_features
 * 
 */
          __pyx_t_13 = __pyx_v_stack_record.impurity;
          __pyx_v_impurity = __pyx_t_13;

          /* "skgarden/mondrian/tree/_tree.pyx":251
 *                 is_left = stack_record.is_left
 *                 impurity = stack_record.impurity
 *                 n_constant_features = stack_record.n_constant_features             # <<<<<<<<<<<<<<
 * 
 *                 n_node_samples = end - start
 */
          __pyx_t_14 = __pyx_v_stack_record.n_constant_features;
          __pyx_v_n_constant_features = __pyx_t_14;

          /* "skgarden/mondrian/tree/_tree.pyx":253
 *                 n_constant_features = stack_record.n_constant_features
 * 
 *                 n_node_samples = end - start             # <<<<<<<<<<<<<<
 *                 splitter.node_reset(start, end, &weighted_n_node_samples)
 * 
 */
          __pyx_v_n_node_samples = (__pyx_v_end - __pyx_v_start);

          /* "skgarden/mondrian/tree/_tree.pyx":254
 * 
 *                 n_node_samples = end - start
 *                 splitter.node_reset(start, end, &weighted_n_node_samples)             # <<<<<<<<<<<<<<
 * 
 *                 if first:
 */
          __pyx_t_5 = ((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_9_splitter_Splitter *)__pyx_v_splitter->__pyx_vtab)->node_reset(__pyx_v_splitter, __pyx_v_start, __pyx_v_end, (&__pyx_v_weighted_n_node_samples)); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 254, __pyx_L8_error)

          /* "skgarden/mondrian/tree/_tree.pyx":256
 *                 splitter.node_reset(start, end, &weighted_n_node_samples)
 * 
 *                 if first:             # <<<<<<<<<<<<<<
 *                     impurity = splitter.node_impurity()
 *                     first = 0
 */
          __pyx_t_9 = (__pyx_v_first != 0);
          if (__pyx_t_9) {

            /* "skgarden/mondrian/tree/_tree.pyx":257
 * 
 *                 if first:
 *                     impurity = splitter.node_impurity()             # <<<<<<<<<<<<<<
 *                     first = 0
 * 
 */
            __pyx_v_impurity = ((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_9_splitter_Splitter *)__pyx_v_splitter->__pyx_vtab)->node_impurity(__pyx_v_splitter);

            /* "skgarden/mondrian/tree/_tree.pyx":258
 *                 if first:
 *                     impurity = splitter.node_impurity()
 *                     first = 0             # <<<<<<<<<<<<<<
 * 
 *                 is_leaf = (depth >= max_depth or
 */
            __pyx_v_first = 0;

            /* "skgarden/mondrian/tree/_tree.pyx":256
 *                 splitter.node_reset(start, end, &weighted_n_node_samples)
 * 
 *                 if first:             # <<<<<<<<<<<<<<
 *                     impurity = splitter.node_impurity()
 *                     first = 0
 */
          }

          /* "skgarden/mondrian/tree/_tree.pyx":260
 *                     first = 0
 * 
 *                 is_leaf = (depth >= max_depth or             # <<<<<<<<<<<<<<
 *                            n_node_samples < min_samples_split)
 * 
 */
          __pyx_t_8 = ((__pyx_v_depth >= __pyx_v_max_depth) != 0);
          if (!__pyx_t_8) {
          } else {
            __pyx_t_9 = __pyx_t_8;
            goto __pyx_L17_bool_binop_done;
          }

          /* "skgarden/mondrian/tree/_tree.pyx":261
 * 
 *                 is_leaf = (depth >= max_depth or
 *                            n_node_samples < min_samples_split)             # <<<<<<<<<<<<<<
 * 
 *                 if not is_leaf:
 */
          __pyx_t_8 = ((__pyx_v_n_node_samples < __pyx_v_min_samples_split) != 0);
          __pyx_t_9 = __pyx_t_8;
          __pyx_L17_bool_binop_done:;
          __pyx_v_is_leaf = __pyx_t_9;

          /* "skgarden/mondrian/tree/_tree.pyx":263
 *                            n_node_samples < min_samples_split)
 * 
 *                 if not is_leaf:             # <<<<<<<<<<<<<<
 *                     is_leaf = splitter.node_split(impurity, &split, &n_constant_features)
 *                     is_leaf = is_leaf or (split.pos >= end)
 */
          __pyx_t_9 = ((!(__pyx_v_is_leaf != 0)) != 0);
          if (__pyx_t_9) {

            /* "skgarden/mondrian/tree/_tree.pyx":264
 * 
 *                 if not is_leaf:
 *                     is_leaf = splitter.node_split(impurity, &split, &n_constant_features)             # <<<<<<<<<<<<<<
 *                     is_leaf = is_leaf or (split.pos >= end)
 *                 else:
 */
            __pyx_t_5 = ((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_9_splitter_Splitter *)__pyx_v_splitter->__pyx_vtab)->node_split(__pyx_v_splitter, __pyx_v_impurity, (&__pyx_v_split), (&__pyx_v_n_constant_features)); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 264, __pyx_L8_error)
            __pyx_v_is_leaf = __pyx_t_5;

            /* "skgarden/mondrian/tree/_tree.pyx":265
 *                 if not is_leaf:
 *                     is_leaf = splitter.node_split(impurity, &split, &n_constant_features)
 *                     is_leaf = is_leaf or (split.pos >= end)             # <<<<<<<<<<<<<<
 *                 else:
 *                     splitter.set_bounds()
 */
            __pyx_t_8 = (__pyx_v_is_leaf != 0);
            if (!__pyx_t_8) {
            } else {
              __pyx_t_9 = __pyx_t_8;
              goto __pyx_L20_bool_binop_done;
            }
            __pyx_t_8 = ((__pyx_v_split.pos >= __pyx_v_end) != 0);
            __pyx_t_9 = __pyx_t_8;
            __pyx_L20_bool_binop_done:;
            __pyx_v_is_leaf = __pyx_t_9;

            /* "skgarden/mondrian/tree/_tree.pyx":263
 *                            n_node_samples < min_samples_split)
 * 
 *                 if not is_leaf:             # <<<<<<<<<<<<<<
 *                     is_leaf = splitter.node_split(impurity, &split, &n_constant_features)
 *                     is_leaf = is_leaf or (split.pos >= end)
 */
            goto __pyx_L19;
          }

          /* "skgarden/mondrian/tree/_tree.pyx":267
 *                     is_leaf = is_leaf or (split.pos >= end)
 *                 else:
 *                     splitter.set_bounds()             # <<<<<<<<<<<<<<
 * 
 *                 # Check if the node is pure.
 */
          /*else*/ {
            ((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_9_splitter_Splitter *)__pyx_v_splitter->__pyx_vtab)->set_bounds(__pyx_v_splitter);
          }
          __pyx_L19:;

          /* "skgarden/mondrian/tree/_tree.pyx":270
 * 
 *                 # Check if the node is pure.
 *                 is_leaf = is_leaf or splitter.criterion.is_pure()             # <<<<<<<<<<<<<<
 * 
 *                 node_id = tree._add_node(parent, is_left, is_leaf, split.feature,
 */
          __pyx_t_8 = (__pyx_v_is_leaf != 0);
          if (!__pyx_t_8) {
          } else {
            __pyx_t_9 = __pyx_t_8;
            goto __pyx_L22_bool_binop_done;
          }
          __pyx_t_8 = (((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_10_criterion_Criterion *)__pyx_v_splitter->criterion->__pyx_vtab)->is_pure(__pyx_v_splitter->criterion) != 0);
          __pyx_t_9 = __pyx_t_8;
          __pyx_L22_bool_binop_done:;
          __pyx_v_is_leaf = __pyx_t_9;

          /* "skgarden/mondrian/tree/_tree.pyx":272
 *                 is_leaf = is_leaf or splitter.criterion.is_pure()
 * 
 *                 node_id = tree._add_node(parent, is_left, is_leaf, split.feature,             # <<<<<<<<<<<<<<
 *                                          split.threshold, impurity, n_node_samples,
 *                                          weighted_n_node_samples,
 */
          __pyx_t_10 = ((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_tree->__pyx_vtab)->_add_node(__pyx_v_tree, __pyx_v_parent, __pyx_v_is_left, __pyx_v_is_leaf, __pyx_v_split.feature, __pyx_v_split.threshold, __pyx_v_impurity, __pyx_v_n_node_samples, __pyx_v_weighted_n_node_samples, __pyx_v_splitter->lower_bounds, __pyx_v_splitter->upper_bounds, __pyx_v_split.E); if (unlikely(__pyx_t_10 == ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t)-1))) __PYX_ERR(0, 272, __pyx_L8_error)
          __pyx_v_node_id = __pyx_t_10;

          /* "skgarden/mondrian/tree/_tree.pyx":279
 *                                          split.E)
 * 
 *                 if node_id == <SIZE_t>(-1):             # <<<<<<<<<<<<<<
 *                     rc = -1
 *                     break
 */
          __pyx_t_9 = ((__pyx_v_node_id == ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t)-1L)) != 0);
          if (__pyx_t_9) {

            /* "skgarden/mondrian/tree/_tree.pyx":280
 * 
 *                 if node_id == <SIZE_t>(-1):
 *                     rc = -1             # <<<<<<<<<<<<<<
 *                     break
 * 
 */
            __pyx_v_rc = -1;

            /* "skgarden/mondrian/tree/_tree.pyx":281
 *                 if node_id == <SIZE_t>(-1):
 *                     rc = -1
 *                     break             # <<<<<<<<<<<<<<
 * 
 *                 # Store value for all nodes, to facilitate tree/model
 */
            goto __pyx_L15_break;

            /* "skgarden/mondrian/tree/_tree.pyx":279
 *                                          split.E)
 * 
 *                 if node_id == <SIZE_t>(-1):             # <<<<<<<<<<<<<<
 *                     rc = -1
 *                     break
 */
          }

          /* "skgarden/mondrian/tree/_tree.pyx":285
 *                 # Store value for all nodes, to facilitate tree/model
 *                 # inspection and interpretation
 *                 splitter.node_value(tree.value + node_id * tree.value_stride)             # <<<<<<<<<<<<<<
 * 
 *                 if not is_leaf:
 */
          ((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_9_splitter_Splitter *)__pyx_v_splitter->__pyx_vtab)->node_value(__pyx_v_splitter, (__pyx_v_tree->value + (__pyx_v_node_id * __pyx_v_tree->value_stride)));

          /* "skgarden/mondrian/tree/_tree.pyx":287
 *                 splitter.node_value(tree.value + node_id * tree.value_stride)
 * 
 *                 if not is_leaf:             # <<<<<<<<<<<<<<
 *                     # Push right child on stack
 *                     rc = stack.push(split.pos, end, depth + 1, node_id, 0,
 */
          __pyx_t_9 = ((!(__pyx_v_is_leaf != 0)) != 0);
          if (__pyx_t_9) {

            /* "skgarden/mondrian/tree/_tree.pyx":289
 *                 if not is_leaf:
 *                     # Push right child on stack
 *                     rc = stack.push(split.pos, end, depth + 1, node_id, 0,             # <<<<<<<<<<<<<<
 *                                     split.impurity_right, n_constant_features)
 *                     if rc == -1:
 */
            __pyx_t_5 = ((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_6_utils_Stack *)__pyx_v_stack->__pyx_vtab)->push(__pyx_v_stack, __pyx_v_split.pos, __pyx_v_end, (__pyx_v_depth + 1), __pyx_v_node_id, 0, __pyx_v_split.impurity_right, __pyx_v_n_constant_features); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 289, __pyx_L8_error)
            __pyx_v_rc = __pyx_t_5;

            /* "skgarden/mondrian/tree/_tree.pyx":291
 *                     rc = stack.push(split.pos, end, depth + 1, node_id, 0,
 *                                     split.impurity_right, n_constant_features)
 *                     if rc == -1:             # <<<<<<<<<<<<<<
 *                         break
 * 
 */
            __pyx_t_9 = ((__pyx_v_rc == -1L) != 0);
            if (__pyx_t_9) {

              /* "skgarden/mondrian/tree/_tree.pyx":292
 *                                     split.impurity_right, n_constant_features)
 *                     if rc == -1:
 *                         break             # <<<<<<<<<<<<<<
 * 
 *                     # Push left child on stack
 */
              goto __pyx_L15_break;

              /* "skgarden/mondrian/tree/_tree.pyx":291
 *                     rc = stack.push(split.pos, end, depth + 1, node_id, 0,
 *                                     split.impurity_right, n_constant_features)
 *                     if rc == -1:             # <<<<<<<<<<<<<<
 *                         break
 * 
 */
            }

            /* "skgarden/mondrian/tree/_tree.pyx":295
 * 
 *                     # Push left child on stack
 *                     rc = stack.push(start, split.pos, depth + 1, node_id, 1,             # <<<<<<<<<<<<<<
 *                                     split.impurity_left, n_constant_features)
 *                     if rc == -1:
 */
            __pyx_t_5 = ((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_6_utils_Stack *)__pyx_v_stack->__pyx_vtab)->push(__pyx_v_stack, __pyx_v_start, __pyx_v_split.pos, (__pyx_v_depth + 1), __pyx_v_node_id, 1, __pyx_v_split.impurity_left, __pyx_v_n_constant_features); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 295, __pyx_L8_error)
            __pyx_v_rc = __pyx_t_5;

            /* "skgarden/mondrian/tree/_tree.pyx":297
 *                     rc = stack.push(start, split.pos, depth + 1, node_id, 1,
 *                                     split.impurity_left, n_constant_features)
 *                     if rc == -1:             # <<<<<<<<<<<<<<
 *                         break
 * 
 */
            __pyx_t_9 = ((__pyx_v_rc == -1L) != 0);
            if (__pyx_t_9) {

              /* "skgarden/mondrian/tree/_tree.pyx":298
 *                                     split.impurity_left, n_constant_features)
 *                     if rc == -1:
 *                         break             # <<<<<<<<<<<<<<
 * 
 *                 if depth > max_depth_seen:
 */
              goto __pyx_L15_break;

              /* "skgarden/mondrian/tree/_tree.pyx":297
 *                     rc = stack.push(start, split.pos, depth + 1, node_id, 1,
 *                                     split.impurity_left, n_constant_features)
 *                     if rc == -1:             # <<<<<<<<<<<<<<
 *                         break
 * 
 */
            }

            /* "skgarden/mondrian/tree/_tree.pyx":287
 *                 splitter.node_value(tree.value + node_id * tree.value_stride)
 * 
 *                 if not is_leaf:             # <<<<<<<<<<<<<<
 *                     # Push right child on stack
 *                     rc = stack.push(split.pos, end, depth + 1, node_id, 0,
 */
          }

          /* "skgarden/mondrian/tree/_tree.pyx":300
 *                         break
 * 
 *                 if depth > max_depth_seen:             # <<<<<<<<<<<<<<
 *                     max_depth_seen = depth
 * 
 */
          __pyx_t_9 = ((__pyx_v_depth > __pyx_v_max_depth_seen) != 0);
          if (__pyx_t_9) {

            /* "skgarden/mondrian/tree/_tree.pyx":301
 * 
 *                 if depth > max_depth_seen:
 *                     max_depth_seen = depth             # <<<<<<<<<<<<<<
 * 
 *             if rc >= 0:
 */
            __pyx_v_max_depth_seen = __pyx_v_depth;

            /* "skgarden/mondrian/tree/_tree.pyx":300
 *                         break
 * 
 *                 if depth > max_depth_seen:             # <<<<<<<<<<<<<<
 *                     max_depth_seen = depth
 * 
 */
          }
        }
        __pyx_L15_break:;

        /* "skgarden/mondrian/tree/_tree.pyx":303
 *                     max_depth_seen = depth
 * 
 *             if rc >= 0:             # <<<<<<<<<<<<<<
 *                 rc = tree._resize_c(tree.node_count)
 * 
 */
        __pyx_t_9 = ((__pyx_v_rc >= 0) != 0);
        if (__pyx_t_9) {

          /* "skgarden/mondrian/tree/_tree.pyx":304
 * 
 *             if rc >= 0:
 *                 rc = tree._resize_c(tree.node_count)             # <<<<<<<<<<<<<<
 * 
 *             if rc >= 0:
 */
          __pyx_t_15.__pyx_n = 1;
          __pyx_t_15.capacity = __pyx_v_tree->node_count;
          __pyx_t_5 = ((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_tree->__pyx_vtab)->_resize_c(__pyx_v_tree, &__pyx_t_15); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 304, __pyx_L8_error)
          __pyx_v_rc = __pyx_t_5;

          /* "skgarden/mondrian/tree/_tree.pyx":303
 *                     max_depth_seen = depth
 * 
 *             if rc >= 0:             # <<<<<<<<<<<<<<
 *                 rc = tree._resize_c(tree.node_count)
 * 
 */
        }

        /* "skgarden/mondrian/tree/_tree.pyx":306
 *                 rc = tree._resize_c(tree.node_count)
 * 
 *             if rc >= 0:             # <<<<<<<<<<<<<<
 *                 tree.max_depth = max_depth_seen
 *         if rc == -1:
 */
        __pyx_t_9 = ((__pyx_v_rc >= 0) != 0);
        if (__pyx_t_9) {

          /* "skgarden/mondrian/tree/_tree.pyx":307
 * 
 *             if rc >= 0:
 *                 tree.max_depth = max_depth_seen             # <<<<<<<<<<<<<<
 *         if rc == -1:
 *             raise MemoryError()
 */
          __pyx_v_tree->max_depth = __pyx_v_max_depth_seen;

          /* "skgarden/mondrian/tree/_tree.pyx":306
 *                 rc = tree._resize_c(tree.node_count)
 * 
 *             if rc >= 0:             # <<<<<<<<<<<<<<
 *                 tree.max_depth = max_depth_seen
 *         if rc == -1:
 */
        }
      }

      /* "skgarden/mondrian/tree/_tree.pyx":234
 *         cdef StackRecord stack_record
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             # push root node onto stack
 *             rc = stack.push(0, n_node_samples, 0, _TREE_UNDEFINED, 0, INFINITY, 0)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L9;
        }
        __pyx_L8_error: {
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L1_error;
        }
        __pyx_L9:;
      }
  }

  /* "skgarden/mondrian/tree/_tree.pyx":308
 *             if rc >= 0:
 *                 tree.max_depth = max_depth_seen
 *         if rc == -1:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_9 = ((__pyx_v_rc == -1L) != 0);
  if (unlikely(__pyx_t_9)) {

    /* "skgarden/mondrian/tree/_tree.pyx":309
 *                 tree.max_depth = max_depth_seen
 *         if rc == -1:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 * # =============================================================================
 */
    PyErr_NoMemory(); __PYX_ERR(0, 309, __pyx_L1_error)

    /* "skgarden/mondrian/tree/_tree.pyx":308
 *             if rc >= 0:
 *                 tree.max_depth = max_depth_seen
 *         if rc == -1:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "skgarden/mondrian/tree/_tree.pyx":184
 *         self.max_depth = max_depth
 * 
 *     cpdef build(self, Tree tree, object X, np.ndarray y,             # <<<<<<<<<<<<<<
 *                 np.ndarray sample_weight=None,
 *                 np.ndarray X_idx_sorted=None):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.DepthFirstTreeBuilder.build", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_splitter);
  __Pyx_XDECREF((PyObject *)__pyx_v_stack);
  __Pyx_XDECREF(__pyx_v_X);
  __Pyx_XDECREF((PyObject *)__pyx_v_y);
  __Pyx_XDECREF((PyObject *)__pyx_v_sample_weight);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder_3build(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder_2build[] = "Build a decision tree from the training set (X, y).";
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder_3build(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_tree = 0;
  PyObject *__pyx_v_X = 0;
  PyArrayObject *__pyx_v_y = 0;
  PyArrayObject *__pyx_v_sample_weight = 0;
  PyArrayObject *__pyx_v_X_idx_sorted = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("build (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_tree,&__pyx_n_s_X,&__pyx_n_s_y,&__pyx_n_s_sample_weight,&__pyx_n_s_X_idx_sorted,0};
    PyObject* values[5] = {0,0,0,0,0};

    /* "skgarden/mondrian/tree/_tree.pyx":185
 * 
 *     cpdef build(self, Tree tree, object X, np.ndarray y,
 *                 np.ndarray sample_weight=None,             # <<<<<<<<<<<<<<
 *                 np.ndarray X_idx_sorted=None):
 *         """Build a decision tree from the training set (X, y)."""
 */
    values[3] = (PyObject *)((PyArrayObject *)Py_None);

    /* "skgarden/mondrian/tree/_tree.pyx":186
 *     cpdef build(self, Tree tree, object X, np.ndarray y,
 *                 np.ndarray sample_weight=None,
 *                 np.ndarray X_idx_sorted=None):             # <<<<<<<<<<<<<<
 *         """Build a decision tree from the training set (X, y)."""
 * 
 */
    values[4] = (PyObject *)((PyArrayObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tree)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("build", 0, 3, 5, 1); __PYX_ERR(0, 184, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("build", 0, 3, 5, 2); __PYX_ERR(0, 184, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sample_weight);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X_idx_sorted);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "build") < 0)) __PYX_ERR(0, 184, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_tree = ((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)values[0]);
    __pyx_v_X = values[1];
    __pyx_v_y = ((PyArrayObject *)values[2]);
    __pyx_v_sample_weight = ((PyArrayObject *)values[3]);
    __pyx_v_X_idx_sorted = ((PyArrayObject *)values[4]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("build", 0, 3, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 184, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.DepthFirstTreeBuilder.build", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_tree), __pyx_ptype_8skgarden_8mondrian_4tree_5_tree_Tree, 1, "tree", 0))) __PYX_ERR(0, 184, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_y), __pyx_ptype_5numpy_ndarray, 1, "y", 0))) __PYX_ERR(0, 184, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_sample_weight), __pyx_ptype_5numpy_ndarray, 1, "sample_weight", 0))) __PYX_ERR(0, 185, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_X_idx_sorted), __pyx_ptype_5numpy_ndarray, 1, "X_idx_sorted", 0))) __PYX_ERR(0, 186, __pyx_L1_error)
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder_2build(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder *)__pyx_v_self), __pyx_v_tree, __pyx_v_X, __pyx_v_y, __pyx_v_sample_weight, __pyx_v_X_idx_sorted);

  /* "skgarden/mondrian/tree/_tree.pyx":184
 *         self.max_depth = max_depth
 * 
 *     cpdef build(self, Tree tree, object X, np.ndarray y,             # <<<<<<<<<<<<<<
 *                 np.ndarray sample_weight=None,
 *                 np.ndarray X_idx_sorted=None):
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder_2build(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder *__pyx_v_self, struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_tree, PyObject *__pyx_v_X, PyArrayObject *__pyx_v_y, PyArrayObject *__pyx_v_sample_weight, PyArrayObject *__pyx_v_X_idx_sorted) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_build __pyx_t_2;
  __Pyx_RefNannySetupContext("build", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 2;
  __pyx_t_2.sample_weight = __pyx_v_sample_weight;
  __pyx_t_2.X_idx_sorted = __pyx_v_X_idx_sorted;
  __pyx_t_1 = __pyx_vtabptr_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder->__pyx_base.build(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *)__pyx_v_self), __pyx_v_tree, __pyx_v_X, __pyx_v_y, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.DepthFirstTreeBuilder.build", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder_4__reduce_cython__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.DepthFirstTreeBuilder.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder_6__setstate_cython__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.DepthFirstTreeBuilder.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":374
 *     # (i.e. through `_resize` or `__setstate__`)
 *     property n_classes:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return sizet_ptr_to_ndarray(self.n_classes, self.n_outputs)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_9n_classes_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_9n_classes_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_9n_classes___get__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_9n_classes___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "skgarden/mondrian/tree/_tree.pyx":375
 *     property n_classes:
 *         def __get__(self):
 *             return sizet_ptr_to_ndarray(self.n_classes, self.n_outputs)             # <<<<<<<<<<<<<<
 * 
 *     property children_left:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_8skgarden_8mondrian_4tree_6_utils_sizet_ptr_to_ndarray(__pyx_v_self->n_classes, __pyx_v_self->n_outputs)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "skgarden/mondrian/tree/_tree.pyx":374
 *     # (i.e. through `_resize` or `__setstate__`)
 *     property n_classes:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return sizet_ptr_to_ndarray(self.n_classes, self.n_outputs)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.n_classes.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":378
 * 
 *     property children_left:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._get_node_ndarray()['left_child'][:self.node_count]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_13children_left_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_13children_left_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_13children_left___get__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_13children_left___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "skgarden/mondrian/tree/_tree.pyx":379
 *     property children_left:
 *         def __get__(self):
 *             return self._get_node_ndarray()['left_child'][:self.node_count]             # <<<<<<<<<<<<<<
 * 
 *     property children_right:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->_get_node_ndarray(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_n_s_left_child); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_t_2, 0, __pyx_v_self->node_count, NULL, NULL, NULL, 0, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "skgarden/mondrian/tree/_tree.pyx":378
 * 
 *     property children_left:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._get_node_ndarray()['left_child'][:self.node_count]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.children_left.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":382
 * 
 *     property children_right:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._get_node_ndarray()['right_child'][:self.node_count]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_14children_right_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_14children_right_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_14children_right___get__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_14children_right___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "skgarden/mondrian/tree/_tree.pyx":383
 *     property children_right:
 *         def __get__(self):
 *             return self._get_node_ndarray()['right_child'][:self.node_count]             # <<<<<<<<<<<<<<
 * 
 *     property feature:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->_get_node_ndarray(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_n_s_right_child); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_t_2, 0, __pyx_v_self->node_count, NULL, NULL, NULL, 0, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "skgarden/mondrian/tree/_tree.pyx":382
 * 
 *     property children_right:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._get_node_ndarray()['right_child'][:self.node_count]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.children_right.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":386
 * 
 *     property feature:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._get_node_ndarray()['feature'][:self.node_count]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_7feature_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_7feature_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_7feature___get__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_7feature___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "skgarden/mondrian/tree/_tree.pyx":387
 *     property feature:
 *         def __get__(self):
 *             return self._get_node_ndarray()['feature'][:self.node_count]             # <<<<<<<<<<<<<<
 * 
 *     property threshold:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->_get_node_ndarray(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_n_s_feature); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_t_2, 0, __pyx_v_self->node_count, NULL, NULL, NULL, 0, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "skgarden/mondrian/tree/_tree.pyx":386
 * 
 *     property feature:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._get_node_ndarray()['feature'][:self.node_count]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.feature.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":390
 * 
 *     property threshold:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._get_node_ndarray()['threshold'][:self.node_count]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_9threshold_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_9threshold_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_9threshold___get__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_9threshold___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "skgarden/mondrian/tree/_tree.pyx":391
 *     property threshold:
 *         def __get__(self):
 *             return self._get_node_ndarray()['threshold'][:self.node_count]             # <<<<<<<<<<<<<<
 * 
 *     property impurity:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->_get_node_ndarray(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_n_s_threshold); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_t_2, 0, __pyx_v_self->node_count, NULL, NULL, NULL, 0, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "skgarden/mondrian/tree/_tree.pyx":390
 * 
 *     property threshold:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._get_node_ndarray()['threshold'][:self.node_count]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.threshold.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":394
 * 
 *     property impurity:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._get_node_ndarray()['impurity'][:self.node_count]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_8impurity_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_8impurity_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_8impurity___get__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_8impurity___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "skgarden/mondrian/tree/_tree.pyx":395
 *     property impurity:
 *         def __get__(self):
 *             return self._get_node_ndarray()['impurity'][:self.node_count]             # <<<<<<<<<<<<<<
 * 
 *     property n_node_samples:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->_get_node_ndarray(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_n_s_impurity); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_t_2, 0, __pyx_v_self->node_count, NULL, NULL, NULL, 0, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "skgarden/mondrian/tree/_tree.pyx":394
 * 
 *     property impurity:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._get_node_ndarray()['impurity'][:self.node_count]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.impurity.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":398
 * 
 *     property n_node_samples:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._get_node_ndarray()['n_node_samples'][:self.node_count]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_14n_node_samples_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_14n_node_samples_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_14n_node_samples___get__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_14n_node_samples___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "skgarden/mondrian/tree/_tree.pyx":399
 *     property n_node_samples:
 *         def __get__(self):
 *             return self._get_node_ndarray()['n_node_samples'][:self.node_count]             # <<<<<<<<<<<<<<
 * 
 *     property weighted_n_node_samples:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->_get_node_ndarray(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_n_s_n_node_samples); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_t_2, 0, __pyx_v_self->node_count, NULL, NULL, NULL, 0, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "skgarden/mondrian/tree/_tree.pyx":398
 * 
 *     property n_node_samples:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._get_node_ndarray()['n_node_samples'][:self.node_count]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.n_node_samples.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":402
 * 
 *     property weighted_n_node_samples:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._get_node_ndarray()['weighted_n_node_samples'][:self.node_count]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_23weighted_n_node_samples_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_23weighted_n_node_samples_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_23weighted_n_node_samples___get__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_23weighted_n_node_samples___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "skgarden/mondrian/tree/_tree.pyx":403
 *     property weighted_n_node_samples:
 *         def __get__(self):
 *             return self._get_node_ndarray()['weighted_n_node_samples'][:self.node_count]             # <<<<<<<<<<<<<<
 * 
 *     property value:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->_get_node_ndarray(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_n_s_weighted_n_node_samples); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_t_2, 0, __pyx_v_self->node_count, NULL, NULL, NULL, 0, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "skgarden/mondrian/tree/_tree.pyx":402
 * 
 *     property weighted_n_node_samples:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._get_node_ndarray()['weighted_n_node_samples'][:self.node_count]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.weighted_n_node_samples.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":406
 * 
 *     property value:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._get_value_ndarray()[:self.node_count]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_5value_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_5value_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_5value___get__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_5value___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "skgarden/mondrian/tree/_tree.pyx":407
 *     property value:
 *         def __get__(self):
 *             return self._get_value_ndarray()[:self.node_count]             # <<<<<<<<<<<<<<
 * 
 *     property tau:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->_get_value_ndarray(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_t_1, 0, __pyx_v_self->node_count, NULL, NULL, NULL, 0, 1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "skgarden/mondrian/tree/_tree.pyx":406
 * 
 *     property value:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._get_value_ndarray()[:self.node_count]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.value.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":410
 * 
 *     property tau:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._get_node_ndarray()["tau"][:self.node_count]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_3tau_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_3tau_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_3tau___get__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_3tau___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "skgarden/mondrian/tree/_tree.pyx":411
 *     property tau:
 *         def __get__(self):
 *             return self._get_node_ndarray()["tau"][:self.node_count]             # <<<<<<<<<<<<<<
 * 
 *     property mean:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->_get_node_ndarray(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_n_s_tau); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_t_2, 0, __pyx_v_self->node_count, NULL, NULL, NULL, 0, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "skgarden/mondrian/tree/_tree.pyx":410
 * 
 *     property tau:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._get_node_ndarray()["tau"][:self.node_count]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.tau.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":414
 * 
 *     property mean:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._get_value_ndarray()[:self.node_count].ravel()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_4mean_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_4mean_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_4mean___get__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_4mean___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "skgarden/mondrian/tree/_tree.pyx":415
 *     property mean:
 *         def __get__(self):
 *             return self._get_value_ndarray()[:self.node_count].ravel()             # <<<<<<<<<<<<<<
 * 
 *     property variance:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->_get_value_ndarray(__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_t_2, 0, __pyx_v_self->node_count, NULL, NULL, NULL, 0, 1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ravel); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "skgarden/mondrian/tree/_tree.pyx":414
 * 
 *     property mean:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._get_value_ndarray()[:self.node_count].ravel()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.mean.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":418
 * 
 *     property variance:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._get_node_ndarray()["variance"][:self.node_count]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_8variance_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_8variance_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_8variance___get__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_8variance___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "skgarden/mondrian/tree/_tree.pyx":419
 *     property variance:
 *         def __get__(self):
 *             return self._get_node_ndarray()["variance"][:self.node_count]             # <<<<<<<<<<<<<<
 * 
 *     property root:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->_get_node_ndarray(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 419, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_n_s_variance); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 419, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_t_2, 0, __pyx_v_self->node_count, NULL, NULL, NULL, 0, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 419, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "skgarden/mondrian/tree/_tree.pyx":418
 * 
 *     property variance:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._get_node_ndarray()["variance"][:self.node_count]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.variance.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":422
 * 
 *     property root:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.root
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_4root_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_4root_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_4root___get__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_4root___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "skgarden/mondrian/tree/_tree.pyx":423
 *     property root:
 *         def __get__(self):
 *             return self.root             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self, int n_features, np.ndarray[SIZE_t, ndim=1] n_classes,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_Py_intptr_t(__pyx_v_self->root); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "skgarden/mondrian/tree/_tree.pyx":422
 * 
 *     property root:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.root
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.root.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":425
 *             return self.root
 * 
 *     def __cinit__(self, int n_features, np.ndarray[SIZE_t, ndim=1] n_classes,             # <<<<<<<<<<<<<<
 *                   int n_outputs):
 *         """Constructor."""
 */

/* Python wrapper */
static int __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_n_features;
  PyArrayObject *__pyx_v_n_classes = 0;
  int __pyx_v_n_outputs;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_n_features,&__pyx_n_s_n_classes,&__pyx_n_s_n_outputs,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n_features)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n_classes)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 1); __PYX_ERR(0, 425, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n_outputs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 2); __PYX_ERR(0, 425, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 425, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_n_features = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_n_features == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 425, __pyx_L3_error)
    __pyx_v_n_classes = ((PyArrayObject *)values[1]);
    __pyx_v_n_outputs = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n_outputs == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 426, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 425, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_n_classes), __pyx_ptype_5numpy_ndarray, 1, "n_classes", 0))) __PYX_ERR(0, 425, __pyx_L1_error)
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree___cinit__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self), __pyx_v_n_features, __pyx_v_n_classes, __pyx_v_n_outputs);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree___cinit__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, int __pyx_v_n_features, PyArrayObject *__pyx_v_n_classes, int __pyx_v_n_outputs) {
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_k;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_n_classes;
  __Pyx_Buffer __pyx_pybuffer_n_classes;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  __Pyx_RefNannySetupContext("__cinit__", 0);
  __pyx_pybuffer_n_classes.pybuffer.buf = NULL;
  __pyx_pybuffer_n_classes.refcount = 0;
  __pyx_pybuffernd_n_classes.data = NULL;
  __pyx_pybuffernd_n_classes.rcbuffer = &__pyx_pybuffer_n_classes;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_n_classes.rcbuffer->pybuffer, (PyObject*)__pyx_v_n_classes, &__Pyx_TypeInfo_nn___pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 425, __pyx_L1_error)
  }
  __pyx_pybuffernd_n_classes.diminfo[0].strides = __pyx_pybuffernd_n_classes.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_n_classes.diminfo[0].shape = __pyx_pybuffernd_n_classes.rcbuffer->pybuffer.shape[0];

  /* "skgarden/mondrian/tree/_tree.pyx":429
 *         """Constructor."""
 *         # Input/Output layout
 *         self.n_features = n_features             # <<<<<<<<<<<<<<
 *         self.n_outputs = n_outputs
 *         self.n_classes = NULL
 */
  __pyx_v_self->n_features = __pyx_v_n_features;

  /* "skgarden/mondrian/tree/_tree.pyx":430
 *         # Input/Output layout
 *         self.n_features = n_features
 *         self.n_outputs = n_outputs             # <<<<<<<<<<<<<<
 *         self.n_classes = NULL
 *         safe_realloc(&self.n_classes, n_outputs)
 */
  __pyx_v_self->n_outputs = __pyx_v_n_outputs;

  /* "skgarden/mondrian/tree/_tree.pyx":431
 *         self.n_features = n_features
 *         self.n_outputs = n_outputs
 *         self.n_classes = NULL             # <<<<<<<<<<<<<<
 *         safe_realloc(&self.n_classes, n_outputs)
 * 
 */
  __pyx_v_self->n_classes = NULL;

  /* "skgarden/mondrian/tree/_tree.pyx":432
 *         self.n_outputs = n_outputs
 *         self.n_classes = NULL
 *         safe_realloc(&self.n_classes, n_outputs)             # <<<<<<<<<<<<<<
 * 
 *         self.max_n_classes = np.max(n_classes)
 */
  __pyx_fuse_1__pyx_f_8skgarden_8mondrian_4tree_6_utils_safe_realloc((&__pyx_v_self->n_classes), __pyx_v_n_outputs); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 432, __pyx_L1_error)

  /* "skgarden/mondrian/tree/_tree.pyx":434
 *         safe_realloc(&self.n_classes, n_outputs)
 * 
 *         self.max_n_classes = np.max(n_classes)             # <<<<<<<<<<<<<<
 *         self.value_stride = n_outputs * self.max_n_classes
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_max); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, ((PyObject *)__pyx_v_n_classes)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_n_classes));
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_1); if (unlikely((__pyx_t_4 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->max_n_classes = __pyx_t_4;

  /* "skgarden/mondrian/tree/_tree.pyx":435
 * 
 *         self.max_n_classes = np.max(n_classes)
 *         self.value_stride = n_outputs * self.max_n_classes             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t k
 */
  __pyx_v_self->value_stride = (__pyx_v_n_outputs * __pyx_v_self->max_n_classes);

  /* "skgarden/mondrian/tree/_tree.pyx":438
 * 
 *         cdef SIZE_t k
 *         for k in range(n_outputs):             # <<<<<<<<<<<<<<
 *             self.n_classes[k] = n_classes[k]
 * 
 */
  __pyx_t_5 = __pyx_v_n_outputs;
  __pyx_t_6 = __pyx_t_5;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_6; __pyx_t_4+=1) {
    __pyx_v_k = __pyx_t_4;

    /* "skgarden/mondrian/tree/_tree.pyx":439
 *         cdef SIZE_t k
 *         for k in range(n_outputs):
 *             self.n_classes[k] = n_classes[k]             # <<<<<<<<<<<<<<
 * 
 *         # Inner structures
 */
    __pyx_t_7 = __pyx_v_k;
    (__pyx_v_self->n_classes[__pyx_v_k]) = (*__Pyx_BufPtrStrided1d(__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t *, __pyx_pybuffernd_n_classes.rcbuffer->pybuffer.buf, __pyx_t_7, __pyx_pybuffernd_n_classes.diminfo[0].strides));
  }

  /* "skgarden/mondrian/tree/_tree.pyx":442
 * 
 *         # Inner structures
 *         self.max_depth = 0             # <<<<<<<<<<<<<<
 *         self.node_count = 0
 *         self.capacity = 0
 */
  __pyx_v_self->max_depth = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":443
 *         # Inner structures
 *         self.max_depth = 0
 *         self.node_count = 0             # <<<<<<<<<<<<<<
 *         self.capacity = 0
 *         self.value = NULL
 */
  __pyx_v_self->node_count = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":444
 *         self.max_depth = 0
 *         self.node_count = 0
 *         self.capacity = 0             # <<<<<<<<<<<<<<
 *         self.value = NULL
 *         self.nodes = NULL
 */
  __pyx_v_self->capacity = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":445
 *         self.node_count = 0
 *         self.capacity = 0
 *         self.value = NULL             # <<<<<<<<<<<<<<
 *         self.nodes = NULL
 * 
 */
  __pyx_v_self->value = NULL;

  /* "skgarden/mondrian/tree/_tree.pyx":446
 *         self.capacity = 0
 *         self.value = NULL
 *         self.nodes = NULL             # <<<<<<<<<<<<<<
 * 
 *         # Used only in partial_fit
 */
  __pyx_v_self->nodes = NULL;

  /* "skgarden/mondrian/tree/_tree.pyx":449
 * 
 *         # Used only in partial_fit
 *         self.root = 0             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_v_self->root = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":425
 *             return self.root
 * 
 *     def __cinit__(self, int n_features, np.ndarray[SIZE_t, ndim=1] n_classes,             # <<<<<<<<<<<<<<
 *                   int n_outputs):
 *         """Constructor."""
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_n_classes.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_n_classes.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":451
 *         self.root = 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         """Destructor."""
 *         # Free all inner structures
 */

/* Python wrapper */
static void __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_2__dealloc__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_2__dealloc__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "skgarden/mondrian/tree/_tree.pyx":454
 *         """Destructor."""
 *         # Free all inner structures
 *         free(self.n_classes)             # <<<<<<<<<<<<<<
 *         free(self.value)
 *         free(self.nodes)
 */
  free(__pyx_v_self->n_classes);

  /* "skgarden/mondrian/tree/_tree.pyx":455
 *         # Free all inner structures
 *         free(self.n_classes)
 *         free(self.value)             # <<<<<<<<<<<<<<
 *         free(self.nodes)
 * 
 */
  free(__pyx_v_self->value);

  /* "skgarden/mondrian/tree/_tree.pyx":456
 *         free(self.n_classes)
 *         free(self.value)
 *         free(self.nodes)             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
  free(__pyx_v_self->nodes);

  /* "skgarden/mondrian/tree/_tree.pyx":451
 *         self.root = 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         """Destructor."""
 *         # Free all inner structures
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "skgarden/mondrian/tree/_tree.pyx":458
 *         free(self.nodes)
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         """Reduce re-implementation, for pickling."""
 *         return (Tree, (self.n_features,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_5__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_8skgarden_8mondrian_4tree_5_tree_4Tree_4__reduce__[] = "Reduce re-implementation, for pickling.";
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_5__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_4__reduce__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_4__reduce__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "skgarden/mondrian/tree/_tree.pyx":460
 *     def __reduce__(self):
 *         """Reduce re-implementation, for pickling."""
 *         return (Tree, (self.n_features,             # <<<<<<<<<<<<<<
 *                        sizet_ptr_to_ndarray(self.n_classes, self.n_outputs),
 *                        self.n_outputs), self.__getstate__())
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_Py_intptr_t(__pyx_v_self->n_features); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "skgarden/mondrian/tree/_tree.pyx":461
 *         """Reduce re-implementation, for pickling."""
 *         return (Tree, (self.n_features,
 *                        sizet_ptr_to_ndarray(self.n_classes, self.n_outputs),             # <<<<<<<<<<<<<<
 *                        self.n_outputs), self.__getstate__())
 * 
 */
  __pyx_t_2 = ((PyObject *)__pyx_f_8skgarden_8mondrian_4tree_6_utils_sizet_ptr_to_ndarray(__pyx_v_self->n_classes, __pyx_v_self->n_outputs)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "skgarden/mondrian/tree/_tree.pyx":462
 *         return (Tree, (self.n_features,
 *                        sizet_ptr_to_ndarray(self.n_classes, self.n_outputs),
 *                        self.n_outputs), self.__getstate__())             # <<<<<<<<<<<<<<
 * 
 *     def __getstate__(self):
 */
  __pyx_t_3 = __Pyx_PyInt_From_Py_intptr_t(__pyx_v_self->n_outputs); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "skgarden/mondrian/tree/_tree.pyx":460
 *     def __reduce__(self):
 *         """Reduce re-implementation, for pickling."""
 *         return (Tree, (self.n_features,             # <<<<<<<<<<<<<<
 *                        sizet_ptr_to_ndarray(self.n_classes, self.n_outputs),
 *                        self.n_outputs), self.__getstate__())
 */
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":462
 *         return (Tree, (self.n_features,
 *                        sizet_ptr_to_ndarray(self.n_classes, self.n_outputs),
 *                        self.n_outputs), self.__getstate__())             # <<<<<<<<<<<<<<
 * 
 *     def __getstate__(self):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_getstate); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":460
 *     def __reduce__(self):
 *         """Reduce re-implementation, for pickling."""
 *         return (Tree, (self.n_features,             # <<<<<<<<<<<<<<
 *                        sizet_ptr_to_ndarray(self.n_classes, self.n_outputs),
 *                        self.n_outputs), self.__getstate__())
 */
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_ptype_8skgarden_8mondrian_4tree_5_tree_Tree));
  __Pyx_GIVEREF(((PyObject *)__pyx_ptype_8skgarden_8mondrian_4tree_5_tree_Tree));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_ptype_8skgarden_8mondrian_4tree_5_tree_Tree));
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_3);
  __pyx_t_4 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "skgarden/mondrian/tree/_tree.pyx":458
 *         free(self.nodes)
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         """Reduce re-implementation, for pickling."""
 *         return (Tree, (self.n_features,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":464
 *                        self.n_outputs), self.__getstate__())
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         """Getstate re-implementation, for pickling."""
 *         d = {}
 */

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_7__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_8skgarden_8mondrian_4tree_5_tree_4Tree_6__getstate__[] = "Getstate re-implementation, for pickling.";
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_7__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_6__getstate__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_6__getstate__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_v_d = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__getstate__", 0);

  /* "skgarden/mondrian/tree/_tree.pyx":466
 *     def __getstate__(self):
 *         """Getstate re-implementation, for pickling."""
 *         d = {}             # <<<<<<<<<<<<<<
 *         # capacity is infered during the __setstate__ using nodes
 *         d["max_depth"] = self.max_depth
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_d = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":468
 *         d = {}
 *         # capacity is infered during the __setstate__ using nodes
 *         d["max_depth"] = self.max_depth             # <<<<<<<<<<<<<<
 *         d["node_count"] = self.node_count
 *         d["nodes"] = self._get_node_ndarray()
 */
  __pyx_t_1 = __Pyx_PyInt_From_Py_intptr_t(__pyx_v_self->max_depth); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(PyDict_SetItem(__pyx_v_d, __pyx_n_s_max_depth, __pyx_t_1) < 0)) __PYX_ERR(0, 468, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":469
 *         # capacity is infered during the __setstate__ using nodes
 *         d["max_depth"] = self.max_depth
 *         d["node_count"] = self.node_count             # <<<<<<<<<<<<<<
 *         d["nodes"] = self._get_node_ndarray()
 *         d["values"] = self._get_value_ndarray()
 */
  __pyx_t_1 = __Pyx_PyInt_From_Py_intptr_t(__pyx_v_self->node_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(PyDict_SetItem(__pyx_v_d, __pyx_n_s_node_count, __pyx_t_1) < 0)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":470
 *         d["max_depth"] = self.max_depth
 *         d["node_count"] = self.node_count
 *         d["nodes"] = self._get_node_ndarray()             # <<<<<<<<<<<<<<
 *         d["values"] = self._get_value_ndarray()
 *         d["root"] = self.root
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->_get_node_ndarray(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(PyDict_SetItem(__pyx_v_d, __pyx_n_s_nodes, __pyx_t_1) < 0)) __PYX_ERR(0, 470, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":471
 *         d["node_count"] = self.node_count
 *         d["nodes"] = self._get_node_ndarray()
 *         d["values"] = self._get_value_ndarray()             # <<<<<<<<<<<<<<
 *         d["root"] = self.root
 *         return d
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->_get_value_ndarray(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(PyDict_SetItem(__pyx_v_d, __pyx_n_s_values, __pyx_t_1) < 0)) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":472
 *         d["nodes"] = self._get_node_ndarray()
 *         d["values"] = self._get_value_ndarray()
 *         d["root"] = self.root             # <<<<<<<<<<<<<<
 *         return d
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_From_Py_intptr_t(__pyx_v_self->root); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 472, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(PyDict_SetItem(__pyx_v_d, __pyx_n_s_root, __pyx_t_1) < 0)) __PYX_ERR(0, 472, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":473
 *         d["values"] = self._get_value_ndarray()
 *         d["root"] = self.root
 *         return d             # <<<<<<<<<<<<<<
 * 
 *     def __setstate__(self, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_d);
  __pyx_r = __pyx_v_d;
  goto __pyx_L0;

  /* "skgarden/mondrian/tree/_tree.pyx":464
 *                        self.n_outputs), self.__getstate__())
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         """Getstate re-implementation, for pickling."""
 *         d = {}
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.__getstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_d);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":475
 *         return d
 * 
 *     def __setstate__(self, d):             # <<<<<<<<<<<<<<
 *         """Setstate re-implementation, for unpickling."""
 *         self.max_depth = d["max_depth"]
 */

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_9__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_d); /*proto*/
static char __pyx_doc_8skgarden_8mondrian_4tree_5_tree_4Tree_8__setstate__[] = "Setstate re-implementation, for unpickling.";
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_9__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_8__setstate__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self), ((PyObject *)__pyx_v_d));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_8__setstate__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_d) {
  PyObject *__pyx_v_node_ndarray = NULL;
  PyObject *__pyx_v_value_ndarray = NULL;
  PyObject *__pyx_v_value_shape = NULL;
  CYTHON_UNUSED void *__pyx_v_nodes;
  CYTHON_UNUSED void *__pyx_v_value;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_4Tree__resize_c __pyx_t_10;
  __Pyx_RefNannySetupContext("__setstate__", 0);

  /* "skgarden/mondrian/tree/_tree.pyx":477
 *     def __setstate__(self, d):
 *         """Setstate re-implementation, for unpickling."""
 *         self.max_depth = d["max_depth"]             # <<<<<<<<<<<<<<
 *         self.node_count = d["node_count"]
 *         self.root = d["root"]
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_d, __pyx_n_s_max_depth); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 477, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 477, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->max_depth = __pyx_t_2;

  /* "skgarden/mondrian/tree/_tree.pyx":478
 *         """Setstate re-implementation, for unpickling."""
 *         self.max_depth = d["max_depth"]
 *         self.node_count = d["node_count"]             # <<<<<<<<<<<<<<
 *         self.root = d["root"]
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_d, __pyx_n_s_node_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 478, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 478, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->node_count = __pyx_t_2;

  /* "skgarden/mondrian/tree/_tree.pyx":479
 *         self.max_depth = d["max_depth"]
 *         self.node_count = d["node_count"]
 *         self.root = d["root"]             # <<<<<<<<<<<<<<
 * 
 *         if 'nodes' not in d:
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_d, __pyx_n_s_root); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 479, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->root = __pyx_t_2;

  /* "skgarden/mondrian/tree/_tree.pyx":481
 *         self.root = d["root"]
 * 
 *         if 'nodes' not in d:             # <<<<<<<<<<<<<<
 *             raise ValueError('You have loaded Tree version which '
 *                              'cannot be imported')
 */
  __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_n_s_nodes, __pyx_v_d, Py_NE)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 481, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (unlikely(__pyx_t_4)) {

    /* "skgarden/mondrian/tree/_tree.pyx":482
 * 
 *         if 'nodes' not in d:
 *             raise ValueError('You have loaded Tree version which '             # <<<<<<<<<<<<<<
 *                              'cannot be imported')
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 482, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 482, __pyx_L1_error)

    /* "skgarden/mondrian/tree/_tree.pyx":481
 *         self.root = d["root"]
 * 
 *         if 'nodes' not in d:             # <<<<<<<<<<<<<<
 *             raise ValueError('You have loaded Tree version which '
 *                              'cannot be imported')
 */
  }

  /* "skgarden/mondrian/tree/_tree.pyx":485
 *                              'cannot be imported')
 * 
 *         node_ndarray = d['nodes']             # <<<<<<<<<<<<<<
 *         value_ndarray = d['values']
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_d, __pyx_n_s_nodes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_node_ndarray = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":486
 * 
 *         node_ndarray = d['nodes']
 *         value_ndarray = d['values']             # <<<<<<<<<<<<<<
 * 
 *         value_shape = (node_ndarray.shape[0], self.n_outputs,
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_d, __pyx_n_s_values); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_value_ndarray = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":488
 *         value_ndarray = d['values']
 * 
 *         value_shape = (node_ndarray.shape[0], self.n_outputs,             # <<<<<<<<<<<<<<
 *                        self.max_n_classes)
 *         if (node_ndarray.ndim != 1 or
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_node_ndarray, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_Py_intptr_t(__pyx_v_self->n_outputs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "skgarden/mondrian/tree/_tree.pyx":489
 * 
 *         value_shape = (node_ndarray.shape[0], self.n_outputs,
 *                        self.max_n_classes)             # <<<<<<<<<<<<<<
 *         if (node_ndarray.ndim != 1 or
 *                 node_ndarray.dtype != NODE_DTYPE or
 */
  __pyx_t_6 = __Pyx_PyInt_From_Py_intptr_t(__pyx_v_self->max_n_classes); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "skgarden/mondrian/tree/_tree.pyx":488
 *         value_ndarray = d['values']
 * 
 *         value_shape = (node_ndarray.shape[0], self.n_outputs,             # <<<<<<<<<<<<<<
 *                        self.max_n_classes)
 *         if (node_ndarray.ndim != 1 or
 */
  __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_6);
  __pyx_t_5 = 0;
  __pyx_t_1 = 0;
  __pyx_t_6 = 0;
  __pyx_v_value_shape = ((PyObject*)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":490
 *         value_shape = (node_ndarray.shape[0], self.n_outputs,
 *                        self.max_n_classes)
 *         if (node_ndarray.ndim != 1 or             # <<<<<<<<<<<<<<
 *                 node_ndarray.dtype != NODE_DTYPE or
 *                 not node_ndarray.flags.c_contiguous or
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_node_ndarray, __pyx_n_s_ndim); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_PyInt_NeObjC(__pyx_t_7, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (!__pyx_t_3) {
  } else {
    __pyx_t_4 = __pyx_t_3;
    goto __pyx_L5_bool_binop_done;
  }

  /* "skgarden/mondrian/tree/_tree.pyx":491
 *                        self.max_n_classes)
 *         if (node_ndarray.ndim != 1 or
 *                 node_ndarray.dtype != NODE_DTYPE or             # <<<<<<<<<<<<<<
 *                 not node_ndarray.flags.c_contiguous or
 *                 value_ndarray.shape != value_shape or
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_node_ndarray, __pyx_n_s_dtype); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_NODE_DTYPE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_6, __pyx_t_7, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!__pyx_t_3) {
  } else {
    __pyx_t_4 = __pyx_t_3;
    goto __pyx_L5_bool_binop_done;
  }

  /* "skgarden/mondrian/tree/_tree.pyx":492
 *         if (node_ndarray.ndim != 1 or
 *                 node_ndarray.dtype != NODE_DTYPE or
 *                 not node_ndarray.flags.c_contiguous or             # <<<<<<<<<<<<<<
 *                 value_ndarray.shape != value_shape or
 *                 not value_ndarray.flags.c_contiguous or
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_node_ndarray, __pyx_n_s_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_c_contiguous); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 492, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_8 = ((!__pyx_t_3) != 0);
  if (!__pyx_t_8) {
  } else {
    __pyx_t_4 = __pyx_t_8;
    goto __pyx_L5_bool_binop_done;
  }

  /* "skgarden/mondrian/tree/_tree.pyx":493
 *                 node_ndarray.dtype != NODE_DTYPE or
 *                 not node_ndarray.flags.c_contiguous or
 *                 value_ndarray.shape != value_shape or             # <<<<<<<<<<<<<<
 *                 not value_ndarray.flags.c_contiguous or
 *                 value_ndarray.dtype != np.float64):
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_value_ndarray, __pyx_n_s_shape); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_7, __pyx_v_value_shape, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!__pyx_t_8) {
  } else {
    __pyx_t_4 = __pyx_t_8;
    goto __pyx_L5_bool_binop_done;
  }

  /* "skgarden/mondrian/tree/_tree.pyx":494
 *                 not node_ndarray.flags.c_contiguous or
 *                 value_ndarray.shape != value_shape or
 *                 not value_ndarray.flags.c_contiguous or             # <<<<<<<<<<<<<<
 *                 value_ndarray.dtype != np.float64):
 *             raise ValueError('Did not recognise loaded array layout')
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_value_ndarray, __pyx_n_s_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_c_contiguous); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_3 = ((!__pyx_t_8) != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_4 = __pyx_t_3;
    goto __pyx_L5_bool_binop_done;
  }

  /* "skgarden/mondrian/tree/_tree.pyx":495
 *                 value_ndarray.shape != value_shape or
 *                 not value_ndarray.flags.c_contiguous or
 *                 value_ndarray.dtype != np.float64):             # <<<<<<<<<<<<<<
 *             raise ValueError('Did not recognise loaded array layout')
 * 
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_value_ndarray, __pyx_n_s_dtype); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_float64); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_7, __pyx_t_6, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __pyx_t_3;
  __pyx_L5_bool_binop_done:;

  /* "skgarden/mondrian/tree/_tree.pyx":490
 *         value_shape = (node_ndarray.shape[0], self.n_outputs,
 *                        self.max_n_classes)
 *         if (node_ndarray.ndim != 1 or             # <<<<<<<<<<<<<<
 *                 node_ndarray.dtype != NODE_DTYPE or
 *                 not node_ndarray.flags.c_contiguous or
 */
  if (unlikely(__pyx_t_4)) {

    /* "skgarden/mondrian/tree/_tree.pyx":496
 *                 not value_ndarray.flags.c_contiguous or
 *                 value_ndarray.dtype != np.float64):
 *             raise ValueError('Did not recognise loaded array layout')             # <<<<<<<<<<<<<<
 * 
 *         self.capacity = node_ndarray.shape[0]
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 496, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 496, __pyx_L1_error)

    /* "skgarden/mondrian/tree/_tree.pyx":490
 *         value_shape = (node_ndarray.shape[0], self.n_outputs,
 *                        self.max_n_classes)
 *         if (node_ndarray.ndim != 1 or             # <<<<<<<<<<<<<<
 *                 node_ndarray.dtype != NODE_DTYPE or
 *                 not node_ndarray.flags.c_contiguous or
 */
  }

  /* "skgarden/mondrian/tree/_tree.pyx":498
 *             raise ValueError('Did not recognise loaded array layout')
 * 
 *         self.capacity = node_ndarray.shape[0]             # <<<<<<<<<<<<<<
 *         if self._resize_c(self.capacity) != 0:
 *             raise MemoryError("resizing tree to %d" % self.capacity)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_node_ndarray, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_6); if (unlikely((__pyx_t_2 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_self->capacity = __pyx_t_2;

  /* "skgarden/mondrian/tree/_tree.pyx":499
 * 
 *         self.capacity = node_ndarray.shape[0]
 *         if self._resize_c(self.capacity) != 0:             # <<<<<<<<<<<<<<
 *             raise MemoryError("resizing tree to %d" % self.capacity)
 *         nodes = memcpy(self.nodes, (<np.ndarray> node_ndarray).data,
 */
  __pyx_t_10.__pyx_n = 1;
  __pyx_t_10.capacity = __pyx_v_self->capacity;
  __pyx_t_9 = ((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->_resize_c(__pyx_v_self, &__pyx_t_10); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 499, __pyx_L1_error)
  __pyx_t_4 = ((__pyx_t_9 != 0) != 0);
  if (unlikely(__pyx_t_4)) {

    /* "skgarden/mondrian/tree/_tree.pyx":500
 *         self.capacity = node_ndarray.shape[0]
 *         if self._resize_c(self.capacity) != 0:
 *             raise MemoryError("resizing tree to %d" % self.capacity)             # <<<<<<<<<<<<<<
 *         nodes = memcpy(self.nodes, (<np.ndarray> node_ndarray).data,
 *                        self.capacity * sizeof(Node))
 */
    __pyx_t_6 = __Pyx_PyInt_From_Py_intptr_t(__pyx_v_self->capacity); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 500, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_resizing_tree_to_d, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 500, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_MemoryError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 500, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 500, __pyx_L1_error)

    /* "skgarden/mondrian/tree/_tree.pyx":499
 * 
 *         self.capacity = node_ndarray.shape[0]
 *         if self._resize_c(self.capacity) != 0:             # <<<<<<<<<<<<<<
 *             raise MemoryError("resizing tree to %d" % self.capacity)
 *         nodes = memcpy(self.nodes, (<np.ndarray> node_ndarray).data,
 */
  }

  /* "skgarden/mondrian/tree/_tree.pyx":501
 *         if self._resize_c(self.capacity) != 0:
 *             raise MemoryError("resizing tree to %d" % self.capacity)
 *         nodes = memcpy(self.nodes, (<np.ndarray> node_ndarray).data,             # <<<<<<<<<<<<<<
 *                        self.capacity * sizeof(Node))
 *         value = memcpy(self.value, (<np.ndarray> value_ndarray).data,
 */
  __pyx_v_nodes = memcpy(__pyx_v_self->nodes, ((PyArrayObject *)__pyx_v_node_ndarray)->data, (__pyx_v_self->capacity * (sizeof(struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node))));

  /* "skgarden/mondrian/tree/_tree.pyx":503
 *         nodes = memcpy(self.nodes, (<np.ndarray> node_ndarray).data,
 *                        self.capacity * sizeof(Node))
 *         value = memcpy(self.value, (<np.ndarray> value_ndarray).data,             # <<<<<<<<<<<<<<
 *                        self.capacity * self.value_stride * sizeof(double))
 * 
 */
  __pyx_v_value = memcpy(__pyx_v_self->value, ((PyArrayObject *)__pyx_v_value_ndarray)->data, ((__pyx_v_self->capacity * __pyx_v_self->value_stride) * (sizeof(double))));

  /* "skgarden/mondrian/tree/_tree.pyx":475
 *         return d
 * 
 *     def __setstate__(self, d):             # <<<<<<<<<<<<<<
 *         """Setstate re-implementation, for unpickling."""
 *         self.max_depth = d["max_depth"]
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_node_ndarray);
  __Pyx_XDECREF(__pyx_v_value_ndarray);
  __Pyx_XDECREF(__pyx_v_value_shape);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":506
 *                        self.capacity * self.value_stride * sizeof(double))
 * 
 *     cdef int _resize(self, SIZE_t capacity) nogil except -1:             # <<<<<<<<<<<<<<
 *         """Resize all inner arrays to `capacity`, if `capacity` == -1, then
 *            double the size of the inner arrays.
 */

static int __pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree__resize(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_capacity) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_4Tree__resize_c __pyx_t_2;
  int __pyx_t_3;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_resize", 0);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif

  /* "skgarden/mondrian/tree/_tree.pyx":507
 * 
 *     cdef int _resize(self, SIZE_t capacity) nogil except -1:
 *         """Resize all inner arrays to `capacity`, if `capacity` == -1, then             # <<<<<<<<<<<<<<
 *            double the size of the inner arrays.
 * 
 */
  /*try:*/ {

    /* "skgarden/mondrian/tree/_tree.pyx":513
 *         or 0 otherwise.
 *         """
 *         if self._resize_c(capacity) != 0:             # <<<<<<<<<<<<<<
 *             # Acquire gil only if we need to raise
 *             with gil:
 */
    __pyx_t_2.__pyx_n = 1;
    __pyx_t_2.capacity = __pyx_v_capacity;
    __pyx_t_1 = ((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->_resize_c(__pyx_v_self, &__pyx_t_2); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 513, __pyx_L4_error)
    __pyx_t_3 = ((__pyx_t_1 != 0) != 0);
    if (__pyx_t_3) {

      /* "skgarden/mondrian/tree/_tree.pyx":515
 *         if self._resize_c(capacity) != 0:
 *             # Acquire gil only if we need to raise
 *             with gil:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
      {
          #ifdef WITH_THREAD
          PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
          #endif
          /*try:*/ {

            /* "skgarden/mondrian/tree/_tree.pyx":516
 *             # Acquire gil only if we need to raise
 *             with gil:
 *                 raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *     # XXX using (size_t)(-1) is ugly, but SIZE_MAX is not available in C89
 */
            PyErr_NoMemory(); __PYX_ERR(0, 516, __pyx_L8_error)
          }

          /* "skgarden/mondrian/tree/_tree.pyx":515
 *         if self._resize_c(capacity) != 0:
 *             # Acquire gil only if we need to raise
 *             with gil:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
          /*finally:*/ {
            __pyx_L8_error: {
              #ifdef WITH_THREAD
              __Pyx_PyGILState_Release(__pyx_gilstate_save);
              #endif
              goto __pyx_L4_error;
            }
          }
      }

      /* "skgarden/mondrian/tree/_tree.pyx":513
 *         or 0 otherwise.
 *         """
 *         if self._resize_c(capacity) != 0:             # <<<<<<<<<<<<<<
 *             # Acquire gil only if we need to raise
 *             with gil:
 */
    }
  }

  /* "skgarden/mondrian/tree/_tree.pyx":507
 * 
 *     cdef int _resize(self, SIZE_t capacity) nogil except -1:
 *         """Resize all inner arrays to `capacity`, if `capacity` == -1, then             # <<<<<<<<<<<<<<
 *            double the size of the inner arrays.
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L5;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "skgarden/mondrian/tree/_tree.pyx":506
 *                        self.capacity * self.value_stride * sizeof(double))
 * 
 *     cdef int _resize(self, SIZE_t capacity) nogil except -1:             # <<<<<<<<<<<<<<
 *         """Resize all inner arrays to `capacity`, if `capacity` == -1, then
 *            double the size of the inner arrays.
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree._resize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":520
 *     # XXX using (size_t)(-1) is ugly, but SIZE_MAX is not available in C89
 *     # (i.e., older MSVC).
 *     cdef int _resize_c(self, SIZE_t capacity=<SIZE_t>(-1)) nogil except -1:             # <<<<<<<<<<<<<<
 *         """Guts of _resize
 * 
 */

static int __pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree__resize_c(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_4Tree__resize_c *__pyx_optional_args) {
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_capacity = __pyx_k__7;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_capacity = __pyx_optional_args->capacity;
    }
  }

  /* "skgarden/mondrian/tree/_tree.pyx":526
 *         or 0 otherwise.
 *         """
 *         if capacity == self.capacity and self.nodes != NULL:             # <<<<<<<<<<<<<<
 *             return 0
 * 
 */
  __pyx_t_2 = ((__pyx_v_capacity == __pyx_v_self->capacity) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_self->nodes != NULL) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "skgarden/mondrian/tree/_tree.pyx":527
 *         """
 *         if capacity == self.capacity and self.nodes != NULL:
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *         if capacity == <SIZE_t>(-1):
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "skgarden/mondrian/tree/_tree.pyx":526
 *         or 0 otherwise.
 *         """
 *         if capacity == self.capacity and self.nodes != NULL:             # <<<<<<<<<<<<<<
 *             return 0
 * 
 */
  }

  /* "skgarden/mondrian/tree/_tree.pyx":529
 *             return 0
 * 
 *         if capacity == <SIZE_t>(-1):             # <<<<<<<<<<<<<<
 *             if self.capacity == 0:
 *                 capacity = 3  # default initial value
 */
  __pyx_t_1 = ((__pyx_v_capacity == ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t)-1L)) != 0);
  if (__pyx_t_1) {

    /* "skgarden/mondrian/tree/_tree.pyx":530
 * 
 *         if capacity == <SIZE_t>(-1):
 *             if self.capacity == 0:             # <<<<<<<<<<<<<<
 *                 capacity = 3  # default initial value
 *             else:
 */
    __pyx_t_1 = ((__pyx_v_self->capacity == 0) != 0);
    if (__pyx_t_1) {

      /* "skgarden/mondrian/tree/_tree.pyx":531
 *         if capacity == <SIZE_t>(-1):
 *             if self.capacity == 0:
 *                 capacity = 3  # default initial value             # <<<<<<<<<<<<<<
 *             else:
 *                 capacity = 2 * self.capacity
 */
      __pyx_v_capacity = 3;

      /* "skgarden/mondrian/tree/_tree.pyx":530
 * 
 *         if capacity == <SIZE_t>(-1):
 *             if self.capacity == 0:             # <<<<<<<<<<<<<<
 *                 capacity = 3  # default initial value
 *             else:
 */
      goto __pyx_L7;
    }

    /* "skgarden/mondrian/tree/_tree.pyx":533
 *                 capacity = 3  # default initial value
 *             else:
 *                 capacity = 2 * self.capacity             # <<<<<<<<<<<<<<
 * 
 *         safe_realloc(&self.nodes, capacity)
 */
    /*else*/ {
      __pyx_v_capacity = (2 * __pyx_v_self->capacity);
    }
    __pyx_L7:;

    /* "skgarden/mondrian/tree/_tree.pyx":529
 *             return 0
 * 
 *         if capacity == <SIZE_t>(-1):             # <<<<<<<<<<<<<<
 *             if self.capacity == 0:
 *                 capacity = 3  # default initial value
 */
  }

  /* "skgarden/mondrian/tree/_tree.pyx":535
 *                 capacity = 2 * self.capacity
 * 
 *         safe_realloc(&self.nodes, capacity)             # <<<<<<<<<<<<<<
 *         safe_realloc(&self.value, capacity * self.value_stride)
 * 
 */
  __pyx_fuse_5__pyx_f_8skgarden_8mondrian_4tree_6_utils_safe_realloc((&__pyx_v_self->nodes), __pyx_v_capacity); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 535, __pyx_L1_error)

  /* "skgarden/mondrian/tree/_tree.pyx":536
 * 
 *         safe_realloc(&self.nodes, capacity)
 *         safe_realloc(&self.value, capacity * self.value_stride)             # <<<<<<<<<<<<<<
 * 
 *         # value memory is initialised to 0 to enable classifier argmax
 */
  __pyx_fuse_3__pyx_f_8skgarden_8mondrian_4tree_6_utils_safe_realloc((&__pyx_v_self->value), (__pyx_v_capacity * __pyx_v_self->value_stride)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 536, __pyx_L1_error)

  /* "skgarden/mondrian/tree/_tree.pyx":539
 * 
 *         # value memory is initialised to 0 to enable classifier argmax
 *         if capacity > self.capacity:             # <<<<<<<<<<<<<<
 *             memset(<void*>(self.value + self.capacity * self.value_stride), 0,
 *                    (capacity - self.capacity) * self.value_stride *
 */
  __pyx_t_1 = ((__pyx_v_capacity > __pyx_v_self->capacity) != 0);
  if (__pyx_t_1) {

    /* "skgarden/mondrian/tree/_tree.pyx":540
 *         # value memory is initialised to 0 to enable classifier argmax
 *         if capacity > self.capacity:
 *             memset(<void*>(self.value + self.capacity * self.value_stride), 0,             # <<<<<<<<<<<<<<
 *                    (capacity - self.capacity) * self.value_stride *
 *                    sizeof(double))
 */
    (void)(memset(((void *)(__pyx_v_self->value + (__pyx_v_self->capacity * __pyx_v_self->value_stride))), 0, (((__pyx_v_capacity - __pyx_v_self->capacity) * __pyx_v_self->value_stride) * (sizeof(double)))));

    /* "skgarden/mondrian/tree/_tree.pyx":539
 * 
 *         # value memory is initialised to 0 to enable classifier argmax
 *         if capacity > self.capacity:             # <<<<<<<<<<<<<<
 *             memset(<void*>(self.value + self.capacity * self.value_stride), 0,
 *                    (capacity - self.capacity) * self.value_stride *
 */
  }

  /* "skgarden/mondrian/tree/_tree.pyx":545
 * 
 *         # if capacity smaller than node_count, adjust the counter
 *         if capacity < self.node_count:             # <<<<<<<<<<<<<<
 *             self.node_count = capacity
 * 
 */
  __pyx_t_1 = ((__pyx_v_capacity < __pyx_v_self->node_count) != 0);
  if (__pyx_t_1) {

    /* "skgarden/mondrian/tree/_tree.pyx":546
 *         # if capacity smaller than node_count, adjust the counter
 *         if capacity < self.node_count:
 *             self.node_count = capacity             # <<<<<<<<<<<<<<
 * 
 *         self.capacity = capacity
 */
    __pyx_v_self->node_count = __pyx_v_capacity;

    /* "skgarden/mondrian/tree/_tree.pyx":545
 * 
 *         # if capacity smaller than node_count, adjust the counter
 *         if capacity < self.node_count:             # <<<<<<<<<<<<<<
 *             self.node_count = capacity
 * 
 */
  }

  /* "skgarden/mondrian/tree/_tree.pyx":548
 *             self.node_count = capacity
 * 
 *         self.capacity = capacity             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_v_self->capacity = __pyx_v_capacity;

  /* "skgarden/mondrian/tree/_tree.pyx":549
 * 
 *         self.capacity = capacity
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef void update_node_extent(self, SIZE_t node_ind, SIZE_t child_ind,
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "skgarden/mondrian/tree/_tree.pyx":520
 *     # XXX using (size_t)(-1) is ugly, but SIZE_MAX is not available in C89
 *     # (i.e., older MSVC).
 *     cdef int _resize_c(self, SIZE_t capacity=<SIZE_t>(-1)) nogil except -1:             # <<<<<<<<<<<<<<
 *         """Guts of _resize
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree._resize_c", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":551
 *         return 0
 * 
 *     cdef void update_node_extent(self, SIZE_t node_ind, SIZE_t child_ind,             # <<<<<<<<<<<<<<
 *                                  DTYPE_t* X_ptr, SIZE_t X_start, SIZE_t X_f_stride):
 *         """
 */

static void __pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree_update_node_extent(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_node_ind, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_child_ind, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *__pyx_v_X_ptr, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_X_start, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_X_f_stride) {
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_f_ind;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t __pyx_v_x_val;
  struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node *__pyx_v_node;
  struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node *__pyx_v_prev_node;
  __Pyx_RefNannyDeclarations
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_1;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_2;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_3;
  __Pyx_RefNannySetupContext("update_node_extent", 0);

  /* "skgarden/mondrian/tree/_tree.pyx":560
 *         cdef SIZE_t f_ind
 *         cdef DTYPE_t x_val
 *         cdef Node* node = &self.nodes[node_ind]             # <<<<<<<<<<<<<<
 *         cdef Node* prev_node = &self.nodes[child_ind]
 * 
 */
  __pyx_v_node = (&(__pyx_v_self->nodes[__pyx_v_node_ind]));

  /* "skgarden/mondrian/tree/_tree.pyx":561
 *         cdef DTYPE_t x_val
 *         cdef Node* node = &self.nodes[node_ind]
 *         cdef Node* prev_node = &self.nodes[child_ind]             # <<<<<<<<<<<<<<
 * 
 *         for f_ind in range(self.n_features):
 */
  __pyx_v_prev_node = (&(__pyx_v_self->nodes[__pyx_v_child_ind]));

  /* "skgarden/mondrian/tree/_tree.pyx":563
 *         cdef Node* prev_node = &self.nodes[child_ind]
 * 
 *         for f_ind in range(self.n_features):             # <<<<<<<<<<<<<<
 *             x_val = X_ptr[X_start + X_f_stride*f_ind]
 *             node.lower_bounds[f_ind] = fmin(x_val, prev_node.lower_bounds[f_ind])
 */
  __pyx_t_1 = __pyx_v_self->n_features;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_f_ind = __pyx_t_3;

    /* "skgarden/mondrian/tree/_tree.pyx":564
 * 
 *         for f_ind in range(self.n_features):
 *             x_val = X_ptr[X_start + X_f_stride*f_ind]             # <<<<<<<<<<<<<<
 *             node.lower_bounds[f_ind] = fmin(x_val, prev_node.lower_bounds[f_ind])
 *             node.upper_bounds[f_ind] = fmax(x_val, prev_node.upper_bounds[f_ind])
 */
    __pyx_v_x_val = (__pyx_v_X_ptr[(__pyx_v_X_start + (__pyx_v_X_f_stride * __pyx_v_f_ind))]);

    /* "skgarden/mondrian/tree/_tree.pyx":565
 *         for f_ind in range(self.n_features):
 *             x_val = X_ptr[X_start + X_f_stride*f_ind]
 *             node.lower_bounds[f_ind] = fmin(x_val, prev_node.lower_bounds[f_ind])             # <<<<<<<<<<<<<<
 *             node.upper_bounds[f_ind] = fmax(x_val, prev_node.upper_bounds[f_ind])
 * 
 */
    (__pyx_v_node->lower_bounds[__pyx_v_f_ind]) = __pyx_f_8skgarden_8mondrian_4tree_5_tree_fmin(__pyx_v_x_val, (__pyx_v_prev_node->lower_bounds[__pyx_v_f_ind]));

    /* "skgarden/mondrian/tree/_tree.pyx":566
 *             x_val = X_ptr[X_start + X_f_stride*f_ind]
 *             node.lower_bounds[f_ind] = fmin(x_val, prev_node.lower_bounds[f_ind])
 *             node.upper_bounds[f_ind] = fmax(x_val, prev_node.upper_bounds[f_ind])             # <<<<<<<<<<<<<<
 * 
 *     cdef void _update_node_info(self, SIZE_t parent_id, SIZE_t child_id,
 */
    (__pyx_v_node->upper_bounds[__pyx_v_f_ind]) = __pyx_f_8skgarden_8mondrian_4tree_5_tree_fmax(__pyx_v_x_val, (__pyx_v_prev_node->upper_bounds[__pyx_v_f_ind]));
  }

  /* "skgarden/mondrian/tree/_tree.pyx":551
 *         return 0
 * 
 *     cdef void update_node_extent(self, SIZE_t node_ind, SIZE_t child_ind,             # <<<<<<<<<<<<<<
 *                                  DTYPE_t* X_ptr, SIZE_t X_start, SIZE_t X_f_stride):
 *         """
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "skgarden/mondrian/tree/_tree.pyx":568
 *             node.upper_bounds[f_ind] = fmax(x_val, prev_node.upper_bounds[f_ind])
 * 
 *     cdef void _update_node_info(self, SIZE_t parent_id, SIZE_t child_id,             # <<<<<<<<<<<<<<
 *                                 DOUBLE_t* y_ptr, SIZE_t y_start):
 *         """
 */

static void __pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree__update_node_info(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_parent_id, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_child_id, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t *__pyx_v_y_ptr, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_y_start) {
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_is_regression;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_child_ptr;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_parent_ptr;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_c_ind;
  struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node *__pyx_v_child;
  struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node *__pyx_v_parent;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t __pyx_v_new_sum;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t __pyx_v_old_mean;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t __pyx_v_new_mean;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t __pyx_v_ss;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_2;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_3;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_4;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_5;
  __Pyx_RefNannySetupContext("_update_node_info", 0);

  /* "skgarden/mondrian/tree/_tree.pyx":574
 *         child_id and y_ptr[y_start]
 *         """
 *         cdef SIZE_t is_regression = self.n_classes[0] == 1             # <<<<<<<<<<<<<<
 *         cdef SIZE_t child_ptr = child_id*self.value_stride
 *         cdef SIZE_t parent_ptr = parent_id*self.value_stride
 */
  __pyx_v_is_regression = ((__pyx_v_self->n_classes[0]) == 1);

  /* "skgarden/mondrian/tree/_tree.pyx":575
 *         """
 *         cdef SIZE_t is_regression = self.n_classes[0] == 1
 *         cdef SIZE_t child_ptr = child_id*self.value_stride             # <<<<<<<<<<<<<<
 *         cdef SIZE_t parent_ptr = parent_id*self.value_stride
 *         cdef SIZE_t c_ind
 */
  __pyx_v_child_ptr = (__pyx_v_child_id * __pyx_v_self->value_stride);

  /* "skgarden/mondrian/tree/_tree.pyx":576
 *         cdef SIZE_t is_regression = self.n_classes[0] == 1
 *         cdef SIZE_t child_ptr = child_id*self.value_stride
 *         cdef SIZE_t parent_ptr = parent_id*self.value_stride             # <<<<<<<<<<<<<<
 *         cdef SIZE_t c_ind
 *         cdef Node* child = &self.nodes[child_id]
 */
  __pyx_v_parent_ptr = (__pyx_v_parent_id * __pyx_v_self->value_stride);

  /* "skgarden/mondrian/tree/_tree.pyx":578
 *         cdef SIZE_t parent_ptr = parent_id*self.value_stride
 *         cdef SIZE_t c_ind
 *         cdef Node* child = &self.nodes[child_id]             # <<<<<<<<<<<<<<
 *         cdef Node* parent = &self.nodes[parent_id]
 *         cdef DTYPE_t new_sum
 */
  __pyx_v_child = (&(__pyx_v_self->nodes[__pyx_v_child_id]));

  /* "skgarden/mondrian/tree/_tree.pyx":579
 *         cdef SIZE_t c_ind
 *         cdef Node* child = &self.nodes[child_id]
 *         cdef Node* parent = &self.nodes[parent_id]             # <<<<<<<<<<<<<<
 *         cdef DTYPE_t new_sum
 *         cdef DTYPE_t old_mean
 */
  __pyx_v_parent = (&(__pyx_v_self->nodes[__pyx_v_parent_id]));

  /* "skgarden/mondrian/tree/_tree.pyx":585
 *         cdef DTYPE_t ss
 * 
 *         if is_regression:             # <<<<<<<<<<<<<<
 *             # Update mean
 *             old_mean = self.value[child_ptr]
 */
  __pyx_t_1 = (__pyx_v_is_regression != 0);
  if (__pyx_t_1) {

    /* "skgarden/mondrian/tree/_tree.pyx":587
 *         if is_regression:
 *             # Update mean
 *             old_mean = self.value[child_ptr]             # <<<<<<<<<<<<<<
 *             new_sum = old_mean*child.n_node_samples + y_ptr[y_start]
 *             new_mean = new_sum / (child.n_node_samples + 1)
 */
    __pyx_v_old_mean = (__pyx_v_self->value[__pyx_v_child_ptr]);

    /* "skgarden/mondrian/tree/_tree.pyx":588
 *             # Update mean
 *             old_mean = self.value[child_ptr]
 *             new_sum = old_mean*child.n_node_samples + y_ptr[y_start]             # <<<<<<<<<<<<<<
 *             new_mean = new_sum / (child.n_node_samples + 1)
 *             self.value[parent_ptr] = new_mean
 */
    __pyx_v_new_sum = ((__pyx_v_old_mean * __pyx_v_child->n_node_samples) + (__pyx_v_y_ptr[__pyx_v_y_start]));

    /* "skgarden/mondrian/tree/_tree.pyx":589
 *             old_mean = self.value[child_ptr]
 *             new_sum = old_mean*child.n_node_samples + y_ptr[y_start]
 *             new_mean = new_sum / (child.n_node_samples + 1)             # <<<<<<<<<<<<<<
 *             self.value[parent_ptr] = new_mean
 * 
 */
    __pyx_v_new_mean = (__pyx_v_new_sum / (__pyx_v_child->n_node_samples + 1));

    /* "skgarden/mondrian/tree/_tree.pyx":590
 *             new_sum = old_mean*child.n_node_samples + y_ptr[y_start]
 *             new_mean = new_sum / (child.n_node_samples + 1)
 *             self.value[parent_ptr] = new_mean             # <<<<<<<<<<<<<<
 * 
 *             # Update variance
 */
    (__pyx_v_self->value[__pyx_v_parent_ptr]) = __pyx_v_new_mean;

    /* "skgarden/mondrian/tree/_tree.pyx":593
 * 
 *             # Update variance
 *             ss = (child.variance + old_mean**2)*child.n_node_samples             # <<<<<<<<<<<<<<
 *             parent.variance = (
 *                 (ss + y_ptr[y_start]**2) / (child.n_node_samples + 1) -
 */
    __pyx_v_ss = ((__pyx_v_child->variance + powf(__pyx_v_old_mean, 2.0)) * __pyx_v_child->n_node_samples);

    /* "skgarden/mondrian/tree/_tree.pyx":594
 *             # Update variance
 *             ss = (child.variance + old_mean**2)*child.n_node_samples
 *             parent.variance = (             # <<<<<<<<<<<<<<
 *                 (ss + y_ptr[y_start]**2) / (child.n_node_samples + 1) -
 *                 new_mean**2)
 */
    __pyx_v_parent->variance = (((__pyx_v_ss + pow((__pyx_v_y_ptr[__pyx_v_y_start]), 2.0)) / (__pyx_v_child->n_node_samples + 1)) - powf(__pyx_v_new_mean, 2.0));

    /* "skgarden/mondrian/tree/_tree.pyx":585
 *         cdef DTYPE_t ss
 * 
 *         if is_regression:             # <<<<<<<<<<<<<<
 *             # Update mean
 *             old_mean = self.value[child_ptr]
 */
    goto __pyx_L3;
  }

  /* "skgarden/mondrian/tree/_tree.pyx":599
 *         else:
 *             # Update class counts.
 *             self.value[parent_ptr + <SIZE_t> y_ptr[y_start]] += 1.0             # <<<<<<<<<<<<<<
 * 
 *             if child_id != parent_id:
 */
  /*else*/ {
    __pyx_t_2 = (__pyx_v_parent_ptr + ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t)(__pyx_v_y_ptr[__pyx_v_y_start])));
    (__pyx_v_self->value[__pyx_t_2]) = ((__pyx_v_self->value[__pyx_t_2]) + 1.0);

    /* "skgarden/mondrian/tree/_tree.pyx":601
 *             self.value[parent_ptr + <SIZE_t> y_ptr[y_start]] += 1.0
 * 
 *             if child_id != parent_id:             # <<<<<<<<<<<<<<
 *                 for c_ind in range(self.n_classes[0]):
 *                     self.value[parent_ptr + c_ind] += self.value[child_ptr + c_ind]
 */
    __pyx_t_1 = ((__pyx_v_child_id != __pyx_v_parent_id) != 0);
    if (__pyx_t_1) {

      /* "skgarden/mondrian/tree/_tree.pyx":602
 * 
 *             if child_id != parent_id:
 *                 for c_ind in range(self.n_classes[0]):             # <<<<<<<<<<<<<<
 *                     self.value[parent_ptr + c_ind] += self.value[child_ptr + c_ind]
 * 
 */
      __pyx_t_2 = (__pyx_v_self->n_classes[0]);
      __pyx_t_3 = __pyx_t_2;
      for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
        __pyx_v_c_ind = __pyx_t_4;

        /* "skgarden/mondrian/tree/_tree.pyx":603
 *             if child_id != parent_id:
 *                 for c_ind in range(self.n_classes[0]):
 *                     self.value[parent_ptr + c_ind] += self.value[child_ptr + c_ind]             # <<<<<<<<<<<<<<
 * 
 *     cdef void set_node_attributes(self, SIZE_t node_ind, SIZE_t left_child,
 */
        __pyx_t_5 = (__pyx_v_parent_ptr + __pyx_v_c_ind);
        (__pyx_v_self->value[__pyx_t_5]) = ((__pyx_v_self->value[__pyx_t_5]) + (__pyx_v_self->value[(__pyx_v_child_ptr + __pyx_v_c_ind)]));
      }

      /* "skgarden/mondrian/tree/_tree.pyx":601
 *             self.value[parent_ptr + <SIZE_t> y_ptr[y_start]] += 1.0
 * 
 *             if child_id != parent_id:             # <<<<<<<<<<<<<<
 *                 for c_ind in range(self.n_classes[0]):
 *                     self.value[parent_ptr + c_ind] += self.value[child_ptr + c_ind]
 */
    }
  }
  __pyx_L3:;

  /* "skgarden/mondrian/tree/_tree.pyx":568
 *             node.upper_bounds[f_ind] = fmax(x_val, prev_node.upper_bounds[f_ind])
 * 
 *     cdef void _update_node_info(self, SIZE_t parent_id, SIZE_t child_id,             # <<<<<<<<<<<<<<
 *                                 DOUBLE_t* y_ptr, SIZE_t y_start):
 *         """
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "skgarden/mondrian/tree/_tree.pyx":605
 *                     self.value[parent_ptr + c_ind] += self.value[child_ptr + c_ind]
 * 
 *     cdef void set_node_attributes(self, SIZE_t node_ind, SIZE_t left_child,             # <<<<<<<<<<<<<<
 *                                   SIZE_t right_child, SIZE_t feature, DOUBLE_t threshold,
 *                                   DTYPE_t tau, SIZE_t n_node_samples,
 */

static void __pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree_set_node_attributes(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_node_ind, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_left_child, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_right_child, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_feature, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t __pyx_v_threshold, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t __pyx_v_tau, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_n_node_samples, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t __pyx_v_weighted_n_node_samples, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t __pyx_v_impurity, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t __pyx_v_variance, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_X_start, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_X_f_stride, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *__pyx_v_X_ptr, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t *__pyx_v_y_ptr, struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_4Tree_set_node_attributes *__pyx_optional_args) {
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_child_ind = ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t)-1);
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_y_start = ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t)0);
  struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node *__pyx_v_node;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t __pyx_v_x_val;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_f_ind;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_val_ptr;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_2;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_3;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_4;
  __Pyx_RefNannySetupContext("set_node_attributes", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_child_ind = __pyx_optional_args->child_ind;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_y_start = __pyx_optional_args->y_start;
      }
    }
  }

  /* "skgarden/mondrian/tree/_tree.pyx":624
 *         X_ptr[X_start: X_start + n_features*X_f_stride]
 *         """
 *         cdef Node* node = &self.nodes[node_ind]             # <<<<<<<<<<<<<<
 *         cdef Node* prev_node
 *         cdef DTYPE_t x_val
 */
  __pyx_v_node = (&(__pyx_v_self->nodes[__pyx_v_node_ind]));

  /* "skgarden/mondrian/tree/_tree.pyx":628
 *         cdef DTYPE_t x_val
 *         cdef SIZE_t f_ind
 *         cdef SIZE_t val_ptr = node_ind*self.value_stride             # <<<<<<<<<<<<<<
 * 
 *         node.left_child = left_child
 */
  __pyx_v_val_ptr = (__pyx_v_node_ind * __pyx_v_self->value_stride);

  /* "skgarden/mondrian/tree/_tree.pyx":630
 *         cdef SIZE_t val_ptr = node_ind*self.value_stride
 * 
 *         node.left_child = left_child             # <<<<<<<<<<<<<<
 *         node.right_child = right_child
 *         node.feature = feature
 */
  __pyx_v_node->left_child = __pyx_v_left_child;

  /* "skgarden/mondrian/tree/_tree.pyx":631
 * 
 *         node.left_child = left_child
 *         node.right_child = right_child             # <<<<<<<<<<<<<<
 *         node.feature = feature
 *         node.threshold = threshold
 */
  __pyx_v_node->right_child = __pyx_v_right_child;

  /* "skgarden/mondrian/tree/_tree.pyx":632
 *         node.left_child = left_child
 *         node.right_child = right_child
 *         node.feature = feature             # <<<<<<<<<<<<<<
 *         node.threshold = threshold
 *         node.tau = tau
 */
  __pyx_v_node->feature = __pyx_v_feature;

  /* "skgarden/mondrian/tree/_tree.pyx":633
 *         node.right_child = right_child
 *         node.feature = feature
 *         node.threshold = threshold             # <<<<<<<<<<<<<<
 *         node.tau = tau
 *         node.n_node_samples = n_node_samples
 */
  __pyx_v_node->threshold = __pyx_v_threshold;

  /* "skgarden/mondrian/tree/_tree.pyx":634
 *         node.feature = feature
 *         node.threshold = threshold
 *         node.tau = tau             # <<<<<<<<<<<<<<
 *         node.n_node_samples = n_node_samples
 *         node.weighted_n_node_samples = weighted_n_node_samples
 */
  __pyx_v_node->tau = __pyx_v_tau;

  /* "skgarden/mondrian/tree/_tree.pyx":635
 *         node.threshold = threshold
 *         node.tau = tau
 *         node.n_node_samples = n_node_samples             # <<<<<<<<<<<<<<
 *         node.weighted_n_node_samples = weighted_n_node_samples
 *         node.impurity = impurity
 */
  __pyx_v_node->n_node_samples = __pyx_v_n_node_samples;

  /* "skgarden/mondrian/tree/_tree.pyx":636
 *         node.tau = tau
 *         node.n_node_samples = n_node_samples
 *         node.weighted_n_node_samples = weighted_n_node_samples             # <<<<<<<<<<<<<<
 *         node.impurity = impurity
 *         node.variance = variance
 */
  __pyx_v_node->weighted_n_node_samples = __pyx_v_weighted_n_node_samples;

  /* "skgarden/mondrian/tree/_tree.pyx":637
 *         node.n_node_samples = n_node_samples
 *         node.weighted_n_node_samples = weighted_n_node_samples
 *         node.impurity = impurity             # <<<<<<<<<<<<<<
 *         node.variance = variance
 *         node.lower_bounds = <DTYPE_t*> malloc(self.n_features * sizeof(DTYPE_t))
 */
  __pyx_v_node->impurity = __pyx_v_impurity;

  /* "skgarden/mondrian/tree/_tree.pyx":638
 *         node.weighted_n_node_samples = weighted_n_node_samples
 *         node.impurity = impurity
 *         node.variance = variance             # <<<<<<<<<<<<<<
 *         node.lower_bounds = <DTYPE_t*> malloc(self.n_features * sizeof(DTYPE_t))
 *         node.upper_bounds = <DTYPE_t*> malloc(self.n_features * sizeof(DTYPE_t))
 */
  __pyx_v_node->variance = __pyx_v_variance;

  /* "skgarden/mondrian/tree/_tree.pyx":639
 *         node.impurity = impurity
 *         node.variance = variance
 *         node.lower_bounds = <DTYPE_t*> malloc(self.n_features * sizeof(DTYPE_t))             # <<<<<<<<<<<<<<
 *         node.upper_bounds = <DTYPE_t*> malloc(self.n_features * sizeof(DTYPE_t))
 * 
 */
  __pyx_v_node->lower_bounds = ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *)malloc((__pyx_v_self->n_features * (sizeof(__pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t)))));

  /* "skgarden/mondrian/tree/_tree.pyx":640
 *         node.variance = variance
 *         node.lower_bounds = <DTYPE_t*> malloc(self.n_features * sizeof(DTYPE_t))
 *         node.upper_bounds = <DTYPE_t*> malloc(self.n_features * sizeof(DTYPE_t))             # <<<<<<<<<<<<<<
 * 
 *         # Set bounds.
 */
  __pyx_v_node->upper_bounds = ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *)malloc((__pyx_v_self->n_features * (sizeof(__pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t)))));

  /* "skgarden/mondrian/tree/_tree.pyx":644
 *         # Set bounds.
 *         # If child_ind is -1, its a leaf, else update the extent of each node.
 *         if child_ind == -1:             # <<<<<<<<<<<<<<
 *             for f_ind in range(self.n_features):
 *                 x_val = X_ptr[X_start + X_f_stride*f_ind]
 */
  __pyx_t_1 = ((__pyx_v_child_ind == -1L) != 0);
  if (__pyx_t_1) {

    /* "skgarden/mondrian/tree/_tree.pyx":645
 *         # If child_ind is -1, its a leaf, else update the extent of each node.
 *         if child_ind == -1:
 *             for f_ind in range(self.n_features):             # <<<<<<<<<<<<<<
 *                 x_val = X_ptr[X_start + X_f_stride*f_ind]
 *                 node.lower_bounds[f_ind] = node.upper_bounds[f_ind] = x_val
 */
    __pyx_t_2 = __pyx_v_self->n_features;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_f_ind = __pyx_t_4;

      /* "skgarden/mondrian/tree/_tree.pyx":646
 *         if child_ind == -1:
 *             for f_ind in range(self.n_features):
 *                 x_val = X_ptr[X_start + X_f_stride*f_ind]             # <<<<<<<<<<<<<<
 *                 node.lower_bounds[f_ind] = node.upper_bounds[f_ind] = x_val
 *         else:
 */
      __pyx_v_x_val = (__pyx_v_X_ptr[(__pyx_v_X_start + (__pyx_v_X_f_stride * __pyx_v_f_ind))]);

      /* "skgarden/mondrian/tree/_tree.pyx":647
 *             for f_ind in range(self.n_features):
 *                 x_val = X_ptr[X_start + X_f_stride*f_ind]
 *                 node.lower_bounds[f_ind] = node.upper_bounds[f_ind] = x_val             # <<<<<<<<<<<<<<
 *         else:
 *             self.update_node_extent(
 */
      (__pyx_v_node->lower_bounds[__pyx_v_f_ind]) = __pyx_v_x_val;
      (__pyx_v_node->upper_bounds[__pyx_v_f_ind]) = __pyx_v_x_val;
    }

    /* "skgarden/mondrian/tree/_tree.pyx":644
 *         # Set bounds.
 *         # If child_ind is -1, its a leaf, else update the extent of each node.
 *         if child_ind == -1:             # <<<<<<<<<<<<<<
 *             for f_ind in range(self.n_features):
 *                 x_val = X_ptr[X_start + X_f_stride*f_ind]
 */
    goto __pyx_L3;
  }

  /* "skgarden/mondrian/tree/_tree.pyx":649
 *                 node.lower_bounds[f_ind] = node.upper_bounds[f_ind] = x_val
 *         else:
 *             self.update_node_extent(             # <<<<<<<<<<<<<<
 *                 node_ind, child_ind, X_ptr, X_start, X_f_stride)
 * 
 */
  /*else*/ {

    /* "skgarden/mondrian/tree/_tree.pyx":650
 *         else:
 *             self.update_node_extent(
 *                 node_ind, child_ind, X_ptr, X_start, X_f_stride)             # <<<<<<<<<<<<<<
 * 
 *         # Set value at node_ind
 */
    ((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->update_node_extent(__pyx_v_self, __pyx_v_node_ind, __pyx_v_child_ind, __pyx_v_X_ptr, __pyx_v_X_start, __pyx_v_X_f_stride);
  }
  __pyx_L3:;

  /* "skgarden/mondrian/tree/_tree.pyx":653
 * 
 *         # Set value at node_ind
 *         if child_ind == -1:             # <<<<<<<<<<<<<<
 *             # Regression
 *             if self.n_classes[0] == 1:
 */
  __pyx_t_1 = ((__pyx_v_child_ind == -1L) != 0);
  if (__pyx_t_1) {

    /* "skgarden/mondrian/tree/_tree.pyx":655
 *         if child_ind == -1:
 *             # Regression
 *             if self.n_classes[0] == 1:             # <<<<<<<<<<<<<<
 *                 self.value[val_ptr] = y_ptr[y_start]
 *             else:
 */
    __pyx_t_1 = (((__pyx_v_self->n_classes[0]) == 1) != 0);
    if (__pyx_t_1) {

      /* "skgarden/mondrian/tree/_tree.pyx":656
 *             # Regression
 *             if self.n_classes[0] == 1:
 *                 self.value[val_ptr] = y_ptr[y_start]             # <<<<<<<<<<<<<<
 *             else:
 *                 self.value[val_ptr + <SIZE_t> y_ptr[y_start]] = 1.0
 */
      (__pyx_v_self->value[__pyx_v_val_ptr]) = (__pyx_v_y_ptr[__pyx_v_y_start]);

      /* "skgarden/mondrian/tree/_tree.pyx":655
 *         if child_ind == -1:
 *             # Regression
 *             if self.n_classes[0] == 1:             # <<<<<<<<<<<<<<
 *                 self.value[val_ptr] = y_ptr[y_start]
 *             else:
 */
      goto __pyx_L7;
    }

    /* "skgarden/mondrian/tree/_tree.pyx":658
 *                 self.value[val_ptr] = y_ptr[y_start]
 *             else:
 *                 self.value[val_ptr + <SIZE_t> y_ptr[y_start]] = 1.0             # <<<<<<<<<<<<<<
 * 
 *     cdef void _init(self, DTYPE_t* X_ptr, DOUBLE_t* y_ptr, SIZE_t X_f_stride):
 */
    /*else*/ {
      (__pyx_v_self->value[(__pyx_v_val_ptr + ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t)(__pyx_v_y_ptr[__pyx_v_y_start])))]) = 1.0;
    }
    __pyx_L7:;

    /* "skgarden/mondrian/tree/_tree.pyx":653
 * 
 *         # Set value at node_ind
 *         if child_ind == -1:             # <<<<<<<<<<<<<<
 *             # Regression
 *             if self.n_classes[0] == 1:
 */
  }

  /* "skgarden/mondrian/tree/_tree.pyx":605
 *                     self.value[parent_ptr + c_ind] += self.value[child_ptr + c_ind]
 * 
 *     cdef void set_node_attributes(self, SIZE_t node_ind, SIZE_t left_child,             # <<<<<<<<<<<<<<
 *                                   SIZE_t right_child, SIZE_t feature, DOUBLE_t threshold,
 *                                   DTYPE_t tau, SIZE_t n_node_samples,
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "skgarden/mondrian/tree/_tree.pyx":660
 *                 self.value[val_ptr + <SIZE_t> y_ptr[y_start]] = 1.0
 * 
 *     cdef void _init(self, DTYPE_t* X_ptr, DOUBLE_t* y_ptr, SIZE_t X_f_stride):             # <<<<<<<<<<<<<<
 *         """
 *         Parameters
 */

static void __pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree__init(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *__pyx_v_X_ptr, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t *__pyx_v_y_ptr, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_X_f_stride) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_init", 0);

  /* "skgarden/mondrian/tree/_tree.pyx":670
 *         X_f_stride: SIZE_t, stride to reach consecutive feature.
 *         """
 *         self.set_node_attributes(0, _TREE_LEAF, _TREE_LEAF, _TREE_UNDEFINED,             # <<<<<<<<<<<<<<
 *                                  _TREE_UNDEFINED, INFINITY, 1, 1, 0.0, 0.0,
 *                                  0, X_f_stride, X_ptr, y_ptr)
 */
  ((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->set_node_attributes(__pyx_v_self, 0, __pyx_v_8skgarden_8mondrian_4tree_5_tree__TREE_LEAF, __pyx_v_8skgarden_8mondrian_4tree_5_tree__TREE_LEAF, __pyx_v_8skgarden_8mondrian_4tree_5_tree__TREE_UNDEFINED, __pyx_v_8skgarden_8mondrian_4tree_5_tree__TREE_UNDEFINED, __pyx_v_8skgarden_8mondrian_4tree_5_tree_INFINITY, 1, 1.0, 0.0, 0.0, 0, __pyx_v_X_f_stride, __pyx_v_X_ptr, __pyx_v_y_ptr, NULL);

  /* "skgarden/mondrian/tree/_tree.pyx":673
 *                                  _TREE_UNDEFINED, INFINITY, 1, 1, 0.0, 0.0,
 *                                  0, X_f_stride, X_ptr, y_ptr)
 *         self.node_count += 1             # <<<<<<<<<<<<<<
 * 
 *     cdef void extend(self, DTYPE_t* X_ptr, DOUBLE_t* y_ptr, SIZE_t X_start,
 */
  __pyx_v_self->node_count = (__pyx_v_self->node_count + 1);

  /* "skgarden/mondrian/tree/_tree.pyx":660
 *                 self.value[val_ptr + <SIZE_t> y_ptr[y_start]] = 1.0
 * 
 *     cdef void _init(self, DTYPE_t* X_ptr, DOUBLE_t* y_ptr, SIZE_t X_f_stride):             # <<<<<<<<<<<<<<
 *         """
 *         Parameters
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "skgarden/mondrian/tree/_tree.pyx":675
 *         self.node_count += 1
 * 
 *     cdef void extend(self, DTYPE_t* X_ptr, DOUBLE_t* y_ptr, SIZE_t X_start,             # <<<<<<<<<<<<<<
 *                      SIZE_t X_f_stride, SIZE_t y_start, UINT32_t random_state,
 *                      SIZE_t min_samples_split):
 */

static void __pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree_extend(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *__pyx_v_X_ptr, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t *__pyx_v_y_ptr, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_X_start, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_X_f_stride, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_y_start, __pyx_t_8skgarden_8mondrian_4tree_5_tree_UINT32_t __pyx_v_random_state, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_min_samples_split) {
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_curr_id;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_parent_id;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_new_child_id;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_new_parent_id;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_f_ind;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_delta;
  struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node *__pyx_v_curr_node;
  struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node *__pyx_v_parent_node;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t __pyx_v_x;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t __pyx_v_x_val;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t __pyx_v_new_rate;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *__pyx_v_e_l;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *__pyx_v_e_u;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *__pyx_v_extent;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t __pyx_v_E;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t __pyx_v_tau_parent;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t __pyx_v_l_b;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t __pyx_v_u_b;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_rc;
  double __pyx_v_xi;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_left_child;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_right_child;
  __Pyx_RefNannyDeclarations
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_1;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_2;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_4Tree__resize_c __pyx_t_7;
  struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_4Tree_set_node_attributes __pyx_t_8;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t __pyx_t_9;
  __Pyx_RefNannySetupContext("extend", 0);

  /* "skgarden/mondrian/tree/_tree.pyx":689
 *         """
 *         # Traverse the tree
 *         cdef SIZE_t curr_id = self.root             # <<<<<<<<<<<<<<
 *         cdef SIZE_t parent_id = -1
 *         cdef SIZE_t left_id
 */
  __pyx_t_1 = __pyx_v_self->root;
  __pyx_v_curr_id = __pyx_t_1;

  /* "skgarden/mondrian/tree/_tree.pyx":690
 *         # Traverse the tree
 *         cdef SIZE_t curr_id = self.root
 *         cdef SIZE_t parent_id = -1             # <<<<<<<<<<<<<<
 *         cdef SIZE_t left_id
 *         cdef SIZE_t right_id
 */
  __pyx_v_parent_id = -1;

  /* "skgarden/mondrian/tree/_tree.pyx":706
 *         cdef DTYPE_t x_val
 *         cdef DTYPE_t new_rate
 *         cdef DTYPE_t* e_l = <DTYPE_t*> malloc(self.n_features * sizeof(DTYPE_t))             # <<<<<<<<<<<<<<
 *         cdef DTYPE_t* e_u = <DTYPE_t*> malloc(self.n_features * sizeof(DTYPE_t))
 *         cdef DTYPE_t* extent = <DTYPE_t*> malloc(self.n_features * sizeof(DTYPE_t))
 */
  __pyx_v_e_l = ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *)malloc((__pyx_v_self->n_features * (sizeof(__pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t)))));

  /* "skgarden/mondrian/tree/_tree.pyx":707
 *         cdef DTYPE_t new_rate
 *         cdef DTYPE_t* e_l = <DTYPE_t*> malloc(self.n_features * sizeof(DTYPE_t))
 *         cdef DTYPE_t* e_u = <DTYPE_t*> malloc(self.n_features * sizeof(DTYPE_t))             # <<<<<<<<<<<<<<
 *         cdef DTYPE_t* extent = <DTYPE_t*> malloc(self.n_features * sizeof(DTYPE_t))
 *         cdef DTYPE_t E
 */
  __pyx_v_e_u = ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *)malloc((__pyx_v_self->n_features * (sizeof(__pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t)))));

  /* "skgarden/mondrian/tree/_tree.pyx":708
 *         cdef DTYPE_t* e_l = <DTYPE_t*> malloc(self.n_features * sizeof(DTYPE_t))
 *         cdef DTYPE_t* e_u = <DTYPE_t*> malloc(self.n_features * sizeof(DTYPE_t))
 *         cdef DTYPE_t* extent = <DTYPE_t*> malloc(self.n_features * sizeof(DTYPE_t))             # <<<<<<<<<<<<<<
 *         cdef DTYPE_t E
 *         cdef DTYPE_t tau_parent = 0.0
 */
  __pyx_v_extent = ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *)malloc((__pyx_v_self->n_features * (sizeof(__pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t)))));

  /* "skgarden/mondrian/tree/_tree.pyx":710
 *         cdef DTYPE_t* extent = <DTYPE_t*> malloc(self.n_features * sizeof(DTYPE_t))
 *         cdef DTYPE_t E
 *         cdef DTYPE_t tau_parent = 0.0             # <<<<<<<<<<<<<<
 *         cdef DTYPE_t threshold
 *         cdef DTYPE_t l_b
 */
  __pyx_v_tau_parent = 0.0;

  /* "skgarden/mondrian/tree/_tree.pyx":717
 *         cdef SIZE_t rc
 * 
 *         while True:             # <<<<<<<<<<<<<<
 *             curr_node = &self.nodes[curr_id]
 * 
 */
  while (1) {

    /* "skgarden/mondrian/tree/_tree.pyx":718
 * 
 *         while True:
 *             curr_node = &self.nodes[curr_id]             # <<<<<<<<<<<<<<
 * 
 *             # Step 1: Calculate e^l, e^u and rate.
 */
    __pyx_v_curr_node = (&(__pyx_v_self->nodes[__pyx_v_curr_id]));

    /* "skgarden/mondrian/tree/_tree.pyx":722
 *             # Step 1: Calculate e^l, e^u and rate.
 *             # If x belongs to the bounding box, this is zero.
 *             new_rate = 0.0             # <<<<<<<<<<<<<<
 *             for f_ind in range(self.n_features):
 *                 x = X_ptr[X_start + f_ind*X_f_stride]
 */
    __pyx_v_new_rate = 0.0;

    /* "skgarden/mondrian/tree/_tree.pyx":723
 *             # If x belongs to the bounding box, this is zero.
 *             new_rate = 0.0
 *             for f_ind in range(self.n_features):             # <<<<<<<<<<<<<<
 *                 x = X_ptr[X_start + f_ind*X_f_stride]
 *                 e_l[f_ind] = fmax(curr_node.lower_bounds[f_ind] - x, 0)
 */
    __pyx_t_1 = __pyx_v_self->n_features;
    __pyx_t_2 = __pyx_t_1;
    for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
      __pyx_v_f_ind = __pyx_t_3;

      /* "skgarden/mondrian/tree/_tree.pyx":724
 *             new_rate = 0.0
 *             for f_ind in range(self.n_features):
 *                 x = X_ptr[X_start + f_ind*X_f_stride]             # <<<<<<<<<<<<<<
 *                 e_l[f_ind] = fmax(curr_node.lower_bounds[f_ind] - x, 0)
 *                 e_u[f_ind] = fmax(x - curr_node.upper_bounds[f_ind], 0)
 */
      __pyx_v_x = (__pyx_v_X_ptr[(__pyx_v_X_start + (__pyx_v_f_ind * __pyx_v_X_f_stride))]);

      /* "skgarden/mondrian/tree/_tree.pyx":725
 *             for f_ind in range(self.n_features):
 *                 x = X_ptr[X_start + f_ind*X_f_stride]
 *                 e_l[f_ind] = fmax(curr_node.lower_bounds[f_ind] - x, 0)             # <<<<<<<<<<<<<<
 *                 e_u[f_ind] = fmax(x - curr_node.upper_bounds[f_ind], 0)
 *                 extent[f_ind] = e_l[f_ind] + e_u[f_ind]
 */
      (__pyx_v_e_l[__pyx_v_f_ind]) = __pyx_f_8skgarden_8mondrian_4tree_5_tree_fmax(((__pyx_v_curr_node->lower_bounds[__pyx_v_f_ind]) - __pyx_v_x), 0.0);

      /* "skgarden/mondrian/tree/_tree.pyx":726
 *                 x = X_ptr[X_start + f_ind*X_f_stride]
 *                 e_l[f_ind] = fmax(curr_node.lower_bounds[f_ind] - x, 0)
 *                 e_u[f_ind] = fmax(x - curr_node.upper_bounds[f_ind], 0)             # <<<<<<<<<<<<<<
 *                 extent[f_ind] = e_l[f_ind] + e_u[f_ind]
 *                 new_rate += extent[f_ind]
 */
      (__pyx_v_e_u[__pyx_v_f_ind]) = __pyx_f_8skgarden_8mondrian_4tree_5_tree_fmax((__pyx_v_x - (__pyx_v_curr_node->upper_bounds[__pyx_v_f_ind])), 0.0);

      /* "skgarden/mondrian/tree/_tree.pyx":727
 *                 e_l[f_ind] = fmax(curr_node.lower_bounds[f_ind] - x, 0)
 *                 e_u[f_ind] = fmax(x - curr_node.upper_bounds[f_ind], 0)
 *                 extent[f_ind] = e_l[f_ind] + e_u[f_ind]             # <<<<<<<<<<<<<<
 *                 new_rate += extent[f_ind]
 * 
 */
      (__pyx_v_extent[__pyx_v_f_ind]) = ((__pyx_v_e_l[__pyx_v_f_ind]) + (__pyx_v_e_u[__pyx_v_f_ind]));

      /* "skgarden/mondrian/tree/_tree.pyx":728
 *                 e_u[f_ind] = fmax(x - curr_node.upper_bounds[f_ind], 0)
 *                 extent[f_ind] = e_l[f_ind] + e_u[f_ind]
 *                 new_rate += extent[f_ind]             # <<<<<<<<<<<<<<
 * 
 *             # Step 2: Sample E from an exponential distribution.
 */
      __pyx_v_new_rate = (__pyx_v_new_rate + (__pyx_v_extent[__pyx_v_f_ind]));
    }

    /* "skgarden/mondrian/tree/_tree.pyx":731
 * 
 *             # Step 2: Sample E from an exponential distribution.
 *             E = rand_exponential(new_rate, &random_state)             # <<<<<<<<<<<<<<
 * 
 *             # Step 3: Induce split.
 */
    __pyx_v_E = __pyx_f_8skgarden_8mondrian_4tree_6_utils_rand_exponential(__pyx_v_new_rate, (&__pyx_v_random_state));

    /* "skgarden/mondrian/tree/_tree.pyx":738
 *             # 2. A parent node with the new child node and the node at
 *             # curr_id as children.
 *             if (tau_parent + E < curr_node.tau and             # <<<<<<<<<<<<<<
 *                 curr_node.n_node_samples + 1 >= min_samples_split):
 * 
 */
    __pyx_t_5 = (((__pyx_v_tau_parent + __pyx_v_E) < __pyx_v_curr_node->tau) != 0);
    if (__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L8_bool_binop_done;
    }

    /* "skgarden/mondrian/tree/_tree.pyx":739
 *             # curr_id as children.
 *             if (tau_parent + E < curr_node.tau and
 *                 curr_node.n_node_samples + 1 >= min_samples_split):             # <<<<<<<<<<<<<<
 * 
 *                 new_child_id = self.node_count
 */
    __pyx_t_5 = (((__pyx_v_curr_node->n_node_samples + 1) >= __pyx_v_min_samples_split) != 0);
    __pyx_t_4 = __pyx_t_5;
    __pyx_L8_bool_binop_done:;

    /* "skgarden/mondrian/tree/_tree.pyx":738
 *             # 2. A parent node with the new child node and the node at
 *             # curr_id as children.
 *             if (tau_parent + E < curr_node.tau and             # <<<<<<<<<<<<<<
 *                 curr_node.n_node_samples + 1 >= min_samples_split):
 * 
 */
    if (__pyx_t_4) {

      /* "skgarden/mondrian/tree/_tree.pyx":741
 *                 curr_node.n_node_samples + 1 >= min_samples_split):
 * 
 *                 new_child_id = self.node_count             # <<<<<<<<<<<<<<
 *                 new_parent_id = self.node_count + 1
 * 
 */
      __pyx_t_1 = __pyx_v_self->node_count;
      __pyx_v_new_child_id = __pyx_t_1;

      /* "skgarden/mondrian/tree/_tree.pyx":742
 * 
 *                 new_child_id = self.node_count
 *                 new_parent_id = self.node_count + 1             # <<<<<<<<<<<<<<
 * 
 *                 # Step 4: Sample delta from a multinomial.
 */
      __pyx_v_new_parent_id = (__pyx_v_self->node_count + 1);

      /* "skgarden/mondrian/tree/_tree.pyx":745
 * 
 *                 # Step 4: Sample delta from a multinomial.
 *                 delta = rand_multinomial(extent, self.n_features, &random_state)             # <<<<<<<<<<<<<<
 * 
 *                 # Step 5: Sample xi uniformly between bounds.
 */
      __pyx_v_delta = __pyx_f_8skgarden_8mondrian_4tree_6_utils_rand_multinomial(__pyx_v_extent, __pyx_v_self->n_features, (&__pyx_v_random_state));

      /* "skgarden/mondrian/tree/_tree.pyx":748
 * 
 *                 # Step 5: Sample xi uniformly between bounds.
 *                 x_val = X_ptr[X_start + delta * X_f_stride]             # <<<<<<<<<<<<<<
 *                 l_b = curr_node.lower_bounds[delta]
 *                 u_b = curr_node.upper_bounds[delta]
 */
      __pyx_v_x_val = (__pyx_v_X_ptr[(__pyx_v_X_start + (__pyx_v_delta * __pyx_v_X_f_stride))]);

      /* "skgarden/mondrian/tree/_tree.pyx":749
 *                 # Step 5: Sample xi uniformly between bounds.
 *                 x_val = X_ptr[X_start + delta * X_f_stride]
 *                 l_b = curr_node.lower_bounds[delta]             # <<<<<<<<<<<<<<
 *                 u_b = curr_node.upper_bounds[delta]
 *                 if x_val > u_b:
 */
      __pyx_v_l_b = (__pyx_v_curr_node->lower_bounds[__pyx_v_delta]);

      /* "skgarden/mondrian/tree/_tree.pyx":750
 *                 x_val = X_ptr[X_start + delta * X_f_stride]
 *                 l_b = curr_node.lower_bounds[delta]
 *                 u_b = curr_node.upper_bounds[delta]             # <<<<<<<<<<<<<<
 *                 if x_val > u_b:
 *                     xi = rand_uniform(u_b, x_val, &random_state)
 */
      __pyx_v_u_b = (__pyx_v_curr_node->upper_bounds[__pyx_v_delta]);

      /* "skgarden/mondrian/tree/_tree.pyx":751
 *                 l_b = curr_node.lower_bounds[delta]
 *                 u_b = curr_node.upper_bounds[delta]
 *                 if x_val > u_b:             # <<<<<<<<<<<<<<
 *                     xi = rand_uniform(u_b, x_val, &random_state)
 *                 else:
 */
      __pyx_t_4 = ((__pyx_v_x_val > __pyx_v_u_b) != 0);
      if (__pyx_t_4) {

        /* "skgarden/mondrian/tree/_tree.pyx":752
 *                 u_b = curr_node.upper_bounds[delta]
 *                 if x_val > u_b:
 *                     xi = rand_uniform(u_b, x_val, &random_state)             # <<<<<<<<<<<<<<
 *                 else:
 *                     xi = rand_uniform(x_val, l_b, &random_state)
 */
        __pyx_v_xi = __pyx_f_8skgarden_8mondrian_4tree_6_utils_rand_uniform(__pyx_v_u_b, __pyx_v_x_val, (&__pyx_v_random_state));

        /* "skgarden/mondrian/tree/_tree.pyx":751
 *                 l_b = curr_node.lower_bounds[delta]
 *                 u_b = curr_node.upper_bounds[delta]
 *                 if x_val > u_b:             # <<<<<<<<<<<<<<
 *                     xi = rand_uniform(u_b, x_val, &random_state)
 *                 else:
 */
        goto __pyx_L10;
      }

      /* "skgarden/mondrian/tree/_tree.pyx":754
 *                     xi = rand_uniform(u_b, x_val, &random_state)
 *                 else:
 *                     xi = rand_uniform(x_val, l_b, &random_state)             # <<<<<<<<<<<<<<
 * 
 *                 # Step 7: Split criteria.
 */
      /*else*/ {
        __pyx_v_xi = __pyx_f_8skgarden_8mondrian_4tree_6_utils_rand_uniform(__pyx_v_x_val, __pyx_v_l_b, (&__pyx_v_random_state));
      }
      __pyx_L10:;

      /* "skgarden/mondrian/tree/_tree.pyx":757
 * 
 *                 # Step 7: Split criteria.
 *                 if x_val < xi:             # <<<<<<<<<<<<<<
 *                     left_child = new_child_id
 *                     right_child = curr_id
 */
      __pyx_t_4 = ((__pyx_v_x_val < __pyx_v_xi) != 0);
      if (__pyx_t_4) {

        /* "skgarden/mondrian/tree/_tree.pyx":758
 *                 # Step 7: Split criteria.
 *                 if x_val < xi:
 *                     left_child = new_child_id             # <<<<<<<<<<<<<<
 *                     right_child = curr_id
 *                 else:
 */
        __pyx_v_left_child = __pyx_v_new_child_id;

        /* "skgarden/mondrian/tree/_tree.pyx":759
 *                 if x_val < xi:
 *                     left_child = new_child_id
 *                     right_child = curr_id             # <<<<<<<<<<<<<<
 *                 else:
 *                     left_child = curr_id
 */
        __pyx_v_right_child = __pyx_v_curr_id;

        /* "skgarden/mondrian/tree/_tree.pyx":757
 * 
 *                 # Step 7: Split criteria.
 *                 if x_val < xi:             # <<<<<<<<<<<<<<
 *                     left_child = new_child_id
 *                     right_child = curr_id
 */
        goto __pyx_L11;
      }

      /* "skgarden/mondrian/tree/_tree.pyx":761
 *                     right_child = curr_id
 *                 else:
 *                     left_child = curr_id             # <<<<<<<<<<<<<<
 *                     right_child = new_child_id
 * 
 */
      /*else*/ {
        __pyx_v_left_child = __pyx_v_curr_id;

        /* "skgarden/mondrian/tree/_tree.pyx":762
 *                 else:
 *                     left_child = curr_id
 *                     right_child = new_child_id             # <<<<<<<<<<<<<<
 * 
 *                 # Allocate memory for the new parent and child.
 */
        __pyx_v_right_child = __pyx_v_new_child_id;
      }
      __pyx_L11:;

      /* "skgarden/mondrian/tree/_tree.pyx":767
 *                 # Store leaf in nodes[self.node_count]
 *                 # Store parent in nodes[self.node_count + 1]
 *                 rc = self._resize_c(self.node_count + 2)             # <<<<<<<<<<<<<<
 *                 if rc == -1:
 *                     raise MemoryError()
 */
      __pyx_t_7.__pyx_n = 1;
      __pyx_t_7.capacity = (__pyx_v_self->node_count + 2);
      __pyx_t_6 = ((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->_resize_c(__pyx_v_self, &__pyx_t_7); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 767, __pyx_L1_error)
      __pyx_v_rc = __pyx_t_6;

      /* "skgarden/mondrian/tree/_tree.pyx":768
 *                 # Store parent in nodes[self.node_count + 1]
 *                 rc = self._resize_c(self.node_count + 2)
 *                 if rc == -1:             # <<<<<<<<<<<<<<
 *                     raise MemoryError()
 * 
 */
      __pyx_t_4 = ((__pyx_v_rc == -1L) != 0);
      if (unlikely(__pyx_t_4)) {

        /* "skgarden/mondrian/tree/_tree.pyx":769
 *                 rc = self._resize_c(self.node_count + 2)
 *                 if rc == -1:
 *                     raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *                 # xxx: We need to get the pointer to curr_id again
 */
        PyErr_NoMemory(); __PYX_ERR(0, 769, __pyx_L1_error)

        /* "skgarden/mondrian/tree/_tree.pyx":768
 *                 # Store parent in nodes[self.node_count + 1]
 *                 rc = self._resize_c(self.node_count + 2)
 *                 if rc == -1:             # <<<<<<<<<<<<<<
 *                     raise MemoryError()
 * 
 */
      }

      /* "skgarden/mondrian/tree/_tree.pyx":773
 *                 # xxx: We need to get the pointer to curr_id again
 *                 # because of the resizing above.
 *                 curr_node = &self.nodes[curr_id]             # <<<<<<<<<<<<<<
 * 
 *                 # Step 7-8: Create new leaf node j'' and update value.
 */
      __pyx_v_curr_node = (&(__pyx_v_self->nodes[__pyx_v_curr_id]));

      /* "skgarden/mondrian/tree/_tree.pyx":776
 * 
 *                 # Step 7-8: Create new leaf node j'' and update value.
 *                 self.set_node_attributes(             # <<<<<<<<<<<<<<
 *                     new_child_id, _TREE_LEAF, _TREE_LEAF, _TREE_UNDEFINED,
 *                     _TREE_UNDEFINED, INFINITY, 1, 1, 0.0, 0.0, X_start,
 */
      __pyx_t_8.__pyx_n = 2;
      __pyx_t_8.child_ind = -1;
      __pyx_t_8.y_start = __pyx_v_y_start;
      ((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->set_node_attributes(__pyx_v_self, __pyx_v_new_child_id, __pyx_v_8skgarden_8mondrian_4tree_5_tree__TREE_LEAF, __pyx_v_8skgarden_8mondrian_4tree_5_tree__TREE_LEAF, __pyx_v_8skgarden_8mondrian_4tree_5_tree__TREE_UNDEFINED, __pyx_v_8skgarden_8mondrian_4tree_5_tree__TREE_UNDEFINED, __pyx_v_8skgarden_8mondrian_4tree_5_tree_INFINITY, 1, 1.0, 0.0, 0.0, __pyx_v_X_start, __pyx_v_X_f_stride, __pyx_v_X_ptr, __pyx_v_y_ptr, &__pyx_t_8); 

      /* "skgarden/mondrian/tree/_tree.pyx":782
 * 
 *                 # Step 6 : Create new parent node j'
 *                 self.set_node_attributes(             # <<<<<<<<<<<<<<
 *                     new_parent_id, left_child, right_child, delta, xi,
 *                     tau_parent + E, curr_node.n_node_samples + 1,
 */
      __pyx_t_8.__pyx_n = 1;
      __pyx_t_8.child_ind = __pyx_v_curr_id;
      ((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->set_node_attributes(__pyx_v_self, __pyx_v_new_parent_id, __pyx_v_left_child, __pyx_v_right_child, __pyx_v_delta, __pyx_v_xi, (__pyx_v_tau_parent + __pyx_v_E), (__pyx_v_curr_node->n_node_samples + 1), (__pyx_v_curr_node->weighted_n_node_samples + 1.0), 0.0, 0.0, __pyx_v_X_start, __pyx_v_X_f_stride, __pyx_v_X_ptr, __pyx_v_y_ptr, &__pyx_t_8); 

      /* "skgarden/mondrian/tree/_tree.pyx":787
 *                     curr_node.weighted_n_node_samples + 1, 0.0, 0.0, X_start,
 *                     X_f_stride, X_ptr, y_ptr, curr_id)
 *                 self._update_node_info(new_parent_id, curr_id, y_ptr, y_start)             # <<<<<<<<<<<<<<
 * 
 *                 # New root if curr_id is root.
 */
      ((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->_update_node_info(__pyx_v_self, __pyx_v_new_parent_id, __pyx_v_curr_id, __pyx_v_y_ptr, __pyx_v_y_start);

      /* "skgarden/mondrian/tree/_tree.pyx":790
 * 
 *                 # New root if curr_id is root.
 *                 if curr_id == self.root:             # <<<<<<<<<<<<<<
 *                     self.root = new_parent_id
 *                 else:
 */
      __pyx_t_4 = ((__pyx_v_curr_id == __pyx_v_self->root) != 0);
      if (__pyx_t_4) {

        /* "skgarden/mondrian/tree/_tree.pyx":791
 *                 # New root if curr_id is root.
 *                 if curr_id == self.root:
 *                     self.root = new_parent_id             # <<<<<<<<<<<<<<
 *                 else:
 *                     # Link to the newly created node j' (new_parent_id)
 */
        __pyx_v_self->root = __pyx_v_new_parent_id;

        /* "skgarden/mondrian/tree/_tree.pyx":790
 * 
 *                 # New root if curr_id is root.
 *                 if curr_id == self.root:             # <<<<<<<<<<<<<<
 *                     self.root = new_parent_id
 *                 else:
 */
        goto __pyx_L13;
      }

      /* "skgarden/mondrian/tree/_tree.pyx":795
 *                     # Link to the newly created node j' (new_parent_id)
 *                     # as the child of the parent of node j (curr_id)
 *                     parent_node = &self.nodes[parent_id]             # <<<<<<<<<<<<<<
 *                     if parent_node.left_child == curr_id:
 *                         parent_node.left_child = new_parent_id
 */
      /*else*/ {
        __pyx_v_parent_node = (&(__pyx_v_self->nodes[__pyx_v_parent_id]));

        /* "skgarden/mondrian/tree/_tree.pyx":796
 *                     # as the child of the parent of node j (curr_id)
 *                     parent_node = &self.nodes[parent_id]
 *                     if parent_node.left_child == curr_id:             # <<<<<<<<<<<<<<
 *                         parent_node.left_child = new_parent_id
 *                     else:
 */
        __pyx_t_4 = ((__pyx_v_parent_node->left_child == __pyx_v_curr_id) != 0);
        if (__pyx_t_4) {

          /* "skgarden/mondrian/tree/_tree.pyx":797
 *                     parent_node = &self.nodes[parent_id]
 *                     if parent_node.left_child == curr_id:
 *                         parent_node.left_child = new_parent_id             # <<<<<<<<<<<<<<
 *                     else:
 *                         parent_node.right_child = new_parent_id
 */
          __pyx_v_parent_node->left_child = __pyx_v_new_parent_id;

          /* "skgarden/mondrian/tree/_tree.pyx":796
 *                     # as the child of the parent of node j (curr_id)
 *                     parent_node = &self.nodes[parent_id]
 *                     if parent_node.left_child == curr_id:             # <<<<<<<<<<<<<<
 *                         parent_node.left_child = new_parent_id
 *                     else:
 */
          goto __pyx_L14;
        }

        /* "skgarden/mondrian/tree/_tree.pyx":799
 *                         parent_node.left_child = new_parent_id
 *                     else:
 *                         parent_node.right_child = new_parent_id             # <<<<<<<<<<<<<<
 *                 self.max_depth += 1
 *                 self.node_count += 2
 */
        /*else*/ {
          __pyx_v_parent_node->right_child = __pyx_v_new_parent_id;
        }
        __pyx_L14:;
      }
      __pyx_L13:;

      /* "skgarden/mondrian/tree/_tree.pyx":800
 *                     else:
 *                         parent_node.right_child = new_parent_id
 *                 self.max_depth += 1             # <<<<<<<<<<<<<<
 *                 self.node_count += 2
 *                 break
 */
      __pyx_v_self->max_depth = (__pyx_v_self->max_depth + 1);

      /* "skgarden/mondrian/tree/_tree.pyx":801
 *                         parent_node.right_child = new_parent_id
 *                 self.max_depth += 1
 *                 self.node_count += 2             # <<<<<<<<<<<<<<
 *                 break
 * 
 */
      __pyx_v_self->node_count = (__pyx_v_self->node_count + 2);

      /* "skgarden/mondrian/tree/_tree.pyx":802
 *                 self.max_depth += 1
 *                 self.node_count += 2
 *                 break             # <<<<<<<<<<<<<<
 * 
 *             # Absorb new sample into curr_id and Traverse further down the
 */
      goto __pyx_L4_break;

      /* "skgarden/mondrian/tree/_tree.pyx":738
 *             # 2. A parent node with the new child node and the node at
 *             # curr_id as children.
 *             if (tau_parent + E < curr_node.tau and             # <<<<<<<<<<<<<<
 *                 curr_node.n_node_samples + 1 >= min_samples_split):
 * 
 */
    }

    /* "skgarden/mondrian/tree/_tree.pyx":809
 *                 # Step 10: Update extent, value at node curr_id and increment
 *                 # the number of samples.
 *                 self.update_node_extent(curr_id, curr_id, X_ptr, X_start, X_f_stride)             # <<<<<<<<<<<<<<
 *                 self._update_node_info(curr_id, curr_id, y_ptr, y_start)
 *                 curr_node.n_node_samples += 1
 */
    /*else*/ {
      ((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->update_node_extent(__pyx_v_self, __pyx_v_curr_id, __pyx_v_curr_id, __pyx_v_X_ptr, __pyx_v_X_start, __pyx_v_X_f_stride);

      /* "skgarden/mondrian/tree/_tree.pyx":810
 *                 # the number of samples.
 *                 self.update_node_extent(curr_id, curr_id, X_ptr, X_start, X_f_stride)
 *                 self._update_node_info(curr_id, curr_id, y_ptr, y_start)             # <<<<<<<<<<<<<<
 *                 curr_node.n_node_samples += 1
 *                 curr_node.weighted_n_node_samples += 1
 */
      ((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->_update_node_info(__pyx_v_self, __pyx_v_curr_id, __pyx_v_curr_id, __pyx_v_y_ptr, __pyx_v_y_start);

      /* "skgarden/mondrian/tree/_tree.pyx":811
 *                 self.update_node_extent(curr_id, curr_id, X_ptr, X_start, X_f_stride)
 *                 self._update_node_info(curr_id, curr_id, y_ptr, y_start)
 *                 curr_node.n_node_samples += 1             # <<<<<<<<<<<<<<
 *                 curr_node.weighted_n_node_samples += 1
 * 
 */
      __pyx_v_curr_node->n_node_samples = (__pyx_v_curr_node->n_node_samples + 1);

      /* "skgarden/mondrian/tree/_tree.pyx":812
 *                 self._update_node_info(curr_id, curr_id, y_ptr, y_start)
 *                 curr_node.n_node_samples += 1
 *                 curr_node.weighted_n_node_samples += 1             # <<<<<<<<<<<<<<
 * 
 *                 if curr_node.left_child == -1:
 */
      __pyx_v_curr_node->weighted_n_node_samples = (__pyx_v_curr_node->weighted_n_node_samples + 1.0);

      /* "skgarden/mondrian/tree/_tree.pyx":814
 *                 curr_node.weighted_n_node_samples += 1
 * 
 *                 if curr_node.left_child == -1:             # <<<<<<<<<<<<<<
 *                     break
 * 
 */
      __pyx_t_4 = ((__pyx_v_curr_node->left_child == -1L) != 0);
      if (__pyx_t_4) {

        /* "skgarden/mondrian/tree/_tree.pyx":815
 * 
 *                 if curr_node.left_child == -1:
 *                     break             # <<<<<<<<<<<<<<
 * 
 *                 # Step 12 - 13: Recurse down the tree.
 */
        goto __pyx_L4_break;

        /* "skgarden/mondrian/tree/_tree.pyx":814
 *                 curr_node.weighted_n_node_samples += 1
 * 
 *                 if curr_node.left_child == -1:             # <<<<<<<<<<<<<<
 *                     break
 * 
 */
      }

      /* "skgarden/mondrian/tree/_tree.pyx":818
 * 
 *                 # Step 12 - 13: Recurse down the tree.
 *                 parent_id = curr_id             # <<<<<<<<<<<<<<
 *                 if X_ptr[X_start + curr_node.feature*X_f_stride] < curr_node.threshold:
 *                     curr_id = curr_node.left_child
 */
      __pyx_v_parent_id = __pyx_v_curr_id;

      /* "skgarden/mondrian/tree/_tree.pyx":819
 *                 # Step 12 - 13: Recurse down the tree.
 *                 parent_id = curr_id
 *                 if X_ptr[X_start + curr_node.feature*X_f_stride] < curr_node.threshold:             # <<<<<<<<<<<<<<
 *                     curr_id = curr_node.left_child
 *                 else:
 */
      __pyx_t_4 = (((__pyx_v_X_ptr[(__pyx_v_X_start + (__pyx_v_curr_node->feature * __pyx_v_X_f_stride))]) < __pyx_v_curr_node->threshold) != 0);
      if (__pyx_t_4) {

        /* "skgarden/mondrian/tree/_tree.pyx":820
 *                 parent_id = curr_id
 *                 if X_ptr[X_start + curr_node.feature*X_f_stride] < curr_node.threshold:
 *                     curr_id = curr_node.left_child             # <<<<<<<<<<<<<<
 *                 else:
 *                     curr_id = curr_node.right_child
 */
        __pyx_t_1 = __pyx_v_curr_node->left_child;
        __pyx_v_curr_id = __pyx_t_1;

        /* "skgarden/mondrian/tree/_tree.pyx":819
 *                 # Step 12 - 13: Recurse down the tree.
 *                 parent_id = curr_id
 *                 if X_ptr[X_start + curr_node.feature*X_f_stride] < curr_node.threshold:             # <<<<<<<<<<<<<<
 *                     curr_id = curr_node.left_child
 *                 else:
 */
        goto __pyx_L16;
      }

      /* "skgarden/mondrian/tree/_tree.pyx":822
 *                     curr_id = curr_node.left_child
 *                 else:
 *                     curr_id = curr_node.right_child             # <<<<<<<<<<<<<<
 *                 tau_parent = curr_node.tau
 *         free(e_l)
 */
      /*else*/ {
        __pyx_t_1 = __pyx_v_curr_node->right_child;
        __pyx_v_curr_id = __pyx_t_1;
      }
      __pyx_L16:;

      /* "skgarden/mondrian/tree/_tree.pyx":823
 *                 else:
 *                     curr_id = curr_node.right_child
 *                 tau_parent = curr_node.tau             # <<<<<<<<<<<<<<
 *         free(e_l)
 *         free(e_u)
 */
      __pyx_t_9 = __pyx_v_curr_node->tau;
      __pyx_v_tau_parent = __pyx_t_9;
    }
  }
  __pyx_L4_break:;

  /* "skgarden/mondrian/tree/_tree.pyx":824
 *                     curr_id = curr_node.right_child
 *                 tau_parent = curr_node.tau
 *         free(e_l)             # <<<<<<<<<<<<<<
 *         free(e_u)
 *         free(extent)
 */
  free(__pyx_v_e_l);

  /* "skgarden/mondrian/tree/_tree.pyx":825
 *                 tau_parent = curr_node.tau
 *         free(e_l)
 *         free(e_u)             # <<<<<<<<<<<<<<
 *         free(extent)
 * 
 */
  free(__pyx_v_e_u);

  /* "skgarden/mondrian/tree/_tree.pyx":826
 *         free(e_l)
 *         free(e_u)
 *         free(extent)             # <<<<<<<<<<<<<<
 * 
 *     cdef SIZE_t _add_node(self, SIZE_t parent, bint is_left, bint is_leaf,
 */
  free(__pyx_v_extent);

  /* "skgarden/mondrian/tree/_tree.pyx":675
 *         self.node_count += 1
 * 
 *     cdef void extend(self, DTYPE_t* X_ptr, DOUBLE_t* y_ptr, SIZE_t X_start,             # <<<<<<<<<<<<<<
 *                      SIZE_t X_f_stride, SIZE_t y_start, UINT32_t random_state,
 *                      SIZE_t min_samples_split):
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("skgarden.mondrian.tree._tree.Tree.extend", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "skgarden/mondrian/tree/_tree.pyx":828
 *         free(extent)
 * 
 *     cdef SIZE_t _add_node(self, SIZE_t parent, bint is_left, bint is_leaf,             # <<<<<<<<<<<<<<
 *                           SIZE_t feature, double threshold, double impurity,
 *                           SIZE_t n_node_samples,
 */

static __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree__add_node(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_parent, int __pyx_v_is_left, int __pyx_v_is_leaf, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_feature, double __pyx_v_threshold, double __pyx_v_impurity, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_n_node_samples, double __pyx_v_weighted_n_node_samples, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *__pyx_v_lower_bounds, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *__pyx_v_upper_bounds, double __pyx_v_E) {
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_node_id;
  struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node *__pyx_v_node;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_r;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;

  /* "skgarden/mondrian/tree/_tree.pyx":841
 *         Returns (size_t)(-1) on error.
 *         """
 *         cdef SIZE_t node_id = self.node_count             # <<<<<<<<<<<<<<
 * 
 *         if node_id >= self.capacity:
 */
  __pyx_t_1 = __pyx_v_self->node_count;
  __pyx_v_node_id = __pyx_t_1;

  /* "skgarden/mondrian/tree/_tree.pyx":843
 *         cdef SIZE_t node_id = self.node_count
 * 
 *         if node_id >= self.capacity:             # <<<<<<<<<<<<<<
 *             if self._resize_c() != 0:
 *                 return <SIZE_t>(-1)
 */
  __pyx_t_2 = ((__pyx_v_node_id >= __pyx_v_self->capacity) != 0);
  if (__pyx_t_2) {

    /* "skgarden/mondrian/tree/_tree.pyx":844
 * 
 *         if node_id >= self.capacity:
 *             if self._resize_c() != 0:             # <<<<<<<<<<<<<<
 *                 return <SIZE_t>(-1)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->_resize_c(__pyx_v_self, NULL); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 844, __pyx_L1_error)
    __pyx_t_2 = ((__pyx_t_3 != 0) != 0);
    if (__pyx_t_2) {

      /* "skgarden/mondrian/tree/_tree.pyx":845
 *         if node_id >= self.capacity:
 *             if self._resize_c() != 0:
 *                 return <SIZE_t>(-1)             # <<<<<<<<<<<<<<
 * 
 *         cdef Node* node = &self.nodes[node_id]
 */
      __pyx_r = ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t)-1L);
      goto __pyx_L0;

      /* "skgarden/mondrian/tree/_tree.pyx":844
 * 
 *         if node_id >= self.capacity:
 *             if self._resize_c() != 0:             # <<<<<<<<<<<<<<
 *                 return <SIZE_t>(-1)
 * 
 */
    }

    /* "skgarden/mondrian/tree/_tree.pyx":843
 *         cdef SIZE_t node_id = self.node_count
 * 
 *         if node_id >= self.capacity:             # <<<<<<<<<<<<<<
 *             if self._resize_c() != 0:
 *                 return <SIZE_t>(-1)
 */
  }

  /* "skgarden/mondrian/tree/_tree.pyx":847
 *                 return <SIZE_t>(-1)
 * 
 *         cdef Node* node = &self.nodes[node_id]             # <<<<<<<<<<<<<<
 *         node.impurity = impurity
 *         node.n_node_samples = n_node_samples
 */
  __pyx_v_node = (&(__pyx_v_self->nodes[__pyx_v_node_id]));

  /* "skgarden/mondrian/tree/_tree.pyx":848
 * 
 *         cdef Node* node = &self.nodes[node_id]
 *         node.impurity = impurity             # <<<<<<<<<<<<<<
 *         node.n_node_samples = n_node_samples
 *         node.weighted_n_node_samples = weighted_n_node_samples
 */
  __pyx_v_node->impurity = __pyx_v_impurity;

  /* "skgarden/mondrian/tree/_tree.pyx":849
 *         cdef Node* node = &self.nodes[node_id]
 *         node.impurity = impurity
 *         node.n_node_samples = n_node_samples             # <<<<<<<<<<<<<<
 *         node.weighted_n_node_samples = weighted_n_node_samples
 * 
 */
  __pyx_v_node->n_node_samples = __pyx_v_n_node_samples;

  /* "skgarden/mondrian/tree/_tree.pyx":850
 *         node.impurity = impurity
 *         node.n_node_samples = n_node_samples
 *         node.weighted_n_node_samples = weighted_n_node_samples             # <<<<<<<<<<<<<<
 * 
 *         if parent != _TREE_UNDEFINED:
 */
  __pyx_v_node->weighted_n_node_samples = __pyx_v_weighted_n_node_samples;

  /* "skgarden/mondrian/tree/_tree.pyx":852
 *         node.weighted_n_node_samples = weighted_n_node_samples
 * 
 *         if parent != _TREE_UNDEFINED:             # <<<<<<<<<<<<<<
 *             if is_left:
 *                 self.nodes[parent].left_child = node_id
 */
  __pyx_t_2 = ((__pyx_v_parent != __pyx_v_8skgarden_8mondrian_4tree_5_tree__TREE_UNDEFINED) != 0);
  if (__pyx_t_2) {

    /* "skgarden/mondrian/tree/_tree.pyx":853
 * 
 *         if parent != _TREE_UNDEFINED:
 *             if is_left:             # <<<<<<<<<<<<<<
 *                 self.nodes[parent].left_child = node_id
 *             else:
 */
    __pyx_t_2 = (__pyx_v_is_left != 0);
    if (__pyx_t_2) {

      /* "skgarden/mondrian/tree/_tree.pyx":854
 *         if parent != _TREE_UNDEFINED:
 *             if is_left:
 *                 self.nodes[parent].left_child = node_id             # <<<<<<<<<<<<<<
 *             else:
 *                 self.nodes[parent].right_child = node_id
 */
      (__pyx_v_self->nodes[__pyx_v_parent]).left_child = __pyx_v_node_id;

      /* "skgarden/mondrian/tree/_tree.pyx":853
 * 
 *         if parent != _TREE_UNDEFINED:
 *             if is_left:             # <<<<<<<<<<<<<<
 *                 self.nodes[parent].left_child = node_id
 *             else:
 */
      goto __pyx_L6;
    }

    /* "skgarden/mondrian/tree/_tree.pyx":856
 *                 self.nodes[parent].left_child = node_id
 *             else:
 *                 self.nodes[parent].right_child = node_id             # <<<<<<<<<<<<<<
 * 
 *         if parent == _TREE_UNDEFINED:
 */
    /*else*/ {
      (__pyx_v_self->nodes[__pyx_v_parent]).right_child = __pyx_v_node_id;
    }
    __pyx_L6:;

    /* "skgarden/mondrian/tree/_tree.pyx":852
 *         node.weighted_n_node_samples = weighted_n_node_samples
 * 
 *         if parent != _TREE_UNDEFINED:             # <<<<<<<<<<<<<<
 *             if is_left:
 *                 self.nodes[parent].left_child = node_id
 */
  }

  /* "skgarden/mondrian/tree/_tree.pyx":858
 *                 self.nodes[parent].right_child = node_id
 * 
 *         if parent == _TREE_UNDEFINED:             # <<<<<<<<<<<<<<
 *             node.tau = E
 *         elif is_leaf:
 */
  __pyx_t_2 = ((__pyx_v_parent == __pyx_v_8skgarden_8mondrian_4tree_5_tree__TREE_UNDEFINED) != 0);
  if (__pyx_t_2) {

    /* "skgarden/mondrian/tree/_tree.pyx":859
 * 
 *         if parent == _TREE_UNDEFINED:
 *             node.tau = E             # <<<<<<<<<<<<<<
 *         elif is_leaf:
 *             node.tau = INFINITY
 */
    __pyx_v_node->tau = __pyx_v_E;

    /* "skgarden/mondrian/tree/_tree.pyx":858
 *                 self.nodes[parent].right_child = node_id
 * 
 *         if parent == _TREE_UNDEFINED:             # <<<<<<<<<<<<<<
 *             node.tau = E
 *         elif is_leaf:
 */
    goto __pyx_L7;
  }

  /* "skgarden/mondrian/tree/_tree.pyx":860
 *         if parent == _TREE_UNDEFINED:
 *             node.tau = E
 *         elif is_leaf:             # <<<<<<<<<<<<<<
 *             node.tau = INFINITY
 *         else:
 */
  __pyx_t_2 = (__pyx_v_is_leaf != 0);
  if (__pyx_t_2) {

    /* "skgarden/mondrian/tree/_tree.pyx":861
 *             node.tau = E
 *         elif is_leaf:
 *             node.tau = INFINITY             # <<<<<<<<<<<<<<
 *         else:
 *             node.tau = E + self.nodes[parent].tau
 */
    __pyx_v_node->tau = __pyx_v_8skgarden_8mondrian_4tree_5_tree_INFINITY;

    /* "skgarden/mondrian/tree/_tree.pyx":860
 *         if parent == _TREE_UNDEFINED:
 *             node.tau = E
 *         elif is_leaf:             # <<<<<<<<<<<<<<
 *             node.tau = INFINITY
 *         else:
 */
    goto __pyx_L7;
  }

  /* "skgarden/mondrian/tree/_tree.pyx":863
 *             node.tau = INFINITY
 *         else:
 *             node.tau = E + self.nodes[parent].tau             # <<<<<<<<<<<<<<
 * 
 *         node.lower_bounds = <DTYPE_t*> malloc(self.n_features * sizeof(DTYPE_t))
 */
  /*else*/ {
    __pyx_v_node->tau = (__pyx_v_E + (__pyx_v_self->nodes[__pyx_v_parent]).tau);
  }
  __pyx_L7:;

  /* "skgarden/mondrian/tree/_tree.pyx":865
 *             node.tau = E + self.nodes[parent].tau
 * 
 *         node.lower_bounds = <DTYPE_t*> malloc(self.n_features * sizeof(DTYPE_t))             # <<<<<<<<<<<<<<
 *         node.upper_bounds = <DTYPE_t*> malloc(self.n_features * sizeof(DTYPE_t))
 *         memcpy(node.lower_bounds, lower_bounds, self.n_features*sizeof(DTYPE_t))
 */
  __pyx_v_node->lower_bounds = ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *)malloc((__pyx_v_self->n_features * (sizeof(__pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t)))));

  /* "skgarden/mondrian/tree/_tree.pyx":866
 * 
 *         node.lower_bounds = <DTYPE_t*> malloc(self.n_features * sizeof(DTYPE_t))
 *         node.upper_bounds = <DTYPE_t*> malloc(self.n_features * sizeof(DTYPE_t))             # <<<<<<<<<<<<<<
 *         memcpy(node.lower_bounds, lower_bounds, self.n_features*sizeof(DTYPE_t))
 *         memcpy(node.upper_bounds, upper_bounds, self.n_features*sizeof(DTYPE_t))
 */
  __pyx_v_node->upper_bounds = ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *)malloc((__pyx_v_self->n_features * (sizeof(__pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t)))));

  /* "skgarden/mondrian/tree/_tree.pyx":867
 *         node.lower_bounds = <DTYPE_t*> malloc(self.n_features * sizeof(DTYPE_t))
 *         node.upper_bounds = <DTYPE_t*> malloc(self.n_features * sizeof(DTYPE_t))
 *         memcpy(node.lower_bounds, lower_bounds, self.n_features*sizeof(DTYPE_t))             # <<<<<<<<<<<<<<
 *         memcpy(node.upper_bounds, upper_bounds, self.n_features*sizeof(DTYPE_t))
 *         node.variance = impurity
 */
  (void)(memcpy(__pyx_v_node->lower_bounds, __pyx_v_lower_bounds, (__pyx_v_self->n_features * (sizeof(__pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t)))));

  /* "skgarden/mondrian/tree/_tree.pyx":868
 *         node.upper_bounds = <DTYPE_t*> malloc(self.n_features * sizeof(DTYPE_t))
 *         memcpy(node.lower_bounds, lower_bounds, self.n_features*sizeof(DTYPE_t))
 *         memcpy(node.upper_bounds, upper_bounds, self.n_features*sizeof(DTYPE_t))             # <<<<<<<<<<<<<<
 *         node.variance = impurity
 * 
 */
  (void)(memcpy(__pyx_v_node->upper_bounds, __pyx_v_upper_bounds, (__pyx_v_self->n_features * (sizeof(__pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t)))));

  /* "skgarden/mondrian/tree/_tree.pyx":869
 *         memcpy(node.lower_bounds, lower_bounds, self.n_features*sizeof(DTYPE_t))
 *         memcpy(node.upper_bounds, upper_bounds, self.n_features*sizeof(DTYPE_t))
 *         node.variance = impurity             # <<<<<<<<<<<<<<
 * 
 *         if is_leaf:
 */
  __pyx_v_node->variance = __pyx_v_impurity;

  /* "skgarden/mondrian/tree/_tree.pyx":871
 *         node.variance = impurity
 * 
 *         if is_leaf:             # <<<<<<<<<<<<<<
 *             node.left_child = _TREE_LEAF
 *             node.right_child = _TREE_LEAF
 */
  __pyx_t_2 = (__pyx_v_is_leaf != 0);
  if (__pyx_t_2) {

    /* "skgarden/mondrian/tree/_tree.pyx":872
 * 
 *         if is_leaf:
 *             node.left_child = _TREE_LEAF             # <<<<<<<<<<<<<<
 *             node.right_child = _TREE_LEAF
 *             node.feature = _TREE_UNDEFINED
 */
    __pyx_v_node->left_child = __pyx_v_8skgarden_8mondrian_4tree_5_tree__TREE_LEAF;

    /* "skgarden/mondrian/tree/_tree.pyx":873
 *         if is_leaf:
 *             node.left_child = _TREE_LEAF
 *             node.right_child = _TREE_LEAF             # <<<<<<<<<<<<<<
 *             node.feature = _TREE_UNDEFINED
 *             node.threshold = _TREE_UNDEFINED
 */
    __pyx_v_node->right_child = __pyx_v_8skgarden_8mondrian_4tree_5_tree__TREE_LEAF;

    /* "skgarden/mondrian/tree/_tree.pyx":874
 *             node.left_child = _TREE_LEAF
 *             node.right_child = _TREE_LEAF
 *             node.feature = _TREE_UNDEFINED             # <<<<<<<<<<<<<<
 *             node.threshold = _TREE_UNDEFINED
 *         else:
 */
    __pyx_v_node->feature = __pyx_v_8skgarden_8mondrian_4tree_5_tree__TREE_UNDEFINED;

    /* "skgarden/mondrian/tree/_tree.pyx":875
 *             node.right_child = _TREE_LEAF
 *             node.feature = _TREE_UNDEFINED
 *             node.threshold = _TREE_UNDEFINED             # <<<<<<<<<<<<<<
 *         else:
 *             # left_child and right_child will be set later
 */
    __pyx_v_node->threshold = __pyx_v_8skgarden_8mondrian_4tree_5_tree__TREE_UNDEFINED;

    /* "skgarden/mondrian/tree/_tree.pyx":871
 *         node.variance = impurity
 * 
 *         if is_leaf:             # <<<<<<<<<<<<<<
 *             node.left_child = _TREE_LEAF
 *             node.right_child = _TREE_LEAF
 */
    goto __pyx_L8;
  }

  /* "skgarden/mondrian/tree/_tree.pyx":878
 *         else:
 *             # left_child and right_child will be set later
 *             node.feature = feature             # <<<<<<<<<<<<<<
 *             node.threshold = threshold
 *         self.node_count += 1
 */
  /*else*/ {
    __pyx_v_node->feature = __pyx_v_feature;

    /* "skgarden/mondrian/tree/_tree.pyx":879
 *             # left_child and right_child will be set later
 *             node.feature = feature
 *             node.threshold = threshold             # <<<<<<<<<<<<<<
 *         self.node_count += 1
 *         return node_id
 */
    __pyx_v_node->threshold = __pyx_v_threshold;
  }
  __pyx_L8:;

  /* "skgarden/mondrian/tree/_tree.pyx":880
 *             node.feature = feature
 *             node.threshold = threshold
 *         self.node_count += 1             # <<<<<<<<<<<<<<
 *         return node_id
 * 
 */
  __pyx_v_self->node_count = (__pyx_v_self->node_count + 1);

  /* "skgarden/mondrian/tree/_tree.pyx":881
 *             node.threshold = threshold
 *         self.node_count += 1
 *         return node_id             # <<<<<<<<<<<<<<
 * 
 *     cpdef np.ndarray apply(self, object X):
 */
  __pyx_r = __pyx_v_node_id;
  goto __pyx_L0;

  /* "skgarden/mondrian/tree/_tree.pyx":828
 *         free(extent)
 * 
 *     cdef SIZE_t _add_node(self, SIZE_t parent, bint is_left, bint is_leaf,             # <<<<<<<<<<<<<<
 *                           SIZE_t feature, double threshold, double impurity,
 *                           SIZE_t n_node_samples,
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree._add_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":883
 *         return node_id
 * 
 *     cpdef np.ndarray apply(self, object X):             # <<<<<<<<<<<<<<
 *         """Finds the terminal region (=leaf node) for each sample in X."""
 *         return self._apply_dense(X)
 */

static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_11apply(PyObject *__pyx_v_self, PyObject *__pyx_v_X); /*proto*/
static PyArrayObject *__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree_apply(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_X, int __pyx_skip_dispatch) {
  PyArrayObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("apply", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    static PY_UINT64_T tp_dict_version = 0, obj_dict_version = 0;
    if (likely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict && tp_dict_version == __PYX_GET_DICT_VERSION(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict) && (!Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset || obj_dict_version == __PYX_GET_DICT_VERSION(_PyObject_GetDictPtr(((PyObject *)__pyx_v_self))))));
    else {
      PY_UINT64_T type_dict_guard = (likely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict)) ? __PYX_GET_DICT_VERSION(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict) : 0;
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_apply); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 883, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_11apply)) {
        __Pyx_XDECREF(((PyObject *)__pyx_r));
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_X) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_X);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 883, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 883, __pyx_L1_error)
        __pyx_r = ((PyArrayObject *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
      tp_dict_version = likely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict) ? __PYX_GET_DICT_VERSION(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict) : 0;
      obj_dict_version = likely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset) ? __PYX_GET_DICT_VERSION(_PyObject_GetDictPtr(((PyObject *)__pyx_v_self))) : 0;
      if (unlikely(type_dict_guard != tp_dict_version)) {
        tp_dict_version = obj_dict_version = 0;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    }
    #endif
  }

  /* "skgarden/mondrian/tree/_tree.pyx":885
 *     cpdef np.ndarray apply(self, object X):
 *         """Finds the terminal region (=leaf node) for each sample in X."""
 *         return self._apply_dense(X)             # <<<<<<<<<<<<<<
 * 
 *     cpdef tuple predict(self, object X, bint return_std=False, bint is_regression=True):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = ((PyObject *)__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree__apply_dense(__pyx_v_self, __pyx_v_X)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "skgarden/mondrian/tree/_tree.pyx":883
 *         return node_id
 * 
 *     cpdef np.ndarray apply(self, object X):             # <<<<<<<<<<<<<<
 *         """Finds the terminal region (=leaf node) for each sample in X."""
 *         return self._apply_dense(X)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.apply", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_11apply(PyObject *__pyx_v_self, PyObject *__pyx_v_X); /*proto*/
static char __pyx_doc_8skgarden_8mondrian_4tree_5_tree_4Tree_10apply[] = "Finds the terminal region (=leaf node) for each sample in X.";
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_11apply(PyObject *__pyx_v_self, PyObject *__pyx_v_X) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("apply (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_10apply(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self), ((PyObject *)__pyx_v_X));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_10apply(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_X) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("apply", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree_apply(__pyx_v_self, __pyx_v_X, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 883, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.apply", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":887
 *         return self._apply_dense(X)
 * 
 *     cpdef tuple predict(self, object X, bint return_std=False, bint is_regression=True):             # <<<<<<<<<<<<<<
 *         """Predicts the regressor and standard deviation for all samples."""
 * 
 */

static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_13predict(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree_predict(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_X, int __pyx_skip_dispatch, struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_4Tree_predict *__pyx_optional_args) {
  int __pyx_v_return_std = ((int)0);
  int __pyx_v_is_regression = ((int)1);
  PyArrayObject *__pyx_v_X_ndarray = 0;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *__pyx_v_X_ptr;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_X_sample_stride;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_X_fx_stride;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_n_samples;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_n_features;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_f_ind;
  PyArrayObject *__pyx_v_node_values = 0;
  PyArrayObject *__pyx_v_mean = 0;
  PyArrayObject *__pyx_v_std = 0;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_n_classes;
  PyArrayObject *__pyx_v_proba = 0;
  PyArrayObject *__pyx_v_n_node_samples = 0;
  struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node *__pyx_v_node;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_i;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_node_id;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t __pyx_v_Delta;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t __pyx_v_parent_tau;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t __pyx_v_p_js;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t __pyx_v_X_val;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t __pyx_v_w_j;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t __pyx_v_p_nsy;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_class_ind;
  double __pyx_v_eta;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mean;
  __Pyx_Buffer __pyx_pybuffer_mean;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_n_node_samples;
  __Pyx_Buffer __pyx_pybuffer_n_node_samples;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_node_values;
  __Pyx_Buffer __pyx_pybuffer_node_values;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_proba;
  __Pyx_Buffer __pyx_pybuffer_proba;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_std;
  __Pyx_Buffer __pyx_pybuffer_std;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_t_10;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_11;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_12;
  PyArrayObject *__pyx_t_13 = NULL;
  PyArrayObject *__pyx_t_14 = NULL;
  PyArrayObject *__pyx_t_15 = NULL;
  PyArrayObject *__pyx_t_16 = NULL;
  PyArrayObject *__pyx_t_17 = NULL;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_18;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_19;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t __pyx_t_20;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_21;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  Py_ssize_t __pyx_t_25;
  Py_ssize_t __pyx_t_26;
  Py_ssize_t __pyx_t_27;
  Py_ssize_t __pyx_t_28;
  Py_ssize_t __pyx_t_29;
  Py_ssize_t __pyx_t_30;
  Py_ssize_t __pyx_t_31;
  Py_ssize_t __pyx_t_32;
  Py_ssize_t __pyx_t_33;
  Py_ssize_t __pyx_t_34;
  Py_ssize_t __pyx_t_35;
  Py_ssize_t __pyx_t_36;
  Py_ssize_t __pyx_t_37;
  Py_ssize_t __pyx_t_38;
  Py_ssize_t __pyx_t_39;
  __Pyx_RefNannySetupContext("predict", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_return_std = __pyx_optional_args->return_std;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_is_regression = __pyx_optional_args->is_regression;
      }
    }
  }
  __pyx_pybuffer_node_values.pybuffer.buf = NULL;
  __pyx_pybuffer_node_values.refcount = 0;
  __pyx_pybuffernd_node_values.data = NULL;
  __pyx_pybuffernd_node_values.rcbuffer = &__pyx_pybuffer_node_values;
  __pyx_pybuffer_mean.pybuffer.buf = NULL;
  __pyx_pybuffer_mean.refcount = 0;
  __pyx_pybuffernd_mean.data = NULL;
  __pyx_pybuffernd_mean.rcbuffer = &__pyx_pybuffer_mean;
  __pyx_pybuffer_std.pybuffer.buf = NULL;
  __pyx_pybuffer_std.refcount = 0;
  __pyx_pybuffernd_std.data = NULL;
  __pyx_pybuffernd_std.rcbuffer = &__pyx_pybuffer_std;
  __pyx_pybuffer_proba.pybuffer.buf = NULL;
  __pyx_pybuffer_proba.refcount = 0;
  __pyx_pybuffernd_proba.data = NULL;
  __pyx_pybuffernd_proba.rcbuffer = &__pyx_pybuffer_proba;
  __pyx_pybuffer_n_node_samples.pybuffer.buf = NULL;
  __pyx_pybuffer_n_node_samples.refcount = 0;
  __pyx_pybuffernd_n_node_samples.data = NULL;
  __pyx_pybuffernd_n_node_samples.rcbuffer = &__pyx_pybuffer_n_node_samples;
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    static PY_UINT64_T tp_dict_version = 0, obj_dict_version = 0;
    if (likely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict && tp_dict_version == __PYX_GET_DICT_VERSION(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict) && (!Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset || obj_dict_version == __PYX_GET_DICT_VERSION(_PyObject_GetDictPtr(((PyObject *)__pyx_v_self))))));
    else {
      PY_UINT64_T type_dict_guard = (likely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict)) ? __PYX_GET_DICT_VERSION(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict) : 0;
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_predict); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 887, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_13predict)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_return_std); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 887, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_is_regression); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 887, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = __pyx_t_1; __pyx_t_6 = NULL;
        __pyx_t_7 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_7 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_v_X, __pyx_t_3, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 3+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 887, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_v_X, __pyx_t_3, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 3+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 887, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_8 = PyTuple_New(3+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 887, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          if (__pyx_t_6) {
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
          }
          __Pyx_INCREF(__pyx_v_X);
          __Pyx_GIVEREF(__pyx_v_X);
          PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_v_X);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_7, __pyx_t_4);
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 887, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (!(likely(PyTuple_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 887, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
      tp_dict_version = likely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict) ? __PYX_GET_DICT_VERSION(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict) : 0;
      obj_dict_version = likely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset) ? __PYX_GET_DICT_VERSION(_PyObject_GetDictPtr(((PyObject *)__pyx_v_self))) : 0;
      if (unlikely(type_dict_guard != tp_dict_version)) {
        tp_dict_version = obj_dict_version = 0;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    }
    #endif
  }

  /* "skgarden/mondrian/tree/_tree.pyx":891
 * 
 *         # Check input
 *         if not isinstance(X, np.ndarray):             # <<<<<<<<<<<<<<
 *             raise ValueError("X should be in np.ndarray format, got %s"
 *                              % type(X))
 */
  __pyx_t_9 = __Pyx_TypeCheck(__pyx_v_X, __pyx_ptype_5numpy_ndarray); 
  __pyx_t_10 = ((!(__pyx_t_9 != 0)) != 0);
  if (unlikely(__pyx_t_10)) {

    /* "skgarden/mondrian/tree/_tree.pyx":893
 *         if not isinstance(X, np.ndarray):
 *             raise ValueError("X should be in np.ndarray format, got %s"
 *                              % type(X))             # <<<<<<<<<<<<<<
 * 
 *         if X.dtype != DTYPE:
 */
    __pyx_t_1 = __Pyx_PyString_FormatSafe(__pyx_kp_s_X_should_be_in_np_ndarray_format, ((PyObject *)Py_TYPE(__pyx_v_X))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 893, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "skgarden/mondrian/tree/_tree.pyx":892
 *         # Check input
 *         if not isinstance(X, np.ndarray):
 *             raise ValueError("X should be in np.ndarray format, got %s"             # <<<<<<<<<<<<<<
 *                              % type(X))
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 892, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 892, __pyx_L1_error)

    /* "skgarden/mondrian/tree/_tree.pyx":891
 * 
 *         # Check input
 *         if not isinstance(X, np.ndarray):             # <<<<<<<<<<<<<<
 *             raise ValueError("X should be in np.ndarray format, got %s"
 *                              % type(X))
 */
  }

  /* "skgarden/mondrian/tree/_tree.pyx":895
 *                              % type(X))
 * 
 *         if X.dtype != DTYPE:             # <<<<<<<<<<<<<<
 *             raise ValueError("X.dtype should be np.float32, got %s" % X.dtype)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_dtype); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 895, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_DTYPE); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 895, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyObject_RichCompare(__pyx_t_2, __pyx_t_1, Py_NE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 895, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 895, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(__pyx_t_10)) {

    /* "skgarden/mondrian/tree/_tree.pyx":896
 * 
 *         if X.dtype != DTYPE:
 *             raise ValueError("X.dtype should be np.float32, got %s" % X.dtype)             # <<<<<<<<<<<<<<
 * 
 *         # Extract input
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_dtype); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 896, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_PyString_FormatSafe(__pyx_kp_s_X_dtype_should_be_np_float32_got, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 896, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 896, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 896, __pyx_L1_error)

    /* "skgarden/mondrian/tree/_tree.pyx":895
 *                              % type(X))
 * 
 *         if X.dtype != DTYPE:             # <<<<<<<<<<<<<<
 *             raise ValueError("X.dtype should be np.float32, got %s" % X.dtype)
 * 
 */
  }

  /* "skgarden/mondrian/tree/_tree.pyx":899
 * 
 *         # Extract input
 *         cdef np.ndarray X_ndarray = X             # <<<<<<<<<<<<<<
 *         cdef DTYPE_t* X_ptr = <DTYPE_t*> X_ndarray.data
 *         cdef SIZE_t X_sample_stride = <SIZE_t> X.strides[0] / <SIZE_t> X.itemsize
 */
  if (!(likely(((__pyx_v_X) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_X, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 899, __pyx_L1_error)
  __pyx_t_5 = __pyx_v_X;
  __Pyx_INCREF(__pyx_t_5);
  __pyx_v_X_ndarray = ((PyArrayObject *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":900
 *         # Extract input
 *         cdef np.ndarray X_ndarray = X
 *         cdef DTYPE_t* X_ptr = <DTYPE_t*> X_ndarray.data             # <<<<<<<<<<<<<<
 *         cdef SIZE_t X_sample_stride = <SIZE_t> X.strides[0] / <SIZE_t> X.itemsize
 *         cdef SIZE_t X_fx_stride = <SIZE_t> X.strides[1] / <SIZE_t> X.itemsize
 */
  __pyx_v_X_ptr = ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *)__pyx_v_X_ndarray->data);

  /* "skgarden/mondrian/tree/_tree.pyx":901
 *         cdef np.ndarray X_ndarray = X
 *         cdef DTYPE_t* X_ptr = <DTYPE_t*> X_ndarray.data
 *         cdef SIZE_t X_sample_stride = <SIZE_t> X.strides[0] / <SIZE_t> X.itemsize             # <<<<<<<<<<<<<<
 *         cdef SIZE_t X_fx_stride = <SIZE_t> X.strides[1] / <SIZE_t> X.itemsize
 *         cdef SIZE_t n_samples = X.shape[0]
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_strides); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_11 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_1); if (unlikely((__pyx_t_11 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 901, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_itemsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_12 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_1); if (unlikely((__pyx_t_12 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 901, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_X_sample_stride = (((__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t)__pyx_t_11) / ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t)__pyx_t_12));

  /* "skgarden/mondrian/tree/_tree.pyx":902
 *         cdef DTYPE_t* X_ptr = <DTYPE_t*> X_ndarray.data
 *         cdef SIZE_t X_sample_stride = <SIZE_t> X.strides[0] / <SIZE_t> X.itemsize
 *         cdef SIZE_t X_fx_stride = <SIZE_t> X.strides[1] / <SIZE_t> X.itemsize             # <<<<<<<<<<<<<<
 *         cdef SIZE_t n_samples = X.shape[0]
 *         cdef SIZE_t n_features = X.shape[1]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_strides); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 902, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_1, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 902, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_12 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_5); if (unlikely((__pyx_t_12 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 902, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_itemsize); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 902, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_11 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_5); if (unlikely((__pyx_t_11 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 902, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_X_fx_stride = (((__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t)__pyx_t_12) / ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t)__pyx_t_11));

  /* "skgarden/mondrian/tree/_tree.pyx":903
 *         cdef SIZE_t X_sample_stride = <SIZE_t> X.strides[0] / <SIZE_t> X.itemsize
 *         cdef SIZE_t X_fx_stride = <SIZE_t> X.strides[1] / <SIZE_t> X.itemsize
 *         cdef SIZE_t n_samples = X.shape[0]             # <<<<<<<<<<<<<<
 *         cdef SIZE_t n_features = X.shape[1]
 *         cdef SIZE_t f_ind
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_shape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 903, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 903, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_11 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_1); if (unlikely((__pyx_t_11 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 903, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_n_samples = __pyx_t_11;

  /* "skgarden/mondrian/tree/_tree.pyx":904
 *         cdef SIZE_t X_fx_stride = <SIZE_t> X.strides[1] / <SIZE_t> X.itemsize
 *         cdef SIZE_t n_samples = X.shape[0]
 *         cdef SIZE_t n_features = X.shape[1]             # <<<<<<<<<<<<<<
 *         cdef SIZE_t f_ind
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_1, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_11 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_5); if (unlikely((__pyx_t_11 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 904, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_n_features = __pyx_t_11;

  /* "skgarden/mondrian/tree/_tree.pyx":910
 *         # These node values are the means in case of regression.
 *         # For classification these are the class counts.
 *         cdef np.ndarray[DOUBLE_t, ndim=2] node_values = self._get_value_ndarray()[:, 0, :]             # <<<<<<<<<<<<<<
 * 
 *         # Initialize output
 */
  __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self->__pyx_vtab)->_get_value_ndarray(__pyx_v_self)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_t_5, __pyx_tuple__9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 910, __pyx_L1_error)
  __pyx_t_13 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_node_values.rcbuffer->pybuffer, (PyObject*)__pyx_t_13, &__Pyx_TypeInfo_nn___pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
      __pyx_v_node_values = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_node_values.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 910, __pyx_L1_error)
    } else {__pyx_pybuffernd_node_values.diminfo[0].strides = __pyx_pybuffernd_node_values.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_node_values.diminfo[0].shape = __pyx_pybuffernd_node_values.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_node_values.diminfo[1].strides = __pyx_pybuffernd_node_values.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_node_values.diminfo[1].shape = __pyx_pybuffernd_node_values.rcbuffer->pybuffer.shape[1];
    }
  }
  __pyx_t_13 = 0;
  __pyx_v_node_values = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":913
 * 
 *         # Initialize output
 *         cdef np.ndarray[DTYPE_t, ndim=1] mean = np.zeros(n_samples, dtype=DTYPE)             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[DTYPE_t, ndim=1] std = np.zeros(n_samples, dtype=DTYPE)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 913, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 913, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_Py_intptr_t(__pyx_v_n_samples); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 913, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 913, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 913, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_DTYPE); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 913, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_8) < 0) __PYX_ERR(0, 913, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 913, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_8) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_8, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 913, __pyx_L1_error)
  __pyx_t_14 = ((PyArrayObject *)__pyx_t_8);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mean.rcbuffer->pybuffer, (PyObject*)__pyx_t_14, &__Pyx_TypeInfo_nn___pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_mean = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_mean.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 913, __pyx_L1_error)
    } else {__pyx_pybuffernd_mean.diminfo[0].strides = __pyx_pybuffernd_mean.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mean.diminfo[0].shape = __pyx_pybuffernd_mean.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_14 = 0;
  __pyx_v_mean = ((PyArrayObject *)__pyx_t_8);
  __pyx_t_8 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":914
 *         # Initialize output
 *         cdef np.ndarray[DTYPE_t, ndim=1] mean = np.zeros(n_samples, dtype=DTYPE)
 *         cdef np.ndarray[DTYPE_t, ndim=1] std = np.zeros(n_samples, dtype=DTYPE)             # <<<<<<<<<<<<<<
 * 
 *         cdef SIZE_t n_classes = node_values.shape[1]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyInt_From_Py_intptr_t(__pyx_v_n_samples); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_8);
  __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_DTYPE); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 914, __pyx_L1_error)
  __pyx_t_15 = ((PyArrayObject *)__pyx_t_5);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_std.rcbuffer->pybuffer, (PyObject*)__pyx_t_15, &__Pyx_TypeInfo_nn___pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_std = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_std.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 914, __pyx_L1_error)
    } else {__pyx_pybuffernd_std.diminfo[0].strides = __pyx_pybuffernd_std.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_std.diminfo[0].shape = __pyx_pybuffernd_std.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_15 = 0;
  __pyx_v_std = ((PyArrayObject *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":916
 *         cdef np.ndarray[DTYPE_t, ndim=1] std = np.zeros(n_samples, dtype=DTYPE)
 * 
 *         cdef SIZE_t n_classes = node_values.shape[1]             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[DTYPE_t, ndim=2] proba = np.zeros((n_samples, n_classes), dtype=DTYPE)
 *         cdef np.ndarray[SIZE_t, ndim=1] n_node_samples = self.n_node_samples
 */
  __pyx_v_n_classes = (__pyx_v_node_values->dimensions[1]);

  /* "skgarden/mondrian/tree/_tree.pyx":917
 * 
 *         cdef SIZE_t n_classes = node_values.shape[1]
 *         cdef np.ndarray[DTYPE_t, ndim=2] proba = np.zeros((n_samples, n_classes), dtype=DTYPE)             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[SIZE_t, ndim=1] n_node_samples = self.n_node_samples
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_From_Py_intptr_t(__pyx_v_n_samples); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyInt_From_Py_intptr_t(__pyx_v_n_classes); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
  __pyx_t_5 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_DTYPE); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 917, __pyx_L1_error)
  __pyx_t_16 = ((PyArrayObject *)__pyx_t_5);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_proba.rcbuffer->pybuffer, (PyObject*)__pyx_t_16, &__Pyx_TypeInfo_nn___pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 2, 0, __pyx_stack) == -1)) {
      __pyx_v_proba = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_proba.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 917, __pyx_L1_error)
    } else {__pyx_pybuffernd_proba.diminfo[0].strides = __pyx_pybuffernd_proba.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_proba.diminfo[0].shape = __pyx_pybuffernd_proba.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_proba.diminfo[1].strides = __pyx_pybuffernd_proba.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_proba.diminfo[1].shape = __pyx_pybuffernd_proba.rcbuffer->pybuffer.shape[1];
    }
  }
  __pyx_t_16 = 0;
  __pyx_v_proba = ((PyArrayObject *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":918
 *         cdef SIZE_t n_classes = node_values.shape[1]
 *         cdef np.ndarray[DTYPE_t, ndim=2] proba = np.zeros((n_samples, n_classes), dtype=DTYPE)
 *         cdef np.ndarray[SIZE_t, ndim=1] n_node_samples = self.n_node_samples             # <<<<<<<<<<<<<<
 * 
 *         # Initialize auxiliary data-structure
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_n_node_samples); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 918, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 918, __pyx_L1_error)
  __pyx_t_17 = ((PyArrayObject *)__pyx_t_5);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_n_node_samples.rcbuffer->pybuffer, (PyObject*)__pyx_t_17, &__Pyx_TypeInfo_nn___pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_n_node_samples = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_n_node_samples.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 918, __pyx_L1_error)
    } else {__pyx_pybuffernd_n_node_samples.diminfo[0].strides = __pyx_pybuffernd_n_node_samples.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_n_node_samples.diminfo[0].shape = __pyx_pybuffernd_n_node_samples.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_17 = 0;
  __pyx_v_n_node_samples = ((PyArrayObject *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":921
 * 
 *         # Initialize auxiliary data-structure
 *         cdef Node* node = NULL             # <<<<<<<<<<<<<<
 *         cdef SIZE_t i = 0
 *         cdef SIZE_t j
 */
  __pyx_v_node = NULL;

  /* "skgarden/mondrian/tree/_tree.pyx":922
 *         # Initialize auxiliary data-structure
 *         cdef Node* node = NULL
 *         cdef SIZE_t i = 0             # <<<<<<<<<<<<<<
 *         cdef SIZE_t j
 *         cdef SIZE_t node_id = 0
 */
  __pyx_v_i = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":924
 *         cdef SIZE_t i = 0
 *         cdef SIZE_t j
 *         cdef SIZE_t node_id = 0             # <<<<<<<<<<<<<<
 *         cdef DOUBLE_t Delta = 0.0
 *         cdef DOUBLE_t parent_tau
 */
  __pyx_v_node_id = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":925
 *         cdef SIZE_t j
 *         cdef SIZE_t node_id = 0
 *         cdef DOUBLE_t Delta = 0.0             # <<<<<<<<<<<<<<
 *         cdef DOUBLE_t parent_tau
 *         cdef DOUBLE_t p_js
 */
  __pyx_v_Delta = 0.0;

  /* "skgarden/mondrian/tree/_tree.pyx":936
 *         cdef SIZE_t class_ind
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             for i in range(n_samples):
 *                 # Step 3
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "skgarden/mondrian/tree/_tree.pyx":937
 * 
 *         with nogil:
 *             for i in range(n_samples):             # <<<<<<<<<<<<<<
 *                 # Step 3
 *                 parent_tau = 0.0
 */
        __pyx_t_11 = __pyx_v_n_samples;
        __pyx_t_12 = __pyx_t_11;
        for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_12; __pyx_t_18+=1) {
          __pyx_v_i = __pyx_t_18;

          /* "skgarden/mondrian/tree/_tree.pyx":939
 *             for i in range(n_samples):
 *                 # Step 3
 *                 parent_tau = 0.0             # <<<<<<<<<<<<<<
 *                 p_nsy = 1.0
 *                 node_id = self.root
 */
          __pyx_v_parent_tau = 0.0;

          /* "skgarden/mondrian/tree/_tree.pyx":940
 *                 # Step 3
 *                 parent_tau = 0.0
 *                 p_nsy = 1.0             # <<<<<<<<<<<<<<
 *                 node_id = self.root
 * 
 */
          __pyx_v_p_nsy = 1.0;

          /* "skgarden/mondrian/tree/_tree.pyx":941
 *                 parent_tau = 0.0
 *                 p_nsy = 1.0
 *                 node_id = self.root             # <<<<<<<<<<<<<<
 * 
 *                 while True:
 */
          __pyx_t_19 = __pyx_v_self->root;
          __pyx_v_node_id = __pyx_t_19;

          /* "skgarden/mondrian/tree/_tree.pyx":943
 *                 node_id = self.root
 * 
 *                 while True:             # <<<<<<<<<<<<<<
 *                     node = &self.nodes[node_id]
 * 
 */
          while (1) {

            /* "skgarden/mondrian/tree/_tree.pyx":944
 * 
 *                 while True:
 *                     node = &self.nodes[node_id]             # <<<<<<<<<<<<<<
 * 
 *                     # Step 5: First part.
 */
            __pyx_v_node = (&(__pyx_v_self->nodes[__pyx_v_node_id]));

            /* "skgarden/mondrian/tree/_tree.pyx":948
 *                     # Step 5: First part.
 *                     # Calculate Delta
 *                     Delta = node.tau - parent_tau             # <<<<<<<<<<<<<<
 *                     parent_tau = node.tau
 * 
 */
            __pyx_v_Delta = (__pyx_v_node->tau - __pyx_v_parent_tau);

            /* "skgarden/mondrian/tree/_tree.pyx":949
 *                     # Calculate Delta
 *                     Delta = node.tau - parent_tau
 *                     parent_tau = node.tau             # <<<<<<<<<<<<<<
 * 
 *                     # Step 5: Second part.
 */
            __pyx_t_20 = __pyx_v_node->tau;
            __pyx_v_parent_tau = __pyx_t_20;

            /* "skgarden/mondrian/tree/_tree.pyx":953
 *                     # Step 5: Second part.
 *                     # Calculate eta
 *                     eta = 0.0             # <<<<<<<<<<<<<<
 *                     for f_ind in range(n_features):
 *                         X_val = X_ptr[X_sample_stride*i + X_fx_stride*f_ind]
 */
            __pyx_v_eta = 0.0;

            /* "skgarden/mondrian/tree/_tree.pyx":954
 *                     # Calculate eta
 *                     eta = 0.0
 *                     for f_ind in range(n_features):             # <<<<<<<<<<<<<<
 *                         X_val = X_ptr[X_sample_stride*i + X_fx_stride*f_ind]
 * 
 */
            __pyx_t_19 = __pyx_v_n_features;
            __pyx_t_21 = __pyx_t_19;
            for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
              __pyx_v_f_ind = __pyx_t_22;

              /* "skgarden/mondrian/tree/_tree.pyx":955
 *                     eta = 0.0
 *                     for f_ind in range(n_features):
 *                         X_val = X_ptr[X_sample_stride*i + X_fx_stride*f_ind]             # <<<<<<<<<<<<<<
 * 
 *                         eta += (fmax(X_val - node.upper_bounds[f_ind], 0) +
 */
              __pyx_v_X_val = (__pyx_v_X_ptr[((__pyx_v_X_sample_stride * __pyx_v_i) + (__pyx_v_X_fx_stride * __pyx_v_f_ind))]);

              /* "skgarden/mondrian/tree/_tree.pyx":957
 *                         X_val = X_ptr[X_sample_stride*i + X_fx_stride*f_ind]
 * 
 *                         eta += (fmax(X_val - node.upper_bounds[f_ind], 0) +             # <<<<<<<<<<<<<<
 *                                 fmax(node.lower_bounds[f_ind] - X_val, 0))
 * 
 */
              __pyx_v_eta = (__pyx_v_eta + (__pyx_f_8skgarden_8mondrian_4tree_5_tree_fmax((__pyx_v_X_val - (__pyx_v_node->upper_bounds[__pyx_v_f_ind])), 0.0) + __pyx_f_8skgarden_8mondrian_4tree_5_tree_fmax(((__pyx_v_node->lower_bounds[__pyx_v_f_ind]) - __pyx_v_X_val), 0.0)));
            }

            /* "skgarden/mondrian/tree/_tree.pyx":962
 *                     # Step 6: Calculate p_j
 *                     # Step 7-11
 *                     if node.left_child == _TREE_LEAF:             # <<<<<<<<<<<<<<
 *                         w_j = p_nsy
 *                     else:
 */
            __pyx_t_10 = ((__pyx_v_node->left_child == __pyx_v_8skgarden_8mondrian_4tree_5_tree__TREE_LEAF) != 0);
            if (__pyx_t_10) {

              /* "skgarden/mondrian/tree/_tree.pyx":963
 *                     # Step 7-11
 *                     if node.left_child == _TREE_LEAF:
 *                         w_j = p_nsy             # <<<<<<<<<<<<<<
 *                     else:
 *                         p_js = 1 - exp(-Delta * eta)
 */
              __pyx_v_w_j = __pyx_v_p_nsy;

              /* "skgarden/mondrian/tree/_tree.pyx":962
 *                     # Step 6: Calculate p_j
 *                     # Step 7-11
 *                     if node.left_child == _TREE_LEAF:             # <<<<<<<<<<<<<<
 *                         w_j = p_nsy
 *                     else:
 */
              goto __pyx_L14;
            }

            /* "skgarden/mondrian/tree/_tree.pyx":965
 *                         w_j = p_nsy
 *                     else:
 *                         p_js = 1 - exp(-Delta * eta)             # <<<<<<<<<<<<<<
 *                         w_j = p_nsy * p_js
 * 
 */
            /*else*/ {
              __pyx_v_p_js = (1.0 - exp(((-__pyx_v_Delta) * __pyx_v_eta)));

              /* "skgarden/mondrian/tree/_tree.pyx":966
 *                     else:
 *                         p_js = 1 - exp(-Delta * eta)
 *                         w_j = p_nsy * p_js             # <<<<<<<<<<<<<<
 * 
 *                     if is_regression:
 */
              __pyx_v_w_j = (__pyx_v_p_nsy * __pyx_v_p_js);
            }
            __pyx_L14:;

            /* "skgarden/mondrian/tree/_tree.pyx":968
 *                         w_j = p_nsy * p_js
 * 
 *                     if is_regression:             # <<<<<<<<<<<<<<
 *                         mean[i] += w_j * node_values[node_id, 0]
 *                     else:
 */
            __pyx_t_10 = (__pyx_v_is_regression != 0);
            if (__pyx_t_10) {

              /* "skgarden/mondrian/tree/_tree.pyx":969
 * 
 *                     if is_regression:
 *                         mean[i] += w_j * node_values[node_id, 0]             # <<<<<<<<<<<<<<
 *                     else:
 *                         for class_ind in range(n_classes):
 */
              __pyx_t_23 = __pyx_v_node_id;
              __pyx_t_24 = 0;
              __pyx_t_25 = __pyx_v_i;
              *__Pyx_BufPtrStrided1d(__pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *, __pyx_pybuffernd_mean.rcbuffer->pybuffer.buf, __pyx_t_25, __pyx_pybuffernd_mean.diminfo[0].strides) += (__pyx_v_w_j * (*__Pyx_BufPtrStrided2d(__pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t *, __pyx_pybuffernd_node_values.rcbuffer->pybuffer.buf, __pyx_t_23, __pyx_pybuffernd_node_values.diminfo[0].strides, __pyx_t_24, __pyx_pybuffernd_node_values.diminfo[1].strides)));

              /* "skgarden/mondrian/tree/_tree.pyx":968
 *                         w_j = p_nsy * p_js
 * 
 *                     if is_regression:             # <<<<<<<<<<<<<<
 *                         mean[i] += w_j * node_values[node_id, 0]
 *                     else:
 */
              goto __pyx_L15;
            }

            /* "skgarden/mondrian/tree/_tree.pyx":971
 *                         mean[i] += w_j * node_values[node_id, 0]
 *                     else:
 *                         for class_ind in range(n_classes):             # <<<<<<<<<<<<<<
 *                             proba[i, class_ind] += w_j * (node_values[node_id, class_ind] / n_node_samples[node_id])
 * 
 */
            /*else*/ {
              __pyx_t_19 = __pyx_v_n_classes;
              __pyx_t_21 = __pyx_t_19;
              for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
                __pyx_v_class_ind = __pyx_t_22;

                /* "skgarden/mondrian/tree/_tree.pyx":972
 *                     else:
 *                         for class_ind in range(n_classes):
 *                             proba[i, class_ind] += w_j * (node_values[node_id, class_ind] / n_node_samples[node_id])             # <<<<<<<<<<<<<<
 * 
 *                     if return_std:
 */
                __pyx_t_26 = __pyx_v_node_id;
                __pyx_t_27 = __pyx_v_class_ind;
                __pyx_t_28 = __pyx_v_node_id;
                __pyx_t_29 = __pyx_v_i;
                __pyx_t_30 = __pyx_v_class_ind;
                *__Pyx_BufPtrStrided2d(__pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *, __pyx_pybuffernd_proba.rcbuffer->pybuffer.buf, __pyx_t_29, __pyx_pybuffernd_proba.diminfo[0].strides, __pyx_t_30, __pyx_pybuffernd_proba.diminfo[1].strides) += (__pyx_v_w_j * ((*__Pyx_BufPtrStrided2d(__pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t *, __pyx_pybuffernd_node_values.rcbuffer->pybuffer.buf, __pyx_t_26, __pyx_pybuffernd_node_values.diminfo[0].strides, __pyx_t_27, __pyx_pybuffernd_node_values.diminfo[1].strides)) / (*__Pyx_BufPtrStrided1d(__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t *, __pyx_pybuffernd_n_node_samples.rcbuffer->pybuffer.buf, __pyx_t_28, __pyx_pybuffernd_n_node_samples.diminfo[0].strides))));
              }
            }
            __pyx_L15:;

            /* "skgarden/mondrian/tree/_tree.pyx":974
 *                             proba[i, class_ind] += w_j * (node_values[node_id, class_ind] / n_node_samples[node_id])
 * 
 *                     if return_std:             # <<<<<<<<<<<<<<
 *                         std[i] += w_j * (node_values[node_id, 0]**2 + node.variance)
 * 
 */
            __pyx_t_10 = (__pyx_v_return_std != 0);
            if (__pyx_t_10) {

              /* "skgarden/mondrian/tree/_tree.pyx":975
 * 
 *                     if return_std:
 *                         std[i] += w_j * (node_values[node_id, 0]**2 + node.variance)             # <<<<<<<<<<<<<<
 * 
 *                     if node.left_child == _TREE_LEAF:
 */
              __pyx_t_31 = __pyx_v_node_id;
              __pyx_t_32 = 0;
              __pyx_t_33 = __pyx_v_i;
              *__Pyx_BufPtrStrided1d(__pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *, __pyx_pybuffernd_std.rcbuffer->pybuffer.buf, __pyx_t_33, __pyx_pybuffernd_std.diminfo[0].strides) += (__pyx_v_w_j * (pow((*__Pyx_BufPtrStrided2d(__pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t *, __pyx_pybuffernd_node_values.rcbuffer->pybuffer.buf, __pyx_t_31, __pyx_pybuffernd_node_values.diminfo[0].strides, __pyx_t_32, __pyx_pybuffernd_node_values.diminfo[1].strides)), 2.0) + __pyx_v_node->variance));

              /* "skgarden/mondrian/tree/_tree.pyx":974
 *                             proba[i, class_ind] += w_j * (node_values[node_id, class_ind] / n_node_samples[node_id])
 * 
 *                     if return_std:             # <<<<<<<<<<<<<<
 *                         std[i] += w_j * (node_values[node_id, 0]**2 + node.variance)
 * 
 */
            }

            /* "skgarden/mondrian/tree/_tree.pyx":977
 *                         std[i] += w_j * (node_values[node_id, 0]**2 + node.variance)
 * 
 *                     if node.left_child == _TREE_LEAF:             # <<<<<<<<<<<<<<
 *                         break
 *                     p_nsy = p_nsy * (1 - p_js)
 */
            __pyx_t_10 = ((__pyx_v_node->left_child == __pyx_v_8skgarden_8mondrian_4tree_5_tree__TREE_LEAF) != 0);
            if (__pyx_t_10) {

              /* "skgarden/mondrian/tree/_tree.pyx":978
 * 
 *                     if node.left_child == _TREE_LEAF:
 *                         break             # <<<<<<<<<<<<<<
 *                     p_nsy = p_nsy * (1 - p_js)
 * 
 */
              goto __pyx_L11_break;

              /* "skgarden/mondrian/tree/_tree.pyx":977
 *                         std[i] += w_j * (node_values[node_id, 0]**2 + node.variance)
 * 
 *                     if node.left_child == _TREE_LEAF:             # <<<<<<<<<<<<<<
 *                         break
 *                     p_nsy = p_nsy * (1 - p_js)
 */
            }

            /* "skgarden/mondrian/tree/_tree.pyx":979
 *                     if node.left_child == _TREE_LEAF:
 *                         break
 *                     p_nsy = p_nsy * (1 - p_js)             # <<<<<<<<<<<<<<
 * 
 *                     # Step 12-14
 */
            __pyx_v_p_nsy = (__pyx_v_p_nsy * (1.0 - __pyx_v_p_js));

            /* "skgarden/mondrian/tree/_tree.pyx":983
 *                     # Step 12-14
 *                     if X_ptr[X_sample_stride * i +
 *                              X_fx_stride * node.feature] <= node.threshold:             # <<<<<<<<<<<<<<
 *                         node_id = node.left_child
 *                     else:
 */
            __pyx_t_10 = (((__pyx_v_X_ptr[((__pyx_v_X_sample_stride * __pyx_v_i) + (__pyx_v_X_fx_stride * __pyx_v_node->feature))]) <= __pyx_v_node->threshold) != 0);

            /* "skgarden/mondrian/tree/_tree.pyx":982
 * 
 *                     # Step 12-14
 *                     if X_ptr[X_sample_stride * i +             # <<<<<<<<<<<<<<
 *                              X_fx_stride * node.feature] <= node.threshold:
 *                         node_id = node.left_child
 */
            if (__pyx_t_10) {

              /* "skgarden/mondrian/tree/_tree.pyx":984
 *                     if X_ptr[X_sample_stride * i +
 *                              X_fx_stride * node.feature] <= node.threshold:
 *                         node_id = node.left_child             # <<<<<<<<<<<<<<
 *                     else:
 *                         node_id = node.right_child
 */
              __pyx_t_19 = __pyx_v_node->left_child;
              __pyx_v_node_id = __pyx_t_19;

              /* "skgarden/mondrian/tree/_tree.pyx":982
 * 
 *                     # Step 12-14
 *                     if X_ptr[X_sample_stride * i +             # <<<<<<<<<<<<<<
 *                              X_fx_stride * node.feature] <= node.threshold:
 *                         node_id = node.left_child
 */
              goto __pyx_L20;
            }

            /* "skgarden/mondrian/tree/_tree.pyx":986
 *                         node_id = node.left_child
 *                     else:
 *                         node_id = node.right_child             # <<<<<<<<<<<<<<
 * 
 *                 if return_std:
 */
            /*else*/ {
              __pyx_t_19 = __pyx_v_node->right_child;
              __pyx_v_node_id = __pyx_t_19;
            }
            __pyx_L20:;
          }
          __pyx_L11_break:;

          /* "skgarden/mondrian/tree/_tree.pyx":988
 *                         node_id = node.right_child
 * 
 *                 if return_std:             # <<<<<<<<<<<<<<
 *                     std[i] -= mean[i]**2
 *                     if std[i] <= 0:
 */
          __pyx_t_10 = (__pyx_v_return_std != 0);
          if (__pyx_t_10) {

            /* "skgarden/mondrian/tree/_tree.pyx":989
 * 
 *                 if return_std:
 *                     std[i] -= mean[i]**2             # <<<<<<<<<<<<<<
 *                     if std[i] <= 0:
 *                         std[i] = 0.0
 */
            __pyx_t_34 = __pyx_v_i;
            __pyx_t_35 = __pyx_v_i;
            *__Pyx_BufPtrStrided1d(__pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *, __pyx_pybuffernd_std.rcbuffer->pybuffer.buf, __pyx_t_35, __pyx_pybuffernd_std.diminfo[0].strides) -= powf((*__Pyx_BufPtrStrided1d(__pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *, __pyx_pybuffernd_mean.rcbuffer->pybuffer.buf, __pyx_t_34, __pyx_pybuffernd_mean.diminfo[0].strides)), 2.0);

            /* "skgarden/mondrian/tree/_tree.pyx":990
 *                 if return_std:
 *                     std[i] -= mean[i]**2
 *                     if std[i] <= 0:             # <<<<<<<<<<<<<<
 *                         std[i] = 0.0
 *                     std[i] = sqrt(std[i])
 */
            __pyx_t_36 = __pyx_v_i;
            __pyx_t_10 = (((*__Pyx_BufPtrStrided1d(__pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *, __pyx_pybuffernd_std.rcbuffer->pybuffer.buf, __pyx_t_36, __pyx_pybuffernd_std.diminfo[0].strides)) <= 0.0) != 0);
            if (__pyx_t_10) {

              /* "skgarden/mondrian/tree/_tree.pyx":991
 *                     std[i] -= mean[i]**2
 *                     if std[i] <= 0:
 *                         std[i] = 0.0             # <<<<<<<<<<<<<<
 *                     std[i] = sqrt(std[i])
 * 
 */
              __pyx_t_37 = __pyx_v_i;
              *__Pyx_BufPtrStrided1d(__pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *, __pyx_pybuffernd_std.rcbuffer->pybuffer.buf, __pyx_t_37, __pyx_pybuffernd_std.diminfo[0].strides) = 0.0;

              /* "skgarden/mondrian/tree/_tree.pyx":990
 *                 if return_std:
 *                     std[i] -= mean[i]**2
 *                     if std[i] <= 0:             # <<<<<<<<<<<<<<
 *                         std[i] = 0.0
 *                     std[i] = sqrt(std[i])
 */
            }

            /* "skgarden/mondrian/tree/_tree.pyx":992
 *                     if std[i] <= 0:
 *                         std[i] = 0.0
 *                     std[i] = sqrt(std[i])             # <<<<<<<<<<<<<<
 * 
 *         if is_regression:
 */
            __pyx_t_38 = __pyx_v_i;
            __pyx_t_39 = __pyx_v_i;
            *__Pyx_BufPtrStrided1d(__pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *, __pyx_pybuffernd_std.rcbuffer->pybuffer.buf, __pyx_t_39, __pyx_pybuffernd_std.diminfo[0].strides) = sqrt((*__Pyx_BufPtrStrided1d(__pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *, __pyx_pybuffernd_std.rcbuffer->pybuffer.buf, __pyx_t_38, __pyx_pybuffernd_std.diminfo[0].strides)));

            /* "skgarden/mondrian/tree/_tree.pyx":988
 *                         node_id = node.right_child
 * 
 *                 if return_std:             # <<<<<<<<<<<<<<
 *                     std[i] -= mean[i]**2
 *                     if std[i] <= 0:
 */
          }
        }
      }

      /* "skgarden/mondrian/tree/_tree.pyx":936
 *         cdef SIZE_t class_ind
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             for i in range(n_samples):
 *                 # Step 3
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L7;
        }
        __pyx_L7:;
      }
  }

  /* "skgarden/mondrian/tree/_tree.pyx":994
 *                     std[i] = sqrt(std[i])
 * 
 *         if is_regression:             # <<<<<<<<<<<<<<
 *             if return_std:
 *                 return mean, std
 */
  __pyx_t_10 = (__pyx_v_is_regression != 0);
  if (__pyx_t_10) {

    /* "skgarden/mondrian/tree/_tree.pyx":995
 * 
 *         if is_regression:
 *             if return_std:             # <<<<<<<<<<<<<<
 *                 return mean, std
 *             return mean,
 */
    __pyx_t_10 = (__pyx_v_return_std != 0);
    if (__pyx_t_10) {

      /* "skgarden/mondrian/tree/_tree.pyx":996
 *         if is_regression:
 *             if return_std:
 *                 return mean, std             # <<<<<<<<<<<<<<
 *             return mean,
 *         else:
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 996, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(((PyObject *)__pyx_v_mean));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_mean));
      PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)__pyx_v_mean));
      __Pyx_INCREF(((PyObject *)__pyx_v_std));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_std));
      PyTuple_SET_ITEM(__pyx_t_5, 1, ((PyObject *)__pyx_v_std));
      __pyx_r = ((PyObject*)__pyx_t_5);
      __pyx_t_5 = 0;
      goto __pyx_L0;

      /* "skgarden/mondrian/tree/_tree.pyx":995
 * 
 *         if is_regression:
 *             if return_std:             # <<<<<<<<<<<<<<
 *                 return mean, std
 *             return mean,
 */
    }

    /* "skgarden/mondrian/tree/_tree.pyx":997
 *             if return_std:
 *                 return mean, std
 *             return mean,             # <<<<<<<<<<<<<<
 *         else:
 *             return proba,
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)__pyx_v_mean));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_mean));
    PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)__pyx_v_mean));
    __pyx_r = ((PyObject*)__pyx_t_5);
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "skgarden/mondrian/tree/_tree.pyx":994
 *                     std[i] = sqrt(std[i])
 * 
 *         if is_regression:             # <<<<<<<<<<<<<<
 *             if return_std:
 *                 return mean, std
 */
  }

  /* "skgarden/mondrian/tree/_tree.pyx":999
 *             return mean,
 *         else:
 *             return proba,             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 999, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)__pyx_v_proba));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_proba));
    PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)__pyx_v_proba));
    __pyx_r = ((PyObject*)__pyx_t_5);
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "skgarden/mondrian/tree/_tree.pyx":887
 *         return self._apply_dense(X)
 * 
 *     cpdef tuple predict(self, object X, bint return_std=False, bint is_regression=True):             # <<<<<<<<<<<<<<
 *         """Predicts the regressor and standard deviation for all samples."""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mean.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_n_node_samples.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_node_values.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_proba.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_std.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.predict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mean.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_n_node_samples.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_node_values.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_proba.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_std.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_X_ndarray);
  __Pyx_XDECREF((PyObject *)__pyx_v_node_values);
  __Pyx_XDECREF((PyObject *)__pyx_v_mean);
  __Pyx_XDECREF((PyObject *)__pyx_v_std);
  __Pyx_XDECREF((PyObject *)__pyx_v_proba);
  __Pyx_XDECREF((PyObject *)__pyx_v_n_node_samples);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_13predict(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8skgarden_8mondrian_4tree_5_tree_4Tree_12predict[] = "Predicts the regressor and standard deviation for all samples.";
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_13predict(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_X = 0;
  int __pyx_v_return_std;
  int __pyx_v_is_regression;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("predict (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_return_std,&__pyx_n_s_is_regression,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_return_std);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_is_regression);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "predict") < 0)) __PYX_ERR(0, 887, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_X = values[0];
    if (values[1]) {
      __pyx_v_return_std = __Pyx_PyObject_IsTrue(values[1]); if (unlikely((__pyx_v_return_std == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 887, __pyx_L3_error)
    } else {
      __pyx_v_return_std = ((int)0);
    }
    if (values[2]) {
      __pyx_v_is_regression = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_is_regression == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 887, __pyx_L3_error)
    } else {
      __pyx_v_is_regression = ((int)1);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("predict", 0, 1, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 887, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.predict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_12predict(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self), __pyx_v_X, __pyx_v_return_std, __pyx_v_is_regression);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_12predict(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_X, int __pyx_v_return_std, int __pyx_v_is_regression) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_4Tree_predict __pyx_t_2;
  __Pyx_RefNannySetupContext("predict", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 2;
  __pyx_t_2.return_std = __pyx_v_return_std;
  __pyx_t_2.is_regression = __pyx_v_is_regression;
  __pyx_t_1 = __pyx_vtabptr_8skgarden_8mondrian_4tree_5_tree_Tree->predict(__pyx_v_self, __pyx_v_X, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 887, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.predict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":1002
 * 
 * 
 *     cdef inline np.ndarray _apply_dense(self, object X):             # <<<<<<<<<<<<<<
 *         """Finds the terminal region (=leaf node) for each sample in X."""
 * 
 */

static PyArrayObject *__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree__apply_dense(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_X) {
  PyArrayObject *__pyx_v_X_ndarray = 0;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *__pyx_v_X_ptr;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_X_sample_stride;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_X_fx_stride;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_n_samples;
  CYTHON_UNUSED __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_n_features;
  PyArrayObject *__pyx_v_out = 0;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t *__pyx_v_out_ptr;
  struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node *__pyx_v_node;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_i;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_curr_node_id;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_out;
  __Pyx_Buffer __pyx_pybuffer_out;
  PyArrayObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_6;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyArrayObject *__pyx_t_10 = NULL;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_11;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_12;
  __Pyx_RefNannySetupContext("_apply_dense", 0);
  __pyx_pybuffer_out.pybuffer.buf = NULL;
  __pyx_pybuffer_out.refcount = 0;
  __pyx_pybuffernd_out.data = NULL;
  __pyx_pybuffernd_out.rcbuffer = &__pyx_pybuffer_out;

  /* "skgarden/mondrian/tree/_tree.pyx":1006
 * 
 *         # Check input
 *         if not isinstance(X, np.ndarray):             # <<<<<<<<<<<<<<
 *             raise ValueError("X should be in np.ndarray format, got %s"
 *                              % type(X))
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_X, __pyx_ptype_5numpy_ndarray); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "skgarden/mondrian/tree/_tree.pyx":1008
 *         if not isinstance(X, np.ndarray):
 *             raise ValueError("X should be in np.ndarray format, got %s"
 *                              % type(X))             # <<<<<<<<<<<<<<
 * 
 *         if X.dtype != DTYPE:
 */
    __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_X_should_be_in_np_ndarray_format, ((PyObject *)Py_TYPE(__pyx_v_X))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1008, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "skgarden/mondrian/tree/_tree.pyx":1007
 *         # Check input
 *         if not isinstance(X, np.ndarray):
 *             raise ValueError("X should be in np.ndarray format, got %s"             # <<<<<<<<<<<<<<
 *                              % type(X))
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1007, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 1007, __pyx_L1_error)

    /* "skgarden/mondrian/tree/_tree.pyx":1006
 * 
 *         # Check input
 *         if not isinstance(X, np.ndarray):             # <<<<<<<<<<<<<<
 *             raise ValueError("X should be in np.ndarray format, got %s"
 *                              % type(X))
 */
  }

  /* "skgarden/mondrian/tree/_tree.pyx":1010
 *                              % type(X))
 * 
 *         if X.dtype != DTYPE:             # <<<<<<<<<<<<<<
 *             raise ValueError("X.dtype should be np.float32, got %s" % X.dtype)
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_dtype); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_DTYPE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = PyObject_RichCompare(__pyx_t_4, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1010, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1010, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(__pyx_t_2)) {

    /* "skgarden/mondrian/tree/_tree.pyx":1011
 * 
 *         if X.dtype != DTYPE:
 *             raise ValueError("X.dtype should be np.float32, got %s" % X.dtype)             # <<<<<<<<<<<<<<
 * 
 *         # Extract input
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_dtype); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1011, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_X_dtype_should_be_np_float32_got, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1011, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1011, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 1011, __pyx_L1_error)

    /* "skgarden/mondrian/tree/_tree.pyx":1010
 *                              % type(X))
 * 
 *         if X.dtype != DTYPE:             # <<<<<<<<<<<<<<
 *             raise ValueError("X.dtype should be np.float32, got %s" % X.dtype)
 * 
 */
  }

  /* "skgarden/mondrian/tree/_tree.pyx":1014
 * 
 *         # Extract input
 *         cdef np.ndarray X_ndarray = X             # <<<<<<<<<<<<<<
 *         cdef DTYPE_t* X_ptr = <DTYPE_t*> X_ndarray.data
 *         cdef SIZE_t X_sample_stride = <SIZE_t> X.strides[0] / <SIZE_t> X.itemsize
 */
  if (!(likely(((__pyx_v_X) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_X, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1014, __pyx_L1_error)
  __pyx_t_5 = __pyx_v_X;
  __Pyx_INCREF(__pyx_t_5);
  __pyx_v_X_ndarray = ((PyArrayObject *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1015
 *         # Extract input
 *         cdef np.ndarray X_ndarray = X
 *         cdef DTYPE_t* X_ptr = <DTYPE_t*> X_ndarray.data             # <<<<<<<<<<<<<<
 *         cdef SIZE_t X_sample_stride = <SIZE_t> X.strides[0] / <SIZE_t> X.itemsize
 *         cdef SIZE_t X_fx_stride = <SIZE_t> X.strides[1] / <SIZE_t> X.itemsize
 */
  __pyx_v_X_ptr = ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *)__pyx_v_X_ndarray->data);

  /* "skgarden/mondrian/tree/_tree.pyx":1016
 *         cdef np.ndarray X_ndarray = X
 *         cdef DTYPE_t* X_ptr = <DTYPE_t*> X_ndarray.data
 *         cdef SIZE_t X_sample_stride = <SIZE_t> X.strides[0] / <SIZE_t> X.itemsize             # <<<<<<<<<<<<<<
 *         cdef SIZE_t X_fx_stride = <SIZE_t> X.strides[1] / <SIZE_t> X.itemsize
 *         cdef SIZE_t n_samples = X.shape[0]
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_strides); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1016, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1016, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_6 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1016, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_itemsize); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1016, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_3); if (unlikely((__pyx_t_7 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1016, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_X_sample_stride = (((__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t)__pyx_t_6) / ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t)__pyx_t_7));

  /* "skgarden/mondrian/tree/_tree.pyx":1017
 *         cdef DTYPE_t* X_ptr = <DTYPE_t*> X_ndarray.data
 *         cdef SIZE_t X_sample_stride = <SIZE_t> X.strides[0] / <SIZE_t> X.itemsize
 *         cdef SIZE_t X_fx_stride = <SIZE_t> X.strides[1] / <SIZE_t> X.itemsize             # <<<<<<<<<<<<<<
 *         cdef SIZE_t n_samples = X.shape[0]
 *         cdef SIZE_t n_features = X.shape[1]
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_strides); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_3, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_7 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_5); if (unlikely((__pyx_t_7 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1017, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_itemsize); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_5); if (unlikely((__pyx_t_6 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1017, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_X_fx_stride = (((__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t)__pyx_t_7) / ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t)__pyx_t_6));

  /* "skgarden/mondrian/tree/_tree.pyx":1018
 *         cdef SIZE_t X_sample_stride = <SIZE_t> X.strides[0] / <SIZE_t> X.itemsize
 *         cdef SIZE_t X_fx_stride = <SIZE_t> X.strides[1] / <SIZE_t> X.itemsize
 *         cdef SIZE_t n_samples = X.shape[0]             # <<<<<<<<<<<<<<
 *         cdef SIZE_t n_features = X.shape[1]
 *         cdef SIZE_t f_ind
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_shape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_6 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1018, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_n_samples = __pyx_t_6;

  /* "skgarden/mondrian/tree/_tree.pyx":1019
 *         cdef SIZE_t X_fx_stride = <SIZE_t> X.strides[1] / <SIZE_t> X.itemsize
 *         cdef SIZE_t n_samples = X.shape[0]
 *         cdef SIZE_t n_features = X.shape[1]             # <<<<<<<<<<<<<<
 *         cdef SIZE_t f_ind
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1019, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_3, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1019, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_5); if (unlikely((__pyx_t_6 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1019, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_n_features = __pyx_t_6;

  /* "skgarden/mondrian/tree/_tree.pyx":1023
 * 
 *         # Initialize output
 *         cdef np.ndarray[SIZE_t] out = np.zeros((n_samples,), dtype=np.intp)             # <<<<<<<<<<<<<<
 *         cdef SIZE_t* out_ptr = <SIZE_t*> out.data
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1023, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1023, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_From_Py_intptr_t(__pyx_v_n_samples); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1023, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1023, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1023, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1023, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1023, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_intp); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1023, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_t_9) < 0) __PYX_ERR(0, 1023, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1023, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!(likely(((__pyx_t_9) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_9, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1023, __pyx_L1_error)
  __pyx_t_10 = ((PyArrayObject *)__pyx_t_9);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_out.rcbuffer->pybuffer, (PyObject*)__pyx_t_10, &__Pyx_TypeInfo_nn___pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_out = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_out.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 1023, __pyx_L1_error)
    } else {__pyx_pybuffernd_out.diminfo[0].strides = __pyx_pybuffernd_out.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_out.diminfo[0].shape = __pyx_pybuffernd_out.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_10 = 0;
  __pyx_v_out = ((PyArrayObject *)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1024
 *         # Initialize output
 *         cdef np.ndarray[SIZE_t] out = np.zeros((n_samples,), dtype=np.intp)
 *         cdef SIZE_t* out_ptr = <SIZE_t*> out.data             # <<<<<<<<<<<<<<
 * 
 *         # Initialize auxiliary data-structure
 */
  __pyx_v_out_ptr = ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t *)__pyx_v_out->data);

  /* "skgarden/mondrian/tree/_tree.pyx":1027
 * 
 *         # Initialize auxiliary data-structure
 *         cdef Node* node = NULL             # <<<<<<<<<<<<<<
 *         cdef SIZE_t i = 0
 *         cdef SIZE_t curr_node_id
 */
  __pyx_v_node = NULL;

  /* "skgarden/mondrian/tree/_tree.pyx":1028
 *         # Initialize auxiliary data-structure
 *         cdef Node* node = NULL
 *         cdef SIZE_t i = 0             # <<<<<<<<<<<<<<
 *         cdef SIZE_t curr_node_id
 * 
 */
  __pyx_v_i = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1031
 *         cdef SIZE_t curr_node_id
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             for i in range(n_samples):
 *                 curr_node_id = self.root
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "skgarden/mondrian/tree/_tree.pyx":1032
 * 
 *         with nogil:
 *             for i in range(n_samples):             # <<<<<<<<<<<<<<
 *                 curr_node_id = self.root
 * 
 */
        __pyx_t_6 = __pyx_v_n_samples;
        __pyx_t_7 = __pyx_t_6;
        for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_7; __pyx_t_11+=1) {
          __pyx_v_i = __pyx_t_11;

          /* "skgarden/mondrian/tree/_tree.pyx":1033
 *         with nogil:
 *             for i in range(n_samples):
 *                 curr_node_id = self.root             # <<<<<<<<<<<<<<
 * 
 *                 while True:
 */
          __pyx_t_12 = __pyx_v_self->root;
          __pyx_v_curr_node_id = __pyx_t_12;

          /* "skgarden/mondrian/tree/_tree.pyx":1035
 *                 curr_node_id = self.root
 * 
 *                 while True:             # <<<<<<<<<<<<<<
 * 
 *                     node = &self.nodes[curr_node_id]
 */
          while (1) {

            /* "skgarden/mondrian/tree/_tree.pyx":1037
 *                 while True:
 * 
 *                     node = &self.nodes[curr_node_id]             # <<<<<<<<<<<<<<
 * 
 *                     if X_ptr[X_sample_stride * i +
 */
            __pyx_v_node = (&(__pyx_v_self->nodes[__pyx_v_curr_node_id]));

            /* "skgarden/mondrian/tree/_tree.pyx":1040
 * 
 *                     if X_ptr[X_sample_stride * i +
 *                              X_fx_stride * node.feature] <= node.threshold:             # <<<<<<<<<<<<<<
 * 
 *                         if node.left_child == _TREE_LEAF:
 */
            __pyx_t_2 = (((__pyx_v_X_ptr[((__pyx_v_X_sample_stride * __pyx_v_i) + (__pyx_v_X_fx_stride * __pyx_v_node->feature))]) <= __pyx_v_node->threshold) != 0);

            /* "skgarden/mondrian/tree/_tree.pyx":1039
 *                     node = &self.nodes[curr_node_id]
 * 
 *                     if X_ptr[X_sample_stride * i +             # <<<<<<<<<<<<<<
 *                              X_fx_stride * node.feature] <= node.threshold:
 * 
 */
            if (__pyx_t_2) {

              /* "skgarden/mondrian/tree/_tree.pyx":1042
 *                              X_fx_stride * node.feature] <= node.threshold:
 * 
 *                         if node.left_child == _TREE_LEAF:             # <<<<<<<<<<<<<<
 *                             break
 *                         curr_node_id = node.left_child
 */
              __pyx_t_2 = ((__pyx_v_node->left_child == __pyx_v_8skgarden_8mondrian_4tree_5_tree__TREE_LEAF) != 0);
              if (__pyx_t_2) {

                /* "skgarden/mondrian/tree/_tree.pyx":1043
 * 
 *                         if node.left_child == _TREE_LEAF:
 *                             break             # <<<<<<<<<<<<<<
 *                         curr_node_id = node.left_child
 * 
 */
                goto __pyx_L11_break;

                /* "skgarden/mondrian/tree/_tree.pyx":1042
 *                              X_fx_stride * node.feature] <= node.threshold:
 * 
 *                         if node.left_child == _TREE_LEAF:             # <<<<<<<<<<<<<<
 *                             break
 *                         curr_node_id = node.left_child
 */
              }

              /* "skgarden/mondrian/tree/_tree.pyx":1044
 *                         if node.left_child == _TREE_LEAF:
 *                             break
 *                         curr_node_id = node.left_child             # <<<<<<<<<<<<<<
 * 
 *                     else:
 */
              __pyx_t_12 = __pyx_v_node->left_child;
              __pyx_v_curr_node_id = __pyx_t_12;

              /* "skgarden/mondrian/tree/_tree.pyx":1039
 *                     node = &self.nodes[curr_node_id]
 * 
 *                     if X_ptr[X_sample_stride * i +             # <<<<<<<<<<<<<<
 *                              X_fx_stride * node.feature] <= node.threshold:
 * 
 */
              goto __pyx_L12;
            }

            /* "skgarden/mondrian/tree/_tree.pyx":1047
 * 
 *                     else:
 *                         if node.right_child == _TREE_LEAF:             # <<<<<<<<<<<<<<
 *                             break
 *                         curr_node_id = node.right_child
 */
            /*else*/ {
              __pyx_t_2 = ((__pyx_v_node->right_child == __pyx_v_8skgarden_8mondrian_4tree_5_tree__TREE_LEAF) != 0);
              if (__pyx_t_2) {

                /* "skgarden/mondrian/tree/_tree.pyx":1048
 *                     else:
 *                         if node.right_child == _TREE_LEAF:
 *                             break             # <<<<<<<<<<<<<<
 *                         curr_node_id = node.right_child
 *                 out_ptr[i] = curr_node_id
 */
                goto __pyx_L11_break;

                /* "skgarden/mondrian/tree/_tree.pyx":1047
 * 
 *                     else:
 *                         if node.right_child == _TREE_LEAF:             # <<<<<<<<<<<<<<
 *                             break
 *                         curr_node_id = node.right_child
 */
              }

              /* "skgarden/mondrian/tree/_tree.pyx":1049
 *                         if node.right_child == _TREE_LEAF:
 *                             break
 *                         curr_node_id = node.right_child             # <<<<<<<<<<<<<<
 *                 out_ptr[i] = curr_node_id
 *         return out
 */
              __pyx_t_12 = __pyx_v_node->right_child;
              __pyx_v_curr_node_id = __pyx_t_12;
            }
            __pyx_L12:;
          }
          __pyx_L11_break:;

          /* "skgarden/mondrian/tree/_tree.pyx":1050
 *                             break
 *                         curr_node_id = node.right_child
 *                 out_ptr[i] = curr_node_id             # <<<<<<<<<<<<<<
 *         return out
 * 
 */
          (__pyx_v_out_ptr[__pyx_v_i]) = __pyx_v_curr_node_id;
        }
      }

      /* "skgarden/mondrian/tree/_tree.pyx":1031
 *         cdef SIZE_t curr_node_id
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             for i in range(n_samples):
 *                 curr_node_id = self.root
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L7;
        }
        __pyx_L7:;
      }
  }

  /* "skgarden/mondrian/tree/_tree.pyx":1051
 *                         curr_node_id = node.right_child
 *                 out_ptr[i] = curr_node_id
 *         return out             # <<<<<<<<<<<<<<
 * 
 *     cpdef object decision_path(self, object X):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_out));
  __pyx_r = ((PyArrayObject *)__pyx_v_out);
  goto __pyx_L0;

  /* "skgarden/mondrian/tree/_tree.pyx":1002
 * 
 * 
 *     cdef inline np.ndarray _apply_dense(self, object X):             # <<<<<<<<<<<<<<
 *         """Finds the terminal region (=leaf node) for each sample in X."""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_out.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree._apply_dense", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_out.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_X_ndarray);
  __Pyx_XDECREF((PyObject *)__pyx_v_out);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":1053
 *         return out
 * 
 *     cpdef object decision_path(self, object X):             # <<<<<<<<<<<<<<
 *         """Finds the decision path (=node) for each sample in X."""
 *         return self._decision_path_dense(X)
 */

static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_15decision_path(PyObject *__pyx_v_self, PyObject *__pyx_v_X); /*proto*/
static PyObject *__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree_decision_path(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_X, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("decision_path", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    static PY_UINT64_T tp_dict_version = 0, obj_dict_version = 0;
    if (likely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict && tp_dict_version == __PYX_GET_DICT_VERSION(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict) && (!Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset || obj_dict_version == __PYX_GET_DICT_VERSION(_PyObject_GetDictPtr(((PyObject *)__pyx_v_self))))));
    else {
      PY_UINT64_T type_dict_guard = (likely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict)) ? __PYX_GET_DICT_VERSION(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict) : 0;
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_decision_path); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1053, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_15decision_path)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_X) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_X);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1053, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
      tp_dict_version = likely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict) ? __PYX_GET_DICT_VERSION(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict) : 0;
      obj_dict_version = likely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset) ? __PYX_GET_DICT_VERSION(_PyObject_GetDictPtr(((PyObject *)__pyx_v_self))) : 0;
      if (unlikely(type_dict_guard != tp_dict_version)) {
        tp_dict_version = obj_dict_version = 0;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    }
    #endif
  }

  /* "skgarden/mondrian/tree/_tree.pyx":1055
 *     cpdef object decision_path(self, object X):
 *         """Finds the decision path (=node) for each sample in X."""
 *         return self._decision_path_dense(X)             # <<<<<<<<<<<<<<
 * 
 *     cpdef object weighted_decision_path(self, object X):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree__decision_path_dense(__pyx_v_self, __pyx_v_X); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1055, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "skgarden/mondrian/tree/_tree.pyx":1053
 *         return out
 * 
 *     cpdef object decision_path(self, object X):             # <<<<<<<<<<<<<<
 *         """Finds the decision path (=node) for each sample in X."""
 *         return self._decision_path_dense(X)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.decision_path", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_15decision_path(PyObject *__pyx_v_self, PyObject *__pyx_v_X); /*proto*/
static char __pyx_doc_8skgarden_8mondrian_4tree_5_tree_4Tree_14decision_path[] = "Finds the decision path (=node) for each sample in X.";
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_15decision_path(PyObject *__pyx_v_self, PyObject *__pyx_v_X) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("decision_path (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_14decision_path(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self), ((PyObject *)__pyx_v_X));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_14decision_path(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_X) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("decision_path", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree_decision_path(__pyx_v_self, __pyx_v_X, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1053, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.decision_path", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":1057
 *         return self._decision_path_dense(X)
 * 
 *     cpdef object weighted_decision_path(self, object X):             # <<<<<<<<<<<<<<
 *         """Returns the weight at each node for each sample in X."""
 *         if not isinstance(X, np.ndarray):
 */

static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_17weighted_decision_path(PyObject *__pyx_v_self, PyObject *__pyx_v_X); /*proto*/
static PyObject *__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree_weighted_decision_path(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_X, int __pyx_skip_dispatch) {
  PyArrayObject *__pyx_v_X_ndarray = 0;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *__pyx_v_X_ptr;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_X_sample_stride;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_X_fx_stride;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_n_samples;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_n_features;
  PyArrayObject *__pyx_v_indptr = 0;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t *__pyx_v_indptr_ptr;
  PyArrayObject *__pyx_v_indices = 0;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t *__pyx_v_indices_ptr;
  PyArrayObject *__pyx_v_values = 0;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *__pyx_v_values_ptr;
  struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node *__pyx_v_node;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_i;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t __pyx_v_parent_tau;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t __pyx_v_delta;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t __pyx_v_eta;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t __pyx_v_X_val;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t __pyx_v_p_s;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t __pyx_v_p_nsy;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_f_ind;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_curr_node_id;
  PyObject *__pyx_v_out = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indptr;
  __Pyx_Buffer __pyx_pybuffer_indptr;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_values;
  __Pyx_Buffer __pyx_pybuffer_values;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_7;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyArrayObject *__pyx_t_10 = NULL;
  PyArrayObject *__pyx_t_11 = NULL;
  PyArrayObject *__pyx_t_12 = NULL;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_13;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_14;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t __pyx_t_15;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_16;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_17;
  long __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  int __pyx_t_20;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  Py_ssize_t __pyx_t_24;
  PyObject *__pyx_t_25 = NULL;
  __Pyx_RefNannySetupContext("weighted_decision_path", 0);
  __pyx_pybuffer_indptr.pybuffer.buf = NULL;
  __pyx_pybuffer_indptr.refcount = 0;
  __pyx_pybuffernd_indptr.data = NULL;
  __pyx_pybuffernd_indptr.rcbuffer = &__pyx_pybuffer_indptr;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  __pyx_pybuffer_values.pybuffer.buf = NULL;
  __pyx_pybuffer_values.refcount = 0;
  __pyx_pybuffernd_values.data = NULL;
  __pyx_pybuffernd_values.rcbuffer = &__pyx_pybuffer_values;
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    static PY_UINT64_T tp_dict_version = 0, obj_dict_version = 0;
    if (likely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict && tp_dict_version == __PYX_GET_DICT_VERSION(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict) && (!Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset || obj_dict_version == __PYX_GET_DICT_VERSION(_PyObject_GetDictPtr(((PyObject *)__pyx_v_self))))));
    else {
      PY_UINT64_T type_dict_guard = (likely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict)) ? __PYX_GET_DICT_VERSION(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict) : 0;
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_weighted_decision_path); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1057, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_17weighted_decision_path)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_X) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_X);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1057, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
      tp_dict_version = likely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict) ? __PYX_GET_DICT_VERSION(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict) : 0;
      obj_dict_version = likely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset) ? __PYX_GET_DICT_VERSION(_PyObject_GetDictPtr(((PyObject *)__pyx_v_self))) : 0;
      if (unlikely(type_dict_guard != tp_dict_version)) {
        tp_dict_version = obj_dict_version = 0;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    }
    #endif
  }

  /* "skgarden/mondrian/tree/_tree.pyx":1059
 *     cpdef object weighted_decision_path(self, object X):
 *         """Returns the weight at each node for each sample in X."""
 *         if not isinstance(X, np.ndarray):             # <<<<<<<<<<<<<<
 *             raise ValueError("X should be in np.ndarray format, got %s"
 *                              % type(X))
 */
  __pyx_t_5 = __Pyx_TypeCheck(__pyx_v_X, __pyx_ptype_5numpy_ndarray); 
  __pyx_t_6 = ((!(__pyx_t_5 != 0)) != 0);
  if (unlikely(__pyx_t_6)) {

    /* "skgarden/mondrian/tree/_tree.pyx":1061
 *         if not isinstance(X, np.ndarray):
 *             raise ValueError("X should be in np.ndarray format, got %s"
 *                              % type(X))             # <<<<<<<<<<<<<<
 * 
 *         if X.dtype != DTYPE:
 */
    __pyx_t_1 = __Pyx_PyString_FormatSafe(__pyx_kp_s_X_should_be_in_np_ndarray_format, ((PyObject *)Py_TYPE(__pyx_v_X))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1061, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "skgarden/mondrian/tree/_tree.pyx":1060
 *         """Returns the weight at each node for each sample in X."""
 *         if not isinstance(X, np.ndarray):
 *             raise ValueError("X should be in np.ndarray format, got %s"             # <<<<<<<<<<<<<<
 *                              % type(X))
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1060, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1060, __pyx_L1_error)

    /* "skgarden/mondrian/tree/_tree.pyx":1059
 *     cpdef object weighted_decision_path(self, object X):
 *         """Returns the weight at each node for each sample in X."""
 *         if not isinstance(X, np.ndarray):             # <<<<<<<<<<<<<<
 *             raise ValueError("X should be in np.ndarray format, got %s"
 *                              % type(X))
 */
  }

  /* "skgarden/mondrian/tree/_tree.pyx":1063
 *                              % type(X))
 * 
 *         if X.dtype != DTYPE:             # <<<<<<<<<<<<<<
 *             raise ValueError("X.dtype should be np.float32, got %s" % X.dtype)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_dtype); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1063, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_DTYPE); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1063, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_t_1, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1063, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1063, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(__pyx_t_6)) {

    /* "skgarden/mondrian/tree/_tree.pyx":1064
 * 
 *         if X.dtype != DTYPE:
 *             raise ValueError("X.dtype should be np.float32, got %s" % X.dtype)             # <<<<<<<<<<<<<<
 * 
 *         # Extract input
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_dtype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1064, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_FormatSafe(__pyx_kp_s_X_dtype_should_be_np_float32_got, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1064, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1064, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1064, __pyx_L1_error)

    /* "skgarden/mondrian/tree/_tree.pyx":1063
 *                              % type(X))
 * 
 *         if X.dtype != DTYPE:             # <<<<<<<<<<<<<<
 *             raise ValueError("X.dtype should be np.float32, got %s" % X.dtype)
 * 
 */
  }

  /* "skgarden/mondrian/tree/_tree.pyx":1067
 * 
 *         # Extract input
 *         cdef np.ndarray X_ndarray = X             # <<<<<<<<<<<<<<
 *         cdef DTYPE_t* X_ptr = <DTYPE_t*> X_ndarray.data
 *         cdef SIZE_t X_sample_stride = <SIZE_t> X.strides[0] / <SIZE_t> X.itemsize
 */
  if (!(likely(((__pyx_v_X) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_X, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1067, __pyx_L1_error)
  __pyx_t_3 = __pyx_v_X;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_X_ndarray = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1068
 *         # Extract input
 *         cdef np.ndarray X_ndarray = X
 *         cdef DTYPE_t* X_ptr = <DTYPE_t*> X_ndarray.data             # <<<<<<<<<<<<<<
 *         cdef SIZE_t X_sample_stride = <SIZE_t> X.strides[0] / <SIZE_t> X.itemsize
 *         cdef SIZE_t X_fx_stride = <SIZE_t> X.strides[1] / <SIZE_t> X.itemsize
 */
  __pyx_v_X_ptr = ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *)__pyx_v_X_ndarray->data);

  /* "skgarden/mondrian/tree/_tree.pyx":1069
 *         cdef np.ndarray X_ndarray = X
 *         cdef DTYPE_t* X_ptr = <DTYPE_t*> X_ndarray.data
 *         cdef SIZE_t X_sample_stride = <SIZE_t> X.strides[0] / <SIZE_t> X.itemsize             # <<<<<<<<<<<<<<
 *         cdef SIZE_t X_fx_stride = <SIZE_t> X.strides[1] / <SIZE_t> X.itemsize
 *         cdef SIZE_t n_samples = X.shape[0]
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_strides); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1069, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1069, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_7 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_1); if (unlikely((__pyx_t_7 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1069, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_itemsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1069, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_1); if (unlikely((__pyx_t_8 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1069, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_X_sample_stride = (((__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t)__pyx_t_7) / ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t)__pyx_t_8));

  /* "skgarden/mondrian/tree/_tree.pyx":1070
 *         cdef DTYPE_t* X_ptr = <DTYPE_t*> X_ndarray.data
 *         cdef SIZE_t X_sample_stride = <SIZE_t> X.strides[0] / <SIZE_t> X.itemsize
 *         cdef SIZE_t X_fx_stride = <SIZE_t> X.strides[1] / <SIZE_t> X.itemsize             # <<<<<<<<<<<<<<
 *         cdef SIZE_t n_samples = X.shape[0]
 *         cdef SIZE_t n_features = X.shape[1]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_strides); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1070, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_1, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1070, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_8 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_3); if (unlikely((__pyx_t_8 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1070, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_itemsize); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1070, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_3); if (unlikely((__pyx_t_7 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1070, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_X_fx_stride = (((__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t)__pyx_t_8) / ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t)__pyx_t_7));

  /* "skgarden/mondrian/tree/_tree.pyx":1071
 *         cdef SIZE_t X_sample_stride = <SIZE_t> X.strides[0] / <SIZE_t> X.itemsize
 *         cdef SIZE_t X_fx_stride = <SIZE_t> X.strides[1] / <SIZE_t> X.itemsize
 *         cdef SIZE_t n_samples = X.shape[0]             # <<<<<<<<<<<<<<
 *         cdef SIZE_t n_features = X.shape[1]
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_7 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_1); if (unlikely((__pyx_t_7 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1071, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_n_samples = __pyx_t_7;

  /* "skgarden/mondrian/tree/_tree.pyx":1072
 *         cdef SIZE_t X_fx_stride = <SIZE_t> X.strides[1] / <SIZE_t> X.itemsize
 *         cdef SIZE_t n_samples = X.shape[0]
 *         cdef SIZE_t n_features = X.shape[1]             # <<<<<<<<<<<<<<
 * 
 *         # Initialize output
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_1, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_3); if (unlikely((__pyx_t_7 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1072, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_n_features = __pyx_t_7;

  /* "skgarden/mondrian/tree/_tree.pyx":1075
 * 
 *         # Initialize output
 *         cdef np.ndarray[SIZE_t] indptr = np.zeros(n_samples + 1, dtype=np.intp)             # <<<<<<<<<<<<<<
 *         cdef SIZE_t* indptr_ptr = <SIZE_t*> indptr.data
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1075, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1075, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_long((__pyx_v_n_samples + 1)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1075, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1075, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1075, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1075, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_intp); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1075, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_9) < 0) __PYX_ERR(0, 1075, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1075, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_9) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_9, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1075, __pyx_L1_error)
  __pyx_t_10 = ((PyArrayObject *)__pyx_t_9);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indptr.rcbuffer->pybuffer, (PyObject*)__pyx_t_10, &__Pyx_TypeInfo_nn___pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_indptr = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_indptr.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 1075, __pyx_L1_error)
    } else {__pyx_pybuffernd_indptr.diminfo[0].strides = __pyx_pybuffernd_indptr.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indptr.diminfo[0].shape = __pyx_pybuffernd_indptr.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_10 = 0;
  __pyx_v_indptr = ((PyArrayObject *)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1076
 *         # Initialize output
 *         cdef np.ndarray[SIZE_t] indptr = np.zeros(n_samples + 1, dtype=np.intp)
 *         cdef SIZE_t* indptr_ptr = <SIZE_t*> indptr.data             # <<<<<<<<<<<<<<
 * 
 *         cdef np.ndarray[SIZE_t] indices = np.zeros(n_samples *
 */
  __pyx_v_indptr_ptr = ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t *)__pyx_v_indptr->data);

  /* "skgarden/mondrian/tree/_tree.pyx":1078
 *         cdef SIZE_t* indptr_ptr = <SIZE_t*> indptr.data
 * 
 *         cdef np.ndarray[SIZE_t] indices = np.zeros(n_samples *             # <<<<<<<<<<<<<<
 *                                                    (1 + self.max_depth),
 *                                                    dtype=np.intp)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_np); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1079
 * 
 *         cdef np.ndarray[SIZE_t] indices = np.zeros(n_samples *
 *                                                    (1 + self.max_depth),             # <<<<<<<<<<<<<<
 *                                                    dtype=np.intp)
 *         cdef SIZE_t* indices_ptr = <SIZE_t*> indices.data
 */
  __pyx_t_9 = __Pyx_PyInt_From_long((__pyx_v_n_samples * (1 + __pyx_v_self->max_depth))); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "skgarden/mondrian/tree/_tree.pyx":1078
 *         cdef SIZE_t* indptr_ptr = <SIZE_t*> indptr.data
 * 
 *         cdef np.ndarray[SIZE_t] indices = np.zeros(n_samples *             # <<<<<<<<<<<<<<
 *                                                    (1 + self.max_depth),
 *                                                    dtype=np.intp)
 */
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_9);
  __pyx_t_9 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1080
 *         cdef np.ndarray[SIZE_t] indices = np.zeros(n_samples *
 *                                                    (1 + self.max_depth),
 *                                                    dtype=np.intp)             # <<<<<<<<<<<<<<
 *         cdef SIZE_t* indices_ptr = <SIZE_t*> indices.data
 * 
 */
  __pyx_t_9 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1080, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1080, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_intp); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1080, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_dtype, __pyx_t_4) < 0) __PYX_ERR(0, 1080, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1078
 *         cdef SIZE_t* indptr_ptr = <SIZE_t*> indptr.data
 * 
 *         cdef np.ndarray[SIZE_t] indices = np.zeros(n_samples *             # <<<<<<<<<<<<<<
 *                                                    (1 + self.max_depth),
 *                                                    dtype=np.intp)
 */
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, __pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1078, __pyx_L1_error)
  __pyx_t_11 = ((PyArrayObject *)__pyx_t_4);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_t_11, &__Pyx_TypeInfo_nn___pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_indices = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_indices.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 1078, __pyx_L1_error)
    } else {__pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_11 = 0;
  __pyx_v_indices = ((PyArrayObject *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1081
 *                                                    (1 + self.max_depth),
 *                                                    dtype=np.intp)
 *         cdef SIZE_t* indices_ptr = <SIZE_t*> indices.data             # <<<<<<<<<<<<<<
 * 
 *         cdef np.ndarray[DTYPE_t] values = np.zeros(n_samples *
 */
  __pyx_v_indices_ptr = ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t *)__pyx_v_indices->data);

  /* "skgarden/mondrian/tree/_tree.pyx":1083
 *         cdef SIZE_t* indices_ptr = <SIZE_t*> indices.data
 * 
 *         cdef np.ndarray[DTYPE_t] values = np.zeros(n_samples *             # <<<<<<<<<<<<<<
 *                                                 (1 + self.max_depth),
 *                                                  dtype=DTYPE)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1083, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1083, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1084
 * 
 *         cdef np.ndarray[DTYPE_t] values = np.zeros(n_samples *
 *                                                 (1 + self.max_depth),             # <<<<<<<<<<<<<<
 *                                                  dtype=DTYPE)
 *         cdef DTYPE_t* values_ptr = <DTYPE_t*> values.data
 */
  __pyx_t_4 = __Pyx_PyInt_From_long((__pyx_v_n_samples * (1 + __pyx_v_self->max_depth))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1083, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "skgarden/mondrian/tree/_tree.pyx":1083
 *         cdef SIZE_t* indices_ptr = <SIZE_t*> indices.data
 * 
 *         cdef np.ndarray[DTYPE_t] values = np.zeros(n_samples *             # <<<<<<<<<<<<<<
 *                                                 (1 + self.max_depth),
 *                                                  dtype=DTYPE)
 */
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1083, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4);
  __pyx_t_4 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1085
 *         cdef np.ndarray[DTYPE_t] values = np.zeros(n_samples *
 *                                                 (1 + self.max_depth),
 *                                                  dtype=DTYPE)             # <<<<<<<<<<<<<<
 *         cdef DTYPE_t* values_ptr = <DTYPE_t*> values.data
 * 
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1085, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_DTYPE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1085, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_t_3) < 0) __PYX_ERR(0, 1085, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1083
 *         cdef SIZE_t* indices_ptr = <SIZE_t*> indices.data
 * 
 *         cdef np.ndarray[DTYPE_t] values = np.zeros(n_samples *             # <<<<<<<<<<<<<<
 *                                                 (1 + self.max_depth),
 *                                                  dtype=DTYPE)
 */
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1083, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1083, __pyx_L1_error)
  __pyx_t_12 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_values.rcbuffer->pybuffer, (PyObject*)__pyx_t_12, &__Pyx_TypeInfo_nn___pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_values = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_values.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 1083, __pyx_L1_error)
    } else {__pyx_pybuffernd_values.diminfo[0].strides = __pyx_pybuffernd_values.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_values.diminfo[0].shape = __pyx_pybuffernd_values.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_12 = 0;
  __pyx_v_values = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1086
 *                                                 (1 + self.max_depth),
 *                                                  dtype=DTYPE)
 *         cdef DTYPE_t* values_ptr = <DTYPE_t*> values.data             # <<<<<<<<<<<<<<
 * 
 *         # Initialize auxiliary data-structure
 */
  __pyx_v_values_ptr = ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *)__pyx_v_values->data);

  /* "skgarden/mondrian/tree/_tree.pyx":1089
 * 
 *         # Initialize auxiliary data-structure
 *         cdef Node* node = NULL             # <<<<<<<<<<<<<<
 *         cdef SIZE_t i = 0
 * 
 */
  __pyx_v_node = NULL;

  /* "skgarden/mondrian/tree/_tree.pyx":1090
 *         # Initialize auxiliary data-structure
 *         cdef Node* node = NULL
 *         cdef SIZE_t i = 0             # <<<<<<<<<<<<<<
 * 
 *         cdef DTYPE_t parent_tau
 */
  __pyx_v_i = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1101
 *         cdef SIZE_t curr_node_id
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             for i in range(n_samples):
 *                 p_nsy = 1.0
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "skgarden/mondrian/tree/_tree.pyx":1102
 * 
 *         with nogil:
 *             for i in range(n_samples):             # <<<<<<<<<<<<<<
 *                 p_nsy = 1.0
 *                 parent_tau = 0.0
 */
        __pyx_t_7 = __pyx_v_n_samples;
        __pyx_t_8 = __pyx_t_7;
        for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_8; __pyx_t_13+=1) {
          __pyx_v_i = __pyx_t_13;

          /* "skgarden/mondrian/tree/_tree.pyx":1103
 *         with nogil:
 *             for i in range(n_samples):
 *                 p_nsy = 1.0             # <<<<<<<<<<<<<<
 *                 parent_tau = 0.0
 *                 indptr_ptr[i + 1] = indptr_ptr[i]
 */
          __pyx_v_p_nsy = 1.0;

          /* "skgarden/mondrian/tree/_tree.pyx":1104
 *             for i in range(n_samples):
 *                 p_nsy = 1.0
 *                 parent_tau = 0.0             # <<<<<<<<<<<<<<
 *                 indptr_ptr[i + 1] = indptr_ptr[i]
 * 
 */
          __pyx_v_parent_tau = 0.0;

          /* "skgarden/mondrian/tree/_tree.pyx":1105
 *                 p_nsy = 1.0
 *                 parent_tau = 0.0
 *                 indptr_ptr[i + 1] = indptr_ptr[i]             # <<<<<<<<<<<<<<
 * 
 *                 curr_node_id = self.root
 */
          (__pyx_v_indptr_ptr[(__pyx_v_i + 1)]) = (__pyx_v_indptr_ptr[__pyx_v_i]);

          /* "skgarden/mondrian/tree/_tree.pyx":1107
 *                 indptr_ptr[i + 1] = indptr_ptr[i]
 * 
 *                 curr_node_id = self.root             # <<<<<<<<<<<<<<
 *                 node = &self.nodes[curr_node_id]
 * 
 */
          __pyx_t_14 = __pyx_v_self->root;
          __pyx_v_curr_node_id = __pyx_t_14;

          /* "skgarden/mondrian/tree/_tree.pyx":1108
 * 
 *                 curr_node_id = self.root
 *                 node = &self.nodes[curr_node_id]             # <<<<<<<<<<<<<<
 * 
 *                 while node.left_child != _TREE_LEAF:
 */
          __pyx_v_node = (&(__pyx_v_self->nodes[__pyx_v_curr_node_id]));

          /* "skgarden/mondrian/tree/_tree.pyx":1110
 *                 node = &self.nodes[curr_node_id]
 * 
 *                 while node.left_child != _TREE_LEAF:             # <<<<<<<<<<<<<<
 * 
 *                     delta = node.tau - parent_tau
 */
          while (1) {
            __pyx_t_6 = ((__pyx_v_node->left_child != __pyx_v_8skgarden_8mondrian_4tree_5_tree__TREE_LEAF) != 0);
            if (!__pyx_t_6) break;

            /* "skgarden/mondrian/tree/_tree.pyx":1112
 *                 while node.left_child != _TREE_LEAF:
 * 
 *                     delta = node.tau - parent_tau             # <<<<<<<<<<<<<<
 *                     parent_tau = node.tau
 * 
 */
            __pyx_v_delta = (__pyx_v_node->tau - __pyx_v_parent_tau);

            /* "skgarden/mondrian/tree/_tree.pyx":1113
 * 
 *                     delta = node.tau - parent_tau
 *                     parent_tau = node.tau             # <<<<<<<<<<<<<<
 * 
 *                     eta = 0.0
 */
            __pyx_t_15 = __pyx_v_node->tau;
            __pyx_v_parent_tau = __pyx_t_15;

            /* "skgarden/mondrian/tree/_tree.pyx":1115
 *                     parent_tau = node.tau
 * 
 *                     eta = 0.0             # <<<<<<<<<<<<<<
 *                     for f_ind in range(n_features):
 *                         X_val = X_ptr[X_sample_stride * i + X_fx_stride * f_ind]
 */
            __pyx_v_eta = 0.0;

            /* "skgarden/mondrian/tree/_tree.pyx":1116
 * 
 *                     eta = 0.0
 *                     for f_ind in range(n_features):             # <<<<<<<<<<<<<<
 *                         X_val = X_ptr[X_sample_stride * i + X_fx_stride * f_ind]
 *                         eta += (fmax(X_val - node.upper_bounds[f_ind], 0.0) +
 */
            __pyx_t_14 = __pyx_v_n_features;
            __pyx_t_16 = __pyx_t_14;
            for (__pyx_t_17 = 0; __pyx_t_17 < __pyx_t_16; __pyx_t_17+=1) {
              __pyx_v_f_ind = __pyx_t_17;

              /* "skgarden/mondrian/tree/_tree.pyx":1117
 *                     eta = 0.0
 *                     for f_ind in range(n_features):
 *                         X_val = X_ptr[X_sample_stride * i + X_fx_stride * f_ind]             # <<<<<<<<<<<<<<
 *                         eta += (fmax(X_val - node.upper_bounds[f_ind], 0.0) +
 *                                 fmax(node.lower_bounds[f_ind] - X_val, 0.0))
 */
              __pyx_v_X_val = (__pyx_v_X_ptr[((__pyx_v_X_sample_stride * __pyx_v_i) + (__pyx_v_X_fx_stride * __pyx_v_f_ind))]);

              /* "skgarden/mondrian/tree/_tree.pyx":1118
 *                     for f_ind in range(n_features):
 *                         X_val = X_ptr[X_sample_stride * i + X_fx_stride * f_ind]
 *                         eta += (fmax(X_val - node.upper_bounds[f_ind], 0.0) +             # <<<<<<<<<<<<<<
 *                                 fmax(node.lower_bounds[f_ind] - X_val, 0.0))
 *                     p_s = 1 - exp(-delta*eta)
 */
              __pyx_v_eta = (__pyx_v_eta + (__pyx_f_8skgarden_8mondrian_4tree_5_tree_fmax((__pyx_v_X_val - (__pyx_v_node->upper_bounds[__pyx_v_f_ind])), 0.0) + __pyx_f_8skgarden_8mondrian_4tree_5_tree_fmax(((__pyx_v_node->lower_bounds[__pyx_v_f_ind]) - __pyx_v_X_val), 0.0)));
            }

            /* "skgarden/mondrian/tree/_tree.pyx":1120
 *                         eta += (fmax(X_val - node.upper_bounds[f_ind], 0.0) +
 *                                 fmax(node.lower_bounds[f_ind] - X_val, 0.0))
 *                     p_s = 1 - exp(-delta*eta)             # <<<<<<<<<<<<<<
 * 
 *                     if p_s > 0:
 */
            __pyx_v_p_s = (1.0 - exp(((-__pyx_v_delta) * __pyx_v_eta)));

            /* "skgarden/mondrian/tree/_tree.pyx":1122
 *                     p_s = 1 - exp(-delta*eta)
 * 
 *                     if p_s > 0:             # <<<<<<<<<<<<<<
 *                         indices_ptr[indptr_ptr[i + 1]] = curr_node_id
 *                         values_ptr[indptr_ptr[i + 1]] = p_s * p_nsy
 */
            __pyx_t_6 = ((__pyx_v_p_s > 0.0) != 0);
            if (__pyx_t_6) {

              /* "skgarden/mondrian/tree/_tree.pyx":1123
 * 
 *                     if p_s > 0:
 *                         indices_ptr[indptr_ptr[i + 1]] = curr_node_id             # <<<<<<<<<<<<<<
 *                         values_ptr[indptr_ptr[i + 1]] = p_s * p_nsy
 *                         indptr_ptr[i + 1] += 1
 */
              (__pyx_v_indices_ptr[(__pyx_v_indptr_ptr[(__pyx_v_i + 1)])]) = __pyx_v_curr_node_id;

              /* "skgarden/mondrian/tree/_tree.pyx":1124
 *                     if p_s > 0:
 *                         indices_ptr[indptr_ptr[i + 1]] = curr_node_id
 *                         values_ptr[indptr_ptr[i + 1]] = p_s * p_nsy             # <<<<<<<<<<<<<<
 *                         indptr_ptr[i + 1] += 1
 * 
 */
              (__pyx_v_values_ptr[(__pyx_v_indptr_ptr[(__pyx_v_i + 1)])]) = (__pyx_v_p_s * __pyx_v_p_nsy);

              /* "skgarden/mondrian/tree/_tree.pyx":1125
 *                         indices_ptr[indptr_ptr[i + 1]] = curr_node_id
 *                         values_ptr[indptr_ptr[i + 1]] = p_s * p_nsy
 *                         indptr_ptr[i + 1] += 1             # <<<<<<<<<<<<<<
 * 
 *                     p_nsy *= (1 - p_s)
 */
              __pyx_t_18 = (__pyx_v_i + 1);
              (__pyx_v_indptr_ptr[__pyx_t_18]) = ((__pyx_v_indptr_ptr[__pyx_t_18]) + 1);

              /* "skgarden/mondrian/tree/_tree.pyx":1122
 *                     p_s = 1 - exp(-delta*eta)
 * 
 *                     if p_s > 0:             # <<<<<<<<<<<<<<
 *                         indices_ptr[indptr_ptr[i + 1]] = curr_node_id
 *                         values_ptr[indptr_ptr[i + 1]] = p_s * p_nsy
 */
            }

            /* "skgarden/mondrian/tree/_tree.pyx":1127
 *                         indptr_ptr[i + 1] += 1
 * 
 *                     p_nsy *= (1 - p_s)             # <<<<<<<<<<<<<<
 *                     if X_ptr[X_sample_stride * i +
 *                              X_fx_stride * node.feature] <= node.threshold:
 */
            __pyx_v_p_nsy = (__pyx_v_p_nsy * (1.0 - __pyx_v_p_s));

            /* "skgarden/mondrian/tree/_tree.pyx":1129
 *                     p_nsy *= (1 - p_s)
 *                     if X_ptr[X_sample_stride * i +
 *                              X_fx_stride * node.feature] <= node.threshold:             # <<<<<<<<<<<<<<
 *                         curr_node_id = node.left_child
 *                     else:
 */
            __pyx_t_6 = (((__pyx_v_X_ptr[((__pyx_v_X_sample_stride * __pyx_v_i) + (__pyx_v_X_fx_stride * __pyx_v_node->feature))]) <= __pyx_v_node->threshold) != 0);

            /* "skgarden/mondrian/tree/_tree.pyx":1128
 * 
 *                     p_nsy *= (1 - p_s)
 *                     if X_ptr[X_sample_stride * i +             # <<<<<<<<<<<<<<
 *                              X_fx_stride * node.feature] <= node.threshold:
 *                         curr_node_id = node.left_child
 */
            if (__pyx_t_6) {

              /* "skgarden/mondrian/tree/_tree.pyx":1130
 *                     if X_ptr[X_sample_stride * i +
 *                              X_fx_stride * node.feature] <= node.threshold:
 *                         curr_node_id = node.left_child             # <<<<<<<<<<<<<<
 *                     else:
 *                         curr_node_id = node.right_child
 */
              __pyx_t_14 = __pyx_v_node->left_child;
              __pyx_v_curr_node_id = __pyx_t_14;

              /* "skgarden/mondrian/tree/_tree.pyx":1128
 * 
 *                     p_nsy *= (1 - p_s)
 *                     if X_ptr[X_sample_stride * i +             # <<<<<<<<<<<<<<
 *                              X_fx_stride * node.feature] <= node.threshold:
 *                         curr_node_id = node.left_child
 */
              goto __pyx_L15;
            }

            /* "skgarden/mondrian/tree/_tree.pyx":1132
 *                         curr_node_id = node.left_child
 *                     else:
 *                         curr_node_id = node.right_child             # <<<<<<<<<<<<<<
 *                     node = &self.nodes[curr_node_id]
 * 
 */
            /*else*/ {
              __pyx_t_14 = __pyx_v_node->right_child;
              __pyx_v_curr_node_id = __pyx_t_14;
            }
            __pyx_L15:;

            /* "skgarden/mondrian/tree/_tree.pyx":1133
 *                     else:
 *                         curr_node_id = node.right_child
 *                     node = &self.nodes[curr_node_id]             # <<<<<<<<<<<<<<
 * 
 *                 # Add the leave node
 */
            __pyx_v_node = (&(__pyx_v_self->nodes[__pyx_v_curr_node_id]));
          }

          /* "skgarden/mondrian/tree/_tree.pyx":1136
 * 
 *                 # Add the leave node
 *                 indices_ptr[indptr_ptr[i + 1]] = curr_node_id             # <<<<<<<<<<<<<<
 *                 values_ptr[indptr_ptr[i + 1]] = p_nsy
 *                 indptr_ptr[i + 1] += 1
 */
          (__pyx_v_indices_ptr[(__pyx_v_indptr_ptr[(__pyx_v_i + 1)])]) = __pyx_v_curr_node_id;

          /* "skgarden/mondrian/tree/_tree.pyx":1137
 *                 # Add the leave node
 *                 indices_ptr[indptr_ptr[i + 1]] = curr_node_id
 *                 values_ptr[indptr_ptr[i + 1]] = p_nsy             # <<<<<<<<<<<<<<
 *                 indptr_ptr[i + 1] += 1
 * 
 */
          (__pyx_v_values_ptr[(__pyx_v_indptr_ptr[(__pyx_v_i + 1)])]) = __pyx_v_p_nsy;

          /* "skgarden/mondrian/tree/_tree.pyx":1138
 *                 indices_ptr[indptr_ptr[i + 1]] = curr_node_id
 *                 values_ptr[indptr_ptr[i + 1]] = p_nsy
 *                 indptr_ptr[i + 1] += 1             # <<<<<<<<<<<<<<
 * 
 *         indices = indices[:indptr[n_samples]]
 */
          __pyx_t_18 = (__pyx_v_i + 1);
          (__pyx_v_indptr_ptr[__pyx_t_18]) = ((__pyx_v_indptr_ptr[__pyx_t_18]) + 1);
        }
      }

      /* "skgarden/mondrian/tree/_tree.pyx":1101
 *         cdef SIZE_t curr_node_id
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             for i in range(n_samples):
 *                 p_nsy = 1.0
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L7;
        }
        __pyx_L7:;
      }
  }

  /* "skgarden/mondrian/tree/_tree.pyx":1140
 *                 indptr_ptr[i + 1] += 1
 * 
 *         indices = indices[:indptr[n_samples]]             # <<<<<<<<<<<<<<
 *         values = values[:indptr[n_samples]]
 *         out = csr_matrix((values, indices, indptr),
 */
  __pyx_t_19 = __pyx_v_n_samples;
  __pyx_t_3 = __Pyx_PyInt_From_Py_intptr_t((*__Pyx_BufPtrStrided1d(__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t *, __pyx_pybuffernd_indptr.rcbuffer->pybuffer.buf, __pyx_t_19, __pyx_pybuffernd_indptr.diminfo[0].strides))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PySlice_New(Py_None, __pyx_t_3, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_indices), __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1140, __pyx_L1_error)
  __pyx_t_11 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __pyx_t_20 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_t_11, &__Pyx_TypeInfo_nn___pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_20 < 0)) {
      PyErr_Fetch(&__pyx_t_21, &__pyx_t_22, &__pyx_t_23);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_21); Py_XDECREF(__pyx_t_22); Py_XDECREF(__pyx_t_23);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_21, __pyx_t_22, __pyx_t_23);
      }
      __pyx_t_21 = __pyx_t_22 = __pyx_t_23 = 0;
    }
    __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_20 < 0)) __PYX_ERR(0, 1140, __pyx_L1_error)
  }
  __pyx_t_11 = 0;
  __Pyx_DECREF_SET(__pyx_v_indices, ((PyArrayObject *)__pyx_t_3));
  __pyx_t_3 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1141
 * 
 *         indices = indices[:indptr[n_samples]]
 *         values = values[:indptr[n_samples]]             # <<<<<<<<<<<<<<
 *         out = csr_matrix((values, indices, indptr),
 *                          shape=(n_samples, self.node_count))
 */
  __pyx_t_24 = __pyx_v_n_samples;
  __pyx_t_3 = __Pyx_PyInt_From_Py_intptr_t((*__Pyx_BufPtrStrided1d(__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t *, __pyx_pybuffernd_indptr.rcbuffer->pybuffer.buf, __pyx_t_24, __pyx_pybuffernd_indptr.diminfo[0].strides))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PySlice_New(Py_None, __pyx_t_3, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_values), __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1141, __pyx_L1_error)
  __pyx_t_12 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_values.rcbuffer->pybuffer);
    __pyx_t_20 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_values.rcbuffer->pybuffer, (PyObject*)__pyx_t_12, &__Pyx_TypeInfo_nn___pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_20 < 0)) {
      PyErr_Fetch(&__pyx_t_23, &__pyx_t_22, &__pyx_t_21);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_values.rcbuffer->pybuffer, (PyObject*)__pyx_v_values, &__Pyx_TypeInfo_nn___pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_23); Py_XDECREF(__pyx_t_22); Py_XDECREF(__pyx_t_21);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_23, __pyx_t_22, __pyx_t_21);
      }
      __pyx_t_23 = __pyx_t_22 = __pyx_t_21 = 0;
    }
    __pyx_pybuffernd_values.diminfo[0].strides = __pyx_pybuffernd_values.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_values.diminfo[0].shape = __pyx_pybuffernd_values.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_20 < 0)) __PYX_ERR(0, 1141, __pyx_L1_error)
  }
  __pyx_t_12 = 0;
  __Pyx_DECREF_SET(__pyx_v_values, ((PyArrayObject *)__pyx_t_3));
  __pyx_t_3 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1142
 *         indices = indices[:indptr[n_samples]]
 *         values = values[:indptr[n_samples]]
 *         out = csr_matrix((values, indices, indptr),             # <<<<<<<<<<<<<<
 *                          shape=(n_samples, self.node_count))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_csr_matrix); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(((PyObject *)__pyx_v_values));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_values));
  PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_v_values));
  __Pyx_INCREF(((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_indices));
  PyTuple_SET_ITEM(__pyx_t_4, 1, ((PyObject *)__pyx_v_indices));
  __Pyx_INCREF(((PyObject *)__pyx_v_indptr));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_indptr));
  PyTuple_SET_ITEM(__pyx_t_4, 2, ((PyObject *)__pyx_v_indptr));
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4);
  __pyx_t_4 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1143
 *         values = values[:indptr[n_samples]]
 *         out = csr_matrix((values, indices, indptr),
 *                          shape=(n_samples, self.node_count))             # <<<<<<<<<<<<<<
 * 
 *         return out
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_9 = __Pyx_PyInt_From_Py_intptr_t(__pyx_v_n_samples); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_1 = __Pyx_PyInt_From_Py_intptr_t(__pyx_v_self->node_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_25 = PyTuple_New(2); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_t_1);
  __pyx_t_9 = 0;
  __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_shape, __pyx_t_25) < 0) __PYX_ERR(0, 1143, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1142
 *         indices = indices[:indptr[n_samples]]
 *         values = values[:indptr[n_samples]]
 *         out = csr_matrix((values, indices, indptr),             # <<<<<<<<<<<<<<
 *                          shape=(n_samples, self.node_count))
 * 
 */
  __pyx_t_25 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_out = __pyx_t_25;
  __pyx_t_25 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1145
 *                          shape=(n_samples, self.node_count))
 * 
 *         return out             # <<<<<<<<<<<<<<
 * 
 *     cdef inline object _decision_path_dense(self, object X):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_out);
  __pyx_r = __pyx_v_out;
  goto __pyx_L0;

  /* "skgarden/mondrian/tree/_tree.pyx":1057
 *         return self._decision_path_dense(X)
 * 
 *     cpdef object weighted_decision_path(self, object X):             # <<<<<<<<<<<<<<
 *         """Returns the weight at each node for each sample in X."""
 *         if not isinstance(X, np.ndarray):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_25);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indptr.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_values.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.weighted_decision_path", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indptr.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_values.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_X_ndarray);
  __Pyx_XDECREF((PyObject *)__pyx_v_indptr);
  __Pyx_XDECREF((PyObject *)__pyx_v_indices);
  __Pyx_XDECREF((PyObject *)__pyx_v_values);
  __Pyx_XDECREF(__pyx_v_out);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_17weighted_decision_path(PyObject *__pyx_v_self, PyObject *__pyx_v_X); /*proto*/
static char __pyx_doc_8skgarden_8mondrian_4tree_5_tree_4Tree_16weighted_decision_path[] = "Returns the weight at each node for each sample in X.";
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_17weighted_decision_path(PyObject *__pyx_v_self, PyObject *__pyx_v_X) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("weighted_decision_path (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_16weighted_decision_path(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self), ((PyObject *)__pyx_v_X));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_16weighted_decision_path(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_X) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("weighted_decision_path", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree_weighted_decision_path(__pyx_v_self, __pyx_v_X, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1057, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.weighted_decision_path", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":1147
 *         return out
 * 
 *     cdef inline object _decision_path_dense(self, object X):             # <<<<<<<<<<<<<<
 *         """Finds the decision path (=node) for each sample in X."""
 * 
 */

static PyObject *__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree__decision_path_dense(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_X) {
  PyArrayObject *__pyx_v_X_ndarray = 0;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *__pyx_v_X_ptr;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_X_sample_stride;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_X_fx_stride;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_n_samples;
  PyArrayObject *__pyx_v_indptr = 0;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t *__pyx_v_indptr_ptr;
  PyArrayObject *__pyx_v_indices = 0;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t *__pyx_v_indices_ptr;
  struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node *__pyx_v_node;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_i;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_v_curr_node_id;
  PyArrayObject *__pyx_v_data = 0;
  PyObject *__pyx_v_out = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_data;
  __Pyx_Buffer __pyx_pybuffer_data;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indptr;
  __Pyx_Buffer __pyx_pybuffer_indptr;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_6;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyArrayObject *__pyx_t_10 = NULL;
  PyArrayObject *__pyx_t_11 = NULL;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_12;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_13;
  long __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  Py_ssize_t __pyx_t_20;
  PyArrayObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  __Pyx_RefNannySetupContext("_decision_path_dense", 0);
  __pyx_pybuffer_indptr.pybuffer.buf = NULL;
  __pyx_pybuffer_indptr.refcount = 0;
  __pyx_pybuffernd_indptr.data = NULL;
  __pyx_pybuffernd_indptr.rcbuffer = &__pyx_pybuffer_indptr;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  __pyx_pybuffer_data.pybuffer.buf = NULL;
  __pyx_pybuffer_data.refcount = 0;
  __pyx_pybuffernd_data.data = NULL;
  __pyx_pybuffernd_data.rcbuffer = &__pyx_pybuffer_data;

  /* "skgarden/mondrian/tree/_tree.pyx":1151
 * 
 *         # Check input
 *         if not isinstance(X, np.ndarray):             # <<<<<<<<<<<<<<
 *             raise ValueError("X should be in np.ndarray format, got %s"
 *                              % type(X))
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_X, __pyx_ptype_5numpy_ndarray); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "skgarden/mondrian/tree/_tree.pyx":1153
 *         if not isinstance(X, np.ndarray):
 *             raise ValueError("X should be in np.ndarray format, got %s"
 *                              % type(X))             # <<<<<<<<<<<<<<
 * 
 *         if X.dtype != DTYPE:
 */
    __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_X_should_be_in_np_ndarray_format, ((PyObject *)Py_TYPE(__pyx_v_X))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "skgarden/mondrian/tree/_tree.pyx":1152
 *         # Check input
 *         if not isinstance(X, np.ndarray):
 *             raise ValueError("X should be in np.ndarray format, got %s"             # <<<<<<<<<<<<<<
 *                              % type(X))
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 1152, __pyx_L1_error)

    /* "skgarden/mondrian/tree/_tree.pyx":1151
 * 
 *         # Check input
 *         if not isinstance(X, np.ndarray):             # <<<<<<<<<<<<<<
 *             raise ValueError("X should be in np.ndarray format, got %s"
 *                              % type(X))
 */
  }

  /* "skgarden/mondrian/tree/_tree.pyx":1155
 *                              % type(X))
 * 
 *         if X.dtype != DTYPE:             # <<<<<<<<<<<<<<
 *             raise ValueError("X.dtype should be np.float32, got %s" % X.dtype)
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_dtype); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_DTYPE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = PyObject_RichCompare(__pyx_t_4, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1155, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1155, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(__pyx_t_2)) {

    /* "skgarden/mondrian/tree/_tree.pyx":1156
 * 
 *         if X.dtype != DTYPE:
 *             raise ValueError("X.dtype should be np.float32, got %s" % X.dtype)             # <<<<<<<<<<<<<<
 * 
 *         # Extract input
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_dtype); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_X_dtype_should_be_np_float32_got, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 1156, __pyx_L1_error)

    /* "skgarden/mondrian/tree/_tree.pyx":1155
 *                              % type(X))
 * 
 *         if X.dtype != DTYPE:             # <<<<<<<<<<<<<<
 *             raise ValueError("X.dtype should be np.float32, got %s" % X.dtype)
 * 
 */
  }

  /* "skgarden/mondrian/tree/_tree.pyx":1159
 * 
 *         # Extract input
 *         cdef np.ndarray X_ndarray = X             # <<<<<<<<<<<<<<
 *         cdef DTYPE_t* X_ptr = <DTYPE_t*> X_ndarray.data
 *         cdef SIZE_t X_sample_stride = <SIZE_t> X.strides[0] / <SIZE_t> X.itemsize
 */
  if (!(likely(((__pyx_v_X) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_X, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1159, __pyx_L1_error)
  __pyx_t_5 = __pyx_v_X;
  __Pyx_INCREF(__pyx_t_5);
  __pyx_v_X_ndarray = ((PyArrayObject *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1160
 *         # Extract input
 *         cdef np.ndarray X_ndarray = X
 *         cdef DTYPE_t* X_ptr = <DTYPE_t*> X_ndarray.data             # <<<<<<<<<<<<<<
 *         cdef SIZE_t X_sample_stride = <SIZE_t> X.strides[0] / <SIZE_t> X.itemsize
 *         cdef SIZE_t X_fx_stride = <SIZE_t> X.strides[1] / <SIZE_t> X.itemsize
 */
  __pyx_v_X_ptr = ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *)__pyx_v_X_ndarray->data);

  /* "skgarden/mondrian/tree/_tree.pyx":1161
 *         cdef np.ndarray X_ndarray = X
 *         cdef DTYPE_t* X_ptr = <DTYPE_t*> X_ndarray.data
 *         cdef SIZE_t X_sample_stride = <SIZE_t> X.strides[0] / <SIZE_t> X.itemsize             # <<<<<<<<<<<<<<
 *         cdef SIZE_t X_fx_stride = <SIZE_t> X.strides[1] / <SIZE_t> X.itemsize
 *         cdef SIZE_t n_samples = X.shape[0]
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_strides); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_6 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1161, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_itemsize); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_3); if (unlikely((__pyx_t_7 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1161, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_X_sample_stride = (((__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t)__pyx_t_6) / ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t)__pyx_t_7));

  /* "skgarden/mondrian/tree/_tree.pyx":1162
 *         cdef DTYPE_t* X_ptr = <DTYPE_t*> X_ndarray.data
 *         cdef SIZE_t X_sample_stride = <SIZE_t> X.strides[0] / <SIZE_t> X.itemsize
 *         cdef SIZE_t X_fx_stride = <SIZE_t> X.strides[1] / <SIZE_t> X.itemsize             # <<<<<<<<<<<<<<
 *         cdef SIZE_t n_samples = X.shape[0]
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_strides); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_3, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_7 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_5); if (unlikely((__pyx_t_7 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1162, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_itemsize); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_5); if (unlikely((__pyx_t_6 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1162, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_X_fx_stride = (((__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t)__pyx_t_7) / ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t)__pyx_t_6));

  /* "skgarden/mondrian/tree/_tree.pyx":1163
 *         cdef SIZE_t X_sample_stride = <SIZE_t> X.strides[0] / <SIZE_t> X.itemsize
 *         cdef SIZE_t X_fx_stride = <SIZE_t> X.strides[1] / <SIZE_t> X.itemsize
 *         cdef SIZE_t n_samples = X.shape[0]             # <<<<<<<<<<<<<<
 * 
 *         # Initialize output
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_X, __pyx_n_s_shape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_6 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1163, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_n_samples = __pyx_t_6;

  /* "skgarden/mondrian/tree/_tree.pyx":1166
 * 
 *         # Initialize output
 *         cdef np.ndarray[SIZE_t] indptr = np.zeros(n_samples + 1, dtype=np.intp)             # <<<<<<<<<<<<<<
 *         cdef SIZE_t* indptr_ptr = <SIZE_t*> indptr.data
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_long((__pyx_v_n_samples + 1)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_intp); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_9) < 0) __PYX_ERR(0, 1166, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_9) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_9, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1166, __pyx_L1_error)
  __pyx_t_10 = ((PyArrayObject *)__pyx_t_9);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indptr.rcbuffer->pybuffer, (PyObject*)__pyx_t_10, &__Pyx_TypeInfo_nn___pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_indptr = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_indptr.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 1166, __pyx_L1_error)
    } else {__pyx_pybuffernd_indptr.diminfo[0].strides = __pyx_pybuffernd_indptr.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indptr.diminfo[0].shape = __pyx_pybuffernd_indptr.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_10 = 0;
  __pyx_v_indptr = ((PyArrayObject *)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1167
 *         # Initialize output
 *         cdef np.ndarray[SIZE_t] indptr = np.zeros(n_samples + 1, dtype=np.intp)
 *         cdef SIZE_t* indptr_ptr = <SIZE_t*> indptr.data             # <<<<<<<<<<<<<<
 * 
 *         cdef np.ndarray[SIZE_t] indices = np.zeros(n_samples *
 */
  __pyx_v_indptr_ptr = ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t *)__pyx_v_indptr->data);

  /* "skgarden/mondrian/tree/_tree.pyx":1169
 *         cdef SIZE_t* indptr_ptr = <SIZE_t*> indptr.data
 * 
 *         cdef np.ndarray[SIZE_t] indices = np.zeros(n_samples *             # <<<<<<<<<<<<<<
 *                                                    (1 + self.max_depth),
 *                                                    dtype=np.intp)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_np); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1170
 * 
 *         cdef np.ndarray[SIZE_t] indices = np.zeros(n_samples *
 *                                                    (1 + self.max_depth),             # <<<<<<<<<<<<<<
 *                                                    dtype=np.intp)
 *         cdef SIZE_t* indices_ptr = <SIZE_t*> indices.data
 */
  __pyx_t_9 = __Pyx_PyInt_From_long((__pyx_v_n_samples * (1 + __pyx_v_self->max_depth))); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "skgarden/mondrian/tree/_tree.pyx":1169
 *         cdef SIZE_t* indptr_ptr = <SIZE_t*> indptr.data
 * 
 *         cdef np.ndarray[SIZE_t] indices = np.zeros(n_samples *             # <<<<<<<<<<<<<<
 *                                                    (1 + self.max_depth),
 *                                                    dtype=np.intp)
 */
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_9);
  __pyx_t_9 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1171
 *         cdef np.ndarray[SIZE_t] indices = np.zeros(n_samples *
 *                                                    (1 + self.max_depth),
 *                                                    dtype=np.intp)             # <<<<<<<<<<<<<<
 *         cdef SIZE_t* indices_ptr = <SIZE_t*> indices.data
 * 
 */
  __pyx_t_9 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_intp); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_dtype, __pyx_t_8) < 0) __PYX_ERR(0, 1171, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1169
 *         cdef SIZE_t* indptr_ptr = <SIZE_t*> indptr.data
 * 
 *         cdef np.ndarray[SIZE_t] indices = np.zeros(n_samples *             # <<<<<<<<<<<<<<
 *                                                    (1 + self.max_depth),
 *                                                    dtype=np.intp)
 */
  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, __pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (!(likely(((__pyx_t_8) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_8, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1169, __pyx_L1_error)
  __pyx_t_11 = ((PyArrayObject *)__pyx_t_8);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_t_11, &__Pyx_TypeInfo_nn___pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_indices = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_indices.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 1169, __pyx_L1_error)
    } else {__pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_11 = 0;
  __pyx_v_indices = ((PyArrayObject *)__pyx_t_8);
  __pyx_t_8 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1172
 *                                                    (1 + self.max_depth),
 *                                                    dtype=np.intp)
 *         cdef SIZE_t* indices_ptr = <SIZE_t*> indices.data             # <<<<<<<<<<<<<<
 * 
 *         # Initialize auxiliary data-structure
 */
  __pyx_v_indices_ptr = ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t *)__pyx_v_indices->data);

  /* "skgarden/mondrian/tree/_tree.pyx":1175
 * 
 *         # Initialize auxiliary data-structure
 *         cdef Node* node = NULL             # <<<<<<<<<<<<<<
 *         cdef SIZE_t i = 0
 *         cdef SIZE_t curr_node_id
 */
  __pyx_v_node = NULL;

  /* "skgarden/mondrian/tree/_tree.pyx":1176
 *         # Initialize auxiliary data-structure
 *         cdef Node* node = NULL
 *         cdef SIZE_t i = 0             # <<<<<<<<<<<<<<
 *         cdef SIZE_t curr_node_id
 * 
 */
  __pyx_v_i = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1179
 *         cdef SIZE_t curr_node_id
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             for i in range(n_samples):
 *                 curr_node_id = self.root
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "skgarden/mondrian/tree/_tree.pyx":1180
 * 
 *         with nogil:
 *             for i in range(n_samples):             # <<<<<<<<<<<<<<
 *                 curr_node_id = self.root
 *                 indptr_ptr[i + 1] = indptr_ptr[i]
 */
        __pyx_t_6 = __pyx_v_n_samples;
        __pyx_t_7 = __pyx_t_6;
        for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_7; __pyx_t_12+=1) {
          __pyx_v_i = __pyx_t_12;

          /* "skgarden/mondrian/tree/_tree.pyx":1181
 *         with nogil:
 *             for i in range(n_samples):
 *                 curr_node_id = self.root             # <<<<<<<<<<<<<<
 *                 indptr_ptr[i + 1] = indptr_ptr[i]
 * 
 */
          __pyx_t_13 = __pyx_v_self->root;
          __pyx_v_curr_node_id = __pyx_t_13;

          /* "skgarden/mondrian/tree/_tree.pyx":1182
 *             for i in range(n_samples):
 *                 curr_node_id = self.root
 *                 indptr_ptr[i + 1] = indptr_ptr[i]             # <<<<<<<<<<<<<<
 * 
 *                 # Add all external nodes
 */
          (__pyx_v_indptr_ptr[(__pyx_v_i + 1)]) = (__pyx_v_indptr_ptr[__pyx_v_i]);

          /* "skgarden/mondrian/tree/_tree.pyx":1185
 * 
 *                 # Add all external nodes
 *                 while curr_node_id != _TREE_LEAF:             # <<<<<<<<<<<<<<
 *                     node = &self.nodes[curr_node_id]
 * 
 */
          while (1) {
            __pyx_t_2 = ((__pyx_v_curr_node_id != __pyx_v_8skgarden_8mondrian_4tree_5_tree__TREE_LEAF) != 0);
            if (!__pyx_t_2) break;

            /* "skgarden/mondrian/tree/_tree.pyx":1186
 *                 # Add all external nodes
 *                 while curr_node_id != _TREE_LEAF:
 *                     node = &self.nodes[curr_node_id]             # <<<<<<<<<<<<<<
 * 
 *                     # ... and node.right_child != _TREE_LEAF:
 */
            __pyx_v_node = (&(__pyx_v_self->nodes[__pyx_v_curr_node_id]));

            /* "skgarden/mondrian/tree/_tree.pyx":1189
 * 
 *                     # ... and node.right_child != _TREE_LEAF:
 *                     indices_ptr[indptr_ptr[i + 1]] = <SIZE_t>(node - self.nodes)             # <<<<<<<<<<<<<<
 *                     indptr_ptr[i + 1] += 1
 * 
 */
            (__pyx_v_indices_ptr[(__pyx_v_indptr_ptr[(__pyx_v_i + 1)])]) = ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t)(__pyx_v_node - __pyx_v_self->nodes));

            /* "skgarden/mondrian/tree/_tree.pyx":1190
 *                     # ... and node.right_child != _TREE_LEAF:
 *                     indices_ptr[indptr_ptr[i + 1]] = <SIZE_t>(node - self.nodes)
 *                     indptr_ptr[i + 1] += 1             # <<<<<<<<<<<<<<
 * 
 *                     if X_ptr[X_sample_stride * i +
 */
            __pyx_t_14 = (__pyx_v_i + 1);
            (__pyx_v_indptr_ptr[__pyx_t_14]) = ((__pyx_v_indptr_ptr[__pyx_t_14]) + 1);

            /* "skgarden/mondrian/tree/_tree.pyx":1193
 * 
 *                     if X_ptr[X_sample_stride * i +
 *                              X_fx_stride * node.feature] <= node.threshold:             # <<<<<<<<<<<<<<
 *                         curr_node_id = node.left_child
 *                     else:
 */
            __pyx_t_2 = (((__pyx_v_X_ptr[((__pyx_v_X_sample_stride * __pyx_v_i) + (__pyx_v_X_fx_stride * __pyx_v_node->feature))]) <= __pyx_v_node->threshold) != 0);

            /* "skgarden/mondrian/tree/_tree.pyx":1192
 *                     indptr_ptr[i + 1] += 1
 * 
 *                     if X_ptr[X_sample_stride * i +             # <<<<<<<<<<<<<<
 *                              X_fx_stride * node.feature] <= node.threshold:
 *                         curr_node_id = node.left_child
 */
            if (__pyx_t_2) {

              /* "skgarden/mondrian/tree/_tree.pyx":1194
 *                     if X_ptr[X_sample_stride * i +
 *                              X_fx_stride * node.feature] <= node.threshold:
 *                         curr_node_id = node.left_child             # <<<<<<<<<<<<<<
 *                     else:
 *                         curr_node_id = node.right_child
 */
              __pyx_t_13 = __pyx_v_node->left_child;
              __pyx_v_curr_node_id = __pyx_t_13;

              /* "skgarden/mondrian/tree/_tree.pyx":1192
 *                     indptr_ptr[i + 1] += 1
 * 
 *                     if X_ptr[X_sample_stride * i +             # <<<<<<<<<<<<<<
 *                              X_fx_stride * node.feature] <= node.threshold:
 *                         curr_node_id = node.left_child
 */
              goto __pyx_L12;
            }

            /* "skgarden/mondrian/tree/_tree.pyx":1196
 *                         curr_node_id = node.left_child
 *                     else:
 *                         curr_node_id = node.right_child             # <<<<<<<<<<<<<<
 * 
 *         indices = indices[:indptr[n_samples]]
 */
            /*else*/ {
              __pyx_t_13 = __pyx_v_node->right_child;
              __pyx_v_curr_node_id = __pyx_t_13;
            }
            __pyx_L12:;
          }
        }
      }

      /* "skgarden/mondrian/tree/_tree.pyx":1179
 *         cdef SIZE_t curr_node_id
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             for i in range(n_samples):
 *                 curr_node_id = self.root
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L7;
        }
        __pyx_L7:;
      }
  }

  /* "skgarden/mondrian/tree/_tree.pyx":1198
 *                         curr_node_id = node.right_child
 * 
 *         indices = indices[:indptr[n_samples]]             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[SIZE_t] data = np.ones(shape=len(indices),
 *                                                dtype=np.intp)
 */
  __pyx_t_15 = __pyx_v_n_samples;
  __pyx_t_8 = __Pyx_PyInt_From_Py_intptr_t((*__Pyx_BufPtrStrided1d(__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t *, __pyx_pybuffernd_indptr.rcbuffer->pybuffer.buf, __pyx_t_15, __pyx_pybuffernd_indptr.diminfo[0].strides))); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PySlice_New(Py_None, __pyx_t_8, Py_None); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_indices), __pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (!(likely(((__pyx_t_8) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_8, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1198, __pyx_L1_error)
  __pyx_t_11 = ((PyArrayObject *)__pyx_t_8);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __pyx_t_16 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_t_11, &__Pyx_TypeInfo_nn___pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_16 < 0)) {
      PyErr_Fetch(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_17); Py_XDECREF(__pyx_t_18); Py_XDECREF(__pyx_t_19);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_17, __pyx_t_18, __pyx_t_19);
      }
      __pyx_t_17 = __pyx_t_18 = __pyx_t_19 = 0;
    }
    __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_16 < 0)) __PYX_ERR(0, 1198, __pyx_L1_error)
  }
  __pyx_t_11 = 0;
  __Pyx_DECREF_SET(__pyx_v_indices, ((PyArrayObject *)__pyx_t_8));
  __pyx_t_8 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1199
 * 
 *         indices = indices[:indptr[n_samples]]
 *         cdef np.ndarray[SIZE_t] data = np.ones(shape=len(indices),             # <<<<<<<<<<<<<<
 *                                                dtype=np.intp)
 *         out = csr_matrix((data, indices, indptr),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_ones); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_20 = PyObject_Length(((PyObject *)__pyx_v_indices)); if (unlikely(__pyx_t_20 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1199, __pyx_L1_error)
  __pyx_t_4 = PyInt_FromSsize_t(__pyx_t_20); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_shape, __pyx_t_4) < 0) __PYX_ERR(0, 1199, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1200
 *         indices = indices[:indptr[n_samples]]
 *         cdef np.ndarray[SIZE_t] data = np.ones(shape=len(indices),
 *                                                dtype=np.intp)             # <<<<<<<<<<<<<<
 *         out = csr_matrix((data, indices, indptr),
 *                          shape=(n_samples, self.node_count))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_intp); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_dtype, __pyx_t_3) < 0) __PYX_ERR(0, 1199, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1199
 * 
 *         indices = indices[:indptr[n_samples]]
 *         cdef np.ndarray[SIZE_t] data = np.ones(shape=len(indices),             # <<<<<<<<<<<<<<
 *                                                dtype=np.intp)
 *         out = csr_matrix((data, indices, indptr),
 */
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_empty_tuple, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1199, __pyx_L1_error)
  __pyx_t_21 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_data.rcbuffer->pybuffer, (PyObject*)__pyx_t_21, &__Pyx_TypeInfo_nn___pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_data = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_data.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 1199, __pyx_L1_error)
    } else {__pyx_pybuffernd_data.diminfo[0].strides = __pyx_pybuffernd_data.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_data.diminfo[0].shape = __pyx_pybuffernd_data.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_21 = 0;
  __pyx_v_data = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1201
 *         cdef np.ndarray[SIZE_t] data = np.ones(shape=len(indices),
 *                                                dtype=np.intp)
 *         out = csr_matrix((data, indices, indptr),             # <<<<<<<<<<<<<<
 *                          shape=(n_samples, self.node_count))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_csr_matrix); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(((PyObject *)__pyx_v_data));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_data));
  PyTuple_SET_ITEM(__pyx_t_8, 0, ((PyObject *)__pyx_v_data));
  __Pyx_INCREF(((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_indices));
  PyTuple_SET_ITEM(__pyx_t_8, 1, ((PyObject *)__pyx_v_indices));
  __Pyx_INCREF(((PyObject *)__pyx_v_indptr));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_indptr));
  PyTuple_SET_ITEM(__pyx_t_8, 2, ((PyObject *)__pyx_v_indptr));
  __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_8);
  __pyx_t_8 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1202
 *                                                dtype=np.intp)
 *         out = csr_matrix((data, indices, indptr),
 *                          shape=(n_samples, self.node_count))             # <<<<<<<<<<<<<<
 * 
 *         return out
 */
  __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_4 = __Pyx_PyInt_From_Py_intptr_t(__pyx_v_n_samples); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_Py_intptr_t(__pyx_v_self->node_count); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_t_5);
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_shape, __pyx_t_22) < 0) __PYX_ERR(0, 1202, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1201
 *         cdef np.ndarray[SIZE_t] data = np.ones(shape=len(indices),
 *                                                dtype=np.intp)
 *         out = csr_matrix((data, indices, indptr),             # <<<<<<<<<<<<<<
 *                          shape=(n_samples, self.node_count))
 * 
 */
  __pyx_t_22 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_9, __pyx_t_8); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_v_out = __pyx_t_22;
  __pyx_t_22 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1204
 *                          shape=(n_samples, self.node_count))
 * 
 *         return out             # <<<<<<<<<<<<<<
 * 
 *     cdef np.ndarray _get_value_ndarray(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_out);
  __pyx_r = __pyx_v_out;
  goto __pyx_L0;

  /* "skgarden/mondrian/tree/_tree.pyx":1147
 *         return out
 * 
 *     cdef inline object _decision_path_dense(self, object X):             # <<<<<<<<<<<<<<
 *         """Finds the decision path (=node) for each sample in X."""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_22);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_data.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indptr.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree._decision_path_dense", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_data.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indptr.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_X_ndarray);
  __Pyx_XDECREF((PyObject *)__pyx_v_indptr);
  __Pyx_XDECREF((PyObject *)__pyx_v_indices);
  __Pyx_XDECREF((PyObject *)__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_out);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":1206
 *         return out
 * 
 *     cdef np.ndarray _get_value_ndarray(self):             # <<<<<<<<<<<<<<
 *         """Wraps value as a 3-d NumPy array.
 * 
 */

static PyArrayObject *__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree__get_value_ndarray(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self) {
  npy_intp __pyx_v_shape[3];
  PyArrayObject *__pyx_v_arr = 0;
  PyArrayObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_get_value_ndarray", 0);

  /* "skgarden/mondrian/tree/_tree.pyx":1213
 *         """
 *         cdef np.npy_intp shape[3]
 *         shape[0] = <np.npy_intp> self.node_count             # <<<<<<<<<<<<<<
 *         shape[1] = <np.npy_intp> self.n_outputs
 *         shape[2] = <np.npy_intp> self.max_n_classes
 */
  (__pyx_v_shape[0]) = ((npy_intp)__pyx_v_self->node_count);

  /* "skgarden/mondrian/tree/_tree.pyx":1214
 *         cdef np.npy_intp shape[3]
 *         shape[0] = <np.npy_intp> self.node_count
 *         shape[1] = <np.npy_intp> self.n_outputs             # <<<<<<<<<<<<<<
 *         shape[2] = <np.npy_intp> self.max_n_classes
 *         cdef np.ndarray arr
 */
  (__pyx_v_shape[1]) = ((npy_intp)__pyx_v_self->n_outputs);

  /* "skgarden/mondrian/tree/_tree.pyx":1215
 *         shape[0] = <np.npy_intp> self.node_count
 *         shape[1] = <np.npy_intp> self.n_outputs
 *         shape[2] = <np.npy_intp> self.max_n_classes             # <<<<<<<<<<<<<<
 *         cdef np.ndarray arr
 *         arr = np.PyArray_SimpleNewFromData(3, shape, np.NPY_DOUBLE, self.value)
 */
  (__pyx_v_shape[2]) = ((npy_intp)__pyx_v_self->max_n_classes);

  /* "skgarden/mondrian/tree/_tree.pyx":1217
 *         shape[2] = <np.npy_intp> self.max_n_classes
 *         cdef np.ndarray arr
 *         arr = np.PyArray_SimpleNewFromData(3, shape, np.NPY_DOUBLE, self.value)             # <<<<<<<<<<<<<<
 *         Py_INCREF(self)
 *         arr.base = <PyObject*> self
 */
  __pyx_t_1 = PyArray_SimpleNewFromData(3, __pyx_v_shape, NPY_DOUBLE, __pyx_v_self->value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1217, __pyx_L1_error)
  __pyx_v_arr = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1218
 *         cdef np.ndarray arr
 *         arr = np.PyArray_SimpleNewFromData(3, shape, np.NPY_DOUBLE, self.value)
 *         Py_INCREF(self)             # <<<<<<<<<<<<<<
 *         arr.base = <PyObject*> self
 *         return arr
 */
  Py_INCREF(((PyObject *)__pyx_v_self));

  /* "skgarden/mondrian/tree/_tree.pyx":1219
 *         arr = np.PyArray_SimpleNewFromData(3, shape, np.NPY_DOUBLE, self.value)
 *         Py_INCREF(self)
 *         arr.base = <PyObject*> self             # <<<<<<<<<<<<<<
 *         return arr
 * 
 */
  __pyx_v_arr->base = ((PyObject *)__pyx_v_self);

  /* "skgarden/mondrian/tree/_tree.pyx":1220
 *         Py_INCREF(self)
 *         arr.base = <PyObject*> self
 *         return arr             # <<<<<<<<<<<<<<
 * 
 *     cdef np.ndarray _get_node_ndarray(self):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_arr));
  __pyx_r = __pyx_v_arr;
  goto __pyx_L0;

  /* "skgarden/mondrian/tree/_tree.pyx":1206
 *         return out
 * 
 *     cdef np.ndarray _get_value_ndarray(self):             # <<<<<<<<<<<<<<
 *         """Wraps value as a 3-d NumPy array.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree._get_value_ndarray", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_arr);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pyx":1222
 *         return arr
 * 
 *     cdef np.ndarray _get_node_ndarray(self):             # <<<<<<<<<<<<<<
 *         """Wraps nodes as a NumPy struct array.
 * 
 */

static PyArrayObject *__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree__get_node_ndarray(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self) {
  npy_intp __pyx_v_shape[1];
  npy_intp __pyx_v_strides[1];
  PyArrayObject *__pyx_v_arr = 0;
  PyArrayObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("_get_node_ndarray", 0);

  /* "skgarden/mondrian/tree/_tree.pyx":1230
 *         """
 *         cdef np.npy_intp shape[1]
 *         shape[0] = <np.npy_intp> self.node_count             # <<<<<<<<<<<<<<
 *         cdef np.npy_intp strides[1]
 *         strides[0] = sizeof(Node)
 */
  (__pyx_v_shape[0]) = ((npy_intp)__pyx_v_self->node_count);

  /* "skgarden/mondrian/tree/_tree.pyx":1232
 *         shape[0] = <np.npy_intp> self.node_count
 *         cdef np.npy_intp strides[1]
 *         strides[0] = sizeof(Node)             # <<<<<<<<<<<<<<
 *         cdef np.ndarray arr
 *         Py_INCREF(NODE_DTYPE)
 */
  (__pyx_v_strides[0]) = (sizeof(struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node));

  /* "skgarden/mondrian/tree/_tree.pyx":1234
 *         strides[0] = sizeof(Node)
 *         cdef np.ndarray arr
 *         Py_INCREF(NODE_DTYPE)             # <<<<<<<<<<<<<<
 *         arr = PyArray_NewFromDescr(np.ndarray, <np.dtype> NODE_DTYPE, 1, shape,
 *                                    strides, <void*> self.nodes,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_NODE_DTYPE); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  Py_INCREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1235
 *         cdef np.ndarray arr
 *         Py_INCREF(NODE_DTYPE)
 *         arr = PyArray_NewFromDescr(np.ndarray, <np.dtype> NODE_DTYPE, 1, shape,             # <<<<<<<<<<<<<<
 *                                    strides, <void*> self.nodes,
 *                                    np.NPY_DEFAULT, None)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_NODE_DTYPE); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "skgarden/mondrian/tree/_tree.pyx":1237
 *         arr = PyArray_NewFromDescr(np.ndarray, <np.dtype> NODE_DTYPE, 1, shape,
 *                                    strides, <void*> self.nodes,
 *                                    np.NPY_DEFAULT, None)             # <<<<<<<<<<<<<<
 *         Py_INCREF(self)
 *         arr.base = <PyObject*> self
 */
  __pyx_t_2 = PyArray_NewFromDescr(((PyObject *)__pyx_ptype_5numpy_ndarray), ((PyArray_Descr *)__pyx_t_1), 1, __pyx_v_shape, __pyx_v_strides, ((void *)__pyx_v_self->nodes), NPY_DEFAULT, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1235
 *         cdef np.ndarray arr
 *         Py_INCREF(NODE_DTYPE)
 *         arr = PyArray_NewFromDescr(np.ndarray, <np.dtype> NODE_DTYPE, 1, shape,             # <<<<<<<<<<<<<<
 *                                    strides, <void*> self.nodes,
 *                                    np.NPY_DEFAULT, None)
 */
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1235, __pyx_L1_error)
  __pyx_v_arr = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1238
 *                                    strides, <void*> self.nodes,
 *                                    np.NPY_DEFAULT, None)
 *         Py_INCREF(self)             # <<<<<<<<<<<<<<
 *         arr.base = <PyObject*> self
 *         return arr
 */
  Py_INCREF(((PyObject *)__pyx_v_self));

  /* "skgarden/mondrian/tree/_tree.pyx":1239
 *                                    np.NPY_DEFAULT, None)
 *         Py_INCREF(self)
 *         arr.base = <PyObject*> self             # <<<<<<<<<<<<<<
 *         return arr
 */
  __pyx_v_arr->base = ((PyObject *)__pyx_v_self);

  /* "skgarden/mondrian/tree/_tree.pyx":1240
 *         Py_INCREF(self)
 *         arr.base = <PyObject*> self
 *         return arr             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_arr));
  __pyx_r = __pyx_v_arr;
  goto __pyx_L0;

  /* "skgarden/mondrian/tree/_tree.pyx":1222
 *         return arr
 * 
 *     cdef np.ndarray _get_node_ndarray(self):             # <<<<<<<<<<<<<<
 *         """Wraps nodes as a NumPy struct array.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree._get_node_ndarray", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_arr);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pxd":47
 * 
 *     # Input/Output layout
 *     cdef public SIZE_t n_features        # Number of features in X             # <<<<<<<<<<<<<<
 *     cdef SIZE_t* n_classes               # Number of classes in y[:, k]
 *     cdef public SIZE_t n_outputs         # Number of outputs in y
 */

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_10n_features_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_10n_features_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_10n_features___get__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_10n_features___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_Py_intptr_t(__pyx_v_self->n_features); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.n_features.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_10n_features_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_10n_features_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_10n_features_2__set__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_10n_features_2__set__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_Py_intptr_t(__pyx_v_value); if (unlikely((__pyx_t_1 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(2, 47, __pyx_L1_error)
  __pyx_v_self->n_features = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.n_features.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pxd":49
 *     cdef public SIZE_t n_features        # Number of features in X
 *     cdef SIZE_t* n_classes               # Number of classes in y[:, k]
 *     cdef public SIZE_t n_outputs         # Number of outputs in y             # <<<<<<<<<<<<<<
 *     cdef public SIZE_t max_n_classes     # max(n_classes)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_9n_outputs_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_9n_outputs_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_9n_outputs___get__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_9n_outputs___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_Py_intptr_t(__pyx_v_self->n_outputs); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.n_outputs.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_9n_outputs_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_9n_outputs_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_9n_outputs_2__set__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_9n_outputs_2__set__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_Py_intptr_t(__pyx_v_value); if (unlikely((__pyx_t_1 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(2, 49, __pyx_L1_error)
  __pyx_v_self->n_outputs = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.n_outputs.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pxd":50
 *     cdef SIZE_t* n_classes               # Number of classes in y[:, k]
 *     cdef public SIZE_t n_outputs         # Number of outputs in y
 *     cdef public SIZE_t max_n_classes     # max(n_classes)             # <<<<<<<<<<<<<<
 * 
 *     # Inner structures: values are stored separately from node structure,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_13max_n_classes_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_13max_n_classes_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_13max_n_classes___get__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_13max_n_classes___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_Py_intptr_t(__pyx_v_self->max_n_classes); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.max_n_classes.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_13max_n_classes_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_13max_n_classes_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_13max_n_classes_2__set__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_13max_n_classes_2__set__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_Py_intptr_t(__pyx_v_value); if (unlikely((__pyx_t_1 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(2, 50, __pyx_L1_error)
  __pyx_v_self->max_n_classes = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.max_n_classes.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pxd":54
 *     # Inner structures: values are stored separately from node structure,
 *     # since size is determined at runtime.
 *     cdef public SIZE_t max_depth         # Max depth of the tree             # <<<<<<<<<<<<<<
 *     cdef public SIZE_t node_count        # Counter for node IDs
 *     cdef public SIZE_t capacity          # Capacity of tree, in terms of nodes
 */

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_9max_depth_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_9max_depth_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_9max_depth___get__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_9max_depth___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_Py_intptr_t(__pyx_v_self->max_depth); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.max_depth.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_9max_depth_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_9max_depth_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_9max_depth_2__set__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_9max_depth_2__set__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_Py_intptr_t(__pyx_v_value); if (unlikely((__pyx_t_1 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(2, 54, __pyx_L1_error)
  __pyx_v_self->max_depth = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.max_depth.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pxd":55
 *     # since size is determined at runtime.
 *     cdef public SIZE_t max_depth         # Max depth of the tree
 *     cdef public SIZE_t node_count        # Counter for node IDs             # <<<<<<<<<<<<<<
 *     cdef public SIZE_t capacity          # Capacity of tree, in terms of nodes
 *     cdef Node* nodes                     # Array of nodes
 */

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_10node_count_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_10node_count_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_10node_count___get__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_10node_count___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_Py_intptr_t(__pyx_v_self->node_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.node_count.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_10node_count_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_10node_count_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_10node_count_2__set__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_10node_count_2__set__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_Py_intptr_t(__pyx_v_value); if (unlikely((__pyx_t_1 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(2, 55, __pyx_L1_error)
  __pyx_v_self->node_count = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.node_count.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "skgarden/mondrian/tree/_tree.pxd":56
 *     cdef public SIZE_t max_depth         # Max depth of the tree
 *     cdef public SIZE_t node_count        # Counter for node IDs
 *     cdef public SIZE_t capacity          # Capacity of tree, in terms of nodes             # <<<<<<<<<<<<<<
 *     cdef Node* nodes                     # Array of nodes
 *     cdef double* value                   # (capacity, n_outputs, max_n_classes) array of values
 */

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_8capacity_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_8capacity_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_8capacity___get__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_8capacity___get__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_Py_intptr_t(__pyx_v_self->capacity); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.capacity.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_8capacity_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_8capacity_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_8capacity_2__set__(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8skgarden_8mondrian_4tree_5_tree_4Tree_8capacity_2__set__(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_Py_intptr_t(__pyx_v_value); if (unlikely((__pyx_t_1 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(2, 56, __pyx_L1_error)
  __pyx_v_self->capacity = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.Tree.capacity.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_TreeBuilder(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     if __pyx_checksum != 0xedae8f4:
 *         from pickle import PickleError as __pyx_PickleError
 */

/* Python wrapper */
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_1__pyx_unpickle_TreeBuilder(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_8skgarden_8mondrian_4tree_5_tree_1__pyx_unpickle_TreeBuilder = {"__pyx_unpickle_TreeBuilder", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8skgarden_8mondrian_4tree_5_tree_1__pyx_unpickle_TreeBuilder, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8skgarden_8mondrian_4tree_5_tree_1__pyx_unpickle_TreeBuilder(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_TreeBuilder (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_TreeBuilder", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_TreeBuilder", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_TreeBuilder") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_TreeBuilder", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.__pyx_unpickle_TreeBuilder", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8skgarden_8mondrian_4tree_5_tree___pyx_unpickle_TreeBuilder(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8skgarden_8mondrian_4tree_5_tree___pyx_unpickle_TreeBuilder(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = NULL;
  PyObject *__pyx_v___pyx_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_TreeBuilder", 0);

  /* "(tree fragment)":2
 * def __pyx_unpickle_TreeBuilder(__pyx_type, long __pyx_checksum, __pyx_state):
 *     if __pyx_checksum != 0xedae8f4:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xedae8f4 = (max_depth, min_samples_leaf, min_samples_split, min_weight_leaf, random_state, splitter))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xedae8f4) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":3
 * def __pyx_unpickle_TreeBuilder(__pyx_type, long __pyx_checksum, __pyx_state):
 *     if __pyx_checksum != 0xedae8f4:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xedae8f4 = (max_depth, min_samples_leaf, min_samples_split, min_weight_leaf, random_state, splitter))" % __pyx_checksum)
 *     __pyx_result = TreeBuilder.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":4
 *     if __pyx_checksum != 0xedae8f4:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xedae8f4 = (max_depth, min_samples_leaf, min_samples_split, min_weight_leaf, random_state, splitter))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = TreeBuilder.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 4, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xed, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 4, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 4, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 4, __pyx_L1_error)

    /* "(tree fragment)":2
 * def __pyx_unpickle_TreeBuilder(__pyx_type, long __pyx_checksum, __pyx_state):
 *     if __pyx_checksum != 0xedae8f4:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xedae8f4 = (max_depth, min_samples_leaf, min_samples_split, min_weight_leaf, random_state, splitter))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":5
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xedae8f4 = (max_depth, min_samples_leaf, min_samples_split, min_weight_leaf, random_state, splitter))" % __pyx_checksum)
 *     __pyx_result = TreeBuilder.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_TreeBuilder__set_state(<TreeBuilder> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_8skgarden_8mondrian_4tree_5_tree_TreeBuilder), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xedae8f4 = (max_depth, min_samples_leaf, min_samples_split, min_weight_leaf, random_state, splitter))" % __pyx_checksum)
 *     __pyx_result = TreeBuilder.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_TreeBuilder__set_state(<TreeBuilder> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":7
 *     __pyx_result = TreeBuilder.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_TreeBuilder__set_state(<TreeBuilder> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_TreeBuilder__set_state(TreeBuilder __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 7, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_8skgarden_8mondrian_4tree_5_tree___pyx_unpickle_TreeBuilder__set_state(((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xedae8f4 = (max_depth, min_samples_leaf, min_samples_split, min_weight_leaf, random_state, splitter))" % __pyx_checksum)
 *     __pyx_result = TreeBuilder.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_TreeBuilder__set_state(<TreeBuilder> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":8
 *     if __pyx_state is not None:
 *         __pyx_unpickle_TreeBuilder__set_state(<TreeBuilder> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_TreeBuilder__set_state(TreeBuilder __pyx_result, tuple __pyx_state):
 *     __pyx_result.max_depth = __pyx_state[0]; __pyx_result.min_samples_leaf = __pyx_state[1]; __pyx_result.min_samples_split = __pyx_state[2]; __pyx_result.min_weight_leaf = __pyx_state[3]; __pyx_result.random_state = __pyx_state[4]; __pyx_result.splitter = __pyx_state[5]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_TreeBuilder(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     if __pyx_checksum != 0xedae8f4:
 *         from pickle import PickleError as __pyx_PickleError
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.__pyx_unpickle_TreeBuilder", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":9
 *         __pyx_unpickle_TreeBuilder__set_state(<TreeBuilder> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_TreeBuilder__set_state(TreeBuilder __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.max_depth = __pyx_state[0]; __pyx_result.min_samples_leaf = __pyx_state[1]; __pyx_result.min_samples_split = __pyx_state[2]; __pyx_result.min_weight_leaf = __pyx_state[3]; __pyx_result.random_state = __pyx_state[4]; __pyx_result.splitter = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_8skgarden_8mondrian_4tree_5_tree___pyx_unpickle_TreeBuilder__set_state(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_1;
  double __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_TreeBuilder__set_state", 0);

  /* "(tree fragment)":10
 *     return __pyx_result
 * cdef __pyx_unpickle_TreeBuilder__set_state(TreeBuilder __pyx_result, tuple __pyx_state):
 *     __pyx_result.max_depth = __pyx_state[0]; __pyx_result.min_samples_leaf = __pyx_state[1]; __pyx_result.min_samples_split = __pyx_state[2]; __pyx_result.min_weight_leaf = __pyx_state[3]; __pyx_result.random_state = __pyx_state[4]; __pyx_result.splitter = __pyx_state[5]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 10, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_As_Py_intptr_t(PyTuple_GET_ITEM(__pyx_v___pyx_state, 0)); if (unlikely((__pyx_t_1 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(1, 10, __pyx_L1_error)
  __pyx_v___pyx_result->max_depth = __pyx_t_1;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 10, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_As_Py_intptr_t(PyTuple_GET_ITEM(__pyx_v___pyx_state, 1)); if (unlikely((__pyx_t_1 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(1, 10, __pyx_L1_error)
  __pyx_v___pyx_result->min_samples_leaf = __pyx_t_1;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 10, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_As_Py_intptr_t(PyTuple_GET_ITEM(__pyx_v___pyx_state, 2)); if (unlikely((__pyx_t_1 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(1, 10, __pyx_L1_error)
  __pyx_v___pyx_result->min_samples_split = __pyx_t_1;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 10, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_PyFloat_AsDouble(PyTuple_GET_ITEM(__pyx_v___pyx_state, 3)); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 10, __pyx_L1_error)
  __pyx_v___pyx_result->min_weight_leaf = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 10, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 4);
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v___pyx_result->random_state);
  __Pyx_DECREF(__pyx_v___pyx_result->random_state);
  __pyx_v___pyx_result->random_state = __pyx_t_3;
  __pyx_t_3 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 10, __pyx_L1_error)
  }
  if (!(likely(((PyTuple_GET_ITEM(__pyx_v___pyx_state, 5)) == Py_None) || likely(__Pyx_TypeTest(PyTuple_GET_ITEM(__pyx_v___pyx_state, 5), __pyx_ptype_8skgarden_8mondrian_4tree_9_splitter_Splitter))))) __PYX_ERR(1, 10, __pyx_L1_error)
  __pyx_t_3 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 5);
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v___pyx_result->splitter);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->splitter));
  __pyx_v___pyx_result->splitter = ((struct __pyx_obj_8skgarden_8mondrian_4tree_9_splitter_Splitter *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":11
 * cdef __pyx_unpickle_TreeBuilder__set_state(TreeBuilder __pyx_result, tuple __pyx_state):
 *     __pyx_result.max_depth = __pyx_state[0]; __pyx_result.min_samples_leaf = __pyx_state[1]; __pyx_result.min_samples_split = __pyx_state[2]; __pyx_result.min_weight_leaf = __pyx_state[3]; __pyx_result.random_state = __pyx_state[4]; __pyx_result.splitter = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 11, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 11, __pyx_L1_error)
  __pyx_t_6 = ((__pyx_t_5 > 6) != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 11, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 != 0);
  __pyx_t_4 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":12
 *     __pyx_result.max_depth = __pyx_state[0]; __pyx_result.min_samples_leaf = __pyx_state[1]; __pyx_result.min_samples_split = __pyx_state[2]; __pyx_result.min_weight_leaf = __pyx_state[3]; __pyx_result.random_state = __pyx_state[4]; __pyx_result.splitter = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[6])             # <<<<<<<<<<<<<<
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 12, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 12, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 12, __pyx_L1_error)
    }
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_3 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_8, PyTuple_GET_ITEM(__pyx_v___pyx_state, 6)) : __Pyx_PyObject_CallOneArg(__pyx_t_9, PyTuple_GET_ITEM(__pyx_v___pyx_state, 6));
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 12, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":11
 * cdef __pyx_unpickle_TreeBuilder__set_state(TreeBuilder __pyx_result, tuple __pyx_state):
 *     __pyx_result.max_depth = __pyx_state[0]; __pyx_result.min_samples_leaf = __pyx_state[1]; __pyx_result.min_samples_split = __pyx_state[2]; __pyx_result.min_weight_leaf = __pyx_state[3]; __pyx_result.random_state = __pyx_state[4]; __pyx_result.splitter = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  }

  /* "(tree fragment)":9
 *         __pyx_unpickle_TreeBuilder__set_state(<TreeBuilder> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_TreeBuilder__set_state(TreeBuilder __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.max_depth = __pyx_state[0]; __pyx_result.min_samples_leaf = __pyx_state[1]; __pyx_result.min_samples_split = __pyx_state[2]; __pyx_result.min_weight_leaf = __pyx_state[3]; __pyx_result.random_state = __pyx_state[4]; __pyx_result.splitter = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("skgarden.mondrian.tree._tree.__pyx_unpickle_TreeBuilder__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":258
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5numpy_7ndarray___getbuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_i;
  int __pyx_v_ndim;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  int __pyx_v_t;
  char *__pyx_v_f;
  PyArray_Descr *__pyx_v_descr = 0;
  int __pyx_v_offset;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyArray_Descr *__pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  char *__pyx_t_9;
  if (__pyx_v_info == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":265
 * 
 *             cdef int i, ndim
 *             cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 */
  __pyx_v_endian_detector = 1;

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":266
 *             cdef int i, ndim
 *             cdef int endian_detector = 1
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 * 
 *             ndim = PyArray_NDIM(self)
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":268
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 *             ndim = PyArray_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 */
  __pyx_v_ndim = PyArray_NDIM(__pyx_v_self);

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  __pyx_t_2 = (((__pyx_v_flags & PyBUF_C_CONTIGUOUS) == PyBUF_C_CONTIGUOUS) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":271
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 */
  __pyx_t_2 = ((!(PyArray_CHKFLAGS(__pyx_v_self, NPY_ARRAY_C_CONTIGUOUS) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  if (unlikely(__pyx_t_1)) {

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":272
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(3, 272, __pyx_L1_error)

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  }

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  __pyx_t_2 = (((__pyx_v_flags & PyBUF_F_CONTIGUOUS) == PyBUF_F_CONTIGUOUS) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":275
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 */
  __pyx_t_2 = ((!(PyArray_CHKFLAGS(__pyx_v_self, NPY_ARRAY_F_CONTIGUOUS) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L7_bool_binop_done:;

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  if (unlikely(__pyx_t_1)) {

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":276
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(3, 276, __pyx_L1_error)

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  }

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":278
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 *             info.buf = PyArray_DATA(self)             # <<<<<<<<<<<<<<
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_v_info->buf = PyArray_DATA(__pyx_v_self);

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":279
 * 
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim             # <<<<<<<<<<<<<<
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 # Allocate new buffer for strides and shape info.
 */
  __pyx_v_info->ndim = __pyx_v_ndim;

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":280
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
  __pyx_t_1 = (((sizeof(npy_intp)) != (sizeof(Py_ssize_t))) != 0);
  if (__pyx_t_1) {

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":283
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)             # <<<<<<<<<<<<<<
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 */
    __pyx_v_info->strides = ((Py_ssize_t *)PyObject_Malloc((((sizeof(Py_ssize_t)) * 2) * ((size_t)__pyx_v_ndim))));

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":284
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)
 *                 info.shape = info.strides + ndim             # <<<<<<<<<<<<<<
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 */
    __pyx_v_info->shape = (__pyx_v_info->strides + __pyx_v_ndim);

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":285
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):             # <<<<<<<<<<<<<<
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 */
    __pyx_t_4 = __pyx_v_ndim;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":286
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]             # <<<<<<<<<<<<<<
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 */
      (__pyx_v_info->strides[__pyx_v_i]) = (PyArray_STRIDES(__pyx_v_self)[__pyx_v_i]);

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":287
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]             # <<<<<<<<<<<<<<
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 */
      (__pyx_v_info->shape[__pyx_v_i]) = (PyArray_DIMS(__pyx_v_self)[__pyx_v_i]);
    }

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":280
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
    goto __pyx_L9;
  }

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":289
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)             # <<<<<<<<<<<<<<
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 */
  /*else*/ {
    __pyx_v_info->strides = ((Py_ssize_t *)PyArray_STRIDES(__pyx_v_self));

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":290
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)             # <<<<<<<<<<<<<<
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 */
    __pyx_v_info->shape = ((Py_ssize_t *)PyArray_DIMS(__pyx_v_self));
  }
  __pyx_L9:;

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":291
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 */
  __pyx_v_info->suboffsets = NULL;

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":292
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)             # <<<<<<<<<<<<<<
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 * 
 */
  __pyx_v_info->itemsize = PyArray_ITEMSIZE(__pyx_v_self);

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":293
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)             # <<<<<<<<<<<<<<
 * 
 *             cdef int t
 */
  __pyx_v_info->readonly = (!(PyArray_ISWRITEABLE(__pyx_v_self) != 0));

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":296
 * 
 *             cdef int t
 *             cdef char* f = NULL             # <<<<<<<<<<<<<<
 *             cdef dtype descr = <dtype>PyArray_DESCR(self)
 *             cdef int offset
 */
  __pyx_v_f = NULL;

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":297
 *             cdef int t
 *             cdef char* f = NULL
 *             cdef dtype descr = <dtype>PyArray_DESCR(self)             # <<<<<<<<<<<<<<
 *             cdef int offset
 * 
 */
  __pyx_t_7 = PyArray_DESCR(__pyx_v_self);
  __pyx_t_3 = ((PyObject *)__pyx_t_7);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_descr = ((PyArray_Descr *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":300
 *             cdef int offset
 * 
 *             info.obj = self             # <<<<<<<<<<<<<<
 * 
 *             if not PyDataType_HASFIELDS(descr):
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":302
 *             info.obj = self
 * 
 *             if not PyDataType_HASFIELDS(descr):             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  __pyx_t_1 = ((!(PyDataType_HASFIELDS(__pyx_v_descr) != 0)) != 0);
  if (__pyx_t_1) {

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":303
 * 
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num             # <<<<<<<<<<<<<<
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 */
    __pyx_t_4 = __pyx_v_descr->type_num;
    __pyx_v_t = __pyx_t_4;

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_2 = ((__pyx_v_descr->byteorder == '>') != 0);
    if (!__pyx_t_2) {
      goto __pyx_L15_next_or;
    } else {
    }
    __pyx_t_2 = (__pyx_v_little_endian != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_L15_next_or:;

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":305
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 */
    __pyx_t_2 = ((__pyx_v_descr->byteorder == '<') != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_t_2 = ((!(__pyx_v_little_endian != 0)) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L14_bool_binop_done:;

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    if (unlikely(__pyx_t_1)) {

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":306
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 306, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(3, 306, __pyx_L1_error)

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    }

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":307
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 */
    switch (__pyx_v_t) {
      case NPY_BYTE:
      __pyx_v_f = ((char *)"b");
      break;
      case NPY_UBYTE:

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":308
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 */
      __pyx_v_f = ((char *)"B");
      break;
      case NPY_SHORT:

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":309
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 */
      __pyx_v_f = ((char *)"h");
      break;
      case NPY_USHORT:

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":310
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 */
      __pyx_v_f = ((char *)"H");
      break;
      case NPY_INT:

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":311
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 */
      __pyx_v_f = ((char *)"i");
      break;
      case NPY_UINT:

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":312
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 */
      __pyx_v_f = ((char *)"I");
      break;
      case NPY_LONG:

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":313
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 */
      __pyx_v_f = ((char *)"l");
      break;
      case NPY_ULONG:

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":314
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 */
      __pyx_v_f = ((char *)"L");
      break;
      case NPY_LONGLONG:

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":315
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 */
      __pyx_v_f = ((char *)"q");
      break;
      case NPY_ULONGLONG:

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":316
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 */
      __pyx_v_f = ((char *)"Q");
      break;
      case NPY_FLOAT:

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":317
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 */
      __pyx_v_f = ((char *)"f");
      break;
      case NPY_DOUBLE:

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":318
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 */
      __pyx_v_f = ((char *)"d");
      break;
      case NPY_LONGDOUBLE:

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":319
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 */
      __pyx_v_f = ((char *)"g");
      break;
      case NPY_CFLOAT:

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":320
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 */
      __pyx_v_f = ((char *)"Zf");
      break;
      case NPY_CDOUBLE:

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":321
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"
 */
      __pyx_v_f = ((char *)"Zd");
      break;
      case NPY_CLONGDOUBLE:

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":322
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 */
      __pyx_v_f = ((char *)"Zg");
      break;
      case NPY_OBJECT:

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":323
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      __pyx_v_f = ((char *)"O");
      break;
      default:

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":325
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *                 info.format = f
 *                 return
 */
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_t); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = PyUnicode_Format(__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(3, 325, __pyx_L1_error)
      break;
    }

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":326
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f             # <<<<<<<<<<<<<<
 *                 return
 *             else:
 */
    __pyx_v_info->format = __pyx_v_f;

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":327
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f
 *                 return             # <<<<<<<<<<<<<<
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":302
 *             info.obj = self
 * 
 *             if not PyDataType_HASFIELDS(descr):             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  }

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":329
 *                 return
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)             # <<<<<<<<<<<<<<
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 */
  /*else*/ {
    __pyx_v_info->format = ((char *)PyObject_Malloc(0xFF));

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":330
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment             # <<<<<<<<<<<<<<
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,
 */
    (__pyx_v_info->format[0]) = '^';

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":331
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0             # <<<<<<<<<<<<<<
 *                 f = _util_dtypestring(descr, info.format + 1,
 *                                       info.format + _buffer_format_string_len,
 */
    __pyx_v_offset = 0;

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":332
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,             # <<<<<<<<<<<<<<
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 */
    __pyx_t_9 = __pyx_f_5numpy__util_dtypestring(__pyx_v_descr, (__pyx_v_info->format + 1), (__pyx_v_info->format + 0xFF), (&__pyx_v_offset)); if (unlikely(__pyx_t_9 == ((char *)NULL))) __PYX_ERR(3, 332, __pyx_L1_error)
    __pyx_v_f = __pyx_t_9;

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":335
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 *                 f[0] = c'\0' # Terminate format string             # <<<<<<<<<<<<<<
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 */
    (__pyx_v_f[0]) = '\x00';
  }

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":258
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.ndarray.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_descr);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":337
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 */

/* Python wrapper */
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info); /*proto*/
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_5numpy_7ndarray_2__releasebuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":338
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_t_1 = (PyArray_HASFIELDS(__pyx_v_self) != 0);
  if (__pyx_t_1) {

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":339
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)             # <<<<<<<<<<<<<<
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 PyObject_Free(info.strides)
 */
    PyObject_Free(__pyx_v_info->format);

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":338
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  }

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":340
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  __pyx_t_1 = (((sizeof(npy_intp)) != (sizeof(Py_ssize_t))) != 0);
  if (__pyx_t_1) {

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":341
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 PyObject_Free(info.strides)             # <<<<<<<<<<<<<<
 *                 # info.shape was stored after info.strides in the same block
 * 
 */
    PyObject_Free(__pyx_v_info->strides);

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":340
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  }

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":337
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":821
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":822
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":821
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":824
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":825
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":824
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":827
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":828
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":827
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":830
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":831
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":830
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":833
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":834
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":833
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":836
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":837
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  __pyx_t_1 = (PyDataType_HASSUBARRAY(__pyx_v_d) != 0);
  if (__pyx_t_1) {

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":838
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);
    goto __pyx_L0;

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":837
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  }

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":840
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_r = __pyx_empty_tuple;
    goto __pyx_L0;
  }

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":836
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":842
 *         return ()
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *__pyx_v_descr, char *__pyx_v_f, char *__pyx_v_end, int *__pyx_v_offset) {
  PyArray_Descr *__pyx_v_child = 0;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  PyObject *__pyx_v_fields = 0;
  PyObject *__pyx_v_childname = NULL;
  PyObject *__pyx_v_new_offset = NULL;
  PyObject *__pyx_v_t = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  long __pyx_t_8;
  char *__pyx_t_9;
  __Pyx_RefNannySetupContext("_util_dtypestring", 0);

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":847
 * 
 *     cdef dtype child
 *     cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 *     cdef tuple fields
 */
  __pyx_v_endian_detector = 1;

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":848
 *     cdef dtype child
 *     cdef int endian_detector = 1
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 *     cdef tuple fields
 * 
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":851
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  if (unlikely(__pyx_v_descr->names == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(3, 851, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_descr->names; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(3, 851, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 851, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_childname, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":852
 * 
 *     for childname in descr.names:
 *         fields = descr.fields[childname]             # <<<<<<<<<<<<<<
 *         child, new_offset = fields
 * 
 */
    if (unlikely(__pyx_v_descr->fields == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(3, 852, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_descr->fields, __pyx_v_childname); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 852, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(PyTuple_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(3, 852, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_fields, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":853
 *     for childname in descr.names:
 *         fields = descr.fields[childname]
 *         child, new_offset = fields             # <<<<<<<<<<<<<<
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 */
    if (likely(__pyx_v_fields != Py_None)) {
      PyObject* sequence = __pyx_v_fields;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(3, 853, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 853, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 853, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
    } else {
      __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(3, 853, __pyx_L1_error)
    }
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_dtype))))) __PYX_ERR(3, 853, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_child, ((PyArray_Descr *)__pyx_t_3));
    __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_new_offset, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":855
 *         child, new_offset = fields
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    __pyx_t_4 = __Pyx_PyInt_From_int((__pyx_v_offset[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 855, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyNumber_Subtract(__pyx_v_new_offset, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 855, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(3, 855, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = ((((__pyx_v_end - __pyx_v_f) - ((int)__pyx_t_5)) < 15) != 0);
    if (unlikely(__pyx_t_6)) {

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":856
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 856, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(3, 856, __pyx_L1_error)

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":855
 *         child, new_offset = fields
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    }

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":858
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_7 = ((__pyx_v_child->byteorder == '>') != 0);
    if (!__pyx_t_7) {
      goto __pyx_L8_next_or;
    } else {
    }
    __pyx_t_7 = (__pyx_v_little_endian != 0);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_L8_next_or:;

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":859
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *             raise ValueError(u"Non-native byte order not supported")
 *             # One could encode it in the format string and have Cython
 */
    __pyx_t_7 = ((__pyx_v_child->byteorder == '<') != 0);
    if (__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_7 = ((!(__pyx_v_little_endian != 0)) != 0);
    __pyx_t_6 = __pyx_t_7;
    __pyx_L7_bool_binop_done:;

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":858
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    if (unlikely(__pyx_t_6)) {

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":860
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *             # One could encode it in the format string and have Cython
 *             # complain instead, BUT: < and > in format strings also imply
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 860, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(3, 860, __pyx_L1_error)

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":858
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    }

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":870
 * 
 *         # Output padding bytes
 *         while offset[0] < new_offset:             # <<<<<<<<<<<<<<
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 */
    while (1) {
      __pyx_t_3 = __Pyx_PyInt_From_int((__pyx_v_offset[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 870, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, __pyx_v_new_offset, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 870, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 870, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (!__pyx_t_6) break;

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":871
 *         # Output padding bytes
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte             # <<<<<<<<<<<<<<
 *             f += 1
 *             offset[0] += 1
 */
      (__pyx_v_f[0]) = 0x78;

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":872
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte
 *             f += 1             # <<<<<<<<<<<<<<
 *             offset[0] += 1
 * 
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":873
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 *             offset[0] += 1             # <<<<<<<<<<<<<<
 * 
 *         offset[0] += child.itemsize
 */
      __pyx_t_8 = 0;
      (__pyx_v_offset[__pyx_t_8]) = ((__pyx_v_offset[__pyx_t_8]) + 1);
    }

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":875
 *             offset[0] += 1
 * 
 *         offset[0] += child.itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not PyDataType_HASFIELDS(child):
 */
    __pyx_t_8 = 0;
    (__pyx_v_offset[__pyx_t_8]) = ((__pyx_v_offset[__pyx_t_8]) + __pyx_v_child->elsize);

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":877
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
    __pyx_t_6 = ((!(PyDataType_HASFIELDS(__pyx_v_child) != 0)) != 0);
    if (__pyx_t_6) {

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":878
 * 
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num             # <<<<<<<<<<<<<<
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")
 */
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_child->type_num); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 878, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_t, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":879
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      __pyx_t_6 = (((__pyx_v_end - __pyx_v_f) < 5) != 0);
      if (unlikely(__pyx_t_6)) {

        /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":880
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 880, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(3, 880, __pyx_L1_error)

        /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":879
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      }

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":883
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_BYTE); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 883, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 883, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 883, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 98;
        goto __pyx_L15;
      }

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":884
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"             # <<<<<<<<<<<<<<
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_UBYTE); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 884, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 884, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 884, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 66;
        goto __pyx_L15;
      }

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":885
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"             # <<<<<<<<<<<<<<
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_SHORT); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 885, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 885, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 885, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x68;
        goto __pyx_L15;
      }

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":886
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"             # <<<<<<<<<<<<<<
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_USHORT); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 886, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 886, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 886, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 72;
        goto __pyx_L15;
      }

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":887
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_INT); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 887, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 887, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 887, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x69;
        goto __pyx_L15;
      }

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":888
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_UINT); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 888, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 888, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 888, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 73;
        goto __pyx_L15;
      }

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":889
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONG); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 889, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 889, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 889, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x6C;
        goto __pyx_L15;
      }

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":890
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_ULONG); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 890, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 890, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 890, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 76;
        goto __pyx_L15;
      }

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":891
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONGLONG); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 891, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 891, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 891, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x71;
        goto __pyx_L15;
      }

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":892
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_ULONGLONG); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 892, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 892, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 892, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 81;
        goto __pyx_L15;
      }

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":893
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"             # <<<<<<<<<<<<<<
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_FLOAT); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 893, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 893, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 893, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x66;
        goto __pyx_L15;
      }

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":894
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_DOUBLE); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 894, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 894, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 894, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x64;
        goto __pyx_L15;
      }

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":895
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"             # <<<<<<<<<<<<<<
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONGDOUBLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 895, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 895, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 895, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x67;
        goto __pyx_L15;
      }

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":896
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf             # <<<<<<<<<<<<<<
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CFLOAT); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 896, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 896, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 896, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x66;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":897
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd             # <<<<<<<<<<<<<<
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CDOUBLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 897, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 897, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 897, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x64;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":898
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg             # <<<<<<<<<<<<<<
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CLONGDOUBLE); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 898, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 898, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 898, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x67;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":899
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"             # <<<<<<<<<<<<<<
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_OBJECT); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 899, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 899, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 899, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (likely(__pyx_t_6)) {
        (__pyx_v_f[0]) = 79;
        goto __pyx_L15;
      }

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":901
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *             f += 1
 *         else:
 */
      /*else*/ {
        __pyx_t_3 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_v_t); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 901, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 901, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(3, 901, __pyx_L1_error)
      }
      __pyx_L15:;

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":902
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *             f += 1             # <<<<<<<<<<<<<<
 *         else:
 *             # Cython ignores struct boundary information ("T{...}"),
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":877
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
      goto __pyx_L13;
    }

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":906
 *             # Cython ignores struct boundary information ("T{...}"),
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)             # <<<<<<<<<<<<<<
 *     return f
 * 
 */
    /*else*/ {
      __pyx_t_9 = __pyx_f_5numpy__util_dtypestring(__pyx_v_child, __pyx_v_f, __pyx_v_end, __pyx_v_offset); if (unlikely(__pyx_t_9 == ((char *)NULL))) __PYX_ERR(3, 906, __pyx_L1_error)
      __pyx_v_f = __pyx_t_9;
    }
    __pyx_L13:;

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":851
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":907
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)
 *     return f             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_f;
  goto __pyx_L0;

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":842
 *         return ()
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("numpy._util_dtypestring", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_child);
  __Pyx_XDECREF(__pyx_v_fields);
  __Pyx_XDECREF(__pyx_v_childname);
  __Pyx_XDECREF(__pyx_v_new_offset);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1022
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1023
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */
  Py_INCREF(__pyx_v_base);

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1024
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  (void)(PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base));

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1022
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1026
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1027
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1028
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  __pyx_t_1 = ((__pyx_v_base == NULL) != 0);
  if (__pyx_t_1) {

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1029
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1028
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  }

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1030
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_base));
  __pyx_r = ((PyObject *)__pyx_v_base);
  goto __pyx_L0;

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1026
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1034
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1035
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1036
 * cdef inline int import_array() except -1:
 *     try:
 *         _import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(3, 1036, __pyx_L3_error)

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1035
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1037
 *     try:
 *         _import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(3, 1037, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1038
 *         _import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 1038, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(3, 1038, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1035
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1034
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1040
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1041
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1042
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(3, 1042, __pyx_L3_error)

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1041
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1043
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(3, 1043, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1044
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 1044, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(3, 1044, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1041
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1040
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1046
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1047
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1048
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(3, 1048, __pyx_L3_error)

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1047
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1049
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(3, 1049, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1050
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 1050, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(3, 1050, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1047
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1046
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_Tree __pyx_vtable_8skgarden_8mondrian_4tree_5_tree_Tree;

static PyObject *__pyx_tp_new_8skgarden_8mondrian_4tree_5_tree_Tree(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *)o);
  p->__pyx_vtab = __pyx_vtabptr_8skgarden_8mondrian_4tree_5_tree_Tree;
  if (unlikely(__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_8skgarden_8mondrian_4tree_5_tree_Tree(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_n_classes(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_9n_classes_1__get__(o);
}

static PyObject *__pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_children_left(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_13children_left_1__get__(o);
}

static PyObject *__pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_children_right(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_14children_right_1__get__(o);
}

static PyObject *__pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_feature(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_7feature_1__get__(o);
}

static PyObject *__pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_threshold(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_9threshold_1__get__(o);
}

static PyObject *__pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_impurity(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_8impurity_1__get__(o);
}

static PyObject *__pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_n_node_samples(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_14n_node_samples_1__get__(o);
}

static PyObject *__pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_weighted_n_node_samples(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_23weighted_n_node_samples_1__get__(o);
}

static PyObject *__pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_value(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_5value_1__get__(o);
}

static PyObject *__pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_tau(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_3tau_1__get__(o);
}

static PyObject *__pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_mean(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_4mean_1__get__(o);
}

static PyObject *__pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_variance(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_8variance_1__get__(o);
}

static PyObject *__pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_root(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_4root_1__get__(o);
}

static PyObject *__pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_n_features(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_10n_features_1__get__(o);
}

static int __pyx_setprop_8skgarden_8mondrian_4tree_5_tree_4Tree_n_features(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_10n_features_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_n_outputs(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_9n_outputs_1__get__(o);
}

static int __pyx_setprop_8skgarden_8mondrian_4tree_5_tree_4Tree_n_outputs(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_9n_outputs_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_max_n_classes(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_13max_n_classes_1__get__(o);
}

static int __pyx_setprop_8skgarden_8mondrian_4tree_5_tree_4Tree_max_n_classes(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_13max_n_classes_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_max_depth(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_9max_depth_1__get__(o);
}

static int __pyx_setprop_8skgarden_8mondrian_4tree_5_tree_4Tree_max_depth(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_9max_depth_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_node_count(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_10node_count_1__get__(o);
}

static int __pyx_setprop_8skgarden_8mondrian_4tree_5_tree_4Tree_node_count(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_10node_count_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_capacity(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_8capacity_1__get__(o);
}

static int __pyx_setprop_8skgarden_8mondrian_4tree_5_tree_4Tree_capacity(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_8capacity_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_8skgarden_8mondrian_4tree_5_tree_Tree[] = {
  {"__reduce__", (PyCFunction)__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_5__reduce__, METH_NOARGS, __pyx_doc_8skgarden_8mondrian_4tree_5_tree_4Tree_4__reduce__},
  {"__getstate__", (PyCFunction)__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_7__getstate__, METH_NOARGS, __pyx_doc_8skgarden_8mondrian_4tree_5_tree_4Tree_6__getstate__},
  {"__setstate__", (PyCFunction)__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_9__setstate__, METH_O, __pyx_doc_8skgarden_8mondrian_4tree_5_tree_4Tree_8__setstate__},
  {"apply", (PyCFunction)__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_11apply, METH_O, __pyx_doc_8skgarden_8mondrian_4tree_5_tree_4Tree_10apply},
  {"predict", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_13predict, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8skgarden_8mondrian_4tree_5_tree_4Tree_12predict},
  {"decision_path", (PyCFunction)__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_15decision_path, METH_O, __pyx_doc_8skgarden_8mondrian_4tree_5_tree_4Tree_14decision_path},
  {"weighted_decision_path", (PyCFunction)__pyx_pw_8skgarden_8mondrian_4tree_5_tree_4Tree_17weighted_decision_path, METH_O, __pyx_doc_8skgarden_8mondrian_4tree_5_tree_4Tree_16weighted_decision_path},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_8skgarden_8mondrian_4tree_5_tree_Tree[] = {
  {(char *)"n_classes", __pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_n_classes, 0, (char *)0, 0},
  {(char *)"children_left", __pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_children_left, 0, (char *)0, 0},
  {(char *)"children_right", __pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_children_right, 0, (char *)0, 0},
  {(char *)"feature", __pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_feature, 0, (char *)0, 0},
  {(char *)"threshold", __pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_threshold, 0, (char *)0, 0},
  {(char *)"impurity", __pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_impurity, 0, (char *)0, 0},
  {(char *)"n_node_samples", __pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_n_node_samples, 0, (char *)0, 0},
  {(char *)"weighted_n_node_samples", __pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_weighted_n_node_samples, 0, (char *)0, 0},
  {(char *)"value", __pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_value, 0, (char *)0, 0},
  {(char *)"tau", __pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_tau, 0, (char *)0, 0},
  {(char *)"mean", __pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_mean, 0, (char *)0, 0},
  {(char *)"variance", __pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_variance, 0, (char *)0, 0},
  {(char *)"root", __pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_root, 0, (char *)0, 0},
  {(char *)"n_features", __pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_n_features, __pyx_setprop_8skgarden_8mondrian_4tree_5_tree_4Tree_n_features, (char *)0, 0},
  {(char *)"n_outputs", __pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_n_outputs, __pyx_setprop_8skgarden_8mondrian_4tree_5_tree_4Tree_n_outputs, (char *)0, 0},
  {(char *)"max_n_classes", __pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_max_n_classes, __pyx_setprop_8skgarden_8mondrian_4tree_5_tree_4Tree_max_n_classes, (char *)0, 0},
  {(char *)"max_depth", __pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_max_depth, __pyx_setprop_8skgarden_8mondrian_4tree_5_tree_4Tree_max_depth, (char *)0, 0},
  {(char *)"node_count", __pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_node_count, __pyx_setprop_8skgarden_8mondrian_4tree_5_tree_4Tree_node_count, (char *)0, 0},
  {(char *)"capacity", __pyx_getprop_8skgarden_8mondrian_4tree_5_tree_4Tree_capacity, __pyx_setprop_8skgarden_8mondrian_4tree_5_tree_4Tree_capacity, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_8skgarden_8mondrian_4tree_5_tree_Tree = {
  PyVarObject_HEAD_INIT(0, 0)
  "skgarden.mondrian.tree._tree.Tree", /*tp_name*/
  sizeof(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8skgarden_8mondrian_4tree_5_tree_Tree, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "Array-based representation of a binary decision tree.\n\n    The binary tree is represented as a number of parallel arrays. The i-th\n    element of each array holds information about the node `i`. Node 0 is the\n    tree's root. You can find a detailed description of all arrays in\n    `_tree.pxd`. NOTE: Some of the arrays only apply to either leaves or split\n    nodes, resp. In this case the values of nodes of the other type are\n    arbitrary!\n\n    Attributes\n    ----------\n    node_count : int\n        The number of nodes (internal nodes + leaves) in the tree.\n\n    capacity : int\n        The current capacity (i.e., size) of the arrays, which is at least as\n        great as `node_count`.\n\n    max_depth : int\n        The maximal depth of the tree.\n\n    children_left : array of int, shape [node_count]\n        children_left[i] holds the node id of the left child of node i.\n        For leaves, children_left[i] == TREE_LEAF. Otherwise,\n        children_left[i] > i. This child handles the case where\n        X[:, feature[i]] <= threshold[i].\n\n    children_right : array of int, shape [node_count]\n        children_right[i] holds the node id of the right child of node i.\n        For leaves, children_right[i] == TREE_LEAF. Otherwise,\n        children_right[i] > i. This child handles the case where\n        X[:, feature[i]] > threshold[i].\n\n    feature : array of int, shape [node_count]\n        feature[i] holds the feature to split on, for the internal node i.\n\n    threshold : array of double, shape [node_count]\n        threshold[i] holds the threshold for the internal node i.\n\n    value : array of double, shape [node_count, n_outputs, max_n_classes]\n        Contains the constant prediction value of each node.\n\n    impurity : array of double, shape [node_count]\n        impurity[i] holds the impurity (i.e., the value of the splitting\n        criterion) at node i.\n\n    n_node_samples : array of int, shape [node_count]\n        n_node_sampl""es[i] holds the number of training samples reaching node i.\n\n    weighted_n_node_samples : array of int, shape [node_count]\n        weighted_n_node_samples[i] holds the weighted number of training samples\n        reaching node i.\n    ", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8skgarden_8mondrian_4tree_5_tree_Tree, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_8skgarden_8mondrian_4tree_5_tree_Tree, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8skgarden_8mondrian_4tree_5_tree_Tree, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_TreeBuilder __pyx_vtable_8skgarden_8mondrian_4tree_5_tree_TreeBuilder;

static PyObject *__pyx_tp_new_8skgarden_8mondrian_4tree_5_tree_TreeBuilder(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *)o);
  p->__pyx_vtab = __pyx_vtabptr_8skgarden_8mondrian_4tree_5_tree_TreeBuilder;
  p->splitter = ((struct __pyx_obj_8skgarden_8mondrian_4tree_9_splitter_Splitter *)Py_None); Py_INCREF(Py_None);
  p->random_state = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_8skgarden_8mondrian_4tree_5_tree_TreeBuilder(PyObject *o) {
  struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *p = (struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->splitter);
  Py_CLEAR(p->random_state);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_8skgarden_8mondrian_4tree_5_tree_TreeBuilder(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *p = (struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *)o;
  if (p->splitter) {
    e = (*v)(((PyObject *)p->splitter), a); if (e) return e;
  }
  if (p->random_state) {
    e = (*v)(p->random_state, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8skgarden_8mondrian_4tree_5_tree_TreeBuilder(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *p = (struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *)o;
  tmp = ((PyObject*)p->splitter);
  p->splitter = ((struct __pyx_obj_8skgarden_8mondrian_4tree_9_splitter_Splitter *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->random_state);
  p->random_state = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_8skgarden_8mondrian_4tree_5_tree_TreeBuilder[] = {
  {"build", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_1build, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_build},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_8skgarden_8mondrian_4tree_5_tree_TreeBuilder = {
  PyVarObject_HEAD_INIT(0, 0)
  "skgarden.mondrian.tree._tree.TreeBuilder", /*tp_name*/
  sizeof(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8skgarden_8mondrian_4tree_5_tree_TreeBuilder, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Interface for different tree building strategies.", /*tp_doc*/
  __pyx_tp_traverse_8skgarden_8mondrian_4tree_5_tree_TreeBuilder, /*tp_traverse*/
  __pyx_tp_clear_8skgarden_8mondrian_4tree_5_tree_TreeBuilder, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8skgarden_8mondrian_4tree_5_tree_TreeBuilder, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8skgarden_8mondrian_4tree_5_tree_TreeBuilder, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder __pyx_vtable_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder;

static PyObject *__pyx_tp_new_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder *p;
  PyObject *o = __pyx_tp_new_8skgarden_8mondrian_4tree_5_tree_TreeBuilder(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_TreeBuilder*)__pyx_vtabptr_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder;
  if (unlikely(__pyx_pw_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static PyMethodDef __pyx_methods_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder[] = {
  {"build", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder_3build, METH_VARARGS|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder_5__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder_7__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder = {
  PyVarObject_HEAD_INIT(0, 0)
  "skgarden.mondrian.tree._tree.PartialFitTreeBuilder", /*tp_name*/
  sizeof(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8skgarden_8mondrian_4tree_5_tree_TreeBuilder, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Build a decision tree incrementally.", /*tp_doc*/
  __pyx_tp_traverse_8skgarden_8mondrian_4tree_5_tree_TreeBuilder, /*tp_traverse*/
  __pyx_tp_clear_8skgarden_8mondrian_4tree_5_tree_TreeBuilder, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder __pyx_vtable_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder;

static PyObject *__pyx_tp_new_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder *p;
  PyObject *o = __pyx_tp_new_8skgarden_8mondrian_4tree_5_tree_TreeBuilder(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_5_tree_TreeBuilder*)__pyx_vtabptr_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder;
  if (unlikely(__pyx_pw_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static PyMethodDef __pyx_methods_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder[] = {
  {"build", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder_3build, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder_2build},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder_5__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder_7__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder = {
  PyVarObject_HEAD_INIT(0, 0)
  "skgarden.mondrian.tree._tree.DepthFirstTreeBuilder", /*tp_name*/
  sizeof(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8skgarden_8mondrian_4tree_5_tree_TreeBuilder, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Build a decision tree in depth-first fashion.", /*tp_doc*/
  __pyx_tp_traverse_8skgarden_8mondrian_4tree_5_tree_TreeBuilder, /*tp_traverse*/
  __pyx_tp_clear_8skgarden_8mondrian_4tree_5_tree_TreeBuilder, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec__tree(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec__tree},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "_tree",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_C, __pyx_k_C, sizeof(__pyx_k_C), 0, 0, 1, 1},
  {&__pyx_n_s_DOUBLE, __pyx_k_DOUBLE, sizeof(__pyx_k_DOUBLE), 0, 0, 1, 1},
  {&__pyx_n_s_DTYPE, __pyx_k_DTYPE, sizeof(__pyx_k_DTYPE), 0, 0, 1, 1},
  {&__pyx_n_s_DepthFirstTreeBuilder, __pyx_k_DepthFirstTreeBuilder, sizeof(__pyx_k_DepthFirstTreeBuilder), 0, 0, 1, 1},
  {&__pyx_kp_s_Did_not_recognise_loaded_array_l, __pyx_k_Did_not_recognise_loaded_array_l, sizeof(__pyx_k_Did_not_recognise_loaded_array_l), 0, 0, 1, 0},
  {&__pyx_kp_u_Format_string_allocated_too_shor, __pyx_k_Format_string_allocated_too_shor, sizeof(__pyx_k_Format_string_allocated_too_shor), 0, 1, 0, 0},
  {&__pyx_kp_u_Format_string_allocated_too_shor_2, __pyx_k_Format_string_allocated_too_shor_2, sizeof(__pyx_k_Format_string_allocated_too_shor_2), 0, 1, 0, 0},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0xed, __pyx_k_Incompatible_checksums_s_vs_0xed, sizeof(__pyx_k_Incompatible_checksums_s_vs_0xed), 0, 0, 1, 0},
  {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {&__pyx_n_s_NODE_DTYPE, __pyx_k_NODE_DTYPE, sizeof(__pyx_k_NODE_DTYPE), 0, 0, 1, 1},
  {&__pyx_kp_u_Non_native_byte_order_not_suppor, __pyx_k_Non_native_byte_order_not_suppor, sizeof(__pyx_k_Non_native_byte_order_not_suppor), 0, 1, 0, 0},
  {&__pyx_n_s_PartialFitTreeBuilder, __pyx_k_PartialFitTreeBuilder, sizeof(__pyx_k_PartialFitTreeBuilder), 0, 0, 1, 1},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s_TREE_LEAF, __pyx_k_TREE_LEAF, sizeof(__pyx_k_TREE_LEAF), 0, 0, 1, 1},
  {&__pyx_n_s_TREE_UNDEFINED, __pyx_k_TREE_UNDEFINED, sizeof(__pyx_k_TREE_UNDEFINED), 0, 0, 1, 1},
  {&__pyx_n_s_Tree, __pyx_k_Tree, sizeof(__pyx_k_Tree), 0, 0, 1, 1},
  {&__pyx_n_s_TreeBuilder, __pyx_k_TreeBuilder, sizeof(__pyx_k_TreeBuilder), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_X, __pyx_k_X, sizeof(__pyx_k_X), 0, 0, 1, 1},
  {&__pyx_kp_s_X_dtype_should_be_np_float32_got, __pyx_k_X_dtype_should_be_np_float32_got, sizeof(__pyx_k_X_dtype_should_be_np_float32_got), 0, 0, 1, 0},
  {&__pyx_n_s_X_idx_sorted, __pyx_k_X_idx_sorted, sizeof(__pyx_k_X_idx_sorted), 0, 0, 1, 1},
  {&__pyx_kp_s_X_should_be_in_np_ndarray_format, __pyx_k_X_should_be_in_np_ndarray_format, sizeof(__pyx_k_X_should_be_in_np_ndarray_format), 0, 0, 1, 0},
  {&__pyx_kp_s_You_have_loaded_Tree_version_whi, __pyx_k_You_have_loaded_Tree_version_whi, sizeof(__pyx_k_You_have_loaded_Tree_version_whi), 0, 0, 1, 0},
  {&__pyx_n_s_apply, __pyx_k_apply, sizeof(__pyx_k_apply), 0, 0, 1, 1},
  {&__pyx_n_s_asarray, __pyx_k_asarray, sizeof(__pyx_k_asarray), 0, 0, 1, 1},
  {&__pyx_n_s_ascontiguousarray, __pyx_k_ascontiguousarray, sizeof(__pyx_k_ascontiguousarray), 0, 0, 1, 1},
  {&__pyx_n_s_asfortranarray, __pyx_k_asfortranarray, sizeof(__pyx_k_asfortranarray), 0, 0, 1, 1},
  {&__pyx_n_s_build, __pyx_k_build, sizeof(__pyx_k_build), 0, 0, 1, 1},
  {&__pyx_n_s_c_contiguous, __pyx_k_c_contiguous, sizeof(__pyx_k_c_contiguous), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_contiguous, __pyx_k_contiguous, sizeof(__pyx_k_contiguous), 0, 0, 1, 1},
  {&__pyx_n_s_csr_matrix, __pyx_k_csr_matrix, sizeof(__pyx_k_csr_matrix), 0, 0, 1, 1},
  {&__pyx_n_s_decision_path, __pyx_k_decision_path, sizeof(__pyx_k_decision_path), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_dtype, __pyx_k_dtype, sizeof(__pyx_k_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_feature, __pyx_k_feature, sizeof(__pyx_k_feature), 0, 0, 1, 1},
  {&__pyx_n_s_flags, __pyx_k_flags, sizeof(__pyx_k_flags), 0, 0, 1, 1},
  {&__pyx_n_s_float32, __pyx_k_float32, sizeof(__pyx_k_float32), 0, 0, 1, 1},
  {&__pyx_n_s_float64, __pyx_k_float64, sizeof(__pyx_k_float64), 0, 0, 1, 1},
  {&__pyx_n_s_formats, __pyx_k_formats, sizeof(__pyx_k_formats), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_impurity, __pyx_k_impurity, sizeof(__pyx_k_impurity), 0, 0, 1, 1},
  {&__pyx_n_s_inf, __pyx_k_inf, sizeof(__pyx_k_inf), 0, 0, 1, 1},
  {&__pyx_n_s_intp, __pyx_k_intp, sizeof(__pyx_k_intp), 0, 0, 1, 1},
  {&__pyx_n_s_is_regression, __pyx_k_is_regression, sizeof(__pyx_k_is_regression), 0, 0, 1, 1},
  {&__pyx_n_s_itemsize, __pyx_k_itemsize, sizeof(__pyx_k_itemsize), 0, 0, 1, 1},
  {&__pyx_n_s_left_child, __pyx_k_left_child, sizeof(__pyx_k_left_child), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_max, __pyx_k_max, sizeof(__pyx_k_max), 0, 0, 1, 1},
  {&__pyx_n_s_max_depth, __pyx_k_max_depth, sizeof(__pyx_k_max_depth), 0, 0, 1, 1},
  {&__pyx_n_s_min_samples_split, __pyx_k_min_samples_split, sizeof(__pyx_k_min_samples_split), 0, 0, 1, 1},
  {&__pyx_n_s_n_classes, __pyx_k_n_classes, sizeof(__pyx_k_n_classes), 0, 0, 1, 1},
  {&__pyx_n_s_n_features, __pyx_k_n_features, sizeof(__pyx_k_n_features), 0, 0, 1, 1},
  {&__pyx_n_s_n_node_samples, __pyx_k_n_node_samples, sizeof(__pyx_k_n_node_samples), 0, 0, 1, 1},
  {&__pyx_n_s_n_outputs, __pyx_k_n_outputs, sizeof(__pyx_k_n_outputs), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_names, __pyx_k_names, sizeof(__pyx_k_names), 0, 0, 1, 1},
  {&__pyx_kp_u_ndarray_is_not_C_contiguous, __pyx_k_ndarray_is_not_C_contiguous, sizeof(__pyx_k_ndarray_is_not_C_contiguous), 0, 1, 0, 0},
  {&__pyx_kp_u_ndarray_is_not_Fortran_contiguou, __pyx_k_ndarray_is_not_Fortran_contiguou, sizeof(__pyx_k_ndarray_is_not_Fortran_contiguou), 0, 1, 0, 0},
  {&__pyx_n_s_ndim, __pyx_k_ndim, sizeof(__pyx_k_ndim), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_s_node_count, __pyx_k_node_count, sizeof(__pyx_k_node_count), 0, 0, 1, 1},
  {&__pyx_n_s_nodes, __pyx_k_nodes, sizeof(__pyx_k_nodes), 0, 0, 1, 1},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_kp_s_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 0, 1, 0},
  {&__pyx_kp_s_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 0, 1, 0},
  {&__pyx_n_s_offsets, __pyx_k_offsets, sizeof(__pyx_k_offsets), 0, 0, 1, 1},
  {&__pyx_n_s_ones, __pyx_k_ones, sizeof(__pyx_k_ones), 0, 0, 1, 1},
  {&__pyx_n_s_order, __pyx_k_order, sizeof(__pyx_k_order), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_predict, __pyx_k_predict, sizeof(__pyx_k_predict), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_TreeBuilder, __pyx_k_pyx_unpickle_TreeBuilder, sizeof(__pyx_k_pyx_unpickle_TreeBuilder), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_randint, __pyx_k_randint, sizeof(__pyx_k_randint), 0, 0, 1, 1},
  {&__pyx_n_s_random_state, __pyx_k_random_state, sizeof(__pyx_k_random_state), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_ravel, __pyx_k_ravel, sizeof(__pyx_k_ravel), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_kp_s_resizing_tree_to_d, __pyx_k_resizing_tree_to_d, sizeof(__pyx_k_resizing_tree_to_d), 0, 0, 1, 0},
  {&__pyx_n_s_return_std, __pyx_k_return_std, sizeof(__pyx_k_return_std), 0, 0, 1, 1},
  {&__pyx_n_s_right_child, __pyx_k_right_child, sizeof(__pyx_k_right_child), 0, 0, 1, 1},
  {&__pyx_n_s_root, __pyx_k_root, sizeof(__pyx_k_root), 0, 0, 1, 1},
  {&__pyx_n_s_sample_weight, __pyx_k_sample_weight, sizeof(__pyx_k_sample_weight), 0, 0, 1, 1},
  {&__pyx_n_s_scipy_sparse, __pyx_k_scipy_sparse, sizeof(__pyx_k_scipy_sparse), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_shape, __pyx_k_shape, sizeof(__pyx_k_shape), 0, 0, 1, 1},
  {&__pyx_n_s_skgarden_mondrian_tree__tree, __pyx_k_skgarden_mondrian_tree__tree, sizeof(__pyx_k_skgarden_mondrian_tree__tree), 0, 0, 1, 1},
  {&__pyx_n_s_splitter, __pyx_k_splitter, sizeof(__pyx_k_splitter), 0, 0, 1, 1},
  {&__pyx_n_s_strides, __pyx_k_strides, sizeof(__pyx_k_strides), 0, 0, 1, 1},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_tau, __pyx_k_tau, sizeof(__pyx_k_tau), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_threshold, __pyx_k_threshold, sizeof(__pyx_k_threshold), 0, 0, 1, 1},
  {&__pyx_n_s_tree, __pyx_k_tree, sizeof(__pyx_k_tree), 0, 0, 1, 1},
  {&__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_k_unknown_dtype_code_in_numpy_pxd, sizeof(__pyx_k_unknown_dtype_code_in_numpy_pxd), 0, 1, 0, 0},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
  {&__pyx_n_s_variance, __pyx_k_variance, sizeof(__pyx_k_variance), 0, 0, 1, 1},
  {&__pyx_n_s_weighted_decision_path, __pyx_k_weighted_decision_path, sizeof(__pyx_k_weighted_decision_path), 0, 0, 1, 1},
  {&__pyx_n_s_weighted_n_node_samples, __pyx_k_weighted_n_node_samples, sizeof(__pyx_k_weighted_n_node_samples), 0, 0, 1, 1},
  {&__pyx_n_s_y, __pyx_k_y, sizeof(__pyx_k_y), 0, 0, 1, 1},
  {&__pyx_n_s_zeros, __pyx_k_zeros, sizeof(__pyx_k_zeros), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 167, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(0, 240, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 482, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(3, 856, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(3, 1038, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple_)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "skgarden/mondrian/tree/_tree.pyx":482
 * 
 *         if 'nodes' not in d:
 *             raise ValueError('You have loaded Tree version which '             # <<<<<<<<<<<<<<
 *                              'cannot be imported')
 * 
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_You_have_loaded_Tree_version_whi); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "skgarden/mondrian/tree/_tree.pyx":496
 *                 not value_ndarray.flags.c_contiguous or
 *                 value_ndarray.dtype != np.float64):
 *             raise ValueError('Did not recognise loaded array layout')             # <<<<<<<<<<<<<<
 * 
 *         self.capacity = node_ndarray.shape[0]
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_Did_not_recognise_loaded_array_l); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "skgarden/mondrian/tree/_tree.pyx":910
 *         # These node values are the means in case of regression.
 *         # For classification these are the class counts.
 *         cdef np.ndarray[DOUBLE_t, ndim=2] node_values = self._get_value_ndarray()[:, 0, :]             # <<<<<<<<<<<<<<
 * 
 *         # Initialize output
 */
  __pyx_slice__8 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__8)) __PYX_ERR(0, 910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__8);
  __Pyx_GIVEREF(__pyx_slice__8);
  __pyx_slice__8 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__8)) __PYX_ERR(0, 910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__8);
  __Pyx_GIVEREF(__pyx_slice__8);
  __pyx_tuple__9 = PyTuple_Pack(3, __pyx_slice__8, __pyx_int_0, __pyx_slice__8); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":272
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_u_ndarray_is_not_C_contiguous); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(3, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":276
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_u_ndarray_is_not_Fortran_contiguou); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(3, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":306
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_u_Non_native_byte_order_not_suppor); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(3, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":856
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_u_Format_string_allocated_too_shor); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(3, 856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":860
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *             # One could encode it in the format string and have Cython
 *             # complain instead, BUT: < and > in format strings also imply
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_u_Non_native_byte_order_not_suppor); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(3, 860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":880
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_u_Format_string_allocated_too_shor_2); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(3, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1038
 *         _import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(3, 1038, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1044
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(3, 1044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1050
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(3, 1050, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "(tree fragment)":1
 * def __pyx_unpickle_TreeBuilder(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     if __pyx_checksum != 0xedae8f4:
 *         from pickle import PickleError as __pyx_PickleError
 */
  __pyx_tuple__17 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);
  __pyx_codeobj__18 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__17, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_TreeBuilder, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__18)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_249227508 = PyInt_FromLong(249227508L); if (unlikely(!__pyx_int_249227508)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_2 = PyInt_FromLong(-2); if (unlikely(!__pyx_int_neg_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_8skgarden_8mondrian_4tree_5_tree_Tree = &__pyx_vtable_8skgarden_8mondrian_4tree_5_tree_Tree;
  __pyx_vtable_8skgarden_8mondrian_4tree_5_tree_Tree._add_node = (__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t (*)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, int, int, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, double, double, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, double, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *, double))__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree__add_node;
  __pyx_vtable_8skgarden_8mondrian_4tree_5_tree_Tree._resize = (int (*)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t))__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree__resize;
  __pyx_vtable_8skgarden_8mondrian_4tree_5_tree_Tree._resize_c = (int (*)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_4Tree__resize_c *__pyx_optional_args))__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree__resize_c;
  __pyx_vtable_8skgarden_8mondrian_4tree_5_tree_Tree._get_value_ndarray = (PyArrayObject *(*)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *))__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree__get_value_ndarray;
  __pyx_vtable_8skgarden_8mondrian_4tree_5_tree_Tree._get_node_ndarray = (PyArrayObject *(*)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *))__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree__get_node_ndarray;
  __pyx_vtable_8skgarden_8mondrian_4tree_5_tree_Tree.predict = (PyObject *(*)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_4Tree_predict *__pyx_optional_args))__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree_predict;
  __pyx_vtable_8skgarden_8mondrian_4tree_5_tree_Tree.apply = (PyArrayObject *(*)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, PyObject *, int __pyx_skip_dispatch))__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree_apply;
  __pyx_vtable_8skgarden_8mondrian_4tree_5_tree_Tree._apply_dense = (PyArrayObject *(*)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, PyObject *))__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree__apply_dense;
  __pyx_vtable_8skgarden_8mondrian_4tree_5_tree_Tree.decision_path = (PyObject *(*)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, PyObject *, int __pyx_skip_dispatch))__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree_decision_path;
  __pyx_vtable_8skgarden_8mondrian_4tree_5_tree_Tree._decision_path_dense = (PyObject *(*)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, PyObject *))__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree__decision_path_dense;
  __pyx_vtable_8skgarden_8mondrian_4tree_5_tree_Tree.weighted_decision_path = (PyObject *(*)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, PyObject *, int __pyx_skip_dispatch))__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree_weighted_decision_path;
  __pyx_vtable_8skgarden_8mondrian_4tree_5_tree_Tree._init = (void (*)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t *, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t))__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree__init;
  __pyx_vtable_8skgarden_8mondrian_4tree_5_tree_Tree.extend = (void (*)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t *, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_UINT32_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t))__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree_extend;
  __pyx_vtable_8skgarden_8mondrian_4tree_5_tree_Tree.set_node_attributes = (void (*)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t *, struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_4Tree_set_node_attributes *__pyx_optional_args))__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree_set_node_attributes;
  __pyx_vtable_8skgarden_8mondrian_4tree_5_tree_Tree.update_node_extent = (void (*)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DTYPE_t *, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t))__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree_update_node_extent;
  __pyx_vtable_8skgarden_8mondrian_4tree_5_tree_Tree._update_node_info = (void (*)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_5_tree_DOUBLE_t *, __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t))__pyx_f_8skgarden_8mondrian_4tree_5_tree_4Tree__update_node_info;
  if (PyType_Ready(&__pyx_type_8skgarden_8mondrian_4tree_5_tree_Tree) < 0) __PYX_ERR(0, 315, __pyx_L1_error)
  __pyx_type_8skgarden_8mondrian_4tree_5_tree_Tree.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_8skgarden_8mondrian_4tree_5_tree_Tree.tp_dictoffset && __pyx_type_8skgarden_8mondrian_4tree_5_tree_Tree.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_8skgarden_8mondrian_4tree_5_tree_Tree.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_8skgarden_8mondrian_4tree_5_tree_Tree.tp_dict, __pyx_vtabptr_8skgarden_8mondrian_4tree_5_tree_Tree) < 0) __PYX_ERR(0, 315, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Tree, (PyObject *)&__pyx_type_8skgarden_8mondrian_4tree_5_tree_Tree) < 0) __PYX_ERR(0, 315, __pyx_L1_error)
  __pyx_ptype_8skgarden_8mondrian_4tree_5_tree_Tree = &__pyx_type_8skgarden_8mondrian_4tree_5_tree_Tree;
  __pyx_vtabptr_8skgarden_8mondrian_4tree_5_tree_TreeBuilder = &__pyx_vtable_8skgarden_8mondrian_4tree_5_tree_TreeBuilder;
  __pyx_vtable_8skgarden_8mondrian_4tree_5_tree_TreeBuilder.build = (PyObject *(*)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *, struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, PyObject *, PyArrayObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_build *__pyx_optional_args))__pyx_f_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_build;
  __pyx_vtable_8skgarden_8mondrian_4tree_5_tree_TreeBuilder._check_input = (PyObject *(*)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *, PyObject *, PyArrayObject *, PyArrayObject *))__pyx_f_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder__check_input;
  if (PyType_Ready(&__pyx_type_8skgarden_8mondrian_4tree_5_tree_TreeBuilder) < 0) __PYX_ERR(0, 101, __pyx_L1_error)
  __pyx_type_8skgarden_8mondrian_4tree_5_tree_TreeBuilder.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_8skgarden_8mondrian_4tree_5_tree_TreeBuilder.tp_dictoffset && __pyx_type_8skgarden_8mondrian_4tree_5_tree_TreeBuilder.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_8skgarden_8mondrian_4tree_5_tree_TreeBuilder.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_8skgarden_8mondrian_4tree_5_tree_TreeBuilder.tp_dict, __pyx_vtabptr_8skgarden_8mondrian_4tree_5_tree_TreeBuilder) < 0) __PYX_ERR(0, 101, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_TreeBuilder, (PyObject *)&__pyx_type_8skgarden_8mondrian_4tree_5_tree_TreeBuilder) < 0) __PYX_ERR(0, 101, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_8skgarden_8mondrian_4tree_5_tree_TreeBuilder) < 0) __PYX_ERR(0, 101, __pyx_L1_error)
  __pyx_ptype_8skgarden_8mondrian_4tree_5_tree_TreeBuilder = &__pyx_type_8skgarden_8mondrian_4tree_5_tree_TreeBuilder;
  __pyx_vtabptr_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder = &__pyx_vtable_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder;
  __pyx_vtable_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder.__pyx_base = *__pyx_vtabptr_8skgarden_8mondrian_4tree_5_tree_TreeBuilder;
  __pyx_vtable_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder.__pyx_base.build = (PyObject *(*)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *, struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, PyObject *, PyArrayObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_build *__pyx_optional_args))__pyx_f_8skgarden_8mondrian_4tree_5_tree_21PartialFitTreeBuilder_build;
  __pyx_type_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder.tp_base = __pyx_ptype_8skgarden_8mondrian_4tree_5_tree_TreeBuilder;
  if (PyType_Ready(&__pyx_type_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder) < 0) __PYX_ERR(0, 128, __pyx_L1_error)
  __pyx_type_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder.tp_dictoffset && __pyx_type_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder.tp_dict, __pyx_vtabptr_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder) < 0) __PYX_ERR(0, 128, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PartialFitTreeBuilder, (PyObject *)&__pyx_type_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder) < 0) __PYX_ERR(0, 128, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder) < 0) __PYX_ERR(0, 128, __pyx_L1_error)
  __pyx_ptype_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder = &__pyx_type_8skgarden_8mondrian_4tree_5_tree_PartialFitTreeBuilder;
  __pyx_vtabptr_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder = &__pyx_vtable_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder;
  __pyx_vtable_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder.__pyx_base = *__pyx_vtabptr_8skgarden_8mondrian_4tree_5_tree_TreeBuilder;
  __pyx_vtable_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder.__pyx_base.build = (PyObject *(*)(struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_TreeBuilder *, struct __pyx_obj_8skgarden_8mondrian_4tree_5_tree_Tree *, PyObject *, PyArrayObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_8skgarden_8mondrian_4tree_5_tree_11TreeBuilder_build *__pyx_optional_args))__pyx_f_8skgarden_8mondrian_4tree_5_tree_21DepthFirstTreeBuilder_build;
  __pyx_type_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder.tp_base = __pyx_ptype_8skgarden_8mondrian_4tree_5_tree_TreeBuilder;
  if (PyType_Ready(&__pyx_type_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder) < 0) __PYX_ERR(0, 175, __pyx_L1_error)
  __pyx_type_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder.tp_dictoffset && __pyx_type_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder.tp_dict, __pyx_vtabptr_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder) < 0) __PYX_ERR(0, 175, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_DepthFirstTreeBuilder, (PyObject *)&__pyx_type_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder) < 0) __PYX_ERR(0, 175, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder) < 0) __PYX_ERR(0, 175, __pyx_L1_error)
  __pyx_ptype_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder = &__pyx_type_8skgarden_8mondrian_4tree_5_tree_DepthFirstTreeBuilder;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(4, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4bool_bool) __PYX_ERR(5, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_7complex_complex) __PYX_ERR(6, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(3, 206, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(3, 229, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(3, 233, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(3, 242, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(3, 918, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("skgarden.mondrian.tree._criterion"); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_8skgarden_8mondrian_4tree_10_criterion_Criterion = __Pyx_ImportType(__pyx_t_1, "skgarden.mondrian.tree._criterion", "Criterion", sizeof(struct __pyx_obj_8skgarden_8mondrian_4tree_10_criterion_Criterion), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_8skgarden_8mondrian_4tree_10_criterion_Criterion) __PYX_ERR(7, 21, __pyx_L1_error)
  __pyx_vtabptr_8skgarden_8mondrian_4tree_10_criterion_Criterion = (struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_10_criterion_Criterion*)__Pyx_GetVtable(__pyx_ptype_8skgarden_8mondrian_4tree_10_criterion_Criterion->tp_dict); if (unlikely(!__pyx_vtabptr_8skgarden_8mondrian_4tree_10_criterion_Criterion)) __PYX_ERR(7, 21, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("skgarden.mondrian.tree._splitter"); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_8skgarden_8mondrian_4tree_9_splitter_Splitter = __Pyx_ImportType(__pyx_t_1, "skgarden.mondrian.tree._splitter", "Splitter", sizeof(struct __pyx_obj_8skgarden_8mondrian_4tree_9_splitter_Splitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_8skgarden_8mondrian_4tree_9_splitter_Splitter) __PYX_ERR(8, 35, __pyx_L1_error)
  __pyx_vtabptr_8skgarden_8mondrian_4tree_9_splitter_Splitter = (struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_9_splitter_Splitter*)__Pyx_GetVtable(__pyx_ptype_8skgarden_8mondrian_4tree_9_splitter_Splitter->tp_dict); if (unlikely(!__pyx_vtabptr_8skgarden_8mondrian_4tree_9_splitter_Splitter)) __PYX_ERR(8, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("skgarden.mondrian.tree._utils"); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_8skgarden_8mondrian_4tree_6_utils_Stack = __Pyx_ImportType(__pyx_t_1, "skgarden.mondrian.tree._utils", "Stack", sizeof(struct __pyx_obj_8skgarden_8mondrian_4tree_6_utils_Stack), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_8skgarden_8mondrian_4tree_6_utils_Stack) __PYX_ERR(9, 78, __pyx_L1_error)
  __pyx_vtabptr_8skgarden_8mondrian_4tree_6_utils_Stack = (struct __pyx_vtabstruct_8skgarden_8mondrian_4tree_6_utils_Stack*)__Pyx_GetVtable(__pyx_ptype_8skgarden_8mondrian_4tree_6_utils_Stack->tp_dict); if (unlikely(!__pyx_vtabptr_8skgarden_8mondrian_4tree_6_utils_Stack)) __PYX_ERR(9, 78, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("skgarden.mondrian.tree._utils"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "sizet_ptr_to_ndarray", (void (**)(void))&__pyx_f_8skgarden_8mondrian_4tree_6_utils_sizet_ptr_to_ndarray, "PyArrayObject *(__pyx_t_8skgarden_8mondrian_4tree_6_utils_SIZE_t *, __pyx_t_8skgarden_8mondrian_4tree_6_utils_SIZE_t)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "rand_uniform", (void (**)(void))&__pyx_f_8skgarden_8mondrian_4tree_6_utils_rand_uniform, "double (double, double, __pyx_t_8skgarden_8mondrian_4tree_6_utils_UINT32_t *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "rand_multinomial", (void (**)(void))&__pyx_f_8skgarden_8mondrian_4tree_6_utils_rand_multinomial, "__pyx_t_8skgarden_8mondrian_4tree_6_utils_SIZE_t (__pyx_t_8skgarden_8mondrian_4tree_6_utils_DTYPE_t *, __pyx_t_8skgarden_8mondrian_4tree_6_utils_SIZE_t, __pyx_t_8skgarden_8mondrian_4tree_6_utils_UINT32_t *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "rand_exponential", (void (**)(void))&__pyx_f_8skgarden_8mondrian_4tree_6_utils_rand_exponential, "double (__pyx_t_8skgarden_8mondrian_4tree_6_utils_DTYPE_t, __pyx_t_8skgarden_8mondrian_4tree_6_utils_UINT32_t *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "__pyx_fuse_0safe_realloc", (void (**)(void))&__pyx_fuse_0__pyx_f_8skgarden_8mondrian_4tree_6_utils_safe_realloc, "__pyx_t_8skgarden_8mondrian_4tree_6_utils_DTYPE_t *(__pyx_t_8skgarden_8mondrian_4tree_6_utils_DTYPE_t **, size_t)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "__pyx_fuse_1safe_realloc", (void (**)(void))&__pyx_fuse_1__pyx_f_8skgarden_8mondrian_4tree_6_utils_safe_realloc, "__pyx_t_8skgarden_8mondrian_4tree_6_utils_SIZE_t *(__pyx_t_8skgarden_8mondrian_4tree_6_utils_SIZE_t **, size_t)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "__pyx_fuse_2safe_realloc", (void (**)(void))&__pyx_fuse_2__pyx_f_8skgarden_8mondrian_4tree_6_utils_safe_realloc, "unsigned char *(unsigned char **, size_t)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "__pyx_fuse_3safe_realloc", (void (**)(void))&__pyx_fuse_3__pyx_f_8skgarden_8mondrian_4tree_6_utils_safe_realloc, "__pyx_t_8skgarden_8mondrian_4tree_6_utils_DOUBLE_t *(__pyx_t_8skgarden_8mondrian_4tree_6_utils_DOUBLE_t **, size_t)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "__pyx_fuse_4safe_realloc", (void (**)(void))&__pyx_fuse_4__pyx_f_8skgarden_8mondrian_4tree_6_utils_safe_realloc, "__pyx_t_8skgarden_8mondrian_4tree_6_utils_DOUBLE_t **(__pyx_t_8skgarden_8mondrian_4tree_6_utils_DOUBLE_t ***, size_t)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "__pyx_fuse_5safe_realloc", (void (**)(void))&__pyx_fuse_5__pyx_f_8skgarden_8mondrian_4tree_6_utils_safe_realloc, "struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node *(struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node **, size_t)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "__pyx_fuse_6safe_realloc", (void (**)(void))&__pyx_fuse_6__pyx_f_8skgarden_8mondrian_4tree_6_utils_safe_realloc, "struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node **(struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node ***, size_t)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "__pyx_fuse_7safe_realloc", (void (**)(void))&__pyx_fuse_7__pyx_f_8skgarden_8mondrian_4tree_6_utils_safe_realloc, "struct __pyx_t_8skgarden_8mondrian_4tree_6_utils_StackRecord *(struct __pyx_t_8skgarden_8mondrian_4tree_6_utils_StackRecord **, size_t)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#if PY_MAJOR_VERSION < 3
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC void
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#else
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC init_tree(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC init_tree(void)
#else
__Pyx_PyMODINIT_FUNC PyInit__tree(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit__tree(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        result = PyDict_SetItemString(moddict, to_name, value);
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__") < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__") < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__") < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__") < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec__tree(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  double __pyx_t_4;
  __pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module '_tree' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit__tree(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("_tree", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_skgarden__mondrian__tree___tree) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "skgarden.mondrian.tree._tree")) {
      if (unlikely(PyDict_SetItemString(modules, "skgarden.mondrian.tree._tree", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_type_import_code() != 0)) goto __pyx_L1_error;
  (void)__Pyx_modinit_variable_import_code();
  if (unlikely(__Pyx_modinit_function_import_code() != 0)) goto __pyx_L1_error;
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "skgarden/mondrian/tree/_tree.pyx":28
 * from libc.string cimport memset
 * 
 * import numpy as np             # <<<<<<<<<<<<<<
 * cimport numpy as np
 * np.import_array()
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numpy, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_1) < 0) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":30
 * import numpy as np
 * cimport numpy as np
 * np.import_array()             # <<<<<<<<<<<<<<
 * 
 * from scipy.sparse import csr_matrix
 */
  __pyx_t_2 = __pyx_f_5numpy_import_array(); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 30, __pyx_L1_error)

  /* "skgarden/mondrian/tree/_tree.pyx":32
 * np.import_array()
 * 
 * from scipy.sparse import csr_matrix             # <<<<<<<<<<<<<<
 * 
 * from ._utils cimport Stack
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_csr_matrix);
  __Pyx_GIVEREF(__pyx_n_s_csr_matrix);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_csr_matrix);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_scipy_sparse, __pyx_t_1, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_csr_matrix); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_csr_matrix, __pyx_t_1) < 0) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":53
 * # =============================================================================
 * 
 * from numpy import float32 as DTYPE             # <<<<<<<<<<<<<<
 * from numpy import float64 as DOUBLE
 * 
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_float32);
  __Pyx_GIVEREF(__pyx_n_s_float32);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_float32);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numpy, __pyx_t_3, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_float32); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_DTYPE, __pyx_t_3) < 0) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":54
 * 
 * from numpy import float32 as DTYPE
 * from numpy import float64 as DOUBLE             # <<<<<<<<<<<<<<
 * 
 * cdef double INFINITY = np.inf
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_float64);
  __Pyx_GIVEREF(__pyx_n_s_float64);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_float64);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_numpy, __pyx_t_1, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_float64); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_DOUBLE, __pyx_t_1) < 0) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":56
 * from numpy import float64 as DOUBLE
 * 
 * cdef double INFINITY = np.inf             # <<<<<<<<<<<<<<
 * 
 * # Some handy constants (BestFirstTreeBuilder)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_inf); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_8skgarden_8mondrian_4tree_5_tree_INFINITY = __pyx_t_4;

  /* "skgarden/mondrian/tree/_tree.pyx":59
 * 
 * # Some handy constants (BestFirstTreeBuilder)
 * cdef int IS_FIRST = 1             # <<<<<<<<<<<<<<
 * cdef int IS_NOT_FIRST = 0
 * cdef int IS_LEFT = 1
 */
  __pyx_v_8skgarden_8mondrian_4tree_5_tree_IS_FIRST = 1;

  /* "skgarden/mondrian/tree/_tree.pyx":60
 * # Some handy constants (BestFirstTreeBuilder)
 * cdef int IS_FIRST = 1
 * cdef int IS_NOT_FIRST = 0             # <<<<<<<<<<<<<<
 * cdef int IS_LEFT = 1
 * cdef int IS_NOT_LEFT = 0
 */
  __pyx_v_8skgarden_8mondrian_4tree_5_tree_IS_NOT_FIRST = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":61
 * cdef int IS_FIRST = 1
 * cdef int IS_NOT_FIRST = 0
 * cdef int IS_LEFT = 1             # <<<<<<<<<<<<<<
 * cdef int IS_NOT_LEFT = 0
 * 
 */
  __pyx_v_8skgarden_8mondrian_4tree_5_tree_IS_LEFT = 1;

  /* "skgarden/mondrian/tree/_tree.pyx":62
 * cdef int IS_NOT_FIRST = 0
 * cdef int IS_LEFT = 1
 * cdef int IS_NOT_LEFT = 0             # <<<<<<<<<<<<<<
 * 
 * TREE_LEAF = -1
 */
  __pyx_v_8skgarden_8mondrian_4tree_5_tree_IS_NOT_LEFT = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":64
 * cdef int IS_NOT_LEFT = 0
 * 
 * TREE_LEAF = -1             # <<<<<<<<<<<<<<
 * TREE_UNDEFINED = -2
 * cdef SIZE_t _TREE_LEAF = TREE_LEAF
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TREE_LEAF, __pyx_int_neg_1) < 0) __PYX_ERR(0, 64, __pyx_L1_error)

  /* "skgarden/mondrian/tree/_tree.pyx":65
 * 
 * TREE_LEAF = -1
 * TREE_UNDEFINED = -2             # <<<<<<<<<<<<<<
 * cdef SIZE_t _TREE_LEAF = TREE_LEAF
 * cdef SIZE_t _TREE_UNDEFINED = TREE_UNDEFINED
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TREE_UNDEFINED, __pyx_int_neg_2) < 0) __PYX_ERR(0, 65, __pyx_L1_error)

  /* "skgarden/mondrian/tree/_tree.pyx":66
 * TREE_LEAF = -1
 * TREE_UNDEFINED = -2
 * cdef SIZE_t _TREE_LEAF = TREE_LEAF             # <<<<<<<<<<<<<<
 * cdef SIZE_t _TREE_UNDEFINED = TREE_UNDEFINED
 * cdef SIZE_t INITIAL_STACK_SIZE = 10
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_TREE_LEAF); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_8skgarden_8mondrian_4tree_5_tree__TREE_LEAF = __pyx_t_5;

  /* "skgarden/mondrian/tree/_tree.pyx":67
 * TREE_UNDEFINED = -2
 * cdef SIZE_t _TREE_LEAF = TREE_LEAF
 * cdef SIZE_t _TREE_UNDEFINED = TREE_UNDEFINED             # <<<<<<<<<<<<<<
 * cdef SIZE_t INITIAL_STACK_SIZE = 10
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_TREE_UNDEFINED); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_Py_intptr_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_8skgarden_8mondrian_4tree_5_tree__TREE_UNDEFINED = __pyx_t_5;

  /* "skgarden/mondrian/tree/_tree.pyx":68
 * cdef SIZE_t _TREE_LEAF = TREE_LEAF
 * cdef SIZE_t _TREE_UNDEFINED = TREE_UNDEFINED
 * cdef SIZE_t INITIAL_STACK_SIZE = 10             # <<<<<<<<<<<<<<
 * 
 * # Repeat struct definition for numpy
 */
  __pyx_v_8skgarden_8mondrian_4tree_5_tree_INITIAL_STACK_SIZE = 10;

  /* "skgarden/mondrian/tree/_tree.pyx":72
 * # Repeat struct definition for numpy
 * NODE_DTYPE = np.dtype({
 *     'names': ['left_child', 'right_child', 'feature', 'threshold', 'impurity',             # <<<<<<<<<<<<<<
 *               'n_node_samples', 'weighted_n_node_samples', 'tau', 'variance'],
 *     'formats': [np.intp, np.intp, np.intp, np.float64, np.float64, np.intp,
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyList_New(9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_left_child);
  __Pyx_GIVEREF(__pyx_n_s_left_child);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_left_child);
  __Pyx_INCREF(__pyx_n_s_right_child);
  __Pyx_GIVEREF(__pyx_n_s_right_child);
  PyList_SET_ITEM(__pyx_t_3, 1, __pyx_n_s_right_child);
  __Pyx_INCREF(__pyx_n_s_feature);
  __Pyx_GIVEREF(__pyx_n_s_feature);
  PyList_SET_ITEM(__pyx_t_3, 2, __pyx_n_s_feature);
  __Pyx_INCREF(__pyx_n_s_threshold);
  __Pyx_GIVEREF(__pyx_n_s_threshold);
  PyList_SET_ITEM(__pyx_t_3, 3, __pyx_n_s_threshold);
  __Pyx_INCREF(__pyx_n_s_impurity);
  __Pyx_GIVEREF(__pyx_n_s_impurity);
  PyList_SET_ITEM(__pyx_t_3, 4, __pyx_n_s_impurity);
  __Pyx_INCREF(__pyx_n_s_n_node_samples);
  __Pyx_GIVEREF(__pyx_n_s_n_node_samples);
  PyList_SET_ITEM(__pyx_t_3, 5, __pyx_n_s_n_node_samples);
  __Pyx_INCREF(__pyx_n_s_weighted_n_node_samples);
  __Pyx_GIVEREF(__pyx_n_s_weighted_n_node_samples);
  PyList_SET_ITEM(__pyx_t_3, 6, __pyx_n_s_weighted_n_node_samples);
  __Pyx_INCREF(__pyx_n_s_tau);
  __Pyx_GIVEREF(__pyx_n_s_tau);
  PyList_SET_ITEM(__pyx_t_3, 7, __pyx_n_s_tau);
  __Pyx_INCREF(__pyx_n_s_variance);
  __Pyx_GIVEREF(__pyx_n_s_variance);
  PyList_SET_ITEM(__pyx_t_3, 8, __pyx_n_s_variance);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_names, __pyx_t_3) < 0) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":74
 *     'names': ['left_child', 'right_child', 'feature', 'threshold', 'impurity',
 *               'n_node_samples', 'weighted_n_node_samples', 'tau', 'variance'],
 *     'formats': [np.intp, np.intp, np.intp, np.float64, np.float64, np.intp,             # <<<<<<<<<<<<<<
 *                 np.float64, np.float32, np.float64],
 *     'offsets': [
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_intp); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_intp); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_intp); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_float64); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_float64); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_intp); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":75
 *               'n_node_samples', 'weighted_n_node_samples', 'tau', 'variance'],
 *     'formats': [np.intp, np.intp, np.intp, np.float64, np.float64, np.intp,
 *                 np.float64, np.float32, np.float64],             # <<<<<<<<<<<<<<
 *     'offsets': [
 *         <Py_ssize_t> &(<Node*> NULL).left_child,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_float64); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_float32); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_float64); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":74
 *     'names': ['left_child', 'right_child', 'feature', 'threshold', 'impurity',
 *               'n_node_samples', 'weighted_n_node_samples', 'tau', 'variance'],
 *     'formats': [np.intp, np.intp, np.intp, np.float64, np.float64, np.intp,             # <<<<<<<<<<<<<<
 *                 np.float64, np.float32, np.float64],
 *     'offsets': [
 */
  __pyx_t_3 = PyList_New(9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_6);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyList_SET_ITEM(__pyx_t_3, 1, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyList_SET_ITEM(__pyx_t_3, 2, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyList_SET_ITEM(__pyx_t_3, 3, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyList_SET_ITEM(__pyx_t_3, 4, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyList_SET_ITEM(__pyx_t_3, 5, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_12);
  PyList_SET_ITEM(__pyx_t_3, 6, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_13);
  PyList_SET_ITEM(__pyx_t_3, 7, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_14);
  PyList_SET_ITEM(__pyx_t_3, 8, __pyx_t_14);
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_formats, __pyx_t_3) < 0) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":77
 *                 np.float64, np.float32, np.float64],
 *     'offsets': [
 *         <Py_ssize_t> &(<Node*> NULL).left_child,             # <<<<<<<<<<<<<<
 *         <Py_ssize_t> &(<Node*> NULL).right_child,
 *         <Py_ssize_t> &(<Node*> NULL).feature,
 */
  __pyx_t_3 = PyInt_FromSsize_t(((Py_ssize_t)(&((struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node *)NULL)->left_child))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "skgarden/mondrian/tree/_tree.pyx":78
 *     'offsets': [
 *         <Py_ssize_t> &(<Node*> NULL).left_child,
 *         <Py_ssize_t> &(<Node*> NULL).right_child,             # <<<<<<<<<<<<<<
 *         <Py_ssize_t> &(<Node*> NULL).feature,
 *         <Py_ssize_t> &(<Node*> NULL).threshold,
 */
  __pyx_t_14 = PyInt_FromSsize_t(((Py_ssize_t)(&((struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node *)NULL)->right_child))); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);

  /* "skgarden/mondrian/tree/_tree.pyx":79
 *         <Py_ssize_t> &(<Node*> NULL).left_child,
 *         <Py_ssize_t> &(<Node*> NULL).right_child,
 *         <Py_ssize_t> &(<Node*> NULL).feature,             # <<<<<<<<<<<<<<
 *         <Py_ssize_t> &(<Node*> NULL).threshold,
 *         <Py_ssize_t> &(<Node*> NULL).impurity,
 */
  __pyx_t_13 = PyInt_FromSsize_t(((Py_ssize_t)(&((struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node *)NULL)->feature))); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);

  /* "skgarden/mondrian/tree/_tree.pyx":80
 *         <Py_ssize_t> &(<Node*> NULL).right_child,
 *         <Py_ssize_t> &(<Node*> NULL).feature,
 *         <Py_ssize_t> &(<Node*> NULL).threshold,             # <<<<<<<<<<<<<<
 *         <Py_ssize_t> &(<Node*> NULL).impurity,
 *         <Py_ssize_t> &(<Node*> NULL).n_node_samples,
 */
  __pyx_t_12 = PyInt_FromSsize_t(((Py_ssize_t)(&((struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node *)NULL)->threshold))); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);

  /* "skgarden/mondrian/tree/_tree.pyx":81
 *         <Py_ssize_t> &(<Node*> NULL).feature,
 *         <Py_ssize_t> &(<Node*> NULL).threshold,
 *         <Py_ssize_t> &(<Node*> NULL).impurity,             # <<<<<<<<<<<<<<
 *         <Py_ssize_t> &(<Node*> NULL).n_node_samples,
 *         <Py_ssize_t> &(<Node*> NULL).weighted_n_node_samples,
 */
  __pyx_t_11 = PyInt_FromSsize_t(((Py_ssize_t)(&((struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node *)NULL)->impurity))); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);

  /* "skgarden/mondrian/tree/_tree.pyx":82
 *         <Py_ssize_t> &(<Node*> NULL).threshold,
 *         <Py_ssize_t> &(<Node*> NULL).impurity,
 *         <Py_ssize_t> &(<Node*> NULL).n_node_samples,             # <<<<<<<<<<<<<<
 *         <Py_ssize_t> &(<Node*> NULL).weighted_n_node_samples,
 *         <Py_ssize_t> &(<Node*> NULL).tau,
 */
  __pyx_t_10 = PyInt_FromSsize_t(((Py_ssize_t)(&((struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node *)NULL)->n_node_samples))); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);

  /* "skgarden/mondrian/tree/_tree.pyx":83
 *         <Py_ssize_t> &(<Node*> NULL).impurity,
 *         <Py_ssize_t> &(<Node*> NULL).n_node_samples,
 *         <Py_ssize_t> &(<Node*> NULL).weighted_n_node_samples,             # <<<<<<<<<<<<<<
 *         <Py_ssize_t> &(<Node*> NULL).tau,
 *         <Py_ssize_t> &(<Node*> NULL).variance,
 */
  __pyx_t_9 = PyInt_FromSsize_t(((Py_ssize_t)(&((struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node *)NULL)->weighted_n_node_samples))); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "skgarden/mondrian/tree/_tree.pyx":84
 *         <Py_ssize_t> &(<Node*> NULL).n_node_samples,
 *         <Py_ssize_t> &(<Node*> NULL).weighted_n_node_samples,
 *         <Py_ssize_t> &(<Node*> NULL).tau,             # <<<<<<<<<<<<<<
 *         <Py_ssize_t> &(<Node*> NULL).variance,
 *     ]
 */
  __pyx_t_8 = PyInt_FromSsize_t(((Py_ssize_t)(&((struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node *)NULL)->tau))); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "skgarden/mondrian/tree/_tree.pyx":85
 *         <Py_ssize_t> &(<Node*> NULL).weighted_n_node_samples,
 *         <Py_ssize_t> &(<Node*> NULL).tau,
 *         <Py_ssize_t> &(<Node*> NULL).variance,             # <<<<<<<<<<<<<<
 *     ]
 * })
 */
  __pyx_t_7 = PyInt_FromSsize_t(((Py_ssize_t)(&((struct __pyx_t_8skgarden_8mondrian_4tree_5_tree_Node *)NULL)->variance))); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "skgarden/mondrian/tree/_tree.pyx":76
 *     'formats': [np.intp, np.intp, np.intp, np.float64, np.float64, np.intp,
 *                 np.float64, np.float32, np.float64],
 *     'offsets': [             # <<<<<<<<<<<<<<
 *         <Py_ssize_t> &(<Node*> NULL).left_child,
 *         <Py_ssize_t> &(<Node*> NULL).right_child,
 */
  __pyx_t_6 = PyList_New(9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_3);
  PyList_SET_ITEM(__pyx_t_6, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_14);
  PyList_SET_ITEM(__pyx_t_6, 1, __pyx_t_14);
  __Pyx_GIVEREF(__pyx_t_13);
  PyList_SET_ITEM(__pyx_t_6, 2, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_12);
  PyList_SET_ITEM(__pyx_t_6, 3, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_11);
  PyList_SET_ITEM(__pyx_t_6, 4, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_10);
  PyList_SET_ITEM(__pyx_t_6, 5, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_9);
  PyList_SET_ITEM(__pyx_t_6, 6, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_8);
  PyList_SET_ITEM(__pyx_t_6, 7, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_7);
  PyList_SET_ITEM(__pyx_t_6, 8, __pyx_t_7);
  __pyx_t_3 = 0;
  __pyx_t_14 = 0;
  __pyx_t_13 = 0;
  __pyx_t_12 = 0;
  __pyx_t_11 = 0;
  __pyx_t_10 = 0;
  __pyx_t_9 = 0;
  __pyx_t_8 = 0;
  __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_offsets, __pyx_t_6) < 0) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":71
 * 
 * # Repeat struct definition for numpy
 * NODE_DTYPE = np.dtype({             # <<<<<<<<<<<<<<
 *     'names': ['left_child', 'right_child', 'feature', 'threshold', 'impurity',
 *               'n_node_samples', 'weighted_n_node_samples', 'tau', 'variance'],
 */
  __pyx_t_6 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5numpy_dtype), __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_NODE_DTYPE, __pyx_t_6) < 0) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":520
 *     # XXX using (size_t)(-1) is ugly, but SIZE_MAX is not available in C89
 *     # (i.e., older MSVC).
 *     cdef int _resize_c(self, SIZE_t capacity=<SIZE_t>(-1)) nogil except -1:             # <<<<<<<<<<<<<<
 *         """Guts of _resize
 * 
 */
  __pyx_k__7 = ((__pyx_t_8skgarden_8mondrian_4tree_5_tree_SIZE_t)-1L);

  /* "(tree fragment)":1
 * def __pyx_unpickle_TreeBuilder(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     if __pyx_checksum != 0xedae8f4:
 *         from pickle import PickleError as __pyx_PickleError
 */
  __pyx_t_6 = PyCFunction_NewEx(&__pyx_mdef_8skgarden_8mondrian_4tree_5_tree_1__pyx_unpickle_TreeBuilder, NULL, __pyx_n_s_skgarden_mondrian_tree__tree); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_TreeBuilder, __pyx_t_6) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "skgarden/mondrian/tree/_tree.pyx":1
 * # cython: cdivision=True             # <<<<<<<<<<<<<<
 * # cython: boundscheck=False
 * # cython: wraparound=False
 */
  __pyx_t_6 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_6) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "../_h_env/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1046
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init skgarden.mondrian.tree._tree", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init skgarden.mondrian.tree._tree");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* IsLittleEndian */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void)
{
  union {
    uint32_t u32;
    uint8_t u8[4];
  } S;
  S.u32 = 0x01020304;
  return S.u8[0] == 4;
}

/* BufferFormatCheck */
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type) {
  stack[0].field = &ctx->root;
  stack[0].parent_offset = 0;
  ctx->root.type = type;
  ctx->root.name = "buffer dtype";
  ctx->root.offset = 0;
  ctx->head = stack;
  ctx->head->field = &ctx->root;
  ctx->fmt_offset = 0;
  ctx->head->parent_offset = 0;
  ctx->new_packmode = '@';
  ctx->enc_packmode = '@';
  ctx->new_count = 1;
  ctx->enc_count = 0;
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  ctx->is_valid_array = 0;
  ctx->struct_alignment = 0;
  while (type->typegroup == 'S') {
    ++ctx->head;
    ctx->head->field = type->fields;
    ctx->head->parent_offset = 0;
    type = type->fields->type;
  }
}
static int __Pyx_BufFmt_ParseNumber(const char** ts) {
    int count;
    const char* t = *ts;
    if (*t < '0' || *t > '9') {
      return -1;
    } else {
        count = *t++ - '0';
        while (*t >= '0' && *t < '9') {
            count *= 10;
            count += *t++ - '0';
        }
    }
    *ts = t;
    return count;
}
static int __Pyx_BufFmt_ExpectNumber(const char **ts) {
    int number = __Pyx_BufFmt_ParseNumber(ts);
    if (number == -1)
        PyErr_Format(PyExc_ValueError,\
                     "Does not understand character buffer dtype format string ('%c')", **ts);
    return number;
}
static void __Pyx_BufFmt_RaiseUnexpectedChar(char ch) {
  PyErr_Format(PyExc_ValueError,
               "Unexpected format string character: '%c'", ch);
}
static const char* __Pyx_BufFmt_DescribeTypeChar(char ch, int is_complex) {
  switch (ch) {
    case 'c': return "'char'";
    case 'b': return "'signed char'";
    case 'B': return "'unsigned char'";
    case 'h': return "'short'";
    case 'H': return "'unsigned short'";
    case 'i': return "'int'";
    case 'I': return "'unsigned int'";
    case 'l': return "'long'";
    case 'L': return "'unsigned long'";
    case 'q': return "'long long'";
    case 'Q': return "'unsigned long long'";
    case 'f': return (is_complex ? "'complex float'" : "'float'");
    case 'd': return (is_complex ? "'complex double'" : "'double'");
    case 'g': return (is_complex ? "'complex long double'" : "'long double'");
    case 'T': return "a struct";
    case 'O': return "Python object";
    case 'P': return "a pointer";
    case 's': case 'p': return "a string";
    case 0: return "end";
    default: return "unparseable format string";
  }
}
static size_t __Pyx_BufFmt_TypeCharToStandardSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return 2;
    case 'i': case 'I': case 'l': case 'L': return 4;
    case 'q': case 'Q': return 8;
    case 'f': return (is_complex ? 8 : 4);
    case 'd': return (is_complex ? 16 : 8);
    case 'g': {
      PyErr_SetString(PyExc_ValueError, "Python does not define a standard format string size for long double ('g')..");
      return 0;
    }
    case 'O': case 'P': return sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static size_t __Pyx_BufFmt_TypeCharToNativeSize(char ch, int is_complex) {
  switch (ch) {
    case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(short);
    case 'i': case 'I': return sizeof(int);
    case 'l': case 'L': return sizeof(long);
    #ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(PY_LONG_LONG);
    #endif
    case 'f': return sizeof(float) * (is_complex ? 2 : 1);
    case 'd': return sizeof(double) * (is_complex ? 2 : 1);
    case 'g': return sizeof(long double) * (is_complex ? 2 : 1);
    case 'O': case 'P': return sizeof(void*);
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
typedef struct { char c; short x; } __Pyx_st_short;
typedef struct { char c; int x; } __Pyx_st_int;
typedef struct { char c; long x; } __Pyx_st_long;
typedef struct { char c; float x; } __Pyx_st_float;
typedef struct { char c; double x; } __Pyx_st_double;
typedef struct { char c; long double x; } __Pyx_st_longdouble;
typedef struct { char c; void *x; } __Pyx_st_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { char c; PY_LONG_LONG x; } __Pyx_st_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToAlignment(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_st_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_st_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_st_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_st_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_st_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_st_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_st_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_st_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
/* These are for computing the padding at the end of the struct to align
   on the first member of the struct. This will probably the same as above,
   but we don't have any guarantees.
 */
typedef struct { short x; char c; } __Pyx_pad_short;
typedef struct { int x; char c; } __Pyx_pad_int;
typedef struct { long x; char c; } __Pyx_pad_long;
typedef struct { float x; char c; } __Pyx_pad_float;
typedef struct { double x; char c; } __Pyx_pad_double;
typedef struct { long double x; char c; } __Pyx_pad_longdouble;
typedef struct { void *x; char c; } __Pyx_pad_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { PY_LONG_LONG x; char c; } __Pyx_pad_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToPadding(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_pad_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_pad_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_pad_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_pad_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_pad_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_pad_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_pad_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_pad_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static char __Pyx_BufFmt_TypeCharToGroup(char ch, int is_complex) {
  switch (ch) {
    case 'c':
        return 'H';
    case 'b': case 'h': case 'i':
    case 'l': case 'q': case 's': case 'p':
        return 'I';
    case 'B': case 'H': case 'I': case 'L': case 'Q':
        return 'U';
    case 'f': case 'd': case 'g':
        return (is_complex ? 'C' : 'R');
    case 'O':
        return 'O';
    case 'P':
        return 'P';
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
static void __Pyx_BufFmt_RaiseExpected(__Pyx_BufFmt_Context* ctx) {
  if (ctx->head == NULL || ctx->head->field == &ctx->root) {
    const char* expected;
    const char* quote;
    if (ctx->head == NULL) {
      expected = "end";
      quote = "";
    } else {
      expected = ctx->head->field->type->name;
      quote = "'";
    }
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected %s%s%s but got %s",
                 quote, expected, quote,
                 __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex));
  } else {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_StructField* parent = (ctx->head - 1)->field;
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected '%s' but got %s in '%s.%s'",
                 field->type->name, __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex),
                 parent->type->name, field->name);
  }
}
static int __Pyx_BufFmt_ProcessTypeChunk(__Pyx_BufFmt_Context* ctx) {
  char group;
  size_t size, offset, arraysize = 1;
  if (ctx->enc_type == 0) return 0;
  if (ctx->head->field->type->arraysize[0]) {
    int i, ndim = 0;
    if (ctx->enc_type == 's' || ctx->enc_type == 'p') {
        ctx->is_valid_array = ctx->head->field->type->ndim == 1;
        ndim = 1;
        if (ctx->enc_count != ctx->head->field->type->arraysize[0]) {
            PyErr_Format(PyExc_ValueError,
                         "Expected a dimension of size %zu, got %zu",
                         ctx->head->field->type->arraysize[0], ctx->enc_count);
            return -1;
        }
    }
    if (!ctx->is_valid_array) {
      PyErr_Format(PyExc_ValueError, "Expected %d dimensions, got %d",
                   ctx->head->field->type->ndim, ndim);
      return -1;
    }
    for (i = 0; i < ctx->head->field->type->ndim; i++) {
      arraysize *= ctx->head->field->type->arraysize[i];
    }
    ctx->is_valid_array = 0;
    ctx->enc_count = 1;
  }
  group = __Pyx_BufFmt_TypeCharToGroup(ctx->enc_type, ctx->is_complex);
  do {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_TypeInfo* type = field->type;
    if (ctx->enc_packmode == '@' || ctx->enc_packmode == '^') {
      size = __Pyx_BufFmt_TypeCharToNativeSize(ctx->enc_type, ctx->is_complex);
    } else {
      size = __Pyx_BufFmt_TypeCharToStandardSize(ctx->enc_type, ctx->is_complex);
    }
    if (ctx->enc_packmode == '@') {
      size_t align_at = __Pyx_BufFmt_TypeCharToAlignment(ctx->enc_type, ctx->is_complex);
      size_t align_mod_offset;
      if (align_at == 0) return -1;
      align_mod_offset = ctx->fmt_offset % align_at;
      if (align_mod_offset > 0) ctx->fmt_offset += align_at - align_mod_offset;
      if (ctx->struct_alignment == 0)
          ctx->struct_alignment = __Pyx_BufFmt_TypeCharToPadding(ctx->enc_type,
                                                                 ctx->is_complex);
    }
    if (type->size != size || type->typegroup != group) {
      if (type->typegroup == 'C' && type->fields != NULL) {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        ++ctx->head;
        ctx->head->field = type->fields;
        ctx->head->parent_offset = parent_offset;
        continue;
      }
      if ((type->typegroup == 'H' || group == 'H') && type->size == size) {
      } else {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
      }
    }
    offset = ctx->head->parent_offset + field->offset;
    if (ctx->fmt_offset != offset) {
      PyErr_Format(PyExc_ValueError,
                   "Buffer dtype mismatch; next field is at offset %" CYTHON_FORMAT_SSIZE_T "d but %" CYTHON_FORMAT_SSIZE_T "d expected",
                   (Py_ssize_t)ctx->fmt_offset, (Py_ssize_t)offset);
      return -1;
    }
    ctx->fmt_offset += size;
    if (arraysize)
      ctx->fmt_offset += (arraysize - 1) * size;
    --ctx->enc_count;
    while (1) {
      if (field == &ctx->root) {
        ctx->head = NULL;
        if (ctx->enc_count != 0) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
        }
        break;
      }
      ctx->head->field = ++field;
      if (field->type == NULL) {
        --ctx->head;
        field = ctx->head->field;
        continue;
      } else if (field->type->typegroup == 'S') {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        if (field->type->fields->type == NULL) continue;
        field = field->type->fields;
        ++ctx->head;
        ctx->head->field = field;
        ctx->head->parent_offset = parent_offset;
        break;
      } else {
        break;
      }
    }
  } while (ctx->enc_count);
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  return 0;
}
static PyObject *
__pyx_buffmt_parse_array(__Pyx_BufFmt_Context* ctx, const char** tsp)
{
    const char *ts = *tsp;
    int i = 0, number;
    int ndim = ctx->head->field->type->ndim;
;
    ++ts;
    if (ctx->new_count != 1) {
        PyErr_SetString(PyExc_ValueError,
                        "Cannot handle repeated arrays in format string");
        return NULL;
    }
    if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
    while (*ts && *ts != ')') {
        switch (*ts) {
            case ' ': case '\f': case '\r': case '\n': case '\t': case '\v':  continue;
            default:  break;
        }
        number = __Pyx_BufFmt_ExpectNumber(&ts);
        if (number == -1) return NULL;
        if (i < ndim && (size_t) number != ctx->head->field->type->arraysize[i])
            return PyErr_Format(PyExc_ValueError,
                        "Expected a dimension of size %zu, got %d",
                        ctx->head->field->type->arraysize[i], number);
        if (*ts != ',' && *ts != ')')
            return PyErr_Format(PyExc_ValueError,
                                "Expected a comma in format string, got '%c'", *ts);
        if (*ts == ',') ts++;
        i++;
    }
    if (i != ndim)
        return PyErr_Format(PyExc_ValueError, "Expected %d dimension(s), got %d",
                            ctx->head->field->type->ndim, i);
    if (!*ts) {
        PyErr_SetString(PyExc_ValueError,
                        "Unexpected end of format string, expected ')'");
        return NULL;
    }
    ctx->is_valid_array = 1;
    ctx->new_count = 1;
    *tsp = ++ts;
    return Py_None;
}
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts) {
  int got_Z = 0;
  while (1) {
    switch(*ts) {
      case 0:
        if (ctx->enc_type != 0 && ctx->head == NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        if (ctx->head != NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        return ts;
      case ' ':
      case '\r':
      case '\n':
        ++ts;
        break;
      case '<':
        if (!__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Little-endian buffer not supported on big-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '>':
      case '!':
        if (__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Big-endian buffer not supported on little-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '=':
      case '@':
      case '^':
        ctx->new_packmode = *ts++;
        break;
      case 'T':
        {
          const char* ts_after_sub;
          size_t i, struct_count = ctx->new_count;
          size_t struct_alignment = ctx->struct_alignment;
          ctx->new_count = 1;
          ++ts;
          if (*ts != '{') {
            PyErr_SetString(PyExc_ValueError, "Buffer acquisition: Expected '{' after 'T'");
            return NULL;
          }
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          ctx->enc_count = 0;
          ctx->struct_alignment = 0;
          ++ts;
          ts_after_sub = ts;
          for (i = 0; i != struct_count; ++i) {
            ts_after_sub = __Pyx_BufFmt_CheckString(ctx, ts);
            if (!ts_after_sub) return NULL;
          }
          ts = ts_after_sub;
          if (struct_alignment) ctx->struct_alignment = struct_alignment;
        }
        break;
      case '}':
        {
          size_t alignment = ctx->struct_alignment;
          ++ts;
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          if (alignment && ctx->fmt_offset % alignment) {
            ctx->fmt_offset += alignment - (ctx->fmt_offset % alignment);
          }
        }
        return ts;
      case 'x':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->fmt_offset += ctx->new_count;
        ctx->new_count = 1;
        ctx->enc_count = 0;
        ctx->enc_type = 0;
        ctx->enc_packmode = ctx->new_packmode;
        ++ts;
        break;
      case 'Z':
        got_Z = 1;
        ++ts;
        if (*ts != 'f' && *ts != 'd' && *ts != 'g') {
          __Pyx_BufFmt_RaiseUnexpectedChar('Z');
          return NULL;
        }
        CYTHON_FALLTHROUGH;
      case 'c': case 'b': case 'B': case 'h': case 'H': case 'i': case 'I':
      case 'l': case 'L': case 'q': case 'Q':
      case 'f': case 'd': case 'g':
      case 'O': case 'p':
        if (ctx->enc_type == *ts && got_Z == ctx->is_complex &&
            ctx->enc_packmode == ctx->new_packmode) {
          ctx->enc_count += ctx->new_count;
          ctx->new_count = 1;
          got_Z = 0;
          ++ts;
          break;
        }
        CYTHON_FALLTHROUGH;
      case 's':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->enc_count = ctx->new_count;
        ctx->enc_packmode = ctx->new_packmode;
        ctx->enc_type = *ts;
        ctx->is_complex = got_Z;
        ++ts;
        ctx->new_count = 1;
        got_Z = 0;
        break;
      case ':':
        ++ts;
        while(*ts != ':') ++ts;
        ++ts;
        break;
      case '(':
        if (!__pyx_buffmt_parse_array(ctx, &ts)) return NULL;
        break;
      default:
        {
          int number = __Pyx_BufFmt_ExpectNumber(&ts);
          if (number == -1) return NULL;
          ctx->new_count = (size_t)number;
        }
    }
  }
}

/* BufferGetAndValidate */
  static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info) {
  if (unlikely(info->buf == NULL)) return;
  if (info->suboffsets == __Pyx_minusones) info->suboffsets = NULL;
  __Pyx_ReleaseBuffer(info);
}
static void __Pyx_ZeroBuffer(Py_buffer* buf) {
  buf->buf = NULL;
  buf->obj = NULL;
  buf->strides = __Pyx_zeros;
  buf->shape = __Pyx_zeros;
  buf->suboffsets = __Pyx_minusones;
}
static int __Pyx__GetBufferAndValidate(
        Py_buffer* buf, PyObject* obj,  __Pyx_TypeInfo* dtype, int flags,
        int nd, int cast, __Pyx_BufFmt_StackElem* stack)
{
  buf->buf = NULL;
  if (unlikely(__Pyx_GetBuffer(obj, buf, flags) == -1)) {
    __Pyx_ZeroBuffer(buf);
    return -1;
  }
  if (unlikely(buf->ndim != nd)) {
    PyErr_Format(PyExc_ValueError,
                 "Buffer has wrong number of dimensions (expected %d, got %d)",
                 nd, buf->ndim);
    goto fail;
  }
  if (!cast) {
    __Pyx_BufFmt_Context ctx;
    __Pyx_BufFmt_Init(&ctx, stack, dtype);
    if (!__Pyx_BufFmt_CheckString(&ctx, buf->format)) goto fail;
  }
  if (unlikely((unsigned)buf->itemsize != dtype->size)) {
    PyErr_Format(PyExc_ValueError,
      "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "d byte%s) does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "d byte%s)",
      buf->itemsize, (buf->itemsize > 1) ? "s" : "",
      dtype->name, (Py_ssize_t)dtype->size, (dtype->size > 1) ? "s" : "");
    goto fail;
  }
  if (buf->suboffsets == NULL) buf->suboffsets = __Pyx_minusones;
  return 0;
fail:;
  __Pyx_SafeReleaseBuffer(buf);
  return -1;
}

/* PyObjectCall2Args */
  static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* DictGetItem */
  #if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* PyIntCompare */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_NeObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_FALSE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a != b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size != 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_TRUE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_TRUE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal != 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a != (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_NE));
}

/* WriteUnraisableException */
  static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* BufferFallbackError */
  static void __Pyx_RaiseBufferFallbackError(void) {
  PyErr_SetString(PyExc_ValueError,
     "Buffer acquisition failed on assignment; and then reacquiring the old buffer failed too!");
}

/* Import */
  static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
  static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* HasAttr */
  static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* RaiseNoneIterError */
  static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* GetTopmostException */
  #if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
  #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* GetException */
  #if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyObject_GenericGetAttrNoDict */
  #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
  #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
  static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* SetupReduce */
  static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_cython); if (unlikely(!reduce_cython)) goto BAD;
            ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto BAD;
            ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto BAD;
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate_cython); if (unlikely(!setstate_cython)) goto BAD;
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto BAD;
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto GOOD;
BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
  #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
  static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* CLineInTraceback */
  #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
  static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
  #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_Py_intptr_t(Py_intptr_t value) {
    const Py_intptr_t neg_one = (Py_intptr_t) ((Py_intptr_t) 0 - (Py_intptr_t) 1), const_zero = (Py_intptr_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(Py_intptr_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(Py_intptr_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(Py_intptr_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(Py_intptr_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(Py_intptr_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(Py_intptr_t),
                                     little, !is_unsigned);
    }
}

/* CIntFromPyVerify */
  #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

#if PY_MAJOR_VERSION < 3
static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags) {
    if (PyObject_CheckBuffer(obj)) return PyObject_GetBuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_ptype_5numpy_ndarray)) return __pyx_pw_5numpy_7ndarray_1__getbuffer__(obj, view, flags);
    PyErr_Format(PyExc_TypeError, "'%.200s' does not have the buffer interface", Py_TYPE(obj)->tp_name);
    return -1;
}
static void __Pyx_ReleaseBuffer(Py_buffer *view) {
    PyObject *obj = view->obj;
    if (!obj) return;
    if (PyObject_CheckBuffer(obj)) {
        PyBuffer_Release(view);
        return;
    }
    if ((0)) {}
        else if (__Pyx_TypeCheck(obj, __pyx_ptype_5numpy_ndarray)) __pyx_pw_5numpy_7ndarray_3__releasebuffer__(obj, view);
    view->obj = NULL;
    Py_DECREF(obj);
}
#endif


  /* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* None */
  static CYTHON_INLINE long __Pyx_pow_long(long b, long e) {
    long t = b;
    switch (e) {
        case 3:
            t *= b;
        CYTHON_FALLTHROUGH;
        case 2:
            t *= b;
        CYTHON_FALLTHROUGH;
        case 1:
            return t;
        case 0:
            return 1;
    }
    #if 1
    if (unlikely(e<0)) return 0;
    #endif
    t = 1;
    while (likely(e)) {
        t *= (b * (e&1)) | ((~e)&1);
        b *= b;
        e >>= 1;
    }
    return t;
}

/* None */
  static CYTHON_INLINE int __Pyx_ErrOccurredWithGIL(void) {
  int err;
  #ifdef WITH_THREAD
  PyGILState_STATE _save = PyGILState_Ensure();
  #endif
  err = !!PyErr_Occurred();
  #ifdef WITH_THREAD
  PyGILState_Release(_save);
  #endif
  return err;
}

/* Declarations */
  #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = 1.0 / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = 1.0 / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0, -1);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
  #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = 1.0 / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = 1.0 / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0, -1);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__NPY_TYPES(enum NPY_TYPES value) {
    const enum NPY_TYPES neg_one = (enum NPY_TYPES) ((enum NPY_TYPES) 0 - (enum NPY_TYPES) 1), const_zero = (enum NPY_TYPES) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum NPY_TYPES) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum NPY_TYPES) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum NPY_TYPES) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum NPY_TYPES) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum NPY_TYPES) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum NPY_TYPES),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
  static CYTHON_INLINE Py_intptr_t __Pyx_PyInt_As_Py_intptr_t(PyObject *x) {
    const Py_intptr_t neg_one = (Py_intptr_t) ((Py_intptr_t) 0 - (Py_intptr_t) 1), const_zero = (Py_intptr_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(Py_intptr_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(Py_intptr_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (Py_intptr_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (Py_intptr_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(Py_intptr_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(Py_intptr_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) >= 2 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((((Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(Py_intptr_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) >= 3 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((((((Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(Py_intptr_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) >= 4 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((((((((Py_intptr_t)digits[3]) << PyLong_SHIFT) | (Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (Py_intptr_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(Py_intptr_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(Py_intptr_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(Py_intptr_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(Py_intptr_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (Py_intptr_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(Py_intptr_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(Py_intptr_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(Py_intptr_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 2 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((Py_intptr_t)-1)*(((((Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(Py_intptr_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 2 * PyLong_SHIFT) {
                            return (Py_intptr_t) ((((((Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(Py_intptr_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 3 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((Py_intptr_t)-1)*(((((((Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(Py_intptr_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 3 * PyLong_SHIFT) {
                            return (Py_intptr_t) ((((((((Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(Py_intptr_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 4 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((Py_intptr_t)-1)*(((((((((Py_intptr_t)digits[3]) << PyLong_SHIFT) | (Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(Py_intptr_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 4 * PyLong_SHIFT) {
                            return (Py_intptr_t) ((((((((((Py_intptr_t)digits[3]) << PyLong_SHIFT) | (Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(Py_intptr_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(Py_intptr_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(Py_intptr_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(Py_intptr_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            Py_intptr_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (Py_intptr_t) -1;
        }
    } else {
        Py_intptr_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (Py_intptr_t) -1;
        val = __Pyx_PyInt_As_Py_intptr_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to Py_intptr_t");
    return (Py_intptr_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to Py_intptr_t");
    return (Py_intptr_t) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE npy_uint32 __Pyx_PyInt_As_npy_uint32(PyObject *x) {
    const npy_uint32 neg_one = (npy_uint32) ((npy_uint32) 0 - (npy_uint32) 1), const_zero = (npy_uint32) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(npy_uint32) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(npy_uint32, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (npy_uint32) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_uint32) 0;
                case  1: __PYX_VERIFY_RETURN_INT(npy_uint32, digit, digits[0])
                case 2:
                    if (8 * sizeof(npy_uint32) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint32, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint32) >= 2 * PyLong_SHIFT) {
                            return (npy_uint32) (((((npy_uint32)digits[1]) << PyLong_SHIFT) | (npy_uint32)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_uint32) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint32, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint32) >= 3 * PyLong_SHIFT) {
                            return (npy_uint32) (((((((npy_uint32)digits[2]) << PyLong_SHIFT) | (npy_uint32)digits[1]) << PyLong_SHIFT) | (npy_uint32)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_uint32) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint32, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint32) >= 4 * PyLong_SHIFT) {
                            return (npy_uint32) (((((((((npy_uint32)digits[3]) << PyLong_SHIFT) | (npy_uint32)digits[2]) << PyLong_SHIFT) | (npy_uint32)digits[1]) << PyLong_SHIFT) | (npy_uint32)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (npy_uint32) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(npy_uint32) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_uint32, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_uint32) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_uint32, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_uint32) 0;
                case -1: __PYX_VERIFY_RETURN_INT(npy_uint32, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(npy_uint32,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(npy_uint32) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint32, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint32) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_uint32) (((npy_uint32)-1)*(((((npy_uint32)digits[1]) << PyLong_SHIFT) | (npy_uint32)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(npy_uint32) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint32, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint32) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_uint32) ((((((npy_uint32)digits[1]) << PyLong_SHIFT) | (npy_uint32)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(npy_uint32) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint32, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint32) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_uint32) (((npy_uint32)-1)*(((((((npy_uint32)digits[2]) << PyLong_SHIFT) | (npy_uint32)digits[1]) << PyLong_SHIFT) | (npy_uint32)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_uint32) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint32, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint32) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_uint32) ((((((((npy_uint32)digits[2]) << PyLong_SHIFT) | (npy_uint32)digits[1]) << PyLong_SHIFT) | (npy_uint32)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(npy_uint32) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint32, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint32) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_uint32) (((npy_uint32)-1)*(((((((((npy_uint32)digits[3]) << PyLong_SHIFT) | (npy_uint32)digits[2]) << PyLong_SHIFT) | (npy_uint32)digits[1]) << PyLong_SHIFT) | (npy_uint32)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_uint32) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint32, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint32) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_uint32) ((((((((((npy_uint32)digits[3]) << PyLong_SHIFT) | (npy_uint32)digits[2]) << PyLong_SHIFT) | (npy_uint32)digits[1]) << PyLong_SHIFT) | (npy_uint32)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(npy_uint32) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_uint32, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_uint32) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_uint32, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            npy_uint32 val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (npy_uint32) -1;
        }
    } else {
        npy_uint32 val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (npy_uint32) -1;
        val = __Pyx_PyInt_As_npy_uint32(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to npy_uint32");
    return (npy_uint32) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to npy_uint32");
    return (npy_uint32) -1;
}

/* FastTypeChecks */
  #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
  static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionImport */
  #ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
  static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
